/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include "cache-updater.h"

#include <algorithm>
#include <map>
#include <sstream>
#include <thread>

#include "apartment-proxy.h"
#include "businterface.h"
#include "cache-db.h"
#include "device-cache.h"
#include "ds-api/route-segments.h"
#include "dss.h"
#include "event.h"
#include "hal/hardware-abstraction.h"
#include "model/apartment.h"
#include "model/set.h"
#include "model/zone.h"
#include "propertysystem_common_paths.h"
#include "src/dsa-http-client.h"
#include "test-device-presence.h"

DS_REFLECT_DEFINE_STRUCT((dss), DeviceCacheMeta);

using dss::cache::CacheState;

namespace dss {

constexpr const char* CacheUpdatePlugin::CacheUpdateEvent;
constexpr const char* CacheUpdatePlugin::CacheUpdateEventFast;
constexpr const char* CacheUpdatePlugin::CacheUpdateEventResume;
constexpr const int CacheUpdatePlugin::CacheUpdateDefaultBackoffInSec;

static ds::LogChannel dsLogChannel("CacheUpdatePlugin");

CacheUpdatePlugin::CacheUpdatePlugin(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db,
                                     EventInterpreter& interpreter, std::shared_ptr<PropertySystem> propSystem,
                                     std::shared_ptr<EventQueue> eventQueue,
                                     std::shared_ptr<EventRunner> eventRunner)
    : EventInterpreterPlugin("device_scene_cache_updater", &interpreter),
      m_apartment(apt),
      m_hal(hal),
      m_pPropSystem(propSystem),
      m_pEventQueue(eventQueue),
      m_pEventRunner(eventRunner),
      m_metaInfoCache(db),
      m_hitBackoffInSec(CacheUpdateDefaultBackoffInSec) {
  if (m_pPropSystem) {
    auto statusProperty = m_pPropSystem->createProperty(pp_device_cache_update_status);
    statusProperty->linkToProxy(PropertyProxyReference<bool>(m_updateInProgress, false));
  }
  migrateDeviceCache();
}

CacheState CacheUpdatePlugin::readSceneFlags(const Dsuid& dsuid, const DeviceSceneId& id, SceneFlags& flags) {
  DS_DEBUG("readSceneFlags", dsuid, id, flags);
  if (!flags.dontCare || !flags.ignoreLocalPrio || !flags.specialHandling || !flags.blink || !flags.ledConfig ||
      !flags.dimTime) {
    m_hal->readBool({.scene = id, .property = SceneProperty::dontCare});
    DS_NOTICE("readSceneFlags miss", dsuid, id, flags);
    return CacheState::miss;
  }

  return CacheState::hit;
}

CacheState CacheUpdatePlugin::readScene(const Dsuid& dsuid, const struct DeviceSceneId& id, SceneEntry& entry) {
  DS_DEBUG("readScene", dsuid, id, entry);
  if (readSceneFlags(dsuid, id, entry.flags) == CacheState::miss) {
    return CacheState::miss;
  }
  if (!entry.action) {
    // Not used
  }
  // Figure out if this scene pertains to device
  for (auto& channelPair : entry.channels) {
    auto channelId = DeviceSceneChannelId{.scene = id, .type = channelPair.first};
    if (!channelPair.second.dontCare) {
      DS_NOTICE("readScene miss", dsuid, channelId, "dontCare");
      m_hal->readBool({channelId, ChannelProperty::dontCare});
      return CacheState::miss;
    } else if (!channelPair.second.value) {
      if (!std::isnan(m_hal->readDouble(dsuid, toPath(channelId)))) {
        DS_NOTICE("readScene miss", dsuid, channelId, "value");
        // NaN is used if the device can't be read ATM
        // Ending up here only says that we used up our powerline budget and
        // should backoff. But The real danger was, if we backoff without having
        // made progress. Then the cache updater could end stuck here.
        return CacheState::miss;
      }
    }
  }

  return CacheState::hit;
}

CacheState CacheUpdatePlugin::readSceneTable(const Dsuid& dsuid, DeviceSceneCache& sceneCache) {
  DS_DEBUG("readSceneTable", dsuid, "sceneCache");
  // TODO: Optimize the number of times we run through
  if (!sceneCache.changeCounter) {
    // Not implemented yet
  }

  DS_DEBUG("Scene count", sceneCache.sceneTable.size());
  for (auto& scenePair : sceneCache.sceneTable) {
    DS_DEBUG("loop", scenePair);
    if (readScene(dsuid, {.dsuid = dsuid, .sceneId = scenePair.first}, scenePair.second) == CacheState::miss) {
      return CacheState::miss;
    }
  }

  return CacheState::hit;
}

CacheState CacheUpdatePlugin::readDevice(const Dsuid& dsuid, DeviceCache& cache) {
  DS_DEBUG("readDevice", dsuid, "cache");
  if (readSceneTable(dsuid, cache.scene) == CacheState::miss) {
    return CacheState::miss;
  }

  for (auto& property : cache.property) {
    DS_DEBUG("readDevice", dsuid, property);
    if (!property.second) {
      DS_NOTICE("readDevice miss", dsuid, property);
      if (property.first == DeviceProperty::outputMode) {
        m_hal->readString(dsuid, ds::str("/property/", property.first));
      } else {
        m_hal->readDouble(dsuid, ds::str("/property/", property.first));
      }
      return CacheState::miss;
    }
  }
  return CacheState::hit;
}

void CacheUpdatePlugin::reschedule() {
  std::shared_ptr<Event> pEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
  pEvent->setProperty(EventProperty::Time, "+" + std::to_string(m_hitBackoffInSec));
  m_pEventQueue->pushEvent(pEvent);

  DS_DEBUG("Cache miss detected, rescheduling in", m_hitBackoffInSec);
}

DeviceCacheMeta CacheUpdatePlugin::getMetaInfo() {
  return m_metaInfoCache.tryLookup(DSUID_NULL);
}

void CacheUpdatePlugin::updateMetaInfo(const DeviceCacheMeta& info) {
  m_metaInfoCache.update(DSUID_NULL, info);
}

void CacheUpdatePlugin::migrateFw387() {
  auto& cacheDb = m_hal->getCache();
  for (auto& dsuid : cacheDb.devices()) {
    auto device = m_apartment->tryGetDeviceByDSID(dsuid);
    if (!device) {
      DS_NOTICE(dsuid, "remove stale entry from cache");
      cacheDb.remove(dsuid);
      continue;
    }

    try {
      if (device->getRevisionID() != 0x387 || device->getDeviceClass() != DeviceClasses_t::DEVICE_CLASS_GR) {
        DS_NOTICE(dsuid, "not affected, skip");
        continue;
      }
    } catch (const ds::Exception& e) {
      DS_NOTICE(dsuid, "assume 3.8.7 device", e.what());
    }

    DS_NOTICE(dsuid, "Firmware 3.8.7 bug. Cached output values are likely corrupted, invalidating cache.");
    auto cacheEntry = cacheDb.tryLookup(dsuid);
    for (auto& scenePair : cacheEntry.scene.sceneTable) {
      auto& sceneEntry = scenePair.second;
      for (auto& channelPair : sceneEntry.channels) {
        // dontCare is fine, only channels are affected
        channelPair.second.value = {};
        channelPair.second.registerValues.clear();
      }
    }
    m_hal->getCache().update(dsuid, cacheEntry);
  }
}

void CacheUpdatePlugin::migrateDeviceCache() {
  auto metaInfo = getMetaInfo();

  if (metaInfo.version == m_deviceCacheVersion) {
    DS_NOTICE("device cache up-to-date: ", metaInfo);
    return;
  }

  if (metaInfo.version > m_deviceCacheVersion) {
    DS_NOTICE("device cache downgrade: ", metaInfo, " to ", m_deviceCacheVersion);
    updateMetaInfo({.version = m_deviceCacheVersion});
    return;
  }

  if (metaInfo.version < 1 && !m_modelReady) {
    // fw not yet cached in device/param cache, hence need online device
    DS_NOTICE("postpone device cache migration until apartment model readout is complete");
    return;
  }

  DS_NOTICE("device-cache:", metaInfo.version, "starting migration");
  auto& cacheDb = m_hal->getCache();
  for (auto& dsuid : cacheDb.devices()) {
    if (!m_hal->deviceCacheable(dsuid)) {
      DS_NOTICE("device no longer cacheable", dsuid);
      cacheDb.remove(dsuid);
      continue;
    }
  }

  DS_NOTICE("devices in cache:", cacheDb.devices().size());
  switch (metaInfo.version) {
    case 0: migrateFw387(); [[fallthrough]];
    default: break;
  }

  DS_NOTICE("migration complete");
  updateMetaInfo({.version = m_deviceCacheVersion});
}

void CacheUpdatePlugin::scanStep() {
  DS_DEBUG("Scanning Device Scene Cache");
  Set devices = m_apartment->getDevices().getByPresence(true);
  m_scanStepCount++;

  DS_DEBUG("Set size", devices.length());

  if (!m_modelReady) {
    if (m_scanStepCount % 100 == 1) {
      // clogs the log if called every 200ms
      DS_NOTICE("scanStep: apartment model not yet ready. postponing scanning...");
    }
    reschedule();
    return;
  }

  if (!m_priorityQueue.empty()) {
    auto dsuid = m_priorityQueue.front();
    m_hal->populateCache(dsuid);

    auto cacheEntry = m_hal->getCache().tryLookup(dsuid);
    if (readDevice(dsuid, cacheEntry) == CacheState::miss) {
      // already used up its fair share of ds485 bus, back-off
      reschedule();
      return;
    }

    // poke ds-model, hence smartphone
    DS_NOTICE("emit ApartmentProxyStateChanged")
    m_apartment->pushEvent(std::make_shared<Event>(EventName::ApartmentProxyStateChanged));
    m_priorityQueue.pop_front();
  }

  int continueIndex = 0;
  if (m_currentDev) {
    int deviceIndex = devices.findByDSID(*m_currentDev);
    continueIndex = deviceIndex != devices.length() ? deviceIndex : 0;
  }

  for (int d = continueIndex; d < devices.length(); d++) {
    std::shared_ptr<Device> device = devices[d].getDevice();
    if (!m_hal->deviceCacheable(device->getDSID())) {
      continue;
    }

    int retriesLeft = 2;
    bool deviceFailHard = false;

    while (retriesLeft > 0 && !deviceFailHard) {
      try {
        if (!m_currentDev || *m_currentDev != device->getDSID()) {
          m_currentDev = device->getDSID();
          m_hal->populateCache(*m_currentDev);
        }
        DeviceCache cacheEntry = m_hal->getCache().tryLookup(device->getDSID());

        if (readDevice(device->getDSID(), cacheEntry) == CacheState::miss) {
          reschedule();
          return;
        }
        // device cache scan finished, continue with next device
        break;
      } catch (const BusApiError& e) {
        if (std::string(e.what()).find("[-20]") != std::string::npos) {
          if (--retriesLeft == 0) {
            auto&& dss = *DSS::getInstance();
            auto dsa = makeDsaHttpClient();
            auto& backendVdcs = dss.backendVdcs();
            DS_WARNING("Device failed to react during cache scan, pinging device", device->getDSID());
            // side effect: if device is offline dsa will update its presence status
            testDevicePresence(std::locale(""), *device, *dsa, backendVdcs);
          } else {
            DS_WARNING("Device failed to react during cache scan", retriesLeft);
          }
        } else {
          deviceFailHard = true;
        }
      } catch (const ds::Exception& e) {
        DS_NOTICE("Could not scan DSUID: ", device->getDSID(), e.what());
        deviceFailHard = true;
      }
    }
  }
  DS_DEBUG("Cache update process completed");
  m_hitBackoffInSec = CacheUpdateDefaultBackoffInSec;
  m_updateInProgress = false;
  m_currentDev = {};
  m_apartment->proxy().updateParameterCacheStatus();
  std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::CacheUpdateFinished);
  m_pEventQueue->pushEvent(pEvent);
}

void CacheUpdatePlugin::scanStart(ScanMode mode) {
  // single-threaded, running from event-loop context
  switch (mode) {
    case ScanMode::NORMAL:
      m_apartment->proxy().updateParameterCacheStatus();
      m_updateInProgress = true;
      scanStep();
      break;
    case ScanMode::FAST:
      m_updateInProgress = true;
      m_hitBackoffInSec = 0;
      m_apartment->proxy().updateParameterCacheStatus();
      scanStep();
      break;
  }
}

void CacheUpdatePlugin::scanDevice(const Dsuid& dsuid) {
  m_priorityQueue.push_back(dsuid); // running from event-loop, single-threaded
  DS_NOTICE("schedule priority readout for device ", dsuid);

  // single-threaded, since runnig from event-loop (aka no concurrent scanCache)
  if (!m_updateInProgress) {
    scanStart(ScanMode::NORMAL);
  }
}

void CacheUpdatePlugin::handleModelReady() {
  // if we need to devices to be online to migrate the cache
  DS_NOTICE("modelReady event received");
  m_modelReady = true;
  migrateDeviceCache();
}

void CacheUpdatePlugin::handleEvent(Event& _event, const EventSubscription& _subscription) {
  try {
    if (_event.getName() == CacheUpdateEvent) {
      if (!m_updateInProgress) {
        scanStart(ScanMode::NORMAL);
      }
    } else if (_event.getName() == CacheUpdateEventFast) {
      if (!m_updateInProgress) { // TODO backoff interval should still be shortened
        scanStart(ScanMode::FAST);
      }
    } else if (_event.getName() == CacheUpdateEventResume) {
      scanStep();
    } else if (_event.getName() == CacheUpdateEventSingleDevice) {
      DS_REQUIRE(_event.hasPropertySet(EventProperty::DSUID));
      scanDevice(str2dsuid(_event.getPropertyByName(EventProperty::DSUID)));
    } else if (_event.getName() == EventName::ModelReady) {
      handleModelReady();
    }
  } catch (const std::exception& e) {
    DS_WARNING("Exception occured during cache update process", e.what());
    reschedule();
  } catch (...) {
    DS_WARNING("Unexpected error encountered in cache updater.");
    reschedule();
  }
}

void CacheUpdatePlugin::subscribe() {
  auto&& interpreter = getEventInterpreter();
  interpreter.subscribe(*this, CacheUpdateEvent);
  interpreter.subscribe(*this, CacheUpdateEventFast);
  interpreter.subscribe(*this, CacheUpdateEventResume);
  interpreter.subscribe(*this, CacheUpdateEventSingleDevice);
  interpreter.subscribe(*this, EventName::ModelReady);

  std::shared_ptr<Event> pEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent);
  pEvent->setProperty(EventProperty::ICalStartTime, DateTime().addMinute(10).toRFC2445IcalDataTime());
  PropertyNodePtr scanIntervalNode = m_pPropSystem->getProperty(pp_device_cache_update_interval);
  int scanDelay = 1440;
  if (scanIntervalNode != nullptr) {
    scanDelay = m_pPropSystem->getProperty(pp_device_cache_update_interval)->getIntegerValue();
  } else {
    DS_WARNING("Could not find interval for cache update in property tree, assuming 1 day.");
  }
  pEvent->setProperty(EventProperty::ICalRRule, "FREQ=MINUTELY;INTERVAL=" + intToString(scanDelay));
  m_pEventQueue->pushEvent(pEvent);
}

} // namespace dss
