/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "device-info.h"

#include <ds-model/model.h>
#include <ds-model/scenarios.h>
#include <ds-scenarios.h>
#include <ds/locale/ds-locale.h>
#include <ds/log.h>
#include <ds/system.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/locale/message.hpp>
#include <boost/range/algorithm/copy.hpp>
#include <boost/range/algorithm/find.hpp>
#include <boost/range/algorithm/find_if.hpp>
#include <boost/range/algorithm/sort.hpp>
#include <boost/range/algorithm_ext/push_back.hpp>

#include "businterface.h"
#include "dss.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "propertysystem.h"
#include "protobufjson.h"
#include "vdc-db.h"
#include "vdc-element-reader.h"
#include "vdc-messages.pb.h"

static ds::LogChannel dsLogChannel("vdcInfo");

namespace dss {

bool DeviceInfo::Filter::operator==(const DeviceInfo::Filter& x) const { return asBoolVector() == x.asBoolVector(); }

DeviceInfo::DeviceInfo(Prefer prefer, StructureQueryBusInterface& query, VdcDb& db, const Device* device,
                       std::string gtin_, const std::locale& locale, JSONWriter& json, const ds::System& system)
    : m_prefer(prefer),
      m_query(query),
      m_db(db),
      m_device(device),
      m_gtin(std::move(gtin_)),
      m_locale(locale),
      m_json(json),
      m_system(system),
      m_hasDynamicDefinitions([&] {
        if (m_device) {
          const auto& device = *m_device;
          if (!device.isVdcDevice()) {
            return false;
          }
          if (!device.isPresent()) {
            // Queries would fail and generate errors.
            // Sadly apartment/getDeviceInfo queries also inactive devices.
            return false;
          }
          switch (m_prefer) {
            case Prefer::none: {
              return DSS::getInstance()
                  ->getApartment()
                  .getDSMeterByDSID(device.getDSMeterDSID())
                  ->getCapability_HasDynamicDefinitions();
            }
            case Prefer::db: return false;
            case Prefer::device: return true;
          }
        }
        return false;
      }()) {
  DS_CONTEXT(*this);

  auto hasVdcDb = true;
  if (DSS::hasInstance()) {
    DSS::getInstance()->getPropertySystem().setBoolValue("/config/hasDeviceDb", true, true, false);
    hasVdcDb = DSS::getInstance()->getPropertySystem().getBoolValue("/config/hasDeviceDb");
  }

  auto useVdcDb = true;
  if (m_device) {
    auto&& device = *m_device;
    m_gtin = device.getOemEanAsString();
    m_vdcTrDomain = device.vdcTrDomain();
    useVdcDb = device.isVdcDevice() ? true : false;
  } else {
    m_vdcTrDomain = ds::str("vdc-common,vdc-gtin-", m_gtin);
  }

  if (hasVdcDb && useVdcDb) {
    try {
      m_dbSpecs = m_db.getSpec(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db spec", e);
    }

    try {
      m_dbStates = m_db.getStates(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db states", e);
    }

    try {
      m_dbEvents = m_db.getEvents(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db events", e);
    }

    try {
      m_dbProperties = m_db.getProperties(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db properties", e);
    }

    try {
      m_dbSensors = m_db.getSensors(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db sensors", e);
    }

    try {
      m_dbActions = m_db.getActions(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db actions", e);
    }

    try {
      m_dbStandardActions = m_db.getStandardActions(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db standard actions", e);
    }

    try {
      m_dbCategories = m_db.getCategories(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db categories", e);
    }

    try {
      m_dbOutputs = m_db.getOutputs(m_gtin);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to query db outputs", e);
    }
  }
}

template <class Items>
typename Items::value_type DeviceInfo::dbItemOrDefault(const Items& items, const std::string& id) {
  typedef typename Items::value_type T;
  auto itemFoundEnd = boost::find_if<boost::return_found_end>(items, [&](const T& x) { return x.id == id; });
  if (itemFoundEnd) {
    return itemFoundEnd.front();
  }

  T item = {};
  item.id = id;
  return item;
}

VdcDb::SensorDesc DeviceInfo::dbSensorOrDefault(unsigned index, const std::string& id) {
  auto itemFoundEnd =
      boost::find_if<boost::return_found_end>(m_dbSensors, [&](const VdcDb::Sensor& x) { return x.prop.id == id; });
  if (itemFoundEnd) {
    return itemFoundEnd.front();
  }
  VdcDb::Sensor sensor = {};
  sensor.sensorIndex = ds::str(index);
  sensor.prop.id = id;
  return sensor;
}

void DeviceInfo::addParameterDescription(const VdcDb::PropertyDesc& prop, gsl::czstring optionsTrContext) {
  switch (prop.typeId) {
    case VdcDb::propertyTypeId::integer:
    case VdcDb::propertyTypeId::numeric:
      m_json.add("type", prop.typePostfix.empty() ? "numeric" : prop.typePostfix);
      // TODO(someday): these fields make no sense for "numeric.boolean"
      m_json.add("min", prop.minValue);
      m_json.add("max", prop.maxValue);
      m_json.add("resolution", prop.resolution);
      m_json.add("siunit", prop.siUnit);
      m_json.add("default", prop.defaultValue);
      break;
    case VdcDb::propertyTypeId::string:
      m_json.add("type", prop.typePostfix.empty() ? "string" : prop.typePostfix);
      m_json.add("default", prop.defaultValue);
      break;
    case VdcDb::propertyTypeId::enumeration:
      m_json.add("type", "enumeration");
      m_json.add("default", prop.defaultValue);
      m_json.startObject("options");
      for (const auto& option : prop.values) {
        m_json.add(option, vdcTr(optionsTrContext, option.c_str()));
      }
      m_json.endObject();
      break;
    case VdcDb::propertyTypeId::boolean:
      m_json.add("type", prop.typePostfix.empty() ? "boolean" : prop.typePostfix);
      m_json.add("default", prop.defaultValue);
      break;
  }
  if (prop.readonly)
    m_json.add("readonly", prop.readonly);
  if (prop.optional)
    m_json.add("optional", prop.optional);
}

static std::string convertHardwareGuid(const std::string& guid) {
  std::string hwid;
  if (beginsWith(guid, "macaddress:")) {
    hwid = "MAC " + guid.substr(11);
  } else if (beginsWith(guid, "enoceanaddress:")) {
    hwid = "EnOcean ID " + guid.substr(15);
  } else if (guid.find(':') != std::string::npos) {
    hwid = guid.substr(guid.find(':') + 1);
  } else {
    hwid = guid;
  }
  return hwid;
}

void DeviceInfo::addSpec() {
  DS_CONTEXT(*this);
  DS_DEBUG_ENTER("");
  std::vector<VdcDb::SpecDesc> specs;

  // Start with specs from db. Db spec values must be translated.
  for (const auto& dbSpec : m_dbSpecs) {
    specs.push_back(dbSpec);
    auto&& spec = specs.back();
    spec.value = vdcTr(ds::str("spec/", spec.id, "/value").c_str(), spec.value.c_str());
  }

  // add / overried spec item
  auto insert = [&](const char* id, std::string value) {
    DS_CONTEXT(id);
    auto specFoundEnd =
        boost::find_if<boost::return_found_end>(specs, [&](const VdcDb::SpecDesc& x) { return x.id == id; });
    if (specFoundEnd) {
      // override db spec by the one in device
      specFoundEnd.front().value = std::move(value);
    } else {
      // create spec, it was not in db
      specs.push_back(VdcDb::SpecDesc({}));
      auto&& spec = specs.back();
      spec.id = id;
      spec.value = std::move(value);
    }
  };

  insert("dsDeviceGTIN", m_gtin);

  if (m_device) {
    auto&& device = *m_device;
    if (device.isVdcDevice()) {
      const auto& spec = device.getVdsdSpec();
      // [dS-Article Data]
      // spec.oemGuid
      // spec.oemModelGuid
      // spec.vendorGuid
      insert("vendorId", spec.vendorId);
      // [Hardware Endpoint description]
      insert("hardwareGuid", convertHardwareGuid(spec.hardwareGuid));
      insert("hardwareModelGuid", spec.hardwareModelGuid);
      // [System ID's]
      // spec.modelUID
      insert("descriptionsGroup", spec.descriptionsGroup);
      insert("descriptionsClass", spec.descriptionsClass);
      // [End-User Info]
      insert("name", device.getName()); // spec.id ignored
      insert("model", spec.model);
      // spec.displayId
      // spec.hardwareVersion
      insert("modelVersion", spec.modelVersion);
      insert("vendorName", spec.vendorName);
    } else {
      insert("dsDeviceGTIN", device.getOemEanAsString());
      insert("vendorId", ds::str(device.getVendorID()));
      insert("vendorName", device.vendorTitle());
      insert("hardwareGuid", device.getDisplayID());
      insert("hardwareModelGuid", device.getGTIN());
      insert("descriptionsGroup", "");
      insert("descriptionsClass", "");
      insert("name", device.getName());
      insert("model", device.getHWInfo());
      insert("modelVersion", device.getRevisionAsString());
    }
  }

  m_json.startObject("spec");
  for (const auto& spec : specs) {
    const auto& id = spec.id;
    m_json.startObject(id);
    m_json.add("title", vdcTr("spec", id.c_str()));
    m_json.add("tags", spec.tags);
    m_json.add("value", spec.value);
    m_json.endObject();
  }
  m_json.endObject();
}

static std::vector<std::string> readValues(VdcElementReader vdcValues) {
  std::vector<std::string> values;
  for (auto&& vdcValue : vdcValues) {
    values.push_back(vdcValue.getName());
  }
  return values;
}

void DeviceInfo::getStateDescriptions(std::vector<VdcDb::StateDesc>& states) {
  if (m_device && m_hasDynamicDefinitions) {
    auto&& device = *m_device;
    auto&& reply = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {
                                              "deviceStateDescriptions" });
    auto&& vdcStates =
        VdcElementReader(reply.vdc_response_get_property().properties())["deviceStateDescriptions"];
    for (auto&& vdcState : vdcStates) {
      auto&& name = vdcState.getName();
      DS_CONTEXT(name);
      auto state = dbItemOrDefault(m_dbStates, name);
      state.values = readValues(vdcState["value"]["values"]);
      states.push_back(std::move(state));
    }
  } else {
    boost::push_back(states, m_dbStates);
  }
}

std::string DeviceInfo::getTranslatedStateName(const std::string& id) {
  DS_CONTEXT(*this);
  DS_DEBUG_ENTER("");
  std::vector<VdcDb::StateDesc> states;

  getStateDescriptions(states);

  for (const auto& state : states) {
    if (state.id == id) {
      return vdcTr("state", id.c_str());
    }
  }
  return id;
}

std::string DeviceInfo::getTranslatedStateValue(const std::string& stateId, const std::string& optionId) {
  DS_CONTEXT(*this);
  DS_DEBUG_ENTER("");
  std::vector<VdcDb::StateDesc> states;

  getStateDescriptions(states);

  for (const auto& state : states) {
    const auto& id = state.id;
    if (state.id == stateId) {
      for (const auto& option : state.values) {
        if (option == optionId) {
          return vdcTr(ds::str("state/", id, "/enum").c_str(), option.c_str());
        }
      }
    }
  }
  return optionId;
}

void DeviceInfo::addStateDescriptions() {
  DS_CONTEXT(*this);
  DS_DEBUG_ENTER("");
  std::vector<VdcDb::StateDesc> states;

  getStateDescriptions(states);

  m_json.startObject("stateDescriptions");
  for (const auto& state : states) {
    const auto& id = state.id;
    m_json.startObject(id);
    m_json.add("title", vdcTr("state", id.c_str()));
    m_json.add("tags", state.tags);
    m_json.startObject("options");
    for (const auto& option : state.values) {
      m_json.add(option, vdcTr(ds::str("state/", id, "/enum").c_str(), option.c_str()));
    }
    m_json.endObject();
    m_json.endObject();
  }
  m_json.endObject();
}

void DeviceInfo::addEventDescriptions() {
  DS_CONTEXT(*this);
  DS_DEBUG_ENTER("");
  std::vector<VdcDb::EventDesc> events;

  if (m_device && m_hasDynamicDefinitions) {
    auto&& device = *m_device;
    auto&& reply = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {"deviceEventDescriptions"});
    auto&& vdcEvents = VdcElementReader(reply.vdc_response_get_property().properties())["deviceEventDescriptions"];

    for (auto&& vdcEvent : vdcEvents) {
      const auto& id = vdcEvent.getName();
      DS_CONTEXT(id);
      auto event = dbItemOrDefault(m_dbEvents, id);
      events.push_back(std::move(event));
    }
  } else {
    boost::push_back(events, m_dbEvents);
  }

  m_json.startObject("eventDescriptions");
  for (const auto& event : events) {
    const auto& id = event.id;
    m_json.startObject(id);
    m_json.add("title", vdcTr("event", id.c_str()));
    if (!event.categoryId.empty()) {
      m_json.add("category", event.categoryId);
    }
    if (event.accessLevel != VdcDb::AccessLevel()) {
      m_json.add("accessLevel", event.accessLevel);
    }
    m_json.endObject();
  }
  m_json.endObject();
}

static VdcDb::PropertyDesc mergeProperty(VdcElementReader vdcProperty, const VdcDb::PropertyDesc& dbProperty) {
  auto property = dbProperty;
  property.id = vdcProperty.getName();
  property.minValue = vdcProperty["min"].getValueAsString();
  property.maxValue = vdcProperty["max"].getValueAsString();
  // TODO(someday): readonly - present in vdc, but not in database and json api
  property.resolution = vdcProperty["resolution"].getValueAsString();
  property.siUnit = vdcProperty["siunit"].getValueAsString();

  property.typePostfix.clear(); // will be restored if vtypeId from vdc typeId matches db typeId
  auto vdcType = vdcProperty["type"].getValueAsString();
  if (vdcType == "numeric") {
    property.typeId = VdcDb::numeric;
  } else if (vdcType == "integer") {
    property.typeId = VdcDb::integer; // postfix lost if type is integer but postfix is numeric
  } else if (vdcType == "boolean") {
    property.typeId = VdcDb::boolean;
  } else if (vdcType == "enumeration") {
    property.typeId = VdcDb::enumeration;
  } else if (vdcType == "string") {
    property.typeId = VdcDb::string;
  } else {
    DS_ERROR("Unrecognized vdc type,  string.", vdcType);
    property.typeId = VdcDb::string;
  }
  // this seems to be the culprit, numeric.duration / numeric.temperature
  // is stored as integer in db, not numeric, hence this comparison fails:
  // solutions: relax the comparisons like numeric ~= integer, or update the db
  // to change type to numeric too
  if (property.typeId == dbProperty.typeId ||
      (property.typeId == VdcDb::numeric && dbProperty.typeId == VdcDb::integer)) {
    property.typePostfix = dbProperty.typePostfix;
  }
  property.siUnit = vdcProperty["siunit"].getValueAsString();
  property.values = readValues(vdcProperty["values"]);
  property.defaultValue = vdcProperty["default"].getValueAsString();

  property.readonly = vdcProperty["readonly"].getValueAsBool();
  property.optional = vdcProperty["optional"].getValueAsBool();
  return property;
}

void DeviceInfo::addPropertyDescriptions() {
  DS_CONTEXT(*this);
  DS_DEBUG_ENTER("");
  std::vector<VdcDb::PropertyDesc> properties;

  if (m_device && m_hasDynamicDefinitions) {
    auto&& device = *m_device;
    auto&& reply = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {"devicePropertyDescriptions"});
    auto&& vdcProperties =
        VdcElementReader(reply.vdc_response_get_property().properties())["devicePropertyDescriptions"];
    for (const auto& vdcProperty : vdcProperties) {
      auto&& name = vdcProperty.getName();
      DS_CONTEXT(name);
      properties.push_back(mergeProperty(vdcProperty, dbItemOrDefault(m_dbProperties, name)));
    }
  } else {
    boost::push_back(properties, m_dbProperties);
  }

  m_json.startObject("propertyDescriptions");
  for (const auto& property : properties) {
    const auto& id = property.id;
    m_json.startObject(id);
    m_json.add("title", vdcTr("property", id.c_str()));
    m_json.add("tags", property.tags);
    addParameterDescription(property, ds::str("property/", id, "/enum").c_str());
    m_json.endObject();
  }
  m_json.endObject();
}

void DeviceInfo::getSensorDescriptions(std::vector<VdcDb::SensorDesc>& sensors) {
  if (m_device && m_hasDynamicDefinitions) {
    auto&& device = *m_device;
    auto&& vdcSensors = m_query.getVdcInputDesc(device.getDSMeterDSID(), device.getDSID());
    for (const auto& vdcSensorPair : vdcSensors.m_sensorInputs) {
      auto&& vdcSensor = vdcSensorPair.second;
      auto&& index = vdcSensor.index;
      auto&& name = vdcSensor.sensorName;
      DS_CONTEXT(index, name);
      auto sensor = dbSensorOrDefault(index, name);
      sensor.dsType = vdcSensor.sensorType; // default: SensorType::UnknownType(=255)
      sensor.prop.typeId = VdcDb::propertyTypeId::numeric;
      sensor.prop.minValue = ds::str(vdcSensor.min);
      sensor.prop.maxValue = ds::str(vdcSensor.max);
      sensor.prop.resolution = ds::str(vdcSensor.resolution);
      sensors.push_back(sensor);
    }
  } else {
    // TODO(someday) check if dsType can be added to database
    if (m_device) {
      for (auto& sensor : m_dbSensors) {
        // SensorType is not stored in database
        // legacy behaviour: dsType is retrieved/cached by device upon readout, use it
        if (static_cast<int>(sensor.dsType) == 0) {
          sensor.dsType = m_device->getSensor(strToInt(sensor.sensorIndex)).m_sensorType;
        }
        sensors.push_back(sensor);
      }
    } else {
      boost::push_back(sensors, m_dbSensors);
    }
  }
}

void DeviceInfo::addSensorDescriptions() {
  DS_CONTEXT(*this);
  std::vector<VdcDb::SensorDesc> sensors;

  getSensorDescriptions(sensors);

  m_json.startObject("sensorDescriptions");
  for (const auto& sensor : sensors) {
    const auto& id = sensor.prop.id;
    m_json.startObject(id);
    m_json.add("title", vdcTr("sensor", id.c_str()));
    m_json.add("tags", sensor.prop.tags);
    addParameterDescription(sensor.prop, "-"); // enum options never used
    m_json.add("dsIndex", strToInt(sensor.sensorIndex));
    if (static_cast<int>(sensor.dsType) != 0) {
      m_json.add("dsType", static_cast<int>(sensor.dsType));
    }
    m_json.endObject();
  }
  m_json.endObject();
}

std::string DeviceInfo::getTranslatedSensorName(const std::string& id) {
  DS_CONTEXT(*this);
  std::vector<VdcDb::SensorDesc> sensors;

  getSensorDescriptions(sensors);

  for (const auto& sensor : sensors) {
    if (sensor.prop.id == id) {
      return vdcTr("sensor", sensor.prop.id.c_str());
    }
  }
  return id;
}

void DeviceInfo::addActionDescriptions() {
  DS_CONTEXT(*this);
  DS_DEBUG_ENTER("");
  std::vector<VdcDb::ActionDesc> actions;

  if (m_device && m_hasDynamicDefinitions) {
    auto&& device = *m_device;
    auto&& reply = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {"deviceActionDescriptions"});
    auto&& vdcActions = VdcElementReader(reply.vdc_response_get_property().properties())["deviceActionDescriptions"];
    for (auto&& vdcAction : vdcActions) {
      auto name = vdcAction.getName();
      DS_CONTEXT(name);
      auto action = dbItemOrDefault(m_dbActions, name);

      const auto& dbParams = action.params;
      auto vdcParams = vdcAction["params"];
      std::vector<VdcDb::PropertyDesc> params;
      for (auto&& vdcParam : vdcParams) {
        auto&& paramName = vdcParam.getName();
        auto&& dbParamFoundEnd = boost::find_if<boost::return_found_end>(
            dbParams, [&](const VdcDb::PropertyDesc& x) { return x.id == paramName; });
        if (dbParamFoundEnd) {
          params.push_back(dbParamFoundEnd.front());
        } else {
          params.push_back(VdcDb::PropertyDesc());
        }
        auto&& param = params.back();
        param = mergeProperty(vdcParam, param);
      }
      action.params = std::move(params);

      actions.push_back(std::move(action));
    }
  } else {
    boost::push_back(actions, m_dbActions);
  }

  m_json.startObject("actionDescriptions");
  for (const auto& action : actions) {
    const auto& id = action.id;
    m_json.startObject(id);
    m_json.add("title", vdcTr("action", id.c_str()));
    if (!action.categoryId.empty()) {
      m_json.add("category", action.categoryId);
    }
    if (!action.tags.empty()) {
      m_json.add("tags", action.tags);
    }
    m_json.startObject("params");
    for (const auto& param : action.params) {
      const auto& paramId = param.id;
      m_json.startObject(paramId);
      if (std::any_of(m_dbOutputs.begin(), m_dbOutputs.end(),
          [&](const VdcDb::OutputDesc& o) { return o.id == paramId; })
      ) {
        m_json.add("title", vdcTr("output", paramId.c_str()));
      } else {
        m_json.add("title", vdcTr(ds::str("action/", id, "/param").c_str(), paramId.c_str()));
      }
      m_json.add("tags", param.tags);
      addParameterDescription(param, ds::str("action/", id, "/param/", paramId, "/enum").c_str());
      m_json.endObject();
    }
    m_json.endObject();
    m_json.endObject();
  }
  m_json.endObject();
}

std::vector<dsModel::DeviceScenario> DeviceInfo::scenarios() {
  std::vector<dsModel::DeviceScenario> result{};
  if(m_device) {
    auto&& device = *m_device;
    for (auto& deviceAction : deviceActions()) {
      const auto actionId = dsModel::DeviceScenarioActionId{deviceAction.id};
      result.push_back(dsModel::DeviceScenario{
          .id = encodeId(DeviceScenarioId{device.id(), actionId}),
          .type = "deviceScenario",
          .lastChanged = {},
          .attributes = {.name = std::move(deviceAction.name), .dsDevice = device.id(), .actionId = actionId}});
    }
  }
  return result;
}

std::vector<DeviceInfoAction> DeviceInfo::deviceActions() {
  std::vector<DeviceInfoAction> result{};
  boost::copy(actions("standardActions"), std::back_inserter(result));
  boost::copy(actions("customActions"), std::back_inserter(result));
  boost::copy(actions("dynamicActionDescriptions"), std::back_inserter(result));
  boost::sort(result, [](const auto& a, const auto& b) { return a.id < b.id; });
  DS_DEBUG_LEAVE();
  return result;
}

std::vector<DeviceInfoAction> DeviceInfo::actions(const std::string& propertyType) {
  DS_DEBUG_ENTER(propertyType);
  std::vector<DeviceInfoAction> actions;
  try {
    auto standardActionName = [&](const auto& id) {
      return dsApi::Translatable{
          .domain = ds::str("vdc-gtin-", m_gtin), .message = id.substr(4).c_str(), .context = "standardAction"};
    };
    if (m_device && ((propertyType != "standardActions") || m_hasDynamicDefinitions) && m_device->isVdcDevice()) {
      auto&& device = *m_device;
      auto&& reply = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {propertyType});
      auto&& vdcActions = VdcElementReader(reply.vdc_response_get_property().properties())[propertyType.c_str()];
      for (auto&& vdcAction : vdcActions) {
        const auto& id = vdcAction.getName();

        auto vdcArgs = vdcAction["params"];
        auto params = std::vector<std::pair<std::string, std::string>>{};
        for (auto&& vdcArg : vdcArgs) {
          params.push_back({vdcArg.getName(), vdcArg.getValueAsString()});
        }
        actions.push_back(
            {.id = id,
             .name = propertyType == "standardActions"
                         ? standardActionName(id)
                         : dsApi::Translatable{.domain = {}, .message = vdcAction["title"].getValueAsString()},
             .title = vdcTr("standardAction", vdcAction["title"].getValueAsString().c_str()),
             .action = vdcAction["action"].getValueAsString(),
             .params = std::move(params)});
      }
    } else {
      for (auto& standardAction : m_dbStandardActions) {
        const auto& id = standardAction.id;
        actions.push_back({.id = id,
                           .name = standardActionName(id),
                           .title = vdcTr("standardAction", id.substr(4).c_str()),
                           .action = std::move(standardAction.action),
                           .params = std::move(standardAction.params),
                           .tags = {}});
      }
    }

    if (m_device && (propertyType == "standardActions")) {
      auto&& device = *m_device;
      for (const auto& deviceAction : device.actions()) {
        actions.push_back(DeviceInfoAction{.id = deviceAction.id,
                                           .name = deviceAction.title,
                                           .title = dsTr(m_locale, deviceAction.title),
                                           .action = {},
                                           .params = {},
                                           .tags = {}});
      }
    }

    DS_DEBUG_LEAVE(propertyType, actions);
  } catch (const std::exception& e) {
    DS_ERROR("Failed to read out device actions.", propertyType, e);
  }
  return actions;
}

void DeviceInfo::addStandardActions() {
  DS_DEBUG_ENTER("");

  std::vector<DeviceInfoAction> standardActions = actions("standardActions");
  auto DS_UNIQUE_NAME(x) = m_json.startScopedObject("standardActions");
  for (const auto& standardAction : standardActions) {
    const auto& id = standardAction.id;
    auto DS_UNIQUE_NAME(x) = m_json.startScopedObject(id);
    m_json.add("title", vdcTr("standardAction", id.substr(4).c_str()));
    if (!standardAction.action.empty()) {
      m_json.add("action", standardAction.action);
    }
    if (!standardAction.tags.empty()) {
      m_json.add("tags", standardAction.tags);
    }
    if (!standardAction.params.empty()) {
      auto DS_UNIQUE_NAME(x) = m_json.startScopedObject("params");
      for (const auto& param : standardAction.params) {
        m_json.add(param.first, param.second);
      }
    }
  }
}

void DeviceInfo::addCustomActions() {
  if (!m_device) {
    return;
  }
  auto&& device = *m_device;
  if (!device.isPresent() || !device.isVdcDevice()) {
    m_json.startObject("customActions");
    m_json.endObject();
    return;
  }

  auto&& message = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {"customActions"});
  VdcElementReader reader(message.vdc_response_get_property().properties());
  m_json.add("customActions");
  ProtobufToJSon::processElementsPretty(reader["customActions"].childElements(), m_json);
}

void DeviceInfo::addDynamicDeviceActions() {
  if (!m_device) {
    return;
  }
  auto&& device = *m_device;
  if (!device.isPresent() || !device.isVdcDevice()) {
    m_json.startObject("deviceActions");
    m_json.endObject();
    return;
  }

  auto&& message = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {"dynamicActionDescriptions"});
  VdcElementReader reader(message.vdc_response_get_property().properties());
  m_json.add("deviceActions");
  ProtobufToJSon::processElementsPretty(reader["dynamicActionDescriptions"].childElements(), m_json);
}

std::vector<dsModel::State> DeviceInfo::deviceStates() {
  std::vector<dsModel::State> result{};

  std::vector<VdcDb::StateDesc> states;
  getStateDescriptions(states);

  for (const auto& state : states) {
    dsModel::State item;
    item.id = dsModel::StateId{state.id};
    item.name = {.domain = ds::str("vdc-gtin-", m_gtin), .message = state.id, .context = "state"};
    for (const auto& option : state.values) {
      item.options.push_back(dsModel::StateOption{.id = dsModel::StateOptionId{option},
                                                  .name = {.domain = ds::str("vdc-gtin-", m_gtin),
                                                           .message = option,
                                                           .context = ds::str("state/", state.id, "/enum")}});
    }
    result.push_back(item);
  }

  return result;
}

std::vector<dsModel::StateStatus> DeviceInfo::deviceStateStatuses() {
  std::vector<dsModel::StateStatus> result{};
  if (m_device && m_device->isVdcDevice()) {
    auto&& device = *m_device;
    auto&& message = m_query.getVdcProperty(device.getDSMeterDSID(), device.getDSID(), {"deviceStates"});
    auto&& vdcStates = VdcElementReader(message.vdc_response_get_property().properties())["deviceStates"];

    for (const auto& vdcState : vdcStates) {
      result.push_back(dsModel::StateStatus{.id = dsModel::StateId{vdcState.getName()},
                                            .value = dsModel::StateOptionId{vdcState["value"].getValueAsString()}});
    }
  }
  return result;
}

void DeviceInfo::addOperational() {
  if (!m_device || !m_device->isPresent() || !m_device->isVdcDevice()) {
    m_json.startObject("outputs");
    m_json.endObject();
    m_json.startObject("states");
    m_json.endObject();
    m_json.startObject("properties");
    m_json.endObject();
    m_json.startObject("sensors");
    m_json.endObject();
    return;
  }
  auto&& device = *m_device;

  const auto& message = m_query.getVdcProperty(
      device.getDSMeterDSID(), device.getDSID(), {"channelStates", "deviceStates", "deviceProperties", "sensorStates"});
  VdcElementReader vdcRoot(message.vdc_response_get_property().properties());

  {
    const auto jsonOutputs = m_json.startScopedObject("outputs");
    for (const auto& output : vdcRoot["channelStates"]) {
      const auto jsonOutput = m_json.startScopedObject(output.getName());
      m_json.add("title", vdcTr("output", output.getName().c_str()));
      ProtobufToJSon::processValuePretty(output["value"].getValue(), "value", m_json);
    }
  }

  {
    m_json.startObject("states");
    auto&& vdcStates = vdcRoot["deviceStates"];
    for (const auto& vdcState : vdcStates) {
      std::string name = vdcState.getName();
      DS_CONTEXT("state", name);
      m_json.startObject(name);
      m_json.add("title", vdcTr("state", name.c_str())); // TODO(someday): this is not operation value
      ProtobufToJSon::processValuePretty(vdcState["value"].getValue(), "value", m_json);
      ProtobufToJSon::processValuePretty(vdcState["age"].getValue(), "age", m_json);
      const auto&& vdcValue = vdcState["value"].getValueAsString();
      m_json.add("displayValue", vdcTr(ds::str("state/", name, "/enum").c_str(),
                                     vdcValue.c_str())); // TODO(someday): this is not operation value
      m_json.endObject();
    }
    m_json.endObject();
  }

  {
    m_json.startObject("properties");
    auto&& vdcProperties = vdcRoot["deviceProperties"];
    for (const auto& vdcProperty : vdcProperties) {
      std::string name = vdcProperty.getName();
      DS_CONTEXT("property", name);
      m_json.startObject(name);
      m_json.add("title", vdcTr("property", name.c_str())); // TODO(someday): this is not operation value
      ProtobufToJSon::processValuePretty(vdcProperty.getValue(), "value", m_json);
      m_json.endObject();
    }
    m_json.endObject();
  }

  {
    m_json.startObject("sensors");
    auto&& vdcSensors = vdcRoot["sensorStates"];
    int index = -1;
    for (const auto& vdcSensor : vdcSensors) {
      index++;
      auto name = vdcSensor.getName();
      DS_CONTEXT("sensor", index, name);
      m_json.startObject(name);
      m_json.add("title", vdcTr("sensor", name.c_str()));
      ProtobufToJSon::processValuePretty(vdcSensor["value"].getValue(), "value", m_json);
      // vdc only reports how many seconds elapsed since the event happened
      auto timestamp = m_system.now() - std::chrono::seconds(int(vdcSensor["age"].getValueAsDouble()));
      m_json.add("timestamp", ds::toString(timestamp));
      m_json.endObject();
    }
    m_json.endObject();
  }
}

void DeviceInfo::addCategories() {
  const auto jsonCategories = m_json.startScopedObject("categories");
  for (const auto& category : m_dbCategories) {
    const auto& id = category.id;
    const auto jsonCategory = m_json.startScopedObject(id);
    m_json.add("orderId", category.orderId);
    m_json.add("title", vdcTr("category", id.c_str()));
    m_json.add("alwaysVisible", category.alwaysVisible);
  }
}

void DeviceInfo::addOutputDescriptions() {
  VdcDb::Outputs outputs;

  if (m_device && m_hasDynamicDefinitions) {
    const auto& message =
        m_query.getVdcProperty(m_device->getDSMeterDSID(), m_device->getDSID(), {"channelDescriptions"});
    const VdcElementReader vdcRoot(message.vdc_response_get_property().properties());
    const auto& vdcProperties = vdcRoot["channelDescriptions"];
    for (const auto& vdcOutput : vdcProperties) {
      outputs.push_back(dbItemOrDefault(m_dbOutputs, vdcOutput.getName()));
      outputs.back().minValue = vdcOutput["min"].getValueAsString();
      outputs.back().maxValue = vdcOutput["max"].getValueAsString();
      outputs.back().resolution = vdcOutput["resolution"].getValueAsString();
      outputs.back().outputType = vdcOutput["channelType"].getValueAsString();
      outputs.back().outputIndex = vdcOutput["dsIndex"].getValueAsString();
      const auto&& values = vdcOutput["values"];
      for (const auto& value : values) {
        outputs.back().options.push_back(std::make_pair(value.getValueAsString(), value.getName()));
      }

    }
  } else {
    boost::push_back(outputs, m_dbOutputs);
  }

  const auto jsonOutputs = m_json.startScopedObject("outputDescriptions");
  for (const auto& output : outputs) {
    const auto& id = output.id;
    DS_CONTEXT("output", id);

    const auto jsonOutput = m_json.startScopedObject(id);
    m_json.add("title", vdcTr("output", id.c_str()));
    m_json.add("category", output.categoryId);
    m_json.add("tags", output.tags);
    m_json.add("outputType", output.outputType);
    m_json.add("outputIndex", output.outputIndex);
    m_json.add("type", output.typePostfix.empty() ? "numeric" : output.typePostfix);
    m_json.add("min", output.minValue);
    m_json.add("max", output.maxValue);
    m_json.add("resolution", output.resolution);
    if (!output.defaultValue.empty()) {
      m_json.add("default", output.defaultValue);
    }
    if (!output.siUnit.empty()) {
      m_json.add("siunit", output.siUnit);
    }
    if (!output.options.empty()) {
      m_json.startObject("options");
      for (const auto& option : output.options) {
        m_json.add(option.first, vdcTr(ds::str("output/", id, "/enum").c_str(), option.second.c_str()));
      }
      m_json.endObject();
    }
  }
}

void DeviceInfo::addOperationalIntern() {
  const auto jsonOperational = m_json.startScopedObject("operational");
  addOperational();
}

DeviceInfo::Filter DeviceInfo::parseFilter(const std::string& filterParam) {
  Filter filter = {};
  if (filterParam.empty()) {
    filter.spec = 1;
    filter.stateDesc = 1;
    filter.eventDesc = 1;
    filter.propertyDesc = 1;
    filter.sensorDesc = 1;
    filter.actionDesc = 1;
    filter.standardActions = 1;
    filter.customActions = 1;
    filter.dynamicDeviceActions = 1;
    filter.operational = 1;
    filter.categories = 1;
    filter.outputDesc = 1;
    return filter;
  }
  std::vector<std::string> items = dss::splitString(filterParam, ',');
  for (const auto& item : items) {
    if (item == "spec") {
      filter.spec = 1;
    } else if (item == "stateDesc") {
      filter.stateDesc = 1;
    } else if (item == "eventDesc") {
      filter.eventDesc = 1;
    } else if (item == "propertyDesc") {
      filter.propertyDesc = 1;
    } else if (item == "sensorDesc") {
      filter.sensorDesc = 1;
    } else if (item == "actionDesc") {
      filter.actionDesc = 1;
    } else if (item == "standardActions") {
      filter.standardActions = 1;
    } else if (item == "dynamicDeviceActions") {
      filter.dynamicDeviceActions = 1;
    } else if (item == "customActions") {
      filter.customActions = 1;
    } else if (item == "operational") {
      filter.operational = 1;
    } else if (item == "categories") {
      filter.categories = 1;
    } else if (item == "outputDesc") {
      filter.outputDesc = 1;
    }
  }
  return filter;
}

void DeviceInfo::addByFilter(Filter filter) {
  if (filter.spec) {
    addSpec();
  }
  if (filter.stateDesc) {
    addStateDescriptions();
  }
  if (filter.eventDesc) {
    addEventDescriptions();
  }
  if (filter.propertyDesc) {
    addPropertyDescriptions();
  }
  if (filter.sensorDesc) {
    addSensorDescriptions();
  }
  if (filter.actionDesc) {
    addActionDescriptions();
  }
  if (filter.standardActions) {
    addStandardActions();
  }
  if (filter.customActions) {
    addCustomActions();
  }
  if (filter.dynamicDeviceActions) {
    addDynamicDeviceActions();
  }
  if (filter.operational) {
    addOperationalIntern();
  }
  if (filter.categories) {
    addCategories();
  }
  if (filter.outputDesc) {
    addOutputDescriptions();
  }
}

std::string DeviceInfo::vdcTr(gsl::czstring context, gsl::czstring id) const {
  return dsDtr(m_locale, m_vdcTrDomain, context, id);
}

std::ostream& operator<<(std::ostream& s, const DeviceInfo& x) {
  s << "(device:";
  if (auto&& device = x.device()) {
    s << *device;
  }
  return s << ds::str("(gtin:", x.gtin(), " locale:", x.locale(), ')');
}

std::ostream& operator<<(std::ostream& s, const DeviceInfo::Filter& x) {
  boost::copy(x.asBoolVector(), std::ostream_iterator<bool>(s, " "));
  return s;
}

} // namespace dss
