#include "remote-connectivity.h"

#include <ds-model/remote-connectivity.h>
#include <ds/asio/resource.h>
#include <ds/http.h>
#include <ds/random.h>
#include <event.h>

#include <boost/algorithm/string.hpp>
#include <fstream>

#include "dss.h"
#include "propertysystem_common_paths.h"
#include "security/security.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsApartmentRemoteConnectivity");

namespace dsApartment {
struct RemoteConnectivityTokenResult {
  std::string apartment_id{};
  std::string relay_name{};
  std::string relay_password{};
  std::string access_token{};
  std::string token_type{};
  std::string user_email{};
};
struct RemoteConnectivityTokenError {
  std::string error{};
};
} // namespace dsApartment
DS_JSON_DECLARE_STRUCT((dsApartment), RemoteConnectivityTokenResult,
                       (apartment_id)(relay_name)(relay_password)(access_token)(token_type)(user_email));
DS_JSON_DEFINE_STRUCT((dsApartment), RemoteConnectivityTokenResult);
DS_JSON_DECLARE_STRUCT((dsApartment), RemoteConnectivityTokenError, (error));
DS_JSON_DEFINE_STRUCT((dsApartment), RemoteConnectivityTokenError);

namespace dsApartment {

bool hasCustomizedPassword(ds::System& system) {
  std::string dssadmin(boost::trim_copy(system.readFileIfExists(ds::str(system.dirs().sysconfdir, "/lighttpd.user")).value_or(std::string{})));
  return dssadmin != R"(dssadmin:dSS11:79f2e01bf54e8a0626f04b139a1decc2)";
}

namespace {

struct RemoteConnectivityImpl : dsModel::RemoteConnectivityApi, ds::asio::Resource<> {
  dss::Apartment &m_apartment;
  dss::SimpleHttpClient &m_httpClient;
  dsBackend::Connector &m_connector;
  dss::PropertySystem &m_propertySystem = *m_apartment.getPropertySystem();

  RemoteConnectivityImpl(ds::asio::ResourceBase &parent, dss::Apartment &apartment, dss::SimpleHttpClient &httpClient,
                         dsBackend::Connector &connector)
      : ds::asio::Resource<>{parent, "remoteConnectivity"},
        m_apartment{apartment},
        m_httpClient{httpClient},
        m_connector{connector} {
    DS_INFO_ENTER("");
  }

  std::string baseUrl() const { return m_propertySystem.getStringValue(dss::pp_websvc_private_url); }
  std::string mshubUrl() const { return m_propertySystem.getStringValue(dss::pp_websvc_url_authority); }

  void updateConfigXml(bool enable) {
    try {
      std::string configureWebsvcFilename = "/usr/share/dss/data/config.d/webservice.xml";
      std::string configureWebsvc = R"XML(<?xml version="1.0" encoding="utf-8"?>
<properties version="1">
  <property name="config">
    <property name="webservice-api/enabled" type="boolean">
      <value>true</value>
    </property>
  </property>
</properties>)XML";
      if (enable == false) {
        configureWebsvc.replace(164, 4, "false");
      }
      std::ofstream configureWebsvcFile("/usr/share/dss/data/config.d/webservice.xml");
      configureWebsvcFile << configureWebsvc;
      configureWebsvcFile.close();
    } catch (const std::exception &e) {
      DS_ERROR("Failed to write config xml for remote connectivity", e);
    }
  }

  [[nodiscard]] dsModel::RemoteConnectivity get() const override {
    try {
      const auto pagekiteName = this->pagekiteName();
      return dsModel::RemoteConnectivity{
          .id = dsModel::RemoteConnectivityId{"0"},
          .type = "remoteConnectivity",
          .lastChanged = {},
          .attributes = {
              .authorized = m_propertySystem.tryValue<bool>(dss::pp_websvc_enabled).value_or(false),
              .connected = m_connector.isConnected(),
              .email = m_propertySystem.tryValue<std::string>(dss::pp_websvc_rc_username).value_or(std::string{}),
              .relayName = pagekiteName,
              .relayUrl = pagekiteName.empty() ? std::string{} : ds::str(pagekiteName, ".digitalstrom.net")}};
    } catch (const std::exception &e) {
      DS_ERROR("Failed to get remote connectivity", e);
      return dsModel::RemoteConnectivity{.id = dsModel::RemoteConnectivityId{"0"}};
    }
  }

  // ds-conf entry enabling mocking of responses from backend
  std::optional<RemoteConnectivityTokenResult> mockedResult() {
    auto result = conf().readOr<RemoteConnectivityTokenResult>("remoteConnectivityTokenResultMock", {});
    if (result.apartment_id.empty()) {
      return {};
    } else {
      return result;
    }
  }

  [[nodiscard]] dsModel::AuthorizeRemoteConnectivityResult authorize(
      dsModel::AuthorizeRemoteConnectivityArgs &&args) override {
    if (const auto mockedResult = this->mockedResult()) {
      DS_NOTICE("Mocked authorize", args);
      if (ds::randint(0, 5) == 0) {
        DS_FAIL_REQUIRE("Test failure");
      }
      return dsModel::AuthorizeRemoteConnectivityResult{
          .device_code = "deviceCode",
          .user_code = "userCode",
          .verification_uri = "www.example.com/authorize",
          .verification_uri_complete = "www.example.com/authorize?userCode=userCode",
          .expires_in = 600,
          .interval = 5};
    }
    DS_NOTICE("Authorize", args);
    DS_REQUIRE(hasCustomizedPassword(context().system) == true, "dssadmin password verification failed");

    auto response = m_httpClient.request(dss::HttpRequest{
        .url = ds::str(baseUrl(), "/v2/accessmanagement/device/authorization"),
        .type = dss::HttpMethod::POST,
        .headers = {{"Content-Type", "application/x-www-form-urlencoded"}},
        .postdata = ds::str("client_id=", ds::urlEncode(args.client_id),
                            "&dsid=", ds::urlEncode(m_propertySystem.getStringValue("/system/dSID")), "&mac=",
                            ds::urlEncode(m_propertySystem.getStringValue("/system/host/interfaces/eth0/mac")))});
    DS_INFO("Authorize response", response);
    DS_REQUIRE(response.code == 200, "Authorization call failed.", response);
    DS_NOTICE("Authorize session", response);
    return ds::fromJson<dsModel::AuthorizeRemoteConnectivityResult>(response.data);
  }

  [[nodiscard]] dsModel::PollAuthorizationRemoteConnectivityResult pollAuthorization(
      dsModel::PollAuthorizationRemoteConnectivityArgs &&args) override {
    if (auto mockedResult = this->mockedResult()) {
      DS_NOTICE("Mocked poll authorization", args);
      if (ds::randint(0, 10) == 0) {
        DS_FAIL_REQUIRE("Test failure");
      }
      if (ds::randint(0, 3) == 0) {
        setRemoteConnectivity(std::move(*mockedResult));
        return {};
      }
      return ds::toJsonFragment(RemoteConnectivityTokenError{.error = [&] {
        switch (ds::randint(0, 12)) {
          case 0: return "slow_down";
          case 1: return "access_denied";
          case 2: return "expired_token";
          default: return "authorization_pending";
        }
      }()});
    }

    DS_NOTICE("Poll authorization", args);
    auto response = m_httpClient.request(dss::HttpRequest{
        .url = ds::str(baseUrl(), "/v2/accessmanagement/device/token"),
        .type = dss::HttpMethod::POST,
        .headers = {{"Content-Type", "application/x-www-form-urlencoded"}},
        .postdata =
            ds::str("client_id=", ds::urlEncode(args.client_id), "&device_code=", ds::urlEncode(args.device_code),
                    "&grant_type=urn%3Aietf%3Aparams%3Aoauth%3Agrant-type%3Adevice_code")});
    DS_NOTICE("Poll authorization response", response);

    const auto data = ds::fromJson<RemoteConnectivityTokenError>(response.data);
    if (!data.error.empty()) {
      return ds::fromJson<dsModel::PollAuthorizationRemoteConnectivityResult>(response.data);
    }
    if (response.code == 200) {
      setRemoteConnectivity(ds::fromJson<RemoteConnectivityTokenResult>(response.data));
      return dsModel::PollAuthorizationRemoteConnectivityResult{};
    }
    DS_FAIL_REQUIRE("Poll authorization failed", response);
  }

  void unauthorize() override {
    DS_NOTICE("Unauthorize");
    unauthorizeApartment();
    unauthorizeRemoteConnectivity();
    unauthorizeDssLocal();
  }

  void unauthorizeDssLocal() {
    DS_NOTICE("UnauthorizeDssLocal");
    auto &security = m_apartment.getDss()->getSecurity();

    // deregister token
    security.revokeToken(m_propertySystem.getStringValue(dss::pp_websvc_rc_osptoken));

    // cleanup remote connectivity status
    if (auto node = m_propertySystem.getProperty(dss::pp_websvc_rc_osptoken)) {
      node->getParentNode()->removeChild(node);
    }
    if (auto node = m_propertySystem.getProperty(dss::pp_websvc_rc_username)) {
      node->getParentNode()->removeChild(node);
    }
    if (auto node = m_propertySystem.getProperty(dss::pp_websvc_dshub_token)) {
      node->getParentNode()->removeChild(node);
    }

    // update rc config
    m_propertySystem.setBoolValue(dss::pp_websvc_enabled, false);
    updateConfigXml(false);

    // shutdown relay link
    std::string deactivate = "/www/pages/cgi-bin/pagekiteconfig.sh off";
    std::system(deactivate.c_str());
  }

  void unauthorizeRemoteConnectivity() {
    DS_NOTICE("UnauthorizeMshub");

    // detach from cloud
    /*std::string parameters = ds::str("?token=", m_propertySystem.getStringValue(dss::pp_websvc_rc_osptoken),
           "&dssid=", ds::urlEncode(m_propertySystem.getStringValue("/system/dSID")));
    auto response = m_httpClient.request(dss::HttpRequest{
        .url = ds::str(mshubUrl(), "/internal/accessmanagement/v1_1/RemoteConnectivity/DeactivateRemoteConnectivity", parameters),
        .type = dss::HttpMethod::POST,
        .headers = {{"Content-Type", "application/x-www-form-urlencoded"}},
        .postdata = {}
    });
    DS_INFO("DeactivateRemoteConnectivity response", response);*/
  }

  void unauthorizeApartment() {
    DS_NOTICE("UnauthorizeApartment");

    if (m_connector.isConnected()) {
      auto request = dsApi::HttpRequest{.path = ds::str("/backend/v1/apartment/disconnectDss"),
                                        .httpArgs = {.method = dsApi::HttpMethod::DELETE}};
      DS_NOTICE("Unlink DSS from Apartment", request);
      m_connector.asyncInvoke(
          std::move(request),
          ds::makePromise<dsApi::HttpResult>(io(), [](std::optional<dsApi::HttpResult>&& result) {
            DS_INFO_ENTER(result);
            if (!result || !result->isSuccess()) {
              DS_ERROR("Failed to unlink DSS", result);
            }
          }));
    }
  }

  void setRemoteConnectivity(RemoteConnectivityTokenResult &&tokenResult) {
    try {
      DS_DEBUG_ENTER(tokenResult);
      DS_NOTICE(tokenResult);
      DS_REQUIRE(!tokenResult.access_token.empty(), "ospToken missing from backend response");
      DS_REQUIRE(!tokenResult.apartment_id.empty(), "apartmentId missing from backend response");
      DS_REQUIRE(!tokenResult.relay_name.empty() && !tokenResult.relay_password.empty(),
                 "Pagekite link or password missing from backend response");

      /// security.xml
      auto &security = m_apartment.getDss()->getSecurity();
      security.createApplicationToken("RemoteConnectivity", tokenResult.access_token);
      security.enableToken(tokenResult.access_token);

      /// system-addon-remote-connectivity.xml
      std::shared_ptr<dss::Event> pEvent = std::make_shared<dss::Event>("remote-connectivity");
      pEvent->setProperty("apartmentId", tokenResult.apartment_id);
      pEvent->setProperty("username", tokenResult.user_email);
      pEvent->setProperty("token", tokenResult.access_token);
      pEvent->setProperty("remoteConnectivity2", "Store");
      m_apartment.getDss()->getEventQueue().pushEvent(pEvent);

      // config.d/webservice.xml
      m_propertySystem.createProperty(dss::pp_websvc_enabled)->setBooleanValue(true);
      updateConfigXml(true);

      std::string configure =
          "/www/pages/cgi-bin/pagekiteconfig.sh configure " + tokenResult.relay_name + " " + tokenResult.relay_password;
      std::string activate = "/www/pages/cgi-bin/pagekiteconfig.sh on";
      std::system(configure.c_str());
      std::system(activate.c_str());

      // finally trigger the update to legacy backend
      // TODO: delay until pagekite link is active?
      /*std::string parameters =
          ds::str("?dssid=", m_propertySystem.getStringValue("/system/dSID"),
                  "&macAddress=", m_propertySystem.getStringValue("/system/host/interfaces/eth0/mac"),
                  "&apartmentId=", tokenResult.apartment_id,
                  "&token=", tokenResult.access_token, "&user=", tokenResult.user_email,
                  "&relayLink=", tokenResult.relay_name, "&dataSharingMode=2");
      auto response = m_httpClient.request(dss::HttpRequest{
        .url = ds::str(mshubUrl(), "/internal/accessmanagement/v3_0/RemoteConnectivity/ActivateRemoteConnectivity", parameters),
        .type = dss::HttpMethod::POST,
        .headers = {{"Content-Type", "application/x-www-form-urlencoded"}},
        .postdata = {}
      });
      DS_INFO("ActivateRemoteConnectivity response", response);

      if (response.code == 200) {
        // trigger an apartmentHasChanged event to MSHub
        dss::ModelChangedEvent e;
        std::shared_ptr<dss::Event> pEventPushUpdates = e.createApartmentChanged();
        m_apartment.getDss()->getEventQueue().pushEvent(pEventPushUpdates);

        // trigger an apartment structure upload to Smarthome Cache
        m_apartment.emitStructureChanged(); // NOTE: seems property tree changes only
      }*/
    } catch (const std::exception &e) {
      DS_ERROR("Failed to set remote connectivity.", e);
      throw;
    }
  }

  std::string pagekiteName() const {
    try {
      std::ifstream pagekiteFile("/etc/default/pagekite", std::ios::in);
      if (!pagekiteFile.is_open()) {
        throw std::runtime_error("Failed to open /etc/default/pagekite");
      }
      std::string line;
      while (std::getline(pagekiteFile, line)) {
        if (boost::starts_with(line, "NAME=")) {
          return line.substr(5);
        }
      }
    } catch (std::exception &e) {
      DS_WARNING("Could not read pagekite name from file.", e);
      return {};
    }
    return {};
  }
};

} // namespace

std::unique_ptr<dsModel::RemoteConnectivityApi> makeRemoteConnectivity(ds::asio::ResourceBase &parent,
                                                                       dss::Apartment &apartment,
                                                                       dss::SimpleHttpClient &httpClient,
                                                                       dsBackend::Connector &connector) {
  return std::make_unique<RemoteConnectivityImpl>(parent, apartment, httpClient, connector);
}
} // namespace dsApartment
