/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "array.h"

#include <ds/catch/catch.h>
#include <ds/json.h>

#include <boost/serialization/strong_typedef.hpp>

namespace dsApiTest {

BOOST_STRONG_TYPEDEF(std::string, ObjectId);

struct Object {
  ObjectId id{};
  std::string type = "TestObject";
  ds::DateTime lastChanged{};
  std::string attributes{};
};

} // namespace dsApiTest
DS_JSON_DECLARE_STRUCT((dsApiTest), Object, (id)(type)(lastChanged)(attributes));
DS_JSON_DEFINE_STRUCT((dsApiTest), Object);

namespace {

static const char* TAGS = "[dsApiArray][ds]";
using namespace std::literals;
using namespace trompeloeil;

using ObjectId = dsApiTest::ObjectId;
using Object = dsApiTest::Object;
TEST_CASE("dsApiNormalizeIdArray", TAGS) {
  SECTION("raw") {
    using Objects = std::vector<Object>;
    const auto object0 = Object{.id = ObjectId{"id0"}};
    const auto object1 = Object{.id = ObjectId{"id1"}};
    const auto object2 = Object{.id = ObjectId{"id2"}};
    const auto object3 = Object{.id = ObjectId{"id3"}};
    const auto object3a = Object{.id = ObjectId{"id3"}, .attributes = "a"};
    const auto object4 = Object{.id = ObjectId{"id4"}};
    SECTION("no change") {
      auto array = Objects{object1, object3};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{object1, object3});
    }
    SECTION("empty") {
      auto array = Objects{};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{});
    }
    SECTION("sorts, deduplicates, drops empty id") {
      auto array = Objects{object2, object3, object1, object3a, {}};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{object1, object2, object3});
    }
  }

  SECTION("boxed") {
    using Objects = dsApi::Array<Object>;
    const auto object0 = ds::Box<Object>{{.id = ObjectId{"id0"}}};
    const auto object1 = ds::Box<Object>{{.id = ObjectId{"id1"}}};
    const auto object2 = ds::Box<Object>{{.id = ObjectId{"id2"}}};
    const auto object3 = ds::Box<Object>{{.id = ObjectId{"id3"}}};
    const auto object3a = ds::Box<Object>{{.id = ObjectId{"id3"}, .attributes = "a"}};
    const auto object4 = ds::Box<Object>{{.id = ObjectId{"id4"}}};

    SECTION("no change") {
      auto array = Objects{object1, object3};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{object1, object3});
    }
    SECTION("empty") {
      auto array = Objects{};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{});
    }
    SECTION("sorts, deduplicates, drops empty id") {
      auto array = Objects{object2, object3, object1, object3a, ds::Box<Object>{{}}};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{object1, object2, object3});
    }
  }

  SECTION("ids") {
    using Objects = std::vector<ObjectId>;
    const auto object0 = ObjectId{"id0"};
    const auto object1 = ObjectId{"id1"};
    const auto object2 = ObjectId{"id2"};
    const auto object3 = ObjectId{"id3"};
    const auto object3a = ObjectId{"id3"};
    const auto object4 = ObjectId{"id4"};
    SECTION("no change") {
      auto array = Objects{object1, object3};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{object1, object3});
    }
    SECTION("empty") {
      auto array = Objects{};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{});
    }
    SECTION("sorts, deduplicates, drops empty id") {
      auto array = Objects{object2, object3, object1, object3a, {}};
      dsApi::normalizeIdArray(array);
      CHECK(array == Objects{object1, object2, object3});
    }
  }
}

TEST_CASE("dsApiArray", TAGS) {
  using Objects = dsApi::Array<Object>;
  const auto object0 = ds::Box<Object>{{.id = ObjectId{"id0"}}};
  const auto object1 = ds::Box<Object>{{.id = ObjectId{"id1"}}};
  const auto object2 = ds::Box<Object>{{.id = ObjectId{"id2"}}};
  const auto object3 = ds::Box<Object>{{.id = ObjectId{"id3"}}};
  const auto object3a = ds::Box<Object>{{.id = ObjectId{"id3"}, .attributes = "a"}};
  const auto object4 = ds::Box<Object>{{.id = ObjectId{"id4"}}};

  SECTION("tryAt") {
    const auto array = dsApi::Array<Object>{{object1, object3}};
    CHECK(array.tryAt(object1->id).value() == object1);
    CHECK(array.tryAt(object2->id) == std::nullopt);
    CHECK(array.tryAt(object3->id).value() == object3);
    CHECK(array.tryAt(object4->id) == std::nullopt);
  }

  SECTION("at") {
    const auto array = dsApi::Array<Object>{{object1, object3}};
    CHECK(array.at(object1->id) == object1);
    CHECK_THROWS_FIND(array.at(object2->id), "Not found");
  }

  SECTION("replace") {
    auto array = dsApi::Array<Object>{object1, object3};
    SECTION("change") {
      array.replace(object3a);
      CHECK(array == dsApi::Array<Object>{object1, object3a});
    }
    SECTION("insert before first") {
      array.replace(object0);
      CHECK(array == dsApi::Array<Object>{object0, object1, object3});
    }
    SECTION("insert in the middle") {
      array.replace(object2);
      CHECK(array == dsApi::Array<Object>{object1, object2, object3});
    }
    SECTION("insert after last") {
      array.replace(object4);
      CHECK(array == dsApi::Array<Object>{object1, object3, object4});
    }
  }

  SECTION("normalize") {
    auto array = dsApi::Array<Object>{object2, object3, object1, object3a, ds::Box<Object>{{}}};
    array.normalize();
    CHECK(array == dsApi::Array<Object>{object1, object2, object3});
  }

  SECTION("boxed") {
    auto array = dsApi::Array<Object>{object3, object1};
    CHECK(std::move(array).boxed() == Objects{object1, object3});
  }

  SECTION("makeArrayBox") {
    CHECK(dsApi::makeArrayBox(std::initializer_list<ds::Box<Object>>{object3, object1}) == Objects{object1, object3});
  }

  SECTION("ArrayBox constructor works") {
    CHECK(dsApi::ArrayBox(dsApi::Array<Object>{object1}) == dsApi::Array<Object>{object1});
  }

  SECTION("ArrayBox constructor requires normalized") {
    REQUIRE_THROWS_FIND(dsApi::ArrayBox(Objects{object3, object1}), "Array is not normalized.");
    REQUIRE_THROWS_FIND(dsApi::ArrayBox(Objects{object3, object3}), "Array is not normalized.");
  }
}

} // namespace
