/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "event.h"

#include <boost/multi_index/hashed_index.hpp>
#include <boost/multi_index/identity.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index_container.hpp>

namespace dsApi {

static ds::LogChannel dsLogChannel("dsApiEvent");

namespace {

struct EventHubImpl final : ds::asio::Resource<>, EventHub {
  ds::Signal<ds::Box<dsApi::Event>> m_event;

  EventHubImpl(ds::asio::ResourceBase& parent) : ds::asio::Resource<>{parent, "eventHub"} {}

  void emitEvent(const ds::Box<dsApi::Event>& x) override {
    DS_DEBUG_ENTER(x);
    m_event.emit(x);
  }

  void subscribe(const Subscription&, EventSink&) override{};

  virtual ds::Slots<ds::Box<dsApi::Event>>& event() override { return m_event; }

  const ds::asio::ResourceInside& inside() override {
    static constexpr ds::asio::ResourceInside instance{
        .description = "Event hub distributing subscribed events to event sinks.",
    };
    return instance;
  }
};

struct TypeEqual {
  template <class A, class B>
  std::size_t operator()(const A& a, const B& b) const {
    return a.type == b.type;
  }
};
struct TypeHash {
  template <class X>
  std::size_t operator()(const X& x) const {
    return boost::hash_value(static_cast<std::string>(x.type));
  }
};

using namespace boost::multi_index;
struct EventSinkImpl final : ds::asio::Resource<>, EventSink {
  EventHub& m_hub;
  EventSinkHooks m_hooks;
  uint64_t m_eventCount = 0;
  ds::Slot<ds::Box<dsApi::Event>> DS_UNIQUE_NAME(slot){*this, hubEvent(m_hub),
                                                       [this](auto&& x) { onEvent(DS_FWD(x)); }};

  EventSinkImpl(ds::asio::ResourceBase& parent, EventHub& hub, EventSinkHooks&& hooks)
      : ds::asio::Resource<>{parent, "eventSink"}, m_hub{hub}, m_hooks{std::move(hooks)} {}

  struct ByType {};

  using Container =
      boost::multi_index_container<Subscription,
                                   indexed_by<hashed_unique<tag<ByType>, identity<Subscription>, TypeHash, TypeEqual>>>;
  Container m_container;

  void subscribe(Subscription&& subscription) override {
    DS_INFO_ENTER(subscription);
    assertSynchronized();
    m_hub.subscribe(subscription, *this);
    auto& byType = m_container.get<ByType>();
    auto it = byType.find(subscription);
    if (it != byType.end()) {
      if (subscription.expiresAt > it->expiresAt) {
        byType.replace(it, std::move(subscription));
      }
    } else {
      m_container.insert(std::move(subscription));
    }
  }

  void onEvent(ds::Box<dsApi::Event>&& event) {
    DS_DEBUG_ENTER(*this, event);
    assertSynchronized();
    auto now = ds::Clock::now();
    auto& byType = m_container.get<ByType>();
    auto it = byType.find(*event);
    if (it != byType.end() && now < it->expiresAt) {
      DS_INFO(*this, "Event.", event->type);
      m_eventCount++;
      m_hooks.event(std::move(event));
    } else {
      DS_DEBUG(*this, "Dropping not subscribed event.", event->type);
      m_hooks.discardedEvent(std::move(event));
    }
  }

  std::vector<InsideSubscription> insideSubscriptions() const {
    std::vector<InsideSubscription> result;
    auto now = ds::Clock::now();
    auto systemNow = system().now();
    for (const auto& subscription : m_container.get<ByType>()) {
      if (subscription.expiresAt > now) {
        result.push_back(
            InsideSubscription{.type = subscription.type,
                               .expiresAt = std::chrono::time_point_cast<std::chrono::system_clock::duration>(
                                   systemNow + (subscription.expiresAt - now))});
      }
    }
    return result;
  }
  const ds::asio::ResourceInside& inside() override {
    using This = EventSinkImpl;
    static constexpr ds::asio::ResourceInside instance{
        .description = "Event sink subscribed to events from event hub.",
        .properties = {
            {.id = "eventCount",
             .description = "Number of passed events",
             .value =
                 [](const ResourceBase& x) { return ds::toJsonFragment(static_cast<const This&>(x).m_eventCount); }},
            {.id = "subscriptions",
             .description = "Active subscriptions",
             .value =
                 [](const ResourceBase& x) {
                   return ds::toJsonFragment(static_cast<const This&>(x).insideSubscriptions());
                 }},
        }};
    return instance;
  }
};

} // namespace

std::unique_ptr<EventHub> makeEventHub(ds::asio::ResourceBase& parent) {
  return std::make_unique<EventHubImpl>(parent);
}

std::unique_ptr<EventSink> makeEventSink(ds::asio::ResourceBase& parent, EventHub& hub, EventSinkHooks&& hooks) {
  return std::make_unique<EventSinkImpl>(parent, hub, std::move(hooks));
}

} // namespace dsApi

DS_JSON_DEFINE_STRUCT((dsApi), Event);
DS_REFLECT_DEFINE_STRUCT((dsApi), Subscription);
DS_JSON_DEFINE_STRUCT((dsApi), InsideSubscription);
