#include "http-server.h"

#include <ds-api/rest.h>
#include <ds/asio/move-only.h>
#include <ds/asio/promise.h>
#include <ds/log.h>

#include <boost/asio.hpp>
#include <boost/beast/core.hpp>
#include <boost/beast/http.hpp>
#include <boost/beast/version.hpp>
#include <chrono>
#include <cstdlib>
#include <memory>
#include <string>

static ds::LogChannel dsLogChannel("dsApiHttpServer");

using namespace std::literals;
namespace ip = boost::asio::ip;      // from <boost/asio.hpp>
using tcp = boost::asio::ip::tcp;    // from <boost/asio.hpp>
namespace http = boost::beast::http; // from <boost/beast/http.hpp>
using error_code = boost::beast::error_code;

namespace dsApi {

namespace {
class Connection : public std::enable_shared_from_this<Connection> {
private:
  tcp::socket m_socket;
  boost::asio::io_service& m_ioService;
  dsApi::AsyncCallee& m_router;
  boost::beast::flat_buffer m_buffer{4096};
  http::request<http::string_body> m_request;
  http::response<http::string_body> m_response;
  bool m_replied{false};

public:
  Connection(tcp::socket&& socket, boost::asio::io_service& ioService, dsApi::AsyncCallee& router) : m_socket(std::move(socket)), m_ioService(ioService), m_router(router) {
    DS_INFO_ENTER("");
  }

  void asyncRun() {
    DS_INFO_ENTER("");
    m_ioService.post([this, self = shared_from_this()] { asyncRead(); });
  }

private:
  // Asynchronously receive a complete request message.
  void asyncRead() {
    DS_DEBUG_ENTER("");

    http::async_read(m_socket, m_buffer, m_request, [this, self = shared_from_this()](error_code ec, std::size_t) {
      auto& request = m_request;
      DS_INFO_ENTER(request.target());
      DS_DEBUG_ENTER(request.target(), request.method_string(), request.body());
      if (ec) {
        DS_ERROR(ec);
        return;
      }
      const auto version = request.version();
      try {
        auto path = request.target();
        auto query = decltype(path){};
        {
          const auto pos = path.find('?');
          if (pos != std::string_view::npos) {
            query = path.substr(pos + 1);
            path = path.substr(0, pos);
          }
        }
        if (request[http::field::content_type].empty()) { // default to json content type
          request.set(http::field::content_type, "application/json");
        }
        request.erase(http::field::content_length);
        auto headers = [&] {
          auto result = dsApi::HttpHeaders{};
          for (auto&& header : request) {
            result[std::string{header.name_string()}].push_back(std::string{header.value()});
          }
          return result;
        };
        auto content = [&] {
          if (request.body().size() == 0) {
            return ds::JsonFragment{};
          }
          return ds::fromJson<ds::JsonFragment>(request.body());
        };

        m_router.asyncInvoke(
            dsApi::HttpRequest{.path = std::string{path},
                               .httpArgs = {.method = ds::requireEnum<dsApi::HttpMethod>(request.method_string()),
                                            .content = content(),
                                            .query = std::string{query},
                                            .headers = headers()}},
            ds::makePromise<dsApi::HttpResult>(m_ioService, [this, self, version](std::optional<dsApi::HttpResult>&& result) {
              m_ioService.post(
                  ds::asio::MoveOnly{[this, self = std::move(self), result = std::move(result), version]() mutable {
                    asyncReply(result ? std::move(*result) : dsApi::cancelledResult(), version);
                  }});
            }));
      } catch (const std::exception& e) {
        DS_ERROR(e);
        asyncReply(dsApi::errorResult(e), version);
      }
    });
  }

  void asyncReply(dsApi::HttpResult&& result, unsigned version) {
    DS_INFO_ENTER(result);
    if (m_replied) {
      return;
    }
    m_replied = true;

    m_response.version(version);
    m_response.keep_alive(false);
    m_response.result(result.status);
    for (const auto& header : result.headers) {
      for (const auto& value : header.second) {
        m_response.insert(header.first, value);
      }
    }
    if (m_response[http::field::content_type].empty()) {
      m_response.set(http::field::content_type, "application/json");
    }
    m_response.body() = ds::toJson(result.content);
    m_response.set(http::field::content_length, m_response.body().size());
    http::async_write(m_socket, m_response, [this, self = shared_from_this()](error_code ec, std::size_t) {
      m_socket.shutdown(tcp::socket::shutdown_send, ec);
    });
  }
};

class HttpServerImpl : public ds::asio::Resource<HttpServerConfig>, public HttpServer {
  boost::asio::io_service& m_ioService;
  dsApi::AsyncCallee& m_api;
  std::unique_ptr<dsNet::TcpAcceptor> m_acceptor =
      dsNet::makeTcpAcceptor(*this, dsNet::TcpAcceptorConfig{config().acceptor});

public:
  HttpServerImpl(ds::asio::ResourceBase& parent, HttpServerConfig&& config, boost::asio::io_service& ioService, dsApi::AsyncCallee& api)
      : ds::asio::Resource<HttpServerConfig>{parent, "httpServer", std::move(config)}, m_ioService(ioService), m_api{api} {
    DS_DEBUG_ENTER("");
    m_acceptor->asyncRun(dsNet::TcpAcceptorHooks{.accept = [this](tcp::socket&& socket) {
      // TcpAcceptorHooks does not call hooks after desctruction
      auto connection = std::make_shared<Connection>(std::move(socket), m_ioService, m_api);
      connection->asyncRun();
    }});
  }

  dsNet::TcpAcceptor& acceptor() override { return *m_acceptor; }
};
} // namespace

std::unique_ptr<HttpServer> makeHttpServer(ds::asio::ResourceBase& parent, HttpServerConfig&& config, boost::asio::io_service& ioService,
                                           dsApi::AsyncCallee& api) {
  return std::make_unique<HttpServerImpl>(parent, std::move(config), ioService, api);
}

} // namespace dsApi

DS_JSON_DEFINE_STRUCT((dsApi), HttpServerConfig);
