/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "service.h"

#include <ds/http.h>
#include <ds/log.h>

#include <boost/algorithm/string.hpp>

#include "route-segments.h"

static ds::LogChannel dsLogChannel("dsApiService");

namespace dsApi {

HttpResult invoke(const std::vector<Route>& routes, HttpRequest&& context) {
  DS_INFO_ENTER(context);
  try {
    struct {
      const Route* route{};
      MatchRouteResult result{};
    } bestMatch;
    for (auto& route : routes) {
      DS_DEBUG(route.pattern, context.path);
      if (auto matchResult = dsApi::matchRoute(route.pattern, context.path)) {
        if (!bestMatch.route || matchResult.fitness > bestMatch.result.fitness) {
          bestMatch = {&route, std::move(matchResult)};
        }
      }
    }
    if (bestMatch.route) {
      context.path = {};
      for (auto& arg : bestMatch.result.args) {
        context.routeArgs[arg.first] = std::move(arg.second);
      }
      auto result = bestMatch.route->invoke(std::move(context));
      DS_INFO_LEAVE(context.path, result);
      return result;
    }
    return resourceNotFoundResult();
  } catch (const std::exception& e) {
    return errorResult(e);
  }
}

HttpResult Service::invoke(HttpRequest&& context) const {
  return dsApi::invoke(routes, std::move(context));
}

Route makeServiceRoute(std::string&& pattern, dsApi::Service& service) {
  return Route{.pattern = ds::str(pattern, "/{*path}"), .invoke = [&service](auto&& context) {
                 auto& routeArgs = context.routeArgs;
                 auto it = routeArgs.find("path");
                 DS_REQUIRE(it != routeArgs.end(), "Required `path` route parameter not present.");
                 context.path = std::move(it->second);
                 routeArgs.erase(it);
                 return service.invoke(std::move(context));
               }};
}

std::vector<HttpMethod> DocumentMethods::allowedMethods() const {
  auto result = std::vector<HttpMethod>{};
  for (const auto& method : this->methods) {
    result.push_back(method.first);
  }
  if (!this->patchRoutes.empty()) {
    result.push_back(HttpMethod::PATCH);
  }
  return result;
}

Route makeDocumentRoute(std::string&& pattern, DocumentMethods&& document) {
  return Route{.pattern = std::move(pattern), .invoke = [document = std::move(document)](auto&& context) {
                 const auto method = context.httpArgs.method;
                 {
                   auto it = document.methods.find(method);
                   if (it != document.methods.end()) {
                     return it->second(std::move(context));
                   }
                 }
                 // default PATCH method implementation when patchRoutes are provided
                 if (method == HttpMethod::PATCH && !document.patchRoutes.empty()) {
                   auto patchOperations = ds::fromJson<PatchOperations>(context.httpArgs.content);
                   context.httpArgs.content = {};
                   bool found = false;
                   for (auto& patchOperation : patchOperations) {
                     auto patchPath = patchOperation.path;
                     DS_CONTEXT(patchPath);
                     for (auto& patchRoute : document.patchRoutes) {
                       if (auto matchResult = dsApi::matchRoute(patchRoute.pattern, patchOperation.path)) {
                         const auto op = patchOperation.op;
                         auto patchRequest =
                             PatchRequest{.operation = std::move(patchOperation), .routeArgs = context.routeArgs};
                         for (auto& arg : matchResult.args) {
                           patchRequest.routeArgs[arg.first] = std::move(arg.second);
                         }
                         DS_DEBUG(patchRequest);
                         auto it = patchRoute.operations.find(op);
                         if (it != patchRoute.operations.end()) {
                           it->second(std::move(patchRequest));
                           found = true;
                           break;
                         }
                         DS_FAIL_REQUIRE("Path operation not supported.", op);
                       }
                     }
                     if (!found) {
                       DS_FAIL_REQUIRE("Patch path not recognized.", patchOperation.path);
                     }
                   }
                   return dsApi::noContentResult();
                 }
                 return methodNotAllowedResult(method, document.allowedMethods());
               }};
}

std::string routeIdImpl(const RouteHttpArgs& routeArgs) {
  const auto it = routeArgs.find("id");
  if (it != routeArgs.end()) {
    return it->second;
  }
  DS_FAIL_REQUIRE("Route argument `id` not found.", routeArgs);
}

GetQuery makeGetQuery(const HttpRequest& request) {
  const auto& query = request.httpArgs.query;
  const auto includeArg = ds::tryHttpQueryArg(query, "include").value_or(std::string{});
  auto include = std::vector<std::string>{};
  if (!includeArg.empty()) {
    boost::split(include, includeArg, boost::is_any_of(","));
  }
  return GetQuery{.include = std::move(include), .includeAll = ds::tryHttpQueryArg(query, "includeAll") == "true"};
}

std::unique_ptr<Service> makeApartmentTestService() {
  auto result = std::make_unique<Service>();
  result->routes.push_back({.pattern = "/echo", .invoke = [](auto&& context) {
                              DS_NOTICE("Echo.", context);
                              return dsApi::okDocumentResult(std::move(context.httpArgs));
                            }});
  result->routes.push_back({.pattern = "/echoRaw", .invoke = [](auto&& context) {
                              DS_NOTICE("Echo raw.", context);
                              return HttpResult{.status = 200, .content = std::move(context.httpArgs.content)};
                            }});
  result->routes.push_back(makeMethodRoute("/ping", []() { DS_NOTICE("Ping."); }));
  result->routes.push_back(makeMethodRoute<>("/fail", []() {
    DS_NOTICE("Fail.");
    DS_FAIL_REQUIRE("Expected fail.");
  }));
  result->routes.push_back({.pattern = "/notImplemented", .invoke = [](auto&&) {
                              DS_NOTICE("Not implemented.");
                              return dsApi::notImplementedResult();
                            }});
  return result;
}

} // namespace dsApi
