/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <avahi-client/publish.h>
#include <ds/asio/timer.h>
#include <ds/common.h>
#include <ds/reflect.h>

#include <boost/noncopyable.hpp>
#include <memory>
#include <string>
#include <vector>

#include "client.h"

namespace dsAvahi {

struct PublishedService {
  AvahiIfIndex interface = AVAHI_IF_UNSPEC;
  AvahiProtocol protocol = AVAHI_PROTO_UNSPEC;
  AvahiPublishFlags flags{};
  std::string name{};
  std::string type{};
  std::string domain{};
  std::string host{};
  uint16_t port{};
  std::vector<std::string> txts{};
};

enum class CollisionMode {
  ALTERNATIVE_NAME, ///< Retry with alternative name on collision
                    // TODO(someday): RETRY,  ///< Retry with the same name on collision
};

class EntryGroup : private Client::Hook {
public:
  EntryGroup(Client& client, CollisionMode collisioMode);
  ~EntryGroup();

  const std::vector<PublishedService>& services() const { return m_services; }
  void setServices(std::vector<PublishedService>&& services);

  /// Convenience function calling setServices() with single service
  void setService(PublishedService&& service);

  /// Is EntryGroup established in avahi deamon?
  bool isEstablished() const { return m_isEstablished; }
  void setIsEstablishedChanged(std::function<void(bool)> x) { m_isEstablishedChanged = std::move(x); }

  /// Avahi deamon sometimes bisbehaves.
  /// So we guard and reconnect to deamon when the establishing times out.
  /// Tests need to tune this timeout, applications shall not need it.
  ds::asio::Timer::duration establishTimeout() const { return m_establishTimeout; }
  void setEstablishTimeout(ds::asio::Timer::duration x) { m_establishTimeout = x; }

private:
  Client& m_client;
  struct Deleter {
    void operator()(::AvahiEntryGroup* x) { avahi_entry_group_free(x); }
  };
  std::unique_ptr< ::AvahiEntryGroup, Deleter> m_ptr;
  ds::asio::Timer m_setServicesModerator;
  ds::asio::Timer::duration m_establishTimeout;
  ds::asio::Timer m_establishTimer;
  CollisionMode m_collisionMode;
  std::vector<PublishedService> m_services;
  bool m_isRunning;
  bool m_isEstablished;
  std::function<void(bool)> m_isEstablishedChanged;

  ::AvahiEntryGroup* ptr() const { return m_ptr.get(); }

  void onRunning() override;
  void freed() override;
  void publish();
  void handleCollision();
  static void sCallback(AvahiEntryGroup* g, AvahiEntryGroupState state, void* data);
  void callback(AvahiEntryGroupState state);

  void setEstablished(bool);
  const std::string& typePrefix() const { return m_client.typePrefix(); }

  friend class Client;
};

} // namespace dsAvahi

DS_REFLECT_DECLARE_STRUCT((dsAvahi), PublishedService,
                          (interface)(protocol)(flags)(name)(type)(domain)(host)(port)(txts))
