/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "connector.h"

#include <json-c/json.h>
#include <ds-api/event.h>
#include <ds-api/rest.h>
#include <ds-model/model.h>
#include <ds-net/buffered-message-writer.h>
#include <ds/asio/move-only.h>
#include <ds/asio/promise.h>
#include <ds/asio/reconnect-timer.h>
#include <ds/asio/resource.h>
#include <ds/asio/timer.h>
#include <ds/clock.h>
#include <ds/elapsed.h>
#include <ds/json.h>
#include <ds/latency-statistics.h>
#include <ds/random.h>
#include <json-c/json_object.h>


#include <boost/algorithm/string/predicate.hpp>
#include <boost/atomic.hpp>
#include <boost/range/numeric.hpp>
#include <sstream>
#include <string>
#include <map>

#include "protocol.h"

static ds::LogChannel dsLogChannel("dsBackendConnector");
using namespace std::literals;

namespace dsBackend {

struct ConnectorConfig {
  ds::Clock::duration keepAlivePeriod = 60s;
  ds::Clock::duration latencyStatisticsPeriod = 3600s;
};

namespace {

struct ConnectorImpl final : public Connector, public ds::asio::Resource<ConnectorConfig>, ConnectorSession {
  dsNet::MessageStreamClient &m_streamClient;
  ConnectorHook &m_hook;
  dsApi::EventHub &m_eventHub;
  std::unique_ptr<dsApi::EventSink> m_eventSink =
      dsApi::makeEventSink(*this, m_eventHub, {.event = [this](auto x) { onEvent(std::move(x)); }});
  boost::atomic_bool m_isActive{};
  boost::atomic_bool m_isConnected{};
  ds::DateTime m_connectedAt;
  std::string m_lastDisconnectReason;
  ds::DateTime m_lastDisconnectAt;
  ds::asio::ReconnectTimer m_reconnectTimer{io()};
  dsNet::ClosePtr<dsNet::MessageStream> m_stream{};
  dsNet::ClosePtr<dsNet::BufferedMessageWriter> m_bufferedWriter;
  ds::asio::Timer m_keepAliveTimer{io()};
  int m_pingId{};
  ds::Elapsed m_pingElapsed{};
  std::string m_pongId{};
  ds::Elapsed m_uptime{};
  ds::LatencyStatistics m_latencyMs{};
  ds::Elapsed m_statisticsNotice{};
  struct CalleeInvocation {
    std::unique_ptr<ds::Promise<dsApi::HttpResult>> promise;
    ds::Clock::time_point startedAt;
    std::string path;
  };
  std::map<RequestId, CalleeInvocation> m_calleeInvocations;
  ds::asio::Timer m_calleeInvocationsDeadline{io()};
  // shared state for one session.
  // TODO(someday) Refactor all classes and methods related to session into Session class
  std::shared_ptr<void> m_session = std::make_shared<bool>(false);

  ConnectorImpl(ds::asio::ResourceBase &parent, std::string &&id, dsNet::MessageStreamClient &streamClient,
                ConnectorHook &hook, dsApi::EventHub &eventHub)
      : ds::asio::Resource<ConnectorConfig>(parent, std::move(id)),
        m_streamClient{streamClient},
        m_hook{hook},
        m_eventHub{eventHub} {
    asyncCalleeInvocationsDeadlineLoop();
  }

  void onEvent(ds::Box<dsApi::Event> &&event) {
    DS_INFO_ENTER(event);
    json_object *copy = nullptr;
    json_object_deep_copy(event->data.get(), &copy, nullptr);
    enqueue(Message{.authenticate = {},
                    .authenticated = {},
                    .ping = {},
                    .pong = {},
                    .request = {},
                    .response = {},
                    .event = Event{.type = event->type, .path = event->path, .data = ds::makeJsonFragment(copy)}});
  }

  bool isActive() const {
    assertSynchronized();
    return m_isActive;
  }

  bool isConnected() const override { return m_isConnected; }

  void setActive(bool isActive) override {
    io().dispatch([this, isActive] {
      if (m_isActive == isActive) {
        return;
      }
      DS_NOTICE("Set active.", isActive);
      m_isActive = isActive;
      m_reconnectTimer.resetInterval();
      asyncReconnect(DS_LOG_STR("(De)activated."));
    });
  }

  const dsNet::MessageStreamClient &streamClient() const { return m_streamClient; }

  ds::Clock::duration uptime() const { return m_isConnected ? m_uptime.elapsed() : ds::Clock::duration{}; }

  void disconnect() { asyncReconnect(DS_LOG_STR("Disconnected.")); }

  void reconnect() {
    m_reconnectTimer.resetInterval();
    asyncReconnect(DS_LOG_STR("Reconnected."));
  }

  const ds::asio::ResourceInside &inside() override {
    using This = ConnectorImpl;
    static constexpr ds::asio::ResourceInside instance{
        .description = "Connector connecting to backend as websocket client.",
        .properties =
            {
                {.id = "isActive",
                 .description = "Is connection being maintained? Or is it intentionally disconnected (e.g. no remote "
                                "connectivity).",
                 .value =
                     [](const ResourceBase &x) { return ds::toJsonFragment(static_cast<const This &>(x).isActive()); },
                 .test = true},
                {.id = "isConnected",
                 .description = "Is connection connected?",
                 .value =
                     [](const ResourceBase &x) {
                       return ds::toJsonFragment(static_cast<const This &>(x).isConnected());
                     },
                 .test = true},
                {.id = "connectedAt",
                 .description = "When was the connection established?",
                 .value =
                     [](const ResourceBase &x) {
                       return ds::toJsonFragment(static_cast<const This &>(x).m_connectedAt);
                     },
                 .test = true},
                {.id = "lastDisconnectReason",
                 .description = "Reason of last disconnect.",
                 .value =
                     [](const ResourceBase &x) {
                       return ds::toJsonFragment(static_cast<const This &>(x).m_lastDisconnectReason);
                     },
                 .test = true},
                {.id = "lastDisconnectAt",
                 .description = "Datetime of last disconnect.",
                 .value =
                     [](const ResourceBase &x) {
                       return ds::toJsonFragment(static_cast<const This &>(x).m_lastDisconnectAt);
                     },
                 .test = true},
                {.id = "latencyMs",
                 .description = "Ping latency statistics since last statistics period tick [ms].",
                 .value =
                     [](const ResourceBase &x) {
                       return ds::toJsonFragment(ds::str(static_cast<const This &>(x).m_latencyMs));
                     }},
                {.id = "url",
                 .description = "Url for next connection. Change triggers reconnect.",
                 .value =
                     [](const ResourceBase &x) {
                       return ds::toJsonFragment(static_cast<const This &>(x).m_streamClient.url());
                     },
                 .test = true,
                 .defaultValue =
                     [](const ResourceBase &x) {
                       return ds::toJsonFragment(static_cast<const This &>(x).m_streamClient.defaultUrl());
                     },
                 .setValue =
                     [](ResourceBase &xx, ds::JsonFragment &&data) {
                       auto &x = static_cast<This &>(xx);
                       auto url = ds::fromJson<std::string>(data);
                       if (x.m_streamClient.url() == url) {
                         return;
                       }
                       x.m_streamClient.setUrl(std::move(url));
                       x.reconnect();
                     }},
            },
        .methods = {
            {.id = "disconnect",
             .description = "Drop existing connection and reconnect after backoff.",
             .invoke = [](ResourceBase &xx) { static_cast<This &>(xx).reconnect(); },
             .test = true},
            {.id = "reconnect",
             .description = "Reconnect now. Resets backoff, drop existing connection and reconnect.",
             .invoke = [](ResourceBase &xx) { static_cast<This &>(xx).reconnect(); },
             .test = true},
        }};
    return instance;
  }

  // 2xkeepAlive to ensure that disconnect is detected before requests start to time out
  ds::Clock::duration calleeInvocationDeadline() { return config().keepAlivePeriod * 2; }

  void asyncReconnect(std::string &&disconnectReason) {
    assertSynchronized();
    DS_INFO_ENTER("");
    try {
      DS_WARNING("Reconnect", disconnectReason, *this);
      if (m_isConnected) {
        DS_NOTICE("Disconnected. latency[ms]:", ds::hideArg(m_latencyMs), "uptime:", ds::hideArg(m_uptime));
        m_isConnected = false;
        m_connectedAt = {};
        m_isConnectedChanged.emit(false);
      }
      m_lastDisconnectReason = std::move(disconnectReason);
      m_lastDisconnectAt = system().now();
      m_keepAliveTimer.cancel();
      m_stream.reset();
      m_bufferedWriter.reset();
      m_pingId = {};
      m_pongId = ds::str(m_pingId);
      m_latencyMs.drop();
      m_calleeInvocations.clear();
      m_session = std::make_shared<bool>(false);

      if (m_isActive) {
        m_reconnectTimer.asyncDelayedReconnect([this, session = std::weak_ptr(m_session)] {
          if (!session.lock()) {
            return;
          }
          asyncConnect();
          asyncKeepAlive(); // keep alive timer guards also initial connection setup
        });
      }
    } catch (const std::exception &e) {
      DS_ASSERT("Asynchronous reconnect loop must not fail.");
    }
  }

  void asyncReconnectOnClose(const Close &closeMessage) {
    m_reconnectTimer.raiseIntervalUpTo(closeMessage.blockedFor);
    asyncReconnect(DS_LOG_STR("Closed.", closeMessage));
    if (closeMessage.permanentlyDisconnect) {
      m_hook.permanentlyDisconnect();
    }
  }

  void asyncConnect() {
    DS_INFO_ENTER();
    try {
      m_stream = m_streamClient.asyncConnect();
      m_stream->asyncRun({.opened =
                              [this, session = std::weak_ptr(m_session)]() {
                                if (!session.lock()) {
                                  return;
                                }
                                DS_INFO_ENTER("opened");
                                asyncWriteAuthenticate();
                              },
                          .closed =
                              [this, session = std::weak_ptr(m_session)](boost::system::error_code ec) {
                                if (!session.lock()) {
                                  return;
                                }
                                DS_INFO_ENTER("closed");
                                asyncReconnect(DS_LOG_STR("Closed.", ec));
                              }});
      m_bufferedWriter = dsNet::makeBufferedMessageWriter(
          *this, *m_stream, dsNet::BufferedMessageWriterConfig{.countCapacity = 256, .sizeCapacity = 4 * 1024 * 1024},
          {.failed = [this, session = std::weak_ptr(m_session)](const std::exception &e) {
            if (!session.lock()) {
              return;
            }
            asyncReconnect(DS_LOG_STR("Buffered write failed.", e));
          }});
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Async resolve failed.", e));
    }
  }

  void asyncWriteAuthenticate() {
    DS_INFO_ENTER("");
    try {
      auto message = ds::toJson(Message{.authenticate = m_hook.authenticate()});
      DS_INFO("Write authenticate", ds::hideArg(message));
      m_stream->asyncWrite(std::move(message),
                           [this, session = std::weak_ptr(m_session)](boost::system::error_code ec) {
                             if (!session.lock()) {
                               return;
                             }
                             if (ec) {
                               asyncReconnect(DS_LOG_STR("Authenticate write failed.", ec));
                               return;
                             }
                             asyncReadAuthenticated();
                           });
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Async write authenticate failed.", e));
    }
  }

  void asyncReadAuthenticated() {
    DS_INFO_ENTER("");
    try {
      m_stream->asyncRead(
          [this, session = std::weak_ptr(m_session)](boost::system::error_code ec, std::string &&message) {
            if (!session.lock()) {
              return;
            }
            if (ec) {
              asyncReconnect(DS_LOG_STR("Authenticated read failed.", ec));
              return;
            }
            asyncReadAuthenticatedConsume(std::move(message));
          });
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Async authenticated read failed.", e));
    }
  }

  void asyncReadAuthenticatedConsume(std::string &&rawMessage) {
    DS_INFO("Read authenticated", ds::hideArg(rawMessage));
    try {
      DS_CONTEXT("readAuthenticatedConsume");
      const auto message = ds::fromJson<Message>(rawMessage);
      if (message.close) {
        asyncReconnectOnClose(*message.close);
        return;
      }
      DS_REQUIRE(message.authenticated, "Authenticated message not received.", rawMessage);
      asyncConnected();
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Async authenticated read failed.", e));
    }
  }

  void asyncConnected() {
    DS_NOTICE("Connected.", *this);
    try {
      DS_ASSERT(!m_isConnected, "Invariant failed");
      m_isConnected = true;
      m_connectedAt = system().now();
      m_uptime.restart();
      m_statisticsNotice.restart();
      m_reconnectTimer.connected();
      m_isConnectedChanged.emit(true);
      asyncRead();
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Async connected failed.", e));
    }
  }

  void asyncKeepAlive() {
    DS_DEBUG_ENTER("");
    try {
      m_keepAliveTimer.randomlyExpiresFromNowPercentDown(config().keepAlivePeriod, 10,
                                                         [this, session = std::weak_ptr(m_session)] {
                                                           if (!session.lock()) {
                                                             return;
                                                           }
                                                           asyncAliveCheck();
                                                         });
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Async keep alive failed.", e));
    }
  }

  void asyncAliveCheck() {
    assertSynchronized();
    DS_INFO_ENTER("");
    try {
      DS_REQUIRE(m_isConnected, "Not connected.");
      DS_REQUIRE(ds::str(m_pingId) == m_pongId, "Pong timeout.", m_pingId, m_pongId);
      enqueue({.authenticate = {}, .authenticated = {}, .ping = Ping{.id = ds::str(++m_pingId)}});
      m_pingElapsed.restart();
      asyncKeepAlive();
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Async alive check failed.", e));
    }
  }

  void enqueue(Message &&message) {
    DS_DEBUG_ENTER(m_isConnected, message);
    assertSynchronized();
    try {
      std::string json = ds::toJson(message);
      if (message.request) {
        Request req = *(message.request);

        if (req.path == "/v1/apartment/weather") {
          WsRequest request;

          request.id = req.id;
          request.path = req.path;
          request.args["method"] = "GET";

          std::stringstream params(req.args.query);

          std::string oneParam;
          while (std::getline(params, oneParam, '&')) {
            std::stringstream oneParamStream(oneParam);
            std::string param;
            std::string paramValue;
            std::getline(oneParamStream, param, '=');
            std::getline(oneParamStream, paramValue, '=');
            DS_DEBUG("Parameter pair", param, paramValue);
            request.args[param] = paramValue;
          }

          MessageV2 altMessage{.request = request};
          json = ds::toJson(altMessage);
        }
      }
      if (!isConnected()) {
        DS_INFO("Discarding message in disconnected state.");
        return;
      }
      m_bufferedWriter->enqueue(std::move(json));
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Enqueue failed.", e));
    }
  }

  void asyncRead() {
    DS_INFO_ENTER("");
    assertSynchronized();
    try {
      m_stream->asyncRead(
          [this, session = std::weak_ptr(m_session)](boost::system::error_code ec, std::string &&message) {
            if (!session.lock()) {
              return;
            }
            if (ec) {
              asyncReconnect(DS_LOG_STR("Read failed.", ec));
              return;
            }
            asyncReadConsume(std::move(message));
          });
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Read failed.", e));
    }
  }

  void asyncReadConsume(std::string &&rawMessage) {
    DS_DEBUG_ENTER();
    try {
      DS_INFO("Read", ds::hideArg(rawMessage));
      auto message = ds::fromJson<Message>(rawMessage);
      if (message.ping) {
        enqueue({.authenticate = {}, .authenticated = {}, .ping = {}, .pong = Pong{.id = message.ping->id}});
      } else if (message.pong) {
        m_pongId = message.pong->id;
        auto latencyMs = std::chrono::duration_cast<std::chrono::milliseconds>(m_pingElapsed.elapsed()).count();
        m_latencyMs(latencyMs);
        DS_INFO(latencyMs, m_latencyMs, m_uptime);
        if (m_statisticsNotice.elapsed() > config().latencyStatisticsPeriod) {
          DS_NOTICE("latency[ms]:", ds::hideArg(m_latencyMs), "uptime:", ds::hideArg(m_uptime));
          m_statisticsNotice.restart();
          m_latencyMs.drop();
        }
      } else if (message.request) {
        auto &request = *message.request;
        const auto id = std::move(request.id);
        try {
          m_hook.asyncInvoke(
              dsApi::HttpRequest{.path = std::move(request.path), .httpArgs = std::move(request.args)},
              ds::makePromise<dsApi::HttpResult>(io(), [this, session = std::weak_ptr(m_session),
                                                        id = std::move(id)](std::optional<dsApi::HttpResult> &&result) {
                if (!session.lock()) {
                  return;
                }

                enqueue(
                    Message{.authenticate = {},
                            .authenticated = {},
                            .ping = {},
                            .pong = {},
                            .request = {},
                            .response = Response{.id = std::move(id),
                                                 .result = result ? std::move(*result) : dsApi::cancelledResult()}});
              }));
        } catch (const std::exception &e) {
          DS_ERROR("Failed to process request", e, id, rawMessage);
          // Response (likely a cancelledResult) is sent when promise is destroyed.
          // We could propagate this error into the response instead of cancelledResult
        }
      } else if (message.response) {
        auto &response = *message.response;
        auto it = m_calleeInvocations.find(response.id);
        if (it != m_calleeInvocations.end()) {
          it->second.promise->value = std::move(response.result);
          m_calleeInvocations.erase(it);
        } else {
          DS_WARNING("Unexpected response ", response.id);
          DS_INFO("Unexpected response ", response);
        }
      } else if (message.subscribe) {
        auto &subscribe = *message.subscribe;
        m_eventSink->subscribe({.type = std::move(subscribe.type), .expiresAt = ds::Clock::now() + subscribe.timeout});
      } else if (message.close) {
        asyncReconnectOnClose(*message.close);
        return;
      } else {
        // ignore unknown message
        DS_WARNING("Unknown message.", ds::hideArg(rawMessage));
      }
      asyncRead();
    } catch (const std::exception &e) {
      asyncReconnect(DS_LOG_STR("Read consume failed.", e));
    }
  }

  void asyncInvoke(dsApi::HttpRequest &&request, std::unique_ptr<ds::Promise<dsApi::HttpResult>> &&promise) override {
    io().dispatch(ds::asio::MoveOnly{[this, session = std::weak_ptr(m_session), request = std::move(request),
                                      promise = std::move(promise)]() mutable {
      if (!session.lock()) {
        return;
      }
      asyncInvokeImpl(std::move(request), std::move(promise));
    }});
  }

  void asyncInvokeImpl(dsApi::HttpRequest &&request, std::unique_ptr<ds::Promise<dsApi::HttpResult>> &&promise) {
    assertSynchronized();
    try {
      DS_INFO("Async invoke", request);
      DS_REQUIRE(isConnected(), "Not connected.");

      // Require and strip /backend prefix in path.
      // This routing may be done in some router later
      auto path = std::move(request.path);
      constexpr auto backendPrefix = "/backend"sv;
      DS_REQUIRE(boost::starts_with(path, backendPrefix), "Paths in backend must start with backendPrefix.", path,
                 backendPrefix);
      path.erase(0, backendPrefix.size());

      const auto id = RequestId{ds::randomId()};
      enqueue(Message{.authenticate = {},
                      .authenticated = {},
                      .ping = {},
                      .pong = {},
                      .request = Request{.id = id, .path = path, .args = std::move(request.httpArgs)}});
      m_calleeInvocations[id] = {.promise = std::move(promise), .startedAt = ds::Clock::now(), .path = std::move(path)};
    } catch (const std::exception &e) {
      DS_NOTICE("Async invoke failed", e);
      promise->value = dsApi::errorResult(e);
    }
  }

  void asyncCalleeInvocationsDeadlineLoop() {
    DS_DEBUG_ENTER("");
    m_calleeInvocationsDeadline.randomlyExpiresFromNow(calleeInvocationDeadline() / 2, [this]() mutable {
      DS_INFO_ENTER("");
      asyncCalleeInvocationsDeadlineLoop();
      try {
        const auto now = ds::Clock::now();
        for (auto it = m_calleeInvocations.begin(); it != m_calleeInvocations.end();) {
          if (now - it->second.startedAt > calleeInvocationDeadline()) {
            DS_WARNING("Cancelling old invocation.", it->first, it->second.path);
            it = m_calleeInvocations.erase(it);
          } else {
            ++it;
          }
        }
      } catch (const std::exception &e) {
        DS_ERROR("Cancelling old invacations failed.", e);
      }
    });
  }

  dsNet::BufferedMessageWriter &bufferedWriter() override {
    DS_REQUIRE(m_bufferedWriter, "Not connected.");
    return *m_bufferedWriter;
  }

  ConnectorSession *maybeSession() override {
    if (m_bufferedWriter) {
      return this;
    }
    return nullptr;
  }
  dsApi::EventSink &eventSink() override { return *m_eventSink; }
  std::string lastDisconnectReason() const override {
    assertSynchronized();
    return m_lastDisconnectReason;
  }
};

} // namespace

std::unique_ptr<Connector> makeConnector(ds::asio::ResourceBase &parent, std::string &&id,
                                         dsNet::MessageStreamClient &streamClient, ConnectorHook &hook,
                                         dsApi::EventHub &eventHub) {
  return std::make_unique<ConnectorImpl>(parent, std::move(id), streamClient, hook, eventHub);
}

} // namespace dsBackend

DS_JSON_DECLARE_STRUCT((dsBackend), ConnectorConfig, (keepAlivePeriod)(latencyStatisticsPeriod));
DS_JSON_DEFINE_STRUCT((dsBackend), ConnectorConfig);
