/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "timed-measurements-source.h"

#include <ds-apartment/ds-devices.h>
#include <ds-backend/connector.h>
#include <ds-model/model.h>
#include <ds/asio/ds-signal.h>
#include <ds/asio/resource.h>
#include <ds/asio/timer.h>
#include <ds/date-time.h>
#include <event/event_create.h>
#include <metering/metering.h>
#include <model/dsmeter.h>
#include <model/group.h>
#include <time.h>

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsBackendTimedMeasurementsSource");

namespace dsBackend {

namespace {
struct TimedMeasurementsSourceImpl : TimedMeasurementsSource,
                                     ds::asio::Resource<>,
                                     dss::RegisteredEventInterpreterPlugin {
  ds::Signal<dsModel::Measurement> m_next;

  ds::Slots<dsModel::Measurement>& next() override { return m_next; }

  ds::asio::Timer m_timer{io()};
  dss::Apartment& m_apartment;

  TimedMeasurementsSourceImpl(ds::asio::ResourceBase& parent, dss::EventInterpreter& eventInterpreter,
                              dss::Apartment& apartment)
      : ds::asio::Resource<>(parent, "timedMeasurementsSource"),
        dss::RegisteredEventInterpreterPlugin("timedMeasurementsSource", eventInterpreter),
        m_apartment(apartment) {
    eventInterpreter.subscribe(*this, dss::EventName::ModelReady);
  }

  void runTimer(ds::Clock::duration d) {
    assertSynchronized();
    m_timer.expiresFromNow(d, [&] { timedUploadLoop(); });
  }

  void handleEvent(dss::Event& event, const dss::EventSubscription&) override {
    DS_INFO_ENTER(event);
    if (event.getName() == dss::EventName::ModelReady) {
      io().dispatch([this] { runTimer(0s); });
    }
  }

  void emitMeasurement(const dsModel::Measurement& measurement) {
    DS_INFO(measurement);
    m_next.emit(measurement);
  }

  void timedUploadLoop() {
    DS_DEBUG_ENTER();

    auto emit = [&](dsModel::MeasurementOrigin origin, dsModel::MeasurementType type, double value) {
      emitMeasurement(dsModel::Measurement{.id = dsModel::MeasurementId{context().system.randomUuidString()},
                                           .type = "measurement",
                                           .lastChanged = system().now(),
                                           .attributes = {.type = type, .origin = origin, .value = value}});
    };

    auto& energyMetering = m_apartment.getMetering();
    auto meters = m_apartment.getDSMeters();
    std::vector<dss::Dsuid> meterIds;
    std::transform(meters.begin(), meters.end(), std::back_inserter(meterIds),
                   [](const auto& meter) { return meter->getDSID(); });

    bool hasProducer = std::any_of(std::cbegin(meters), std::cend(meters),
                                   [](const auto& meter) { return meter->getCapability_HasMeteringProducer(); });

    try {
      double apartmentEnergy = 0;
      double apartmentProducedEnergy = 0;
      double apartmentEnergyDelta = 0;
      double apartmentProducedEnergyDelta = 0;

      std::optional<double> apartmentModbusEnergy;
      std::optional<double> apartmentModbusProducedEnergy;
      std::optional<double> apartmentModbusEnergyDelta;
      std::optional<double> apartmentModbusProducedEnergyDelta;

      //open metering system
      auto omsOrigin = energyMetering.openMeteringSystemOrigin();
      if (omsOrigin) {
        apartmentEnergy = energyMetering.get15minConsumedOMS(dss::metering::SeriesTypes::energy);
        apartmentProducedEnergy = energyMetering.get15minProducedOMS(dss::metering::SeriesTypes::energy);
        apartmentEnergyDelta = energyMetering.get15minConsumedOMS(dss::metering::SeriesTypes::energyDelta);
        apartmentProducedEnergyDelta = energyMetering.get15minProducedOMS(dss::metering::SeriesTypes::energyDelta);
      } else {
        //Modbus looks like dsm to minimze changes to mobile apps
        auto modbusMeters = energyMetering.getModbusAttachedMeters();
        for (const auto& meter : modbusMeters) {
          if (std::string(meter.id) != "") {
            std::string fakeId = ds::str("modbus-", meter.serial);

            double meterEnergy = energyMetering.get15minConsumed(std::vector<std::string>{meter.serial},
                                                                 dss::metering::SeriesTypes::energy);
            double meterProducedEnergy = energyMetering.get15minProduced(std::vector<std::string>{meter.serial},
                                                                         dss::metering::SeriesTypes::energy);
            double meterEnergyDelta = energyMetering.get15minConsumed(std::vector<std::string>{meter.serial},
                                                                      dss::metering::SeriesTypes::energyDelta);
            double meterProducedEnergyDelta = energyMetering.get15minProduced(std::vector<std::string>{meter.serial},
                                                                              dss::metering::SeriesTypes::energyDelta);
            emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
                dsModel::MeasurementType::energyCounter, meterEnergy);  

            emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
                dsModel::MeasurementType::energyProducedCounter, meterProducedEnergy);

            emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
                dsModel::MeasurementType::energy, meterEnergyDelta);
            
            emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
                dsModel::MeasurementType::energyProduced, meterProducedEnergyDelta);

            if (meter.isGlobal) {
              apartmentModbusEnergy = meterEnergy;
              apartmentModbusProducedEnergy = meterProducedEnergy;
              apartmentModbusEnergyDelta = meterEnergyDelta;
              apartmentModbusProducedEnergyDelta = meterProducedEnergyDelta;
            } else {
              apartmentEnergy += meterEnergy;
              apartmentProducedEnergy += meterProducedEnergy;
              apartmentEnergyDelta += meterEnergyDelta;
              apartmentProducedEnergyDelta += meterProducedEnergyDelta;
            }
          }
        }

        if (apartmentModbusEnergy) {
          apartmentEnergy = *apartmentModbusEnergy;
          apartmentProducedEnergy = *apartmentModbusProducedEnergy;
          apartmentEnergyDelta = *apartmentModbusEnergyDelta;
          apartmentProducedEnergyDelta = *apartmentModbusProducedEnergyDelta;
        } else {
          // Add dsms to have aggregate apartment power
          apartmentEnergy += energyMetering.get15minConsumed(meterIds, dss::metering::SeriesTypes::energy);
          apartmentProducedEnergy += energyMetering.get15minProduced(meterIds, dss::metering::SeriesTypes::energy);
          apartmentEnergyDelta += energyMetering.get15minConsumed(meterIds, dss::metering::SeriesTypes::energyDelta);
          apartmentProducedEnergyDelta += energyMetering.get15minProduced(meterIds, dss::metering::SeriesTypes::energyDelta);
        }
      }

      /// Apartment energy counter
      emit(dsModel::MeasurementOriginApartment{.type = "apartment"}, dsModel::MeasurementType::energyCounter,
           apartmentEnergy);

      if (hasProducer) {
        emit(dsModel::MeasurementOriginApartment{.type = "apartment"}, dsModel::MeasurementType::energyProducedCounter,
             apartmentProducedEnergy);
      }

      /// Apartment energy
      emit(dsModel::MeasurementOriginApartment{.type = "apartment"}, dsModel::MeasurementType::energy,
           apartmentEnergyDelta);

      if (hasProducer) {
        emit(dsModel::MeasurementOriginApartment{.type = "apartment"}, dsModel::MeasurementType::energyProduced,
             apartmentProducedEnergyDelta);
      }

      //Modbus looks like dsm to minimze changes to mobile apps
      auto modbusMeters = energyMetering.getModbusAttachedMeters();
      for (const auto& meter : modbusMeters) {
        if (std::string(meter.id) != "") {
          std::string fakeId = ds::str("modbus-", meter.serial);
          emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
               dsModel::MeasurementType::energyCounter,
               energyMetering.get15minConsumed(std::vector<std::string>{meter.serial},
                                               dss::metering::SeriesTypes::energy));  

          emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
               dsModel::MeasurementType::energyProducedCounter,
               energyMetering.get15minProduced(std::vector<std::string>{meter.serial},
                                               dss::metering::SeriesTypes::energy));

          emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
               dsModel::MeasurementType::energy,
               energyMetering.get15minConsumed(std::vector<std::string>{meter.serial},
                                               dss::metering::SeriesTypes::energyDelta));
          
          emit(dsModel::MeasurementOriginController{.type = "controller", .id = dsModel::ControllerId(fakeId)},
               dsModel::MeasurementType::energyProduced,
               energyMetering.get15minProduced(std::vector<std::string>{meter.serial},
                                               dss::metering::SeriesTypes::energyDelta));
        }
      } 

      /// Controllers energy counter
      for (auto meterPtr : m_apartment.getDSMeters()) {
        auto& meter = *meterPtr;
        if (!meter.isPresent() || !meter.getCapability_HasMetering()) {
          continue;
        }
        emit(dsModel::MeasurementOriginController{.type = "controller", .id = meter.id()},
             dsModel::MeasurementType::energyCounter,
             energyMetering.get15minConsumed(std::vector<dss::Dsuid>{meterPtr->getDSID()},
                                             dss::metering::SeriesTypes::energy));

        if (meter.getCapability_HasMeteringProducer()) {
          emit(dsModel::MeasurementOriginController{.type = "controller", .id = meter.id()},
               dsModel::MeasurementType::energyProducedCounter,
               energyMetering.get15minProduced(std::vector<dss::Dsuid>{meterPtr->getDSID()},
                                               dss::metering::SeriesTypes::energy));
        }
      }

      /// Controllers energy
      for (auto meterPtr : m_apartment.getDSMeters()) {
        auto& meter = *meterPtr;
        if (!meter.isPresent() || !meter.getCapability_HasMetering()) {
          continue;
        }
        emit(dsModel::MeasurementOriginController{.type = "controller", .id = meter.id()},
             dsModel::MeasurementType::energy,
             energyMetering.get15minConsumed(std::vector<dss::Dsuid>{meterPtr->getDSID()},
                                             dss::metering::SeriesTypes::energyDelta));

        if (meter.getCapability_HasMeteringProducer()) {
          emit(dsModel::MeasurementOriginController{.type = "controller", .id = meter.id()},
               dsModel::MeasurementType::energyProduced,
               energyMetering.get15minProduced(std::vector<dss::Dsuid>{meterPtr->getDSID()},
                                               dss::metering::SeriesTypes::energyDelta));
        }
      }

    } catch (const std::exception& e) {
      DS_ERROR("Measurement failed.", e)
    }
    io().dispatch([this] { runTimer(15min); });
  }
};
} // namespace

std::unique_ptr<TimedMeasurementsSource> makeTimedMeasurementsSource(ds::asio::ResourceBase& parent,
                                                                     dss::EventInterpreter& eventInterpreter,
                                                                     dss::Apartment& apartment) {
  return std::make_unique<TimedMeasurementsSourceImpl>(parent, eventInterpreter, apartment);
}

} // namespace dsBackend
