#pragma once

#include <ds-api/service-fwd.h>
#include <ds/reflect.h>

#include <boost/noncopyable.hpp>

#include "apartment-fwd.h"
#include "api-revision-fwd.h"
#include "devices-fwd.h"
#include "groups-fwd.h"
#include "installation-fwd.h"
#include "meterings-fwd.h"
#include "model-fwd.h"
#include "remote-connectivity-fwd.h"
#include "scenarios-fwd.h"
#include "user-defined-states-fwd.h"

DS_REFLECTED_ENUM((dsModel), ZoneActionId, (none)(deepOff)(standby));
DS_REFLECTED_ENUM((dsModel), LightsActionId, (none)(off)(on)(brighter)(darker)(stop)(forcedOff)(fadeOff));
DS_REFLECTED_ENUM((dsModel), ShadesActionId, (none)(down)(up)(stepDown)(stepUp)(stop)(sunProtection));
DS_REFLECTED_ENUM((dsModel), AwningsActionId, (none)(out)(in)(stepOut)(stepIn)(stop)(sunProtection));
DS_REFLECTED_ENUM((dsModel), AudioActionId, (none)(stop)(play)(pause)(mute)(unmute)(volumeUp)(volumeDown));
DS_REFLECTED_ENUM((dsModel), VideoActionId, (none)(stop)(play)(pause)(mute)(unmute)(volumeUp)(volumeDown));
DS_REFLECTED_ENUM((dsModel), VentilationActionId, (none)(off)(level1)(level2)(level3)(level4)(boost));
DS_REFLECTED_ENUM((dsModel), RecirculationActionId, (none)(off)(level1)(level2)(level3)(level4)(boost));
DS_REFLECTED_ENUM((dsModel), DeviceActionId, (none)(off)(on)(localOff)(localOn)(stop)(impulse)(inc)(dec));

DS_REFLECTED_ENUM((dsModel), ParameterCacheUpdateMode, (none)(slow)(fast));

namespace dsModel {

struct InvokeStandardScenarioArgs {
  InvokeContextId context{};
  ActionId actionId{};
  ApplicationId application{};
  DSDeviceId dsDevice{};
  ZoneId zone{};
  AreaId area{};
  ClusterId cluster{};
};

struct ScenarioArgs {
  ScenarioId scenarioId{};
};

class ScenariosApi : public boost::noncopyable {
public:
  virtual ~ScenariosApi() = default;

  [[nodiscard]] virtual std::vector<Scenario> applicationZoneScenarios() const = 0;
  [[nodiscard]] virtual std::vector<Scenario> userDefinedActions() const = 0;
  [[nodiscard]] virtual std::vector<Scenario> deviceScenarios() const = 0;
  [[nodiscard]] virtual std::vector<Scenario> scenarios() const = 0;
  virtual void invokeStandardScenario(const InvokeStandardScenarioArgs &) = 0;
  virtual ScenarioId createScenario(const Scenario &) = 0;
  virtual void deleteScenario(const ScenarioId &) = 0;
  virtual void setScenarioName(const ScenarioId &, const std::string &value) = 0;
  virtual void invokeScenario(const ScenarioId &) = 0;
  virtual void persistScenario(const ScenarioId &) = 0;
};

std::vector<dsApi::Route> apartmentApiRoutes(ScenariosApi &api);

struct MeteringArgs {
  dsModel::MeteringId meteringId{};
};

struct MeteringSeriesArgs {
  int resolution;
  dsModel::MeteringId meteringId{};
};

class MeteringsApi : public boost::noncopyable {
public:
  virtual ~MeteringsApi() = default;

  [[nodiscard]] virtual Meterings meterings() const = 0;
  [[nodiscard]] virtual Metering metering(const MeteringArgs &) const = 0;

  [[nodiscard]] virtual MeteringValues meteringValues() const = 0;
  [[nodiscard]] virtual MeteringValue meteringValue(const MeteringArgs &) const = 0;

  [[nodiscard]] virtual SeriesCompatValues seriesCompatValues() const = 0;
  [[nodiscard]] virtual SeriesCompatValue seriesCompatValue(const MeteringSeriesArgs &) const = 0;

  [[nodiscard]] virtual dsModel::SerialPortConfig getSerialPortConfig() const = 0;

  virtual void setSerialPortConfig(const dsModel::SerialPortConfig& config) = 0;

  [[nodiscard]] virtual std::vector<dsModel::ModbusMeter> getModbusAttachedMeters() const = 0;

  virtual void addModbusAttachedMeter(const dsModel::ModbusMeter& meter) = 0;

  virtual void deleteModbusAttachedMeter(const dsModel::MeteringId& id) = 0;

  [[nodiscard]] virtual std::vector<dsModel::ElectricityMeterConfig> getModbusSupportedMeters() const = 0;

  [[nodiscard]] virtual std::vector<dsModel::ElectricityMeterConfig> getModbusUserMeters() const = 0;

  virtual void addModbusSupportedMeter(const dsModel::ElectricityMeterConfig& meterConfig) = 0;

  virtual void deleteModbusSupportedMeter(const dsModel::MeteringConfigId& id) = 0;
};

std::vector<dsApi::Route> apartmentApiRoutes(MeteringsApi &api);

struct ApartmentArgs {
  std::string dummy{}; // TODO(someday): remove once reflection supports empty structs
};

struct ZoneArgs {
  ZoneId zoneId{};
};

struct AreaArgs {
  ZoneId zoneId{};
  std::string zoneApplicationType{};
  std::string areaId{};
};

struct DeviceArgs {
  DSDeviceId deviceId{};
};

struct SubmoduleArgs {
  SubmoduleId submoduleId{};
};

struct FunctionBlockArgs {
  FunctionBlockId functionBlockId{};
};

struct OutputStatusArgs {
  DSDeviceId deviceId{};
  std::string outputId{};
};

struct UserDefinedStateStatusArgs {
  std::string userDefinedStateId{};
};

struct UpdateDeviceParameterCacheArgs {
  ParameterCacheUpdateMode mode{};
};

struct FloorArgs {
  FloorId floorId{};
};

struct ClusterArgs {
  ClusterId clusterId{};
};

struct TemperatureSetpointArgs {
  ApplicationId applicationId{};
  TemperatureLevel temperatureLevel{};
};

struct TemperatureStatusArgs {
  ZoneId zoneId{};
};

struct ApplicationTokenArgs {
  ApplicationTokenId id{};
};


//TODO - where to go with this
struct ModbusMeterDescription {
  uint8_t slaveAddress;
  std::string meterSerial;
  std::string meterName;
  std::string meterTechnicalName;
};

struct RemoteConnectivityApi {
  virtual ~RemoteConnectivityApi() = default;

  [[nodiscard]] virtual RemoteConnectivity get() const = 0;
  [[nodiscard]] virtual AuthorizeRemoteConnectivityResult authorize(AuthorizeRemoteConnectivityArgs &&) = 0;
  [[nodiscard]] virtual PollAuthorizationRemoteConnectivityResult pollAuthorization(
      PollAuthorizationRemoteConnectivityArgs &&) = 0;
  virtual void unauthorize() = 0;
};

class ApartmentApi : public boost::noncopyable {
public:
  virtual ~ApartmentApi() = default;

  virtual ScenariosApi &scenarios() = 0;
  [[nodiscard]] virtual const ScenariosApi &scenarios() const = 0;

  [[nodiscard]] virtual Zone zone(const ZoneId &id) const = 0;
  [[nodiscard]] virtual std::vector<Zone> zones() const = 0;
  [[nodiscard]] virtual dsModel::Area zoneApplicationAreaAttributes(const dsModel::ZoneId &zoneId,
                                                            const std::string &zoneApplicationType,
                                                            const std::string &areaId) const = 0;
  virtual void setZoneName(ZoneArgs &&, const std::string &value) = 0;
  virtual void setAreaName(dsModel::ZoneArgs&& args, dsModel::ZoneApplicationType zoneApplicationType, const std::string areaId, const std::string& value) = 0;
  [[nodiscard]] virtual ZoneStatus zoneStatus(const ZoneId &id) const = 0;
  [[nodiscard]] virtual std::vector<ZoneStatus> zoneStatuses() const = 0;

  [[nodiscard]] virtual Cluster cluster(const ClusterId &id) const = 0;
  [[nodiscard]] virtual std::vector<Cluster> clusters() const = 0;
  [[nodiscard]] virtual ClusterStatus clusterStatus(const ClusterId &id) const = 0;
  [[nodiscard]] virtual std::vector<ClusterStatus> clusterStatuses() const = 0;
  virtual void setClusterName(const ClusterArgs &args, std::string &&name) = 0;

  [[nodiscard]] virtual DSDevice device(const DSDeviceId &) const = 0;
  [[nodiscard]] virtual std::vector<DSDevice> devices() const = 0;
  virtual void setShadeAttributes(const DSDeviceId &, ShadeAttributes &&attributes) = 0;

  [[nodiscard]] virtual DSDeviceStatus deviceStatus(const DSDeviceId &) const = 0;
  [[nodiscard]] virtual std::vector<DSDeviceStatus> deviceStatuses() const = 0;
  virtual void setOutputValue(OutputStatusArgs &&, double value) = 0;

  [[nodiscard]] virtual Submodule submodule(const SubmoduleId &) const = 0;
  [[nodiscard]] virtual std::vector<Submodule> submodules() const = 0;

  [[nodiscard]] virtual FunctionBlock functionBlock(const FunctionBlockId &) const = 0;
  [[nodiscard]] virtual std::vector<FunctionBlock> functionBlocks() const = 0;

  [[nodiscard]] virtual ApiRevision apiRevision() const = 0;

  [[nodiscard]] virtual ApartmentAccessStatus apartmentAccessStatus() const = 0;
  [[nodiscard]] virtual ApartmentUserStatus apartmentUserStatus() const = 0;
  [[nodiscard]] virtual ApartmentSecurity apartmentSecurity() const = 0;
  virtual void setApartmentAlarm1Name(std::string &&) = 0;
  virtual void setApartmentAlarm2Name(std::string &&) = 0;
  virtual void setApartmentAlarm3Name(std::string &&) = 0;
  virtual void setApartmentAlarm4Name(std::string &&) = 0;
  [[nodiscard]] virtual ApartmentSecurityStatus apartmentSecurityStatus() const = 0;
  [[nodiscard]] virtual ApartmentTemperatureStatus apartmentTemperatureStatus() const = 0;
  virtual void setApartmentTemperatureStatusEcoPreferred(const ApartmentArgs &, bool) = 0;
  [[nodiscard]] virtual ApartmentWeatherStatus apartmentWeatherStatus() const = 0;
  [[nodiscard]] virtual ApartmentEnvironmentStatus apartmentEnvironmentStatus() const = 0;
  [[nodiscard]] virtual ApartmentMeasurementStatus apartmentMeasurementStatus() const = 0;

  [[nodiscard]] virtual ZoneMeasurementsStatus zoneMeasurementsStatus(const ZoneId &) const = 0;
  [[nodiscard]] virtual TemperatureStatus temperatureStatus(const ZoneId &) const = 0;
  [[nodiscard]] virtual ZoneApplicationsStatus zoneApplicationsStatus(const ZoneId &) const = 0;
  [[nodiscard]] virtual ApartmentApplicationsStatus apartmentApplicationsStatus() const = 0;
  [[nodiscard]] virtual Application temperatureApplication(const ZoneId &) const = 0;
  [[nodiscard]] virtual Temperature temperature(const ZoneId &) const = 0;
  virtual void setTemperatureSetpoint(const TemperatureSetpointArgs&, double) = 0;
  virtual void setTemperatureStatusSetpoint(const TemperatureStatusArgs&, double) = 0;
  virtual void setTemperatureStatusSetbackRequested(const TemperatureStatusArgs&, bool) = 0;
  virtual void setTemperatureStatusEcoPreferred(const TemperatureStatusArgs&, bool) = 0;
  virtual void setTemperatureStatusPassiveCooling(const TemperatureStatusArgs&, bool) = 0;

  [[nodiscard]] virtual std::vector<Application> applications() const = 0;
  [[nodiscard]] virtual ZoneApplications zoneApplications(const ZoneId &) const = 0;
  [[nodiscard]] virtual ApartmentApplications apartmentApplications() const = 0;

  [[nodiscard]] virtual Apartment apartment(const dsApi::GetQuery &) const = 0;
  [[nodiscard]] virtual dsApi::HttpResult apartmentCached(const dsApi::HttpRequest &) const = 0;
  [[nodiscard]] virtual dsApi::HttpResult apartmentStatusCached(const dsApi::HttpRequest &) const = 0;
  virtual void pollApartment() = 0;

  [[nodiscard]] virtual ApartmentStatus apartmentStatus(const dsApi::GetQuery &) const = 0;
  virtual void pollApartmentStatus() = 0;

  virtual MeteringsApi &meterings() = 0;
  [[nodiscard]] virtual const MeteringsApi &meterings() const = 0;

  [[nodiscard]] virtual UserDefinedStates userDefinedStates() const = 0;
  [[nodiscard]] virtual std::vector<UserDefinedStateStatus> userDefinedStateStatuses() const = 0;
  virtual void setUserDefinedStateStatus(UserDefinedStateStatusArgs &&, const std::string &value) = 0;

  virtual void updateParameterCache(const UpdateDeviceParameterCacheArgs &) = 0;
  [[nodiscard]] virtual ParameterCache parameterCacheStatus() const = 0;

  [[nodiscard]] virtual Floor floor(const FloorArgs &args) const = 0;
  [[nodiscard]] virtual Floors floors() const = 0;
  virtual void setFloorName(const FloorArgs &args, const std::string &value) = 0;

  [[nodiscard]] virtual Installation installation() const = 0;
  [[nodiscard]] virtual DSServer dsServer() const = 0;

  [[nodiscard]] virtual Controller controller(const ControllerId &) const = 0;
  virtual AuthorizeControllerResult authorizeController(const ControllerId &, AuthorizeControllerArgs &&) = 0;
  virtual void toggleMaintenance(dsModel::MaintenanceData& maintenanceData) = 0;
  virtual void storeControllerAccessToken(const ControllerId &, StoreControllerAccessTokenArgs &&) = 0;
  [[nodiscard]] virtual Controllers controllers() const = 0;

  virtual ApplicationTokenId addApplicationToken(const ApplicationToken &) = 0;
  virtual void renameApplicationToken(const dsModel::ApplicationTokenId& id, const std::string& newName) = 0;
  virtual void revokeApplicationToken(const ApplicationTokenId &) = 0;

  [[nodiscard]] virtual RemoteConnectivityApi &remoteConnectivity() = 0;

  void requireApiRevisionRemovalsLessThan(int x);
};

std::unique_ptr<dsApi::Service> makeInstallationService(ApartmentApi &);
std::unique_ptr<dsApi::Service> makeApartmentService(ApartmentApi &);
std::unique_ptr<dsApi::Service> makeDssService(ApartmentApi &);

} // namespace dsModel

DS_JSON_DECLARE_STRUCT((dsModel), InvokeStandardScenarioArgs, (context)(actionId)(application)(dsDevice)(zone)(area)(cluster));
DS_JSON_DECLARE_STRUCT((dsModel), ScenarioArgs, (scenarioId));

DS_JSON_DECLARE_STRUCT((dsModel), ApartmentArgs, (dummy));
DS_JSON_DECLARE_STRUCT((dsModel), ZoneArgs, (zoneId));
DS_JSON_DECLARE_STRUCT((dsModel), AreaArgs, (zoneId)(zoneApplicationType)(areaId));
DS_JSON_DECLARE_STRUCT((dsModel), MeteringArgs, (meteringId));
DS_JSON_DECLARE_STRUCT((dsModel), MeteringSeriesArgs, (resolution)(meteringId));
DS_JSON_DECLARE_STRUCT((dsModel), DeviceArgs, (deviceId));
DS_JSON_DECLARE_STRUCT((dsModel), SubmoduleArgs, (submoduleId));
DS_JSON_DECLARE_STRUCT((dsModel), FunctionBlockArgs, (functionBlockId));
DS_JSON_DECLARE_STRUCT((dsModel), OutputStatusArgs, (deviceId)(outputId));
DS_JSON_DECLARE_STRUCT((dsModel), UserDefinedStateStatusArgs, (userDefinedStateId));
DS_JSON_DECLARE_STRUCT((dsModel), UpdateDeviceParameterCacheArgs, (mode));
DS_JSON_DECLARE_STRUCT((dsModel), FloorArgs, (floorId));
DS_JSON_DECLARE_STRUCT((dsModel), ClusterArgs, (clusterId));
DS_JSON_DECLARE_STRUCT((dsModel), TemperatureSetpointArgs, (applicationId)(temperatureLevel));
DS_JSON_DECLARE_STRUCT((dsModel), TemperatureStatusArgs, (zoneId));
DS_JSON_DECLARE_STRUCT((dsModel), ApplicationTokenArgs, (id));
