#include "reconnect-timer.h"
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/log.h>
#include <boost/range/algorithm/adjacent_find.hpp>
#include <trompeloeil.hpp>

namespace {

static const char* TAGS = "[dsAsioReconnectTimer][dsAsio][ds]";

using namespace std::literals;

struct Connect {
public:
  MAKE_MOCK0(call, void());
};

SCENARIO("dsAsioReconnectTimer", TAGS) {
  auto clock = ds::ClockMocker{};
  auto io = boost::asio::io_service{};
  auto work = boost::asio::io_service::work{io};
  auto reconnectTimer = ds::asio::ReconnectTimer{io};
  auto connect = Connect{};

  auto ioResetRun = [&] {
    io.reset();
    io.run();
  };

  THEN("default intervals are sane") {
    auto intervals = reconnectTimer.intervals();
    CHECK(intervals.front() == 0s);
    CHECK(intervals.back() >= 10min);
    // exponential backoff
    CHECK(boost::adjacent_find(intervals, [](auto&& a, auto&& b) { return !(a < b && b - a > a); }) == intervals.end());
  }

  SECTION("raiseInterval increases up to lastInterval") {
    for (const auto& interval : reconnectTimer.intervals()) {
      CHECK(reconnectTimer.interval() == interval);
      reconnectTimer.raiseInterval();
    }
    CHECK(reconnectTimer.interval() == reconnectTimer.intervals().back());
  }

  SECTION("raiseIntervalUpTo") {
    SECTION("raise interval raises to first higher interval") {
      reconnectTimer.raiseIntervalUpTo(10s);
      CHECK(reconnectTimer.interval() > 10s);

      SECTION("raise interval does not decrease") {
        reconnectTimer.raiseIntervalUpTo(1s);
        CHECK(reconnectTimer.interval() > 10s);
      }
    }
    SECTION("raise interval raises only up to last interval") {
      reconnectTimer.raiseIntervalUpTo(reconnectTimer.intervals().back() * 10);
      CHECK(reconnectTimer.interval() == reconnectTimer.intervals().back());
    }
  }

  auto checkDelayedReconnect = [&](auto delay) {
    reconnectTimer.asyncDelayedReconnect([&] {
      io.stop();
      connect.call();
    });

    // not called too early
    if (delay > 0s) {
      clock.advance(delay * 3 / 4);
      ds::asio::Timer timer{io};
      timer.expiresNow([&] { io.stop(); });
      ioResetRun();
    }
    // called +- at time
    REQUIRE_CALL(connect, call());
    clock.advance(delay / 4);
    ioResetRun();

    CHECK(true);
  };

  THEN("asyncDelayedReconnect delayed according to intervals") {
    auto intervals = reconnectTimer.intervals();
    auto maxDelay = intervals.back();

    // default schedue with last value repeating
    intervals.push_back(maxDelay);
    intervals.push_back(maxDelay);

    for (const auto& delay : intervals) {
      INFO(delay.count());
      checkDelayedReconnect(delay);
    }

    WHEN("connected") {
      reconnectTimer.connected();
      THEN("reconnected with max delay if connected for short time") {
        clock.advance(reconnectTimer.resetIntervalsAfterConnectedDuration() / 2);
        checkDelayedReconnect(intervals.back());
      }
      THEN("reconnected fast if connected long enough") {
        clock.advance(reconnectTimer.resetIntervalsAfterConnectedDuration() * 3 / 2);
        checkDelayedReconnect(intervals.front());
      }
    }
  }
}

} // namespace
