/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <optional>
#include <tl/expected.hpp>

#include "exception.h"
#include "log.h"

namespace ds {

/// tl::expected (future std::expected) specialized
/// with ds::Exception as error to support ds code style.
///
/// Contrary to tl(std)::expected
/// * `operator *()` - throws `ds::Exception` on error
/// * `value()` throws `ds::Exception` and not `bad_expected_access<ds::Exception>`
/// * `error()` - not provided because it is unsafe. TODO(someday): Reevaluate when needed?
///
/// See also
/// * https://github.com/TartanLlama/expected
/// * http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2017/p0323r3.pdf
template <class T>
struct Expected : tl::expected<T, ds::Exception> {
  using Super = tl::expected<T, ds::Exception>;
  using typename Super::error_type;
  using typename Super::unexpected_type;
  using typename Super::value_type;

  using tl::expected<T, ds::Exception>::expected;
  using Super::operator=;
  using Super::emplace;
  using Super::swap;

  // Returns the contained value or throws ds::Exception
  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr const U& value() const& {
    if (!has_value())
      throw Super::err().value();
    return *static_cast<const Super&>(*this);
  }
  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr U& value() & {
    if (!has_value())
      throw Super::error();
    return *static_cast<Super&>(*this);
  }
  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr const U&& value() const&& {
    if (!has_value())
      throw Super::error();
    return *std::move(static_cast<const Super&&>(*this));
  }
  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr U&& value() && {
    if (!has_value())
      throw Super::error();
    return *std::move(static_cast<Super&&>(*this));
  }

  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr const U& operator*() const& {
    return value();
  }
  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr U& operator*() & {
    return value();
  }
  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr const U&& operator*() const&& {
    return value();
  }
  template <class U = T, tl::detail::enable_if_t<!std::is_void<U>::value>* = nullptr>
  constexpr U&& operator*() && {
    return value();
  }

  using Super::operator bool;
  using Super::has_value;
  using Super::value_or;

  // Carries out some operation on the stored object if there is one.
  using Super::map;

  // Carries out some operation on the unexpected object if there is one.
  using Super::map_error;

  // Like map, but for operations which return a tl::expected
  using Super::and_then;

  // Calls some function if there is no value stored.
  using Super::or_else;

  // not safe - could be reimplemented
  T* operator->() const = delete;
  T* operator->() = delete;

  // not safe - could be reimplemented
  const error_type& error() const& = delete;
  error_type& error() & = delete;
  const error_type&& error() const&& = delete;
  error_type&& error() && = delete;

  // implicit conversion to std::optional for easy migration
  // TODO(someday): convert to explicit conversion? It loses error.
  // TODO(someday): lref variants
  operator std::optional<T>() const&& {
    if (has_value()) {
      return value();
    } else {
      return std::nullopt;
    }
  }
  operator std::optional<T>() && {
    if (has_value()) {
      return value();
    } else {
      return std::nullopt;
    }
  }
};

// Make unexpected error with captured context
#define DS_UNEXPECTED(...) \
  tl::make_unexpected(ds::Exception(::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__), DS_CONTEXT_ARGS)))

} // namespace ds

namespace tl::detail {
// Trait for checking if a type is a tl::expected
// Required by `and_then` methods.
// Could be avoided by reimplementing `and_then`
template <class T>
struct is_expected_impl<ds::Expected<T>> : std::true_type {};

} // namespace tl::detail
