/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <cstring>
#include <iosfwd>
#include <optional>
#include <vector>
#include "exception.h"
#include "str.h"

namespace ds {

/// Logging severities
enum class LogSeverity {
  /// Unconditional debug information. Logging at this level shall not be committed.
  PRINT,

  /// Something went wrong, and execution cannot continue.
  FATAL,

  /// Something is wrong, but execution can continue with garbage output.
  ERROR,

  /// A problem was detected but execution can continue with correct output.
  WARNING,

  /// Important information, does not indicate a problem. Printed by default.
  NOTICE,

  /// Information describing what the code is up to.
  /// E.g enters to public methods.
  /// Not printed by default.
  INFO,

  /// Detailed debug information. E.g. enters to private methods. Not printed by default.
  DEBUG,
};
ds::Enum<LogSeverity> dsEnum(LogSeverity);
std::ostream& operator<<(std::ostream& stream, LogSeverity x);

// parses severity at the start of the string
std::optional<LogSeverity> tryParseLogSeverityChar(const char*);
// interprets syslog severity value
LogSeverity logSeverityFromSyslog(int level);

/// All logging macros stringify passed expressions together with their value.
/// String literals and expressions starting with char '_' are not stringified.
///
/// Example:
///
///     DS_NOTICE("Connected to server.", address, port);
///
/// This will log message like like:
///
///     Connected to server. address:128.55.2.2 port:844
///
/// Macros `DS_REQUIRE` and `DS_ASSERT` check for preconditions and invariants
/// instead of `throw` and `assert`.
///
/// Macro `DS_PRINT` logs unconditionally without using any channgel (like printf).
/// Such code shall not be committed.
///
/// Symbol dsLogChannel in current scope is used  in logging macros
/// `DS_ERROR`, `DS_WARNING`, ... `DS_DEBUG`, `DS_[DEBUG|INFO]_[ENTER|LEAVE]`.
///
/// Macro `DS_CONTEXT` defines context that is attached to messages in current scope.
///
/// All logging channels default to NOTICE severity. Use `DS_LOG` environmental variable to
/// tune channels severity. `F` = FATAL, `E` = ERROR, etc
///
/// `DS_LOG` environmental variable syntax:
///
///    [DEFAULT_SEVERITY],[CHANNEL_NAME_PREFIX1]:[SEVERITY1],[CHANNEL_NAME_PREFIX2]:[SEVERITY2]
///
/// Examples:
/// ````
/// DS_LOG=W ./binary
/// DS_LOG=D ./binary
/// DS_LOG=W,dsAsioWatchdog:D,dsSqlite3:I ./binary
/// ````
///
/// Logs are colorified by default if and only if they output to teminal.
/// Use `DS_LOG_COLORS` environmental variable to override the default.
///
/// Examples:
/// ````
/// DS_LOG_COLORS=auto ./binary
/// DS_LOG_COLORS=yes./binary
/// DS_LOG_COLORS=no ./binary
/// ````
///
/// `DS_LOG`, `DS_LOG_COLORS` are searched in file `/etc/libds`
/// if not passed as environmental variable.
///
/// Overload std::ostream << operator in the same namespace as your class
/// to make your class serializable in logging.
///
/// It is good practice to align the channel name with source file name
/// and with namespace + class name.
/// The common prefix with channel name can be then ommitted in the log message
///
/// Example file src/foo/bar.cpp:
/// ````cpp
/// #include "bar.h"
/// #include <ds/log.h>
///
/// static ds::LogChannel dsLogChannel("fooBar");
///
/// namespace foo {
/// Bar::Bar() { DS_INFO_ENTER(); }
/// Bar::listen(const char* address, int port) {
///     DS_DEBUG_ENTER(address, port);
///     DS_REQUIRE(port != 0, "Port must be explicit.", address, port);
///     DS_NOTICE("Listening on ", address, port);
/// }
/// ````

/// Provide a for-each construct for variadic macros.
/// Source https://codecraft.co/2014/11/25/variadic-macros-tricks/

// Accept any number of args >= N, but expand to just the Nth one.
// Here, N == 22.
#define DS_GET_NTH_ARG(                                                                                     \
    _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12, _13, _14, _15, _16, _17, _18, _19, _20, _21, N, ...) \
  N

// Define some macros to help us create overrides based on the
// arity of a for-each-style macro.
#define DS_MFE_0(_call, ...)
#define DS_MFE_1(_call, x) _call(x)
#define DS_MFE_2(_call, x, ...) _call(x) DS_MFE_1(_call, __VA_ARGS__)
#define DS_MFE_3(_call, x, ...) _call(x) DS_MFE_2(_call, __VA_ARGS__)
#define DS_MFE_4(_call, x, ...) _call(x) DS_MFE_3(_call, __VA_ARGS__)
#define DS_MFE_5(_call, x, ...) _call(x) DS_MFE_4(_call, __VA_ARGS__)
#define DS_MFE_6(_call, x, ...) _call(x) DS_MFE_5(_call, __VA_ARGS__)
#define DS_MFE_7(_call, x, ...) _call(x) DS_MFE_6(_call, __VA_ARGS__)
#define DS_MFE_8(_call, x, ...) _call(x) DS_MFE_7(_call, __VA_ARGS__)
#define DS_MFE_9(_call, x, ...) _call(x) DS_MFE_8(_call, __VA_ARGS__)
#define DS_MFE_10(_call, x, ...) _call(x) DS_MFE_9(_call, __VA_ARGS__)
#define DS_MFE_11(_call, x, ...) _call(x) DS_MFE_10(_call, __VA_ARGS__)
#define DS_MFE_12(_call, x, ...) _call(x) DS_MFE_11(_call, __VA_ARGS__)
#define DS_MFE_13(_call, x, ...) _call(x) DS_MFE_12(_call, __VA_ARGS__)
#define DS_MFE_14(_call, x, ...) _call(x) DS_MFE_13(_call, __VA_ARGS__)
#define DS_MFE_15(_call, x, ...) _call(x) DS_MFE_14(_call, __VA_ARGS__)
#define DS_MFE_16(_call, x, ...) _call(x) DS_MFE_15(_call, __VA_ARGS__)
#define DS_MFE_17(_call, x, ...) _call(x) DS_MFE_16(_call, __VA_ARGS__)
#define DS_MFE_18(_call, x, ...) _call(x) DS_MFE_17(_call, __VA_ARGS__)
#define DS_MFE_19(_call, x, ...) _call(x) DS_MFE_18(_call, __VA_ARGS__)
#define DS_MFE_20(_call, x, ...) _call(x) DS_MFE_19(_call, __VA_ARGS__)

#define DS_MACRO_FOR_EACH(x, ...)                                                                              \
  DS_GET_NTH_ARG("ignored", ##__VA_ARGS__, DS_MFE_20, DS_MFE_19, DS_MFE_18, DS_MFE_17, DS_MFE_16, DS_MFE_15,   \
      DS_MFE_14, DS_MFE_13, DS_MFE_12, DS_MFE_11, DS_MFE_10, DS_MFE_9, DS_MFE_8, DS_MFE_7, DS_MFE_6, DS_MFE_5, \
      DS_MFE_4, DS_MFE_3, DS_MFE_2, DS_MFE_1, DS_MFE_0)                                                        \
  (x, ##__VA_ARGS__)

/// `DS_CONTEXT(...)`:  Adds additional contextual information
/// to logs and exceptions thrown from within the current scope.
/// Note that the parameters to DS_CONTEXT() are evaluated lazily,
/// any variables used must remain valid until the end of the scope.
///
#define DS_CONTEXT(...)                                                                                 \
  auto DS_UNIQUE_NAME(_dsContextFunc) = [&](std::ostream& ostream) {                                    \
    ::ds::ostream(ostream, "" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__));                                \
  };                                                                                                    \
  ::ds::detail::LogContextImpl<decltype(DS_UNIQUE_NAME(_dsContextFunc))> DS_UNIQUE_NAME(_dsLogContext)( \
      DS_UNIQUE_NAME(_dsContextFunc))

/// Linked list of thread local contexts.
/// Use `DS_CONTEXT` macro to make new log context in current scope
/// Use `logContext()` to get current log context.
class LogContext : boost::noncopyable {
protected:
  /// Adds the instance to the thread local linked list of contexts
  LogContext();

  /// Removes the instance from the thread local linked list of contexts
  virtual ~LogContext();

  // Serializes this context item
  virtual void ostream(std::ostream&) = 0;

private:
  LogContext* m_next;
  friend std::ostream& operator<<(std::ostream&, LogContext*);
};
std::ostream& operator<<(std::ostream&, LogContext*);

/// Get current context. Can be serialized by `ostream <<`.
LogContext* logContext();

/// Wraps argument to mark it the argument name shall not be logged.
/// It is interpreted in preprecessor, so it must be used
/// with explicit `ds::` or `::ds::` namespace prefix.
template <typename T>
T&& hideArg(T&& t) {
  return std::forward<T>(t);
}

/// Return passed stringified argument if it is worth to log or an empty string othwerwise.
///
/// Example:
/// trimArgName(##hello) -> "hello"
/// trimArgName(##"hello") -> ""
/// trimArgName(##ds::hideArg(x)) -> ""
/// trimArgName(##::ds::hideArg(x)) -> ""
constexpr const char* trimArgName(const char* x) {
  return (x[0] != '"') && (std::strncmp(x, "ds::hideArg", 11) != 0) && (std::strncmp(x, "::ds::hideArg", 13) != 0) ? x
                                                                                                                   : "";
}

#define DS_ARG(x) , ::ds::trimArgName(#x ":"), x, " "
#define DS_CONTEXT_ARGS ::ds::trimFile(__FILE__), ':', __LINE__, ' ', ::ds::logContext()
#define DS_LOG_CHANNEL_CONTEXT_ARGS(channel) ::ds::logContext(), ::ds::trimFile(__FILE__, channel), ':', __LINE__

// * `DS_REQUIRE(condition, ...)`:  Check external input and preconditions
// e.g. to validate parameters passed from a caller.
// A failure indicates that the caller is buggy.
//
// Compared to an explicit `throw` clause, this macro captures more context
// and enforces the `ds` exception policy.
//
// Example:
//
// DS_REQUIRE(value >= 0, "Value cannot be negative.", value);
// DS_FAIL_REQUIRE("Unknown", value);
//
#define DS_REQUIRE(condition, ...)                                                                                     \
  [&] {                                                                                                                \
    if (DS_LIKELY(condition)) {                                                                                        \
    } else {                                                                                                           \
      throw ds::Exception(                                                                                             \
          ::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__), "condition:", #condition, ' ', DS_CONTEXT_ARGS)); \
    }                                                                                                                  \
  }()
#define DS_FAIL_REQUIRE(...) \
  throw ds::Exception(::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__), DS_CONTEXT_ARGS))

/// * `DS_FAIL_REQUIRE(userMessage, ...)`:
/// Throws exception with translated user message.
#define DS_FAIL_WITH_USER_MESSAGE(userMessage, ...) \
  throw ::ds::Exception(                            \
      ::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__), DS_CONTEXT_ARGS), ::std::move(userMessage))

// Assert the `condition` is true.
// This macro should be used to check for bugs in the surrounding code (broken invariant, ...)
// and its dependencies, but NOT to check for invalid input.
// Example:
//
// DS_ASSERT(m_state == State::CONNECTED, m_state);
// DS_FAIL_ASSERT("Invalid", m_state);
//
#define DS_ASSERT(condition, ...)                                                                        \
  [&] {                                                                                                  \
    if (DS_LIKELY(condition)) {                                                                          \
    } else {                                                                                             \
      ::ds::assertFailed(#condition, ::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__)).c_str(), \
          __PRETTY_FUNCTION__, __FILE__, __LINE__);                                                      \
    }                                                                                                    \
  }()

#define DS_FAIL_ASSERT(...) DS_ASSERT(false, ##__VA_ARGS__);

// Logging channel registered into Logger singleton.
//
// Use DS_DEBUG, DS_INFO, etc for logging.
//
class LogChannel : boost::noncopyable {
public:
  // Registers this log channel to `Logger::instance()`.
  //
  // Log channels shall default to NOTICE severity for consistency.
  explicit LogChannel(std::string name, LogSeverity defaultSeverity = LogSeverity::NOTICE);

  // Unregisters this log channel from `Logger::instance()`
  ~LogChannel();

  const std::string& name() const { return m_name; }

  LogSeverity severity() const { return m_severity; }
  void setSeverity(LogSeverity x) { m_severity = x; }

  // Set severity, but only if `Logger::defaultSeverity() == none`.
  // Otherwise it may reset the severity to Logger::defaultSeverity().
  //
  // Intended for places that don't know the initial severity in their constructor.
  void setInitialSeverity(LogSeverity x);

  bool shouldLog(LogSeverity severity) const { return severity <= m_severity; }
  void log(LogSeverity severity, std::string message) const;

private:
  // copy of severity from channel to make method `shouldLog`
  // method as fast as possible.
  const std::string m_name;
  LogSeverity m_severity;
};
std::ostream& operator<<(std::ostream& stream, const LogChannel& x);

/// Rule specifying severity for all channels
/// with name starting with given prefix.
struct LogRule {
  std::string channelNamePrefix;
  LogSeverity severity;
  bool operator==(const LogRule& x) const { return channelNamePrefix == x.channelNamePrefix && severity == x.severity; }
};

/// Parse rule in format "channelNamePrefix:severity" / "severity"
std::optional<LogRule> tryParseLogRule(char* rule);

/// Parse rules in format "severity,channelNamePrefix1:severity,channelNamePrefix2:severity"
/// Parsing failures are ignored.
std::vector<LogRule> tryParseLogRules(const char*);

/// Logger singleton holding reference to all channels.
/// Methods are thread safe.
class Logger : boost::noncopyable {
  Logger();

public:
  static Logger& instance();

  typedef std::function<void(const char* channelName, LogSeverity severity, std::string)> LogFunction;

  // Synchronize to internal mutex and log the message by calling `logFunction`.
  void log(const char* channelName, LogSeverity severity, std::string) const;

  /// Set log function called by `log()`.
  void setLogFunction(LogFunction);

  /// Default log function writing messages to error output
  static void defaultLogFunction(const char* channelName, LogSeverity severity, std::string);

  /// Default log channels severity.
  ///
  /// none -> log channel's initial severity is used
  /// !none -> overrides log channel's initial severity
  ///
  /// Initially parsed DS_LOG environmental variable: `DS_LOG=I`
  std::optional<LogSeverity> defaultSeverity() const;
  void setDefaultSeverity(std::optional<LogSeverity> x);

  // There is no getter for channels, it would require clients to hold our mutex.

  /// Returns channel severity. Throws if channel does not exist.
  LogSeverity channelSeverity(std::string_view channelName) const;

  /// Sets channel severity. Throws if channel does not exist.
  void setChannelSeverity(std::string_view channelName, LogSeverity);

  struct ChannelSeverity {
    std::string name;
    LogSeverity severity;
  };

  // Make vector of channels with their severities.
  std::vector<ChannelSeverity> channelSeverities() const;

private:
  struct Impl;
  const std::unique_ptr<Impl> m_impl;

  void updateChannelSeverity(LogChannel& channel);
  // adds channel and sets its severity
  void addChannel(LogChannel& channel);
  void removeChannel(LogChannel& channel);

  friend class LogChannel;
};

#define DS_LOG_STR_CHANNEL(channel, ...) \
  ::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__), DS_LOG_CHANNEL_CONTEXT_ARGS(channel))
#define DS_LOG_STR(...) ::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__), DS_CONTEXT_ARGS)

#define DS_CHANNEL_LOG(channel, severity, ...)                           \
  if (!channel.shouldLog(severity)) {                                    \
  } else {                                                               \
    channel.log((severity), DS_LOG_STR_CHANNEL(channel, ##__VA_ARGS__)); \
  }

#define DS_CHANNEL_DEBUG(channel, ...) DS_CHANNEL_LOG(channel, ::ds::LogSeverity::DEBUG, ##__VA_ARGS__)
#define DS_CHANNEL_DEBUG_ENTER(channel, ...) \
  DS_CHANNEL_LOG(channel, ::ds::LogSeverity::DEBUG, "Enter", ::ds::hideArg(__PRETTY_FUNCTION__), ##__VA_ARGS__)
#define DS_CHANNEL_DEBUG_LEAVE(channel, ...) \
  DS_CHANNEL_LOG(channel, ::ds::LogSeverity::DEBUG, "Leave", ::ds::hideArg(__PRETTY_FUNCTION__), ##__VA_ARGS__)
#define DS_CHANNEL_INFO(channel, ...) DS_CHANNEL_LOG(channel, ::ds::LogSeverity::INFO, ##__VA_ARGS__)
#define DS_CHANNEL_INFO_ENTER(channel, ...) \
  DS_CHANNEL_LOG(channel, ::ds::LogSeverity::INFO, "Enter", ::ds::hideArg(__PRETTY_FUNCTION__), ##__VA_ARGS__)
#define DS_CHANNEL_INFO_LEAVE(channel, ...) \
  DS_CHANNEL_LOG(channel, ::ds::LogSeverity::INFO, "Leave", ::ds::hideArg(__PRETTY_FUNCTION__), ##__VA_ARGS__)
#define DS_CHANNEL_NOTICE(channel, ...) DS_CHANNEL_LOG(channel, ::ds::LogSeverity::NOTICE, ##__VA_ARGS__)
#define DS_CHANNEL_WARNING(channel, ...) DS_CHANNEL_LOG(channel, ::ds::LogSeverity::WARNING, ##__VA_ARGS__)
#define DS_CHANNEL_ERROR(channel, ...) DS_CHANNEL_LOG(channel, ::ds::LogSeverity::ERROR, ##__VA_ARGS__)
#define DS_CHANNEL_FATAL(channel, ...) DS_CHANNEL_LOG(channel, ::ds::LogSeverity::FATAL, ##__VA_ARGS__)

// Macros logging to channel available at symbol `dsLogChannel`.
//
// Symbol `dsLogChannel` can be declared at any scope:
// * object static
// * class static
// * class member
// * local
#define DS_LOG(...) DS_CHANNEL_LOG(dsLogChannel, ##__VA_ARGS__)
#define DS_DEBUG(...) DS_CHANNEL_DEBUG(dsLogChannel, ##__VA_ARGS__)
#define DS_DEBUG_ENTER(...) DS_CHANNEL_DEBUG_ENTER(dsLogChannel, ##__VA_ARGS__)
#define DS_DEBUG_LEAVE(...) DS_CHANNEL_DEBUG_LEAVE(dsLogChannel, ##__VA_ARGS__)
#define DS_INFO(...) DS_CHANNEL_INFO(dsLogChannel, ##__VA_ARGS__)
#define DS_INFO_ENTER(...) DS_CHANNEL_INFO_ENTER(dsLogChannel, ##__VA_ARGS__)
#define DS_INFO_LEAVE(...) DS_CHANNEL_INFO_LEAVE(dsLogChannel, ##__VA_ARGS__)
#define DS_NOTICE(...) DS_CHANNEL_NOTICE(dsLogChannel, ##__VA_ARGS__)
#define DS_WARNING(...) DS_CHANNEL_WARNING(dsLogChannel, ##__VA_ARGS__)
#define DS_ERROR(...) DS_CHANNEL_ERROR(dsLogChannel, ##__VA_ARGS__)
#define DS_FATAL(...) DS_CHANNEL_FATAL(dsLogChannel, ##__VA_ARGS__)

// * DS_PRINT(...)`:  print temporary log messsage unconditionally.
// Useful for ad-hoc logging. Code with this macro shall not be committed.
#define DS_PRINT(...)           \
  ::ds::Logger::instance().log( \
      "", ::ds::LogSeverity::PRINT, ::ds::strArgs("" DS_MACRO_FOR_EACH(DS_ARG, ##__VA_ARGS__), DS_CONTEXT_ARGS))

/// Stringify function tuned for DS_ARG arguments
template <typename... Args>
std::string strArgs(Args&&... args) {
  auto s = ::ds::str(std::forward<Args>(args)...);
  if (auto size = s.size()) {
    if (s[size - 1] == ' ') {
      s.resize(size - 1); // last character is extra ' ' due to macro expansion problems
    }
  }
  return s;
}

/// Trim file to meaningfull part.
std::string trimFile(std::string file);
std::string trimFile(std::string file, const ds::LogChannel& channel);

/// Assert (abort or throw) with given message.
///
/// assert is macro, so we cannot reuse its name here
[[noreturn]] void assertFailed(char const* condition, char const* message, char const* function, char const* file,
                               long line);

// test assert. Asserts DS_TEST_ASSERT is set. Does nothing othwerwise.
void testAssert();

namespace detail {

template <typename Func>
class LogContextImpl : public LogContext {
public:
  LogContextImpl(Func& func) : m_func(func) {}

private:
  Func& m_func;
  void ostream(std::ostream& s) override { m_func(s); }
};

} // namespace detail

// Converts one number to another, throws on narrowing error.
// Like gsl::narrow, but throws ds::Exception and captures context in it.
//
// TODO(someday): find better place for this function.
// We need better place for DS_REQUIRE first.
//
template <class T, class U>
inline T narrow(U u) {
  T t = gsl::narrow_cast<T>(u);
  DS_REQUIRE(static_cast<U>(t) == u && (t < 0) == (u < 0), "Narrowing error.", t, u);
  return t;
}

} // namespace ds
