/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "system.h"
#include <ds/catch/catch.h>
#include <ds/log.h>
#include <boost/algorithm/string/predicate.hpp>

static const char* TAGS = "[dsSystem][ds]";

namespace {

TEST_CASE("dsSystem", TAGS) {
  SECTION("dirs from build system by default") {
    auto& system = ds::system();
    const auto& dirs = system.dirs();
    CHECK(dirs.sysconfdir == DS_SYSCONFDIR);
    CHECK(dirs.datadir == DS_DATADIR);
  }

  SECTION("dirs passed in constructor") {
    auto system = ds::makeSystem({.sysconfdir = "/sysconfdir", .datadir = "/datadir"});
    const auto& dirs = system->dirs();
    CHECK(dirs.sysconfdir == "/sysconfdir");
    CHECK(dirs.datadir == "/datadir");
  }

  SECTION("testSystem") {
    // DS_SRC_DIR is passed as -D... by build system
    const auto& path = DS_SRC_DIR "/test";
    auto& testSystem = ds::system();

    SECTION("check existsFile") { CHECK(testSystem.existsFile(ds::str(path, "/true-line"))); }
    SECTION("check !existsFile") { CHECK(!testSystem.existsFile(ds::str(path, "/non-existing"))); }

    SECTION("readFile reads existing file") { CHECK(testSystem.readFile(ds::str(path, "/true-line")) == "true\n"); }

    SECTION("readFile throws on non-existing file") {
      CHECK_THROWS(testSystem.readFile(ds::str(path, "/non-existing")));
    }

    SECTION("readFileIfExists reads existing file") {
      CHECK(testSystem.readFile(ds::str(path, "/true-line")) == "true\n");
    }

    SECTION("readFileIfExists returns none on non-existing file") {
      CHECK(testSystem.readFileIfExists(ds::str(path, "non-existing")) == std::nullopt);
    }

    SECTION("directory files exist") {
      auto files = testSystem.lsRegularFiles(path, "");
      CHECK(files.size() == 1);
      CHECK(boost::contains(files.at(0), "true-line"));
    }

    SECTION("directory files do not exist") {
      auto files = testSystem.lsRegularFiles(path, ".test-file-extension");
      CHECK(files.size() == 0);
    }

    SECTION("directory files directory does not exists") {
      CHECK(testSystem.lsRegularFiles(ds::str(path, "not-existing"), "") == std::vector<std::string>{});
    }

    SECTION("directory files directory is a file") {
      CHECK_THROWS_FIND(testSystem.lsRegularFiles(ds::str(path, "/true-line"), ""), "Not a directory");
    }
  }
  SECTION("now returns system_clock") {
    auto& system = ds::system();
    auto dsNow = system.now();
    auto stdNow = std::chrono::system_clock::now();
    CHECK(dsNow <= stdNow);
    CHECK(stdNow - dsNow <= std::chrono::seconds(5));
  }

  SECTION("randomId") {
    auto& system = ds::system();
    auto id = system.randomId();
    INFO(id);
    SECTION("is long enough") { CHECK(id.size() >= 16); }
    SECTION("changes") { CHECK(id != system.randomId()); }
  }

  SECTION("randomUuidString") {
    auto& system = ds::system();
    auto id = system.randomUuidString();
    INFO(id);
    SECTION("is in uuid format") {
      CHECK(id.size() == 36);
      CHECK(id[8] == '-');
      CHECK(id[13] == '-');
      CHECK(id[18] == '-');
      CHECK(id[23] == '-');
    }
  }
}

} // namespace
