/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "dsbusinterface.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

#include "event-mock.h"
#include "model/modelmaintenance-mock.h"
#include "propertysystem-mock.h"
#include "security/security-mock.h"

static ds::LogChannel dsLogChannel("ds485DsbusinterfaceTest");

using namespace dss;
using namespace trompeloeil;
using namespace std::literals;
using trompeloeil::_;

namespace {

static const char* TAGS = "[ds485][dsbusinterface]";

TEST_CASE("dsbusinterface capture callbacks and convert to events", TAGS) {
  ds::asio::TestLoopResource loop; // RAII, ensure destructor runs last

  SecurityMock security{};
  EventSinkMock eventSink{};
  PropertySystemMock propertySystem{};
  ModelMaintenanceMock modelMaintenance{propertySystem, eventSink};

  DSBusInterface dsbusinterface(nullptr, &modelMaintenance, security);

  SECTION("handleSensorValueExEvent") {
    SECTION("float conversion") {
      struct X {
        uint64_t raw;
        double value;
      };
      // https://en.wikipedia.org/wiki/Double-precision_floating-point_format
      const auto& x = GENERATE(
          values(std::initializer_list<X>{{0b0, 0.0},
                                          {0b0100000000001000000000000000000000000000000000000000000000000000, 3},
                                          {0x3F88000000000000, 0.01171875},
                                          {0x7FEFFFFFFFFFFFFF, std::numeric_limits<double>::max()},
                                          {0x0010000000000000, std::numeric_limits<double>::min()}}));

      auto context = std::array<uint8_t, 4>{'f', '0', '0', 0};
      ALLOW_CALL(security, signIn(_)).RETURN(true);
      DS_NOTICE(x.value);
      REQUIRE_CALL(modelMaintenance, addModelEvent(_))
          .LR_SIDE_EFFECT(CHECK(dynamic_cast<const ModelEventWithSensorEx*>(_1.get())->m_sensorValue == x.value));
      dsbusinterface.handleSensorValueExEvent(0, dsuid_t{}, dsuid_t{}, 1, 1, x.raw, 1, 1, context.data());
    }
  }

  WHEN("ds485 member is lost") {
    dsuid_t id{};
    ds::Slot<dsuid_t> ds485MemberLost(loop, dsbusinterface.ds485MemberLost, [&id](auto arg) { id = arg; });
    REQUIRE_CALL(security, signIn(_)).RETURN(true);
    Dsuid dsmid{"1111111111111111111111111111111100"s};
    dsbusinterface.handleBusChange(&dsmid, 1); // 1 -> lost
    loop.postStopRun();
    CHECK(id == dsmid);
  }

  WHEN("ds485 member joins") {
    dsuid_t id{};
    ds::Slot<dsuid_t> ds485MemberLost(loop, dsbusinterface.ds485MemberJoin, [&id](auto arg) { id = arg; });
    REQUIRE_CALL(security, signIn(_)).RETURN(true);

    Dsuid dsmid{"1111111111111111111111111111111100"s};
    dsbusinterface.handleBusChange(&dsmid, 0); // 0 -> join
    loop.postStopRun();
    CHECK(id == dsmid);
  }
}

} // namespace
