/*
    Copyright (c) 2009 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "ds485types.h"

#include <ds/log.h>
#include <stdlib.h>
#include <cstring>
#include <iomanip>
#include <sstream>
#include <stdexcept>
#include <string>

#include <digitalSTROM/dsuid.h>

std::ostream &operator<<(std::ostream &stream, const dsuid_t &x) { return stream << strdsuid(&x); }
std::ostream &operator<<(std::ostream &stream, const dsid_t &x) {
  return stream << dss::dsid2str(x);
}

namespace dss {

std::string dsuid2str(const dsuid_t &dsuid) {
  std::ostringstream str;
  for (int i = 0; i < DSUID_SIZE; ++i) {
    str << std::hex << std::setw(2) << std::setfill('0') << (int)dsuid.id[i];
  }
  return str.str();
}

dsuid_t str2dsuid(const char *s) {
  dsuid_t dsuid;

  if (::dsuid_from_string(s, &dsuid) != DSUID_RC_OK) {
    DS_FAIL_REQUIRE("Failed to parse dsuid", s);
  }
  return dsuid;
}

/**
 * Logic:
 * - If there is a value in dsuidStr it is expected to be a dSUID
 * - dsidStr contains a dSID
 */
dsuid_t dsidOrDsuid2dsuid(std::string dsidStr, std::string dsuidStr) {
  dsuid_t dsuid;
  if (dsuidStr.empty()) {
    dsid_t dsid = str2dsid(dsidStr);
    dsuid = dsuid_from_dsid(&dsid);
  } else {
    dsuid = str2dsuid(dsuidStr);
  }
  return dsuid;
}

dsid_t str2dsid(std::string dsid_str) {
  dsid_t dsid;

  if (dsid_str.length() != 24) {
    DS_FAIL_REQUIRE("Failed to parse dsid", dsid_str);
  }

  size_t i = 0;
  size_t count = 0;
  while (i < 24) {
    std::string number = dsid_str.substr(i, 2);
    dsid.id[count] = (unsigned char)strtol(number.c_str(), NULL, 16);
    i = i + 2;
    count++;
  }

  return dsid;
}

std::string dsid2str(const dsid_t &dsid) {
  std::ostringstream str;
  for (int i = 0; i < 12; ++i) {
    str << std::hex << std::setw(2) << std::setfill('0') << (int)dsid.id[i];
  }
  return str.str();
}

uint32_t dsuid2serial(dsuid_t dsuid) {
  uint32_t serial = 0;
  if (::dsuid_get_serial_number(&dsuid, &serial) != DSUID_RC_OK) {
    DS_FAIL_REQUIRE("Cannot extract serial number from dSUID", dsuid);
  }
  return serial;
}

bool IsEvenDsuid(dsuid_t dsuid) {
  uint32_t serial = 0;
  if (dsuid_get_serial_number(&dsuid, &serial) != DSUID_RC_OK) {
    // probably a vDC device, go by submodule
    return !(dsuid.id[DSUID_SIZE - 1] % 2);
  }

  return !(serial % 2);
}

void dsuid_get_next_dsuid(const dsuid_t dsuid, dsuid_t *out) {
  DS_ASSERT(out, "");
  int ret = ::dsuid_get_next_dsuid(&dsuid, out);
  if (ret != DSUID_RC_OK) {
    // probably a vDC, go by submodule
    memcpy(out->id, dsuid.id, DSUID_SIZE);
    out->id[DSUID_SIZE - 1]++;
  }
}

void dsuid_get_previous_dsuid(const dsuid_t dsuid, dsuid_t *out) {
  DS_ASSERT(out, "");
  int ret = ::dsuid_get_previous_dsuid(&dsuid, out);
  if (ret != DSUID_RC_OK) {
    // probably a vDC, go by submodule
    memcpy(out->id, dsuid.id, DSUID_SIZE);
    out->id[DSUID_SIZE - 1]--;
  }
}

bool dsuid_to_dsid(const dsuid_t dsuid, dsid_t *dsid) {
  if (::dsuid_to_dsid(&dsuid, dsid) != DSUID_RC_OK) {
    memset(dsid, 0, sizeof(dsid_t));
    return false;
  }
  return true;
}

dsid_t dsuidToDsid(const dsuid_t &dsuid) {
  dsid_t dsid{};
  // dsid is gtin-96, only real ds device support it, vdc can not be converted -> exception
  DS_REQUIRE(dsuid_to_dsid(dsuid, &dsid), dsuid);
  return dsid;
}

dsuid_t dsuid_from_dsid(const dsid_t &dsid) {
  dsuid_t dsuid;
  dsuid = ::dsuid_from_dsid(&dsid);
  return dsuid;
}

} // namespace
