/*
    Copyright (c) 2009,2010,2012 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Michael Troß, aizo GmbH <michael.tross@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#include "eventinterpreterplugins.h"

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <fcntl.h>
#include <json.h>
#include <limits.h>
#include <signal.h>
#include <spawn.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <algorithm>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/function.hpp>
#include <boost/scoped_ptr.hpp>

#include "base.h"
#include "businterface.h"
#include "dss.h"
#include "event.h"
#include "event/event_create.h"
#include "event/event_fields.h"
#include "internaleventrelaytarget.h"
#include "logger.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/group.h"
#include "model/modelmaintenance.h"
#include "model/set.h"
#include "model/state.h"
#include "model/zone.h"
#include "monitor_tasks.h"
#include "propertysystem_common_paths.h"
#include "scanner/databasedownload.h"
#include "security/security.h"
#include "setbuilder.h"
#include "subscription_profiler.h"
#include "unix/systeminfo.h"
#include "webservice_api.h"

static ds::LogChannel dsLogChannel("eventInterpreterPlugins");

namespace dss {


  //================================================== EventInterpreterPluginRaiseEvent

  EventInterpreterPluginRaiseEvent::EventInterpreterPluginRaiseEvent(EventInterpreter* _pInterpreter)
  : EventInterpreterPlugin("raise_event", _pInterpreter)
  { } // ctor

  EventInterpreterPluginRaiseEvent::~EventInterpreterPluginRaiseEvent()
  { } // dtor

  void EventInterpreterPluginRaiseEvent::handleEvent(Event& _event, const EventSubscription& _subscription) {
    std::shared_ptr<Event> newEvent = std::make_shared<Event>(_subscription.getOptions()->getParameter("event_name"));
    if(_subscription.getOptions()->hasParameter("time")) {
      std::string timeParam = _subscription.getOptions()->getParameter("time");
      if(!timeParam.empty()) {
        Logger::getInstance()->log("RaiseEvent: Event has time");
        newEvent->setTime(timeParam);
      }
    }
    applyOptionsWithSuffix(_subscription.getOptions(), "_default", newEvent, false);
    if (_subscription.getOptions()->hasParameter(EventProperty::Location)) {
      std::string location = _subscription.getOptions()->getParameter(EventProperty::Location);
      if(!location.empty()) {
        Logger::getInstance()->log("RaiseEvent: Event has location");
        newEvent->setLocation(location);
      }
    }
    if (newEvent->hasPropertySet(EventProperty::ICalStartTime)) {
      auto now = DateTime().toRFC2445IcalDataTime();
      auto today = now.substr(0, now.find("T"));
      auto target = newEvent->getPropertyByName(EventProperty::ICalStartTime);
      auto targetTime = target.substr(target.find("T")+1);
      auto newStart = today + "T" + targetTime;
      newEvent->setProperty(EventProperty::ICalStartTime, newStart);
      Logger::getInstance()->log("RaiseEvent: reset iCalStartTime from " + target + " to " + newStart);
    }
    newEvent->applyProperties(_event.getProperties());
    applyOptionsWithSuffix(_subscription.getOptions(), "_override", newEvent, true);
    getEventInterpreter().getQueue().pushEvent(newEvent);
  } // handleEvent

  void EventInterpreterPluginRaiseEvent::applyOptionsWithSuffix(std::shared_ptr<const SubscriptionOptions> _options, const std::string& _suffix, std::shared_ptr<Event> _event, bool _onlyOverride) {
    if(_options != NULL) {
      for (auto&& kv : _options->getParameters().getContainer()) {
        if(endsWith(kv.first, _suffix)) {
          std::string propName = kv.first.substr(0, kv.first.length() - _suffix.length());
          if(_event->hasPropertySet(propName) || !_onlyOverride) {
            Logger::getInstance()->log("RaiseEvent: setting property " + propName + " to " + kv.second);
            _event->setProperty(propName, kv.second);
          }
        }
      }
    }
  } // applyOptionsWithSuffix

  //================================================== EventRelayTarget

  EventRelayTarget::~EventRelayTarget() {
    while(!m_Subscriptions.empty()) {
      unsubscribeFrom(m_Subscriptions.front()->getID());
    }
  } // dtor

  void EventRelayTarget::subscribeTo(std::shared_ptr<EventSubscription> _pSubscription) {
    m_Relay.registerSubscription(this, _pSubscription->getID());
    m_Subscriptions.push_back(_pSubscription);
    getRelay().getEventInterpreter().subscribe(_pSubscription);
  } // subscribeTo

  void EventRelayTarget::unsubscribeFrom(const std::string& _subscriptionID) {
    m_Relay.removeSubscription(_subscriptionID);
    SubscriptionPtrList::iterator it =
      std::find_if(m_Subscriptions.begin(), m_Subscriptions.end(),
                   boost::bind(&EventSubscription::getID, _1) == _subscriptionID);
    getRelay().getEventInterpreter().unsubscribe(_subscriptionID);
    if(it != m_Subscriptions.end()) {
      m_Subscriptions.erase(it);
    }
  } // unsubscribeFrom


  //================================================== EventInterpreterInternalRelay

  EventInterpreterInternalRelay::EventInterpreterInternalRelay(EventInterpreter* _pInterpreter)
  : EventInterpreterPlugin(getPluginName(), _pInterpreter)
  { } // ctor

  EventInterpreterInternalRelay::~EventInterpreterInternalRelay() {
  } // dtor

  void EventInterpreterInternalRelay::handleEvent(Event& _event, const EventSubscription& _subscription) {
    EventRelayTarget* target = m_IDTargetMap[_subscription.getID()];
    if(target != NULL) {
      target->handleEvent(_event, _subscription);
    }
  } // handleEvent

  void EventInterpreterInternalRelay::registerSubscription(EventRelayTarget* _pTarget, const std::string& _subscriptionID) {
    m_IDTargetMap[_subscriptionID] = _pTarget;
  } // registerSubscription

  void EventInterpreterInternalRelay::removeSubscription(const std::string& _subscriptionID) {
    m_IDTargetMap[_subscriptionID] = NULL;
  } // removeSubscription

  //================================================== EventInterpreterPluginSendmail

  bool EventInterpreterPluginSendmail::m_active = false;

  EventInterpreterPluginSendmail::EventInterpreterPluginSendmail(EventInterpreter* _pInterpreter)
      : EventInterpreterPlugin(EventName::SendMail, _pInterpreter), m_mailq_dir{MAILDIR} {
    if (m_mailq_dir.empty()) {
      if (DSS::hasInstance()) {
        m_mailq_dir = DSS::getInstance()->getDataDirectory() + "/mail/";
      }
    }
    boost::filesystem::path p(m_mailq_dir);
    try {
      if (!boost::filesystem::is_directory(p)) {
        boost::filesystem::create_directory(p);
      }
    } catch (boost::filesystem::filesystem_error& e) {
      Logger::getInstance()->log("EventInterpreterPluginSendmail: cannot create mail "
          "queue folder: " + std::string(e.what()), lsFatal);
    }

    Logger::getInstance()->log("EventInterpreterPluginSendmail: "
        "writing e-mail to spool directory \"" + m_mailq_dir + "\"", lsInfo);
  } // ctor

  EventInterpreterPluginSendmail::~EventInterpreterPluginSendmail() {
  }

  void EventInterpreterPluginSendmail::handleEvent(Event& _event, const EventSubscription& _subscription) {
    std::string sender;
    std::string recipient, recipient_cc, recipient_bcc;
    std::string header;
    std::string subject;
    std::string body;

    if(_event.hasPropertySet("from")) {
      sender = _event.getPropertyByName("from");
    } else if(_subscription.getOptions()->hasParameter("from")) {
      sender = _subscription.getOptions()->getParameter("from");
    }

    if(_event.hasPropertySet("to")) {
      recipient = _event.getPropertyByName("to");
    } else if(_subscription.getOptions()->hasParameter("to")) {
      recipient = _subscription.getOptions()->getParameter("to");
    }

    if(_event.hasPropertySet("cc")) {
      recipient_cc = _event.getPropertyByName("cc");
    } else if(_subscription.getOptions()->hasParameter("cc")) {
      recipient_cc = _subscription.getOptions()->getParameter("cc");
    }

    if(_event.hasPropertySet("bcc")) {
      recipient_bcc = _event.getPropertyByName("bcc");
    } else if(_subscription.getOptions()->hasParameter("bcc")) {
      recipient_bcc = _subscription.getOptions()->getParameter("bcc");
    }

    if (_event.hasPropertySet("header")) {
      header = _event.getPropertyByName("header");
    } else if(_subscription.getOptions()->hasParameter("header")) {
      header = _subscription.getOptions()->getParameter("header");
    }

    if(_event.hasPropertySet("subject")) {
      subject = _event.getPropertyByName("subject");
    } else if(_subscription.getOptions()->hasParameter("subject")) {
      subject = _subscription.getOptions()->getParameter("subject");
    }

    if (_event.hasPropertySet("body")) {
      body = _event.getPropertyByName("body");
    } else if(_subscription.getOptions()->hasParameter("body")) {
      body = _subscription.getOptions()->getParameter("body");
    }

    char mailText[PATH_MAX];
    strncpy(mailText, m_mailq_dir.c_str(), ((m_mailq_dir.length() + 1) < PATH_MAX) ? m_mailq_dir.length() + 1 : PATH_MAX);
    mailText[PATH_MAX-1] = '\0';
    strcat(mailText, "/mailXXXXXX");

    int prevUmask = umask(S_IRWXG | S_IRWXO);
    int mailFile = mkstemp((char *) mailText);
    umask(prevUmask);
    if (mailFile < 0) {
      Logger::getInstance()->log(std::string("EventInterpreterPluginSendmail: generating temporary file ") + mailText + " failed [" +
          intToString(errno) + "]: " + strerror(errno), lsFatal);
      return;
    }
    FILE* mailStream = fdopen(mailFile, "w");
    if (mailStream == NULL) {
      Logger::getInstance()->log("EventInterpreterPluginSendmail: writing to temporary file failed [" +
          intToString(errno) + "]", lsFatal);
      return;
    }

    try {
      Logger::getInstance()->log("EventInterpreterPluginSendmail::handleEvent: Sendmail", lsDebug);

      DateTime now;
      std::ostringstream mail;
      mail << "Date: " << now.toRFC2822String() << "\n"
          << "To: " << recipient << "\n"
          << "From: " << sender << "\n"
          << "Subject: " << subject << "\n"
          << "X-Mailer: digitalSTROM Server (v" DSS_VERSION ")" << "\n";

      bool hasContentType = false;
      if (!header.empty()) {
        std::vector<std::string> slist = dss::splitString(header, '\n', true);
        for (std::string line : slist) {
          mail << line << "\n";
          if (strncasecmp(line.c_str(), "Content-Type:", 13) == 0) {
            hasContentType = true;
          }
        }
      }
      if (!hasContentType) {
        mail << "Content-Type: text/plain; charset=utf-8\n";
      }
      mail << "\n";
      mail << body;
      fputs(mail.str().c_str(), mailStream);
      fclose(mailStream);

    } catch (std::exception& e) {
      Logger::getInstance()->log("EventInterpreterPluginSendmail: failed to send mail: " +
          std::string(e.what()), lsFatal);
    }

    Logger::getInstance()->log("EventInterpreterPluginSendmail: new mail in " +
        std::string(mailText), lsDebug);
  } // handleEvent

  EventInterpreterPluginExecutionDeniedDigest::EventInterpreterPluginExecutionDeniedDigest(EventInterpreter* _pInterpreter)
        : EventInterpreterPlugin("execution_denied_digest", _pInterpreter) {
    m_timestamp = DateTime();
    m_check_scheduled = false;
  }

  void EventInterpreterPluginExecutionDeniedDigest::handleEvent(Event& _event, const EventSubscription& _subscription) {
    if (_event.getName() == EventName::ExecutionDenied) {
      boost::mutex::scoped_lock lock(m_digest_mutex);
      std::string action;
      std::string reason;
      if (_event.hasPropertySet("source-name")) {
        action = _event.getPropertyByName("source-name");
      }
      if (action.empty()) {
        if (_event.hasPropertySet("action-name")) {
          action = _event.getPropertyByName("action-name");
        }
      }
      if (_event.hasPropertySet("reason")) {
        reason = _event.getPropertyByName("reason");
      }

      m_timestamp = DateTime();

      m_digest << "[" << m_timestamp.toString() << "] "
               << "Action \"" << action << "\" was not executed. "
               << reason << "!" << std::endl << std::endl;

      if (!m_check_scheduled) {
        std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::ExecutionDeniedDigestCheck);
        pEvent->setProperty("time", "+60");
        if (DSS::hasInstance()) {
          DSS::getInstance()->getEventQueue().pushEvent(pEvent);
          m_check_scheduled = true;
        }
      }

    } else if  (_event.getName() == EventName::ExecutionDeniedDigestCheck) {
      boost::mutex::scoped_lock lock(m_digest_mutex);
      m_check_scheduled = false;

      DateTime now = DateTime();
      DateTime last = DateTime(m_timestamp);
      last.addSeconds(60);
      if ((last >= now) && (m_digest.str().length() < 2048)) {
        return;
      }

      if (m_digest.str().length() == 0) {
        return;
      }
      std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::SendMail);
      pEvent->setProperty("body", m_digest.str());
      pEvent->setProperty("subject",
                          "digitalSTROM server actions not executed");
      pEvent->setProperty("from", "dSS");
      if (DSS::hasInstance()) {
        pEvent->setProperty("to",
                            DSS::getInstance()->getApartment().getName());
        DSS::getInstance()->getEventQueue().pushEvent(pEvent);
      }
      m_digest.clear();
      m_digest.str("");
      m_timestamp = DateTime();
    }
  }

  EventInterpreterPluginApartmentChange::EventInterpreterPluginApartmentChange(EventInterpreter*
                                                                               _pInterpreter)
        : EventInterpreterPlugin("apartment_model_change", _pInterpreter)
  {
  }

  void EventInterpreterPluginApartmentChange::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    if (!WebserviceMsHub::isAuthorized()) {
      return;
    }

    for (auto&& elt : _event.getProperties().getContainer()) {
      Logger::getInstance()->log(" name " + elt.first + " : " + elt.second);
    }

    WebserviceMsHub::ChangeType type;
    /* momentan ist definiert: 1=Apartment, 2=TimedEvent, 3=UDA */
    if (_event.getName() == ModelChangedEvent::Apartment) {
      type = WebserviceMsHub::ApartmentChange;
    } else if (_event.getName() == ModelChangedEvent::TimedEvent) {
      type = WebserviceMsHub::TimedEventChange;
    } else if (_event.getName() == ModelChangedEvent::UserDefinedAction) {
      type = WebserviceMsHub::UDAChange;
    } else {
      DS_ERROR("Unknown ApartmentChange event.", _event);
      return;
    }

    /* no retval, no error handling, just log entry */
    WebserviceMsHub::doModelChanged(type, WebserviceCallDone_t());
  }

  namespace EventName {
  std::string WebserviceKeepAlive = "keepWebserviceAlive";
  std::string WebserviceGetWeatherInformation = "getWeatherInformation";
  } // namespace EventName

  __DEFINE_LOG_CHANNEL__(EventInterpreterWebservicePlugin, lsInfo);

  EventInterpreterWebservicePlugin::EventInterpreterWebservicePlugin(EventInterpreter*
                                                                     _pInterpreter)
    : EventInterpreterPlugin("EventInterpreterWebservicePlugin", _pInterpreter)
  {
    websvcEnabledNode =
      DSS::getInstance()->getPropertySystem().getProperty(pp_websvc_enabled);
    websvcEnabledNode ->addListener(this);
  }

  EventInterpreterWebservicePlugin::~EventInterpreterWebservicePlugin() {
    websvcEnabledNode->removeListener(this);
  }

  void EventInterpreterWebservicePlugin::propertyChanged(PropertyNodePtr _caller,
                                                                  PropertyNodePtr _changedNode) {
    // initiate connection as soon as webservice got enabled
    if (_changedNode->getBoolValue() == true) {
      std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::WebserviceKeepAlive);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=100");
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
      std::shared_ptr<Event> pEventWeather = std::make_shared<Event>(EventName::WebserviceGetWeatherInformation);
      pEventWeather->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEventWeather->setProperty(EventProperty::ICalRRule, "FREQ=MINUTELY;INTERVAL=60");
      DSS::getInstance()->getEventQueue().pushEvent(pEventWeather);
    } else {
      DSS::getInstance()->getEventRunner().removeEventByName(EventName::WebserviceKeepAlive);
      DSS::getInstance()->getEventRunner().removeEventByName(EventName::WebserviceGetWeatherInformation);
    }
  }

  void EventInterpreterWebservicePlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::Running));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::WebserviceKeepAlive));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::WebserviceGetWeatherInformation));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ApplicationTokenDeleted));
  }

  void EventInterpreterWebservicePlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    DS_DEBUG("Webservice handle", _event);
    if (_event.getName() == EventName::Running) {
      if (WebserviceMsHub::isAuthorized()) {
        std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::WebserviceKeepAlive);
        pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
        pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=100");
        DSS::getInstance()->getEventQueue().pushEvent(pEvent);
        std::shared_ptr<Event> pEventWeather = std::make_shared<Event>(EventName::WebserviceGetWeatherInformation);
        pEventWeather->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
        pEventWeather->setProperty(EventProperty::ICalRRule, "FREQ=MINUTELY;INTERVAL=60");
        DSS::getInstance()->getEventQueue().pushEvent(pEventWeather);
        return;
      }
    }

    if (!WebserviceMsHub::isAuthorized()) {
      log("!webservice_communication_authorized", lsWarning);
      return;
    }

    if (_event.getName() == EventName::WebserviceKeepAlive) {
      std::shared_ptr<URLRequestCallback> cb;
      //WebserviceConnection::getInstanceMsHub()->request(
      //    "public/accessmanagement/v1_0/RemoteConnectivity/TestConnection", "", HttpMethod::GET, cb, false);
    } else if (_event.getName() == EventName::ApplicationTokenDeleted) {
      if (!_event.hasPropertySet(EventProperty::ApplicationToken)) {
        log("Invalid token deleted event missing token", lsWarning);
        return;
      }
      std::string token = _event.getPropertyByName(EventProperty::ApplicationToken);
      WebserviceMsHub::doNotifyTokenDeleted(token, WebserviceCallDone_t());
      return;
    } else if (_event.getName() == EventName::WebserviceGetWeatherInformation) {
      WebserviceMsHub::doGetWeatherInformation();
    }
  }

  EventInterpreterSensorMonitorPlugin::EventInterpreterSensorMonitorPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("EventInterpreterSensorMonitorPlugin", _pInterpreter) {}

  __DEFINE_LOG_CHANNEL__(EventInterpreterSensorMonitorPlugin, lsInfo);

  void EventInterpreterSensorMonitorPlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ModelReady));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::CheckSensorValues));
  }

  void EventInterpreterSensorMonitorPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    log("handle: " + _event.getName(), lsDebug);

    // start the sensor timeout detection with a delay to allow for synchronization to take place
    if (_event.getName() == EventName::ModelReady) {
      std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::CheckSensorValues);
      pEvent->setProperty("time", "+3600");
      if (DSS::hasInstance()) {
        DSS::getInstance()->getEventQueue().pushEvent(pEvent);
      }
      return;
    }

    if (DSS::hasInstance()) {
      TaskProcessor &tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
      tp.addEvent(std::make_shared<SensorMonitorTask>(&(DSS::getInstance()->getApartment())));
    }
  }

  EventInterpreterStateSensorPlugin::EventInterpreterStateSensorPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("EventInterpreterStateSensorPlugin", _pInterpreter) {}

  __DEFINE_LOG_CHANNEL__(EventInterpreterStateSensorPlugin, lsInfo);

  void EventInterpreterStateSensorPlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::DeviceSensorValue));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ZoneSensorValue));
  }

  void EventInterpreterStateSensorPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    log("handle: " + _event.getName(), lsDebug);

    if (!DSS::hasInstance()) {
      return;
    }

    if (_event.getName() == EventName::DeviceSensorValue && _event.getRaisedAtDevice()) {
      std::shared_ptr<const DeviceReference> pDevRev = _event.getRaisedAtDevice();

      std::string sensorValueFloat = _event.getPropertyByName("sensorValueFloat");
      double fValue = strToDouble(sensorValueFloat);

      auto sensorType = _event.getPropertyByName<SensorType>("sensorType");
      if (sensorType == SensorType::UnknownType) {
        try {
          int index = strToInt(_event.getPropertyByName("sensorIndex"));
          sensorType = pDevRev->getDevice()->getSensor(index).m_sensorType;
        } catch (ItemNotFoundException& ex) {}
      }

      DSS::getInstance()->getApartment().updateSensorStates(pDevRev->getDSID(), sensorType, fValue, CallOrigin::dsmApi);
      return;
    }

    if (_event.getName() == EventName::ZoneSensorValue && _event.getRaisedAtGroup()) {
      std::shared_ptr<const Group> pGroup = _event.getRaisedAtGroup();

      auto sensorType = _event.getPropertyByName<SensorType>("sensorType");
      std::string sensorValueFloat = _event.getPropertyByName("sensorValueFloat");
      double fValue = strToDouble(sensorValueFloat);

      DSS::getInstance()->getApartment().updateSensorStates(pGroup->getZoneID(), pGroup->getID(), sensorType, fValue, CallOrigin::dsmApi);
      return;
    }
  }

  EventInterpreterHeatingMonitorPlugin::EventInterpreterHeatingMonitorPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("EventInterpreterHeatingMonitorPlugin", _pInterpreter) {}

  __DEFINE_LOG_CHANNEL__(EventInterpreterHeatingMonitorPlugin, lsInfo);

  void EventInterpreterHeatingMonitorPlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(*this, EventName::ModelReady);
    interpreter.subscribe(*this, EventName::DSMeterReady);
    interpreter.subscribe(*this, EventName::HeatingControllerState);
    interpreter.subscribe(*this, EventName::TemperatureControllersElection);
    interpreter.subscribe(*this, EventName::ZoneSensorValue);
  }

  void EventInterpreterHeatingMonitorPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    log("handle: " + _event.getName(), lsDebug);
    if (DSS::hasInstance()) {
      auto& tp = DSS::getInstance()->getModelMaintenance().getTaskProcessor();
      tp.addEvent(std::make_shared<HeatingMonitorTask>(&(DSS::getInstance()->getApartment()), _event.shared_from_this()));
    }
  }

  EventInterpreterHeatingValveProtectionPlugin::EventInterpreterHeatingValveProtectionPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("EventInterpreterHeatingValveProtectionPlugin", _pInterpreter) {}

  __DEFINE_LOG_CHANNEL__(EventInterpreterHeatingValveProtectionPlugin, lsInfo);

  void EventInterpreterHeatingValveProtectionPlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ModelReady));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::HeatingValveProtection));
  }

  void EventInterpreterHeatingValveProtectionPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    log("handle: " + _event.getName(), lsDebug);
    if (DSS::hasInstance()) {
      TaskProcessor &tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
      tp.addEvent(std::make_shared<HeatingValveProtectionTask>(&(DSS::getInstance()->getApartment()), _event.shared_from_this()));
    }
  }

  EventInterpreterDebugMonitorPlugin::EventInterpreterDebugMonitorPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("EventInterpreterDebugMonitorPlugin", _pInterpreter) {}

  __DEFINE_LOG_CHANNEL__(EventInterpreterDebugMonitorPlugin, lsInfo);

  void EventInterpreterDebugMonitorPlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ModelReady));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::DebugMonitorUpdate));
  }

  void EventInterpreterDebugMonitorPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    log("handle: " + _event.getName(), lsDebug);

    try {
      SystemInfo info;
      info.memory();
    } catch (...) {
      Logger::getInstance()->log("EventInterpreterDebugMonitorPlugin: error executing memory update", lsWarning);
    }

    std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::DebugMonitorUpdate);
    pEvent->setProperty(EventProperty::Time, "+" + intToString(600));
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    }
  }

  EventInterpreterDatabaseUpdatePlugin::EventInterpreterDatabaseUpdatePlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("import_database", _pInterpreter) {}

  void EventInterpreterDatabaseUpdatePlugin::subscribe() {
  }

  void EventInterpreterDatabaseUpdatePlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    DS_DEBUG("DatabaseUpdate handle", _event);

    std::string id;
    std::string url;

    if (_subscription.getOptions()->hasParameter("script_id")) {
      // legacy code
      id = _subscription.getOptions()->getParameter("script_id");
    }

    if (_subscription.getOptions()->hasParameter("id")) {
      // overwrites script_id
      id = _subscription.getOptions()->getParameter("id");
    }

    if (_subscription.getOptions()->hasParameter("url")) {
      url = _subscription.getOptions()->getParameter("url");
    }


    if (id.empty()) {
      Logger::getInstance()->log("EventInterpreterDatabaseUpdatePlugin: not "
            "downloading database because <id> parameter is empty");
      return;
    }

    if (url.empty()) {
      Logger::getInstance()->log("EventInterpreterDatabaseUpdatePlugin: not "
            "downloading database because <url> parameter is empty");
      return;
    }

    if (DSS::hasInstance()) {
      TaskProcessor &tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
      tp.addEvent(std::make_shared<DatabaseDownload>(id, url));
    }
  }

  AutoclusterUpdatePlugin::AutoclusterUpdatePlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("AutoclusterUpdatePlugin", _pInterpreter) {}

  __DEFINE_LOG_CHANNEL__(AutoclusterUpdatePlugin, lsInfo);

  void AutoclusterUpdatePlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ModelReady));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::AutoClusterUpdate));
  }

  void AutoclusterUpdatePlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    log("handle: " + _event.getName(), lsDebug);
    if (DSS::hasInstance()) {
      TaskProcessor &tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
      tp.addEvent(std::make_shared<AutoClusterTask>(&(DSS::getInstance()->getApartment()), _event.shared_from_this()));
    }
  }

  LockClusterPlugin::LockClusterPlugin(EventInterpreter* _pInterpreter)
      : EventInterpreterPlugin("LockClusterPlugin", _pInterpreter) {}

    __DEFINE_LOG_CHANNEL__(LockClusterPlugin, lsInfo);

    void LockClusterPlugin::subscribe() {
      auto&& interpreter = getEventInterpreter();
      interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ModelReady));
      interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::StateChange));
      interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::LockClusterUpdate));
      interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::LockClusterRefresh));
    }

    void LockClusterPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
    {
      DS_INFO("handle: " + _event.getName());
      if (DSS::hasInstance()) {
        TaskProcessor &tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
        tp.addEvent(std::make_shared<LockClusterTask>(&(DSS::getInstance()->getApartment()), _event.shared_from_this()));
      }
    }

  ClusterWeatherProtectionPlugin::ClusterWeatherProtectionPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("ClusterWeatherProtectionPlugin", _pInterpreter) {}

  __DEFINE_LOG_CHANNEL__(ClusterWeatherProtectionPlugin, lsInfo);

  void ClusterWeatherProtectionPlugin::subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ModelReady));
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::ClusterCheckWind));
  }

  void ClusterWeatherProtectionPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    log("handle: " + _event.getName(), lsDebug);
    if (!DSS::hasInstance()) {
      return;
    }

    if (_event.getName() == EventName::ModelReady) {
      handleModelReadyEvent();
    } else if (_event.getName() == EventName::ClusterCheckWind) {
      handleClusterCheckWindEvent(_event);
    }
  }

  void ClusterWeatherProtectionPlugin::handleModelReadyEvent() {
    auto& apartment = DSS::getInstance()->getApartment();
    auto states = apartment.getStates();
    states.erase(std::remove_if(states.begin(), states.end(), [](const auto &state) {
      return !(boost::starts_with(state->getName(), "wind.group") && state->getState() == State_Active);
    }), states.end());

    DS_DEBUG("Sanitizing active cluster wind states.", states);
    for (const auto& state : states) {
      auto clusterId = state->tryGetClusterID();
      if (!clusterId) continue;
      try {
        auto cluster = apartment.getCluster(*clusterId);
        cluster->setWindProtectionTimer(DateTime());
        DSS::getInstance()->raiseEvent(
          createClusterCheckWindEvent(*clusterId, ClusterWindTimeout));
      } catch (std::exception& e) {
        DS_WARNING("No cluster with clusterId.", clusterId, e.what());
        apartment.removeState(state);
      }
    }
  }

  void ClusterWeatherProtectionPlugin::handleClusterCheckWindEvent(Event& _event) {
    auto clusterId = _event.getPropertyByName("clusterId");
    auto wind = DSS::getInstance()->getApartment().tryGetNonScriptState("wind.group" + clusterId);
    if (!wind && wind->getState() == State_Active)
      return;

    try {
      DateTime now;
      auto cluster = DSS::getInstance()->getApartment().getCluster(strToInt(clusterId));
      auto clusterWindProtectionTimer = cluster->getWindProtectionTimer().value();
      if (now.difference(clusterWindProtectionTimer) >= ClusterWindTimeout) {
        cluster->callScene(CallOrigin::json, SceneAccessCategory::manual, SceneWindInactive, "", true);
        cluster->setWindProtectionTimer(std::nullopt);
      } else {
        DSS::getInstance()->raiseEvent(
          createClusterCheckWindEvent(strToInt(clusterId), ClusterWindTimeout - now.difference(clusterWindProtectionTimer)));
      }
    } catch (std::exception& e) {
      DS_WARNING("Cluster weather protection plugin failed. Setting wind state to inactive.", clusterId, wind, e.what());
      wind->setState(CallOrigin::system, State_Inactive);
    }
  }

  UpdateShadeStopTimingsPlugin::UpdateShadeStopTimingsPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("UpdateShadeStopTimingsPlugin", _pInterpreter) {}

  void UpdateShadeStopTimingsPlugin::subscribe() {
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::UpdateShadeStopTimings));
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::UpdateShadeStopTimingsDelayed));
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::ModelReady));
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::DeviceEvent));
  }

  void UpdateShadeStopTimingsPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    if (!DSS::hasInstance()) {
      return;
    }

    std::shared_ptr<Event> pEvent =
            std::make_shared<Event>(EventName::UpdateShadeStopTimings);
    pEvent->setProperty(EventProperty::UniqueFirst, "Yes");

    if (_event.getName() == EventName::ModelReady) {
      pEvent->setProperty(EventProperty::Time, "+1200");
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    } else if (_event.getName() == EventName::DeviceEvent) {
      std::string action = _event.getPropertyByName("action");
      if ((action != "moved") && (action != "groupAdd") &&
          (action != "groupRemove")) {
        return;
      }

      std::shared_ptr<const Device> device = _event.getRaisedAtDevice()->getDevice();
      if (!device->getDeviceFeatures().posTime) {
        return;
      }

      pEvent->setProperty(EventProperty::Time, "+600");
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    } else if (_event.getName() == EventName::UpdateShadeStopTimingsDelayed) {
      pEvent->setProperty(EventProperty::Time, "+600");
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    } else if (_event.getName() == EventName::UpdateShadeStopTimings) {
      TaskProcessor &tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
      tp.addEvent(std::make_shared<UpdateShadeStopTimingsTask>(&(DSS::getInstance()->getApartment())));
    }
  }

  UpdateButtonUsagePlugin::UpdateButtonUsagePlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("UpdateButtonUsagePlugin", _pInterpreter) {}

  void UpdateButtonUsagePlugin::subscribe() {
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::CallScene));
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::ButtonClickBus));
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::ButtonDeviceAction));
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::UndoScene));
  }

  void UpdateButtonUsagePlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    if (!DSS::hasInstance()) {
      return;
    }

    dsuid_t dsuid;

    Properties props = _event.getProperties();
    if (!props.has(ef_originDSUID)) {
      if (!(_event.getRaiseLocation() == erlDevice)) {
        return;
      }

      dsuid = _event.getRaisedAtDevice()->getDSID();
    } else {
      dsuid = str2dsuid(props.get(ef_originDSUID));
    }

    std::shared_ptr<Device> device = DSS::getInstance()->getApartment().getDeviceByDSID(dsuid);

    if (device->getButtonUsage() == ButtonUsage::auto_unused) {
      device->setButtonUsage(ButtonUsage::used);
    }
  }

  SunPositionSensorPlugin::SunPositionSensorPlugin(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("SunPositionSensorPlugin", _pInterpreter) {}

  void SunPositionSensorPlugin::subscribe() {
    getEventInterpreter().subscribe(std::make_shared<EventSubscription>(*this, EventName::SunPositionSensorUpdate));

    std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::SunPositionSensorUpdate);
    pEvent->setProperty(EventProperty::ICalStartTime, DateTime().addMinute(1).toRFC2445IcalDataTime());
    pEvent->setProperty(EventProperty::ICalRRule, "FREQ=MINUTELY;INTERVAL=1");
    DSS::getInstance()->getEventQueue().pushEvent(pEvent);
  }

  void SunPositionSensorPlugin::handleEvent(Event& _event, const EventSubscription& _subscription)
  {
    if (!DSS::hasInstance()) {
      return;
    }

    TaskProcessor &tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
    tp.addEvent(std::make_shared<SunPositionSensorTask>(&(DSS::getInstance()->getApartment())));
  }

} // namespace dss
