/*
    Copyright (c) 2009,2010,2011 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "archives-connection.h"

#include <ds/log.h>

#include "metering/archives-storage.h"
#include "metering/rrd-interface.h"

static ds::LogChannel dsLogChannel("dssMeteringArchivesConnection");

namespace dss::metering {

namespace {

class ArchivesConnectionImpl final : public ArchivesConnection {
  std::shared_ptr<RRDInterface> m_rrdInterface;
  std::string m_rrdCachedPath{};
  std::unique_ptr<ArchivesStorage> m_archivesStorage;

public:
  ArchivesConnectionImpl(std::shared_ptr<RRDInterface> rrdInterface, ds::System& system)
      : m_rrdInterface(std::move(rrdInterface)), m_archivesStorage(makeArchivesStorage(system)) {}
  ~ArchivesConnectionImpl() override = default;

  void flush(const std::vector<std::string>& paths) const override;
  std::optional<ds::DateTime> lastTimestamp(const std::string& path) const override;
  void setRRDCachedPath(std::string rrdCachedPath) override { m_rrdCachedPath = std::move(rrdCachedPath); }
  void restartCached() const override;

  ArchivesStorage& getArchivesStorage() const override { return *m_archivesStorage; }

  RRDInterface& rrdInterface() const override { return *m_rrdInterface; }
  std::string rrdCachedPath() const override { return m_rrdCachedPath; };
};

void ArchivesConnectionImpl::flush(const std::vector<std::string>& paths) const {
  if (m_rrdCachedPath.empty()) {
    DS_DEBUG("flush: no rrdcached path set");
    return;
  }

  std::vector<std::string> lines;
  lines.push_back("flushcached");
  lines.push_back("--daemon");
  lines.push_back(m_rrdCachedPath);

  std::copy(std::cbegin(paths), std::cend(paths), std::back_inserter(lines));
  DS_DEBUG("flushing cached rrd data to file");
  m_rrdInterface->flush(lines);
}

std::optional<ds::DateTime> ArchivesConnectionImpl::lastTimestamp(const std::string& path) const {
  auto lastAccess = m_rrdInterface->lastTimestamp(path);
  if (!lastAccess) {
    // probably invalid file format, rely on later checks to fail as well
    return std::nullopt;
  }
  return std::chrono::system_clock::from_time_t(*lastAccess);
}

void ArchivesConnectionImpl::restartCached() const {
  if (!m_rrdCachedPath.empty()) {
    DS_NOTICE("Restarting rrdcached....");
    /* TODO sv restart rrdcached might hang */
    if (system("killall -9 rrdcached") != 0) {
      DS_WARNING("Restarting rrdcached was not successful.");
    }
  }
}

} // namespace

std::shared_ptr<ArchivesConnection> makeArchivesConnection(std::shared_ptr<RRDInterface> rrdInterface,
                                                           ds::System& system) {
  return std::make_shared<ArchivesConnectionImpl>(rrdInterface, system);
};

} // namespace dss::metering
