/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "archives-storage.h"

#include <ds/catch/catch.h>
#include <ds/system-mock.h>

#include "archives-connection-mock.h"

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;

static ds::LogChannel dsLogChannel("dssMeteringArchivesStorageTest");

inline bool operator==(const dsid_t &l, const dsid_t &r) {
  return memcmp(l.id, r.id, sizeof(std::declval<dsid_t>().id)) == 0;
}

namespace {

static const char* TAGS = "[dSS][metering][archivesStorage]";

SCENARIO("archivesStorage", TAGS) {
  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto system = ds::SystemMock();
  auto storage = makeArchivesStorage(system);

  GIVEN("storageLocation") {
    WHEN("not set") { CHECK(storage->storageLocation() == "/invalid-store/"s); }
    WHEN("set") {
      storage->setMeteringStorageLocation("/path/to/dir"s);
      CHECK(storage->storageLocation() == "/path/to/dir"s);
    }
  }
  SECTION("filename") {
    CHECK(filename({Dsuid{controllerId1}, Category::consumption}) == "1111111111111111111111111111111100.rrd"s);
    CHECK(filename({Dsuid{controllerId1}, Category::production}) ==
          "1111111111111111111111111111111100-production.rrd"s);
  }
  SECTION("hasArchive") {
    GIVEN("archive exists") {
      REQUIRE_CALL(system, existsFile(_)).RETURN(true); // xx
      CHECK(storage->hasArchive({Dsuid{controllerId1}, Category::consumption}));
    }
    GIVEN("archive does not exists") {
      REQUIRE_CALL(system, existsFile(_)).RETURN(false);
      CHECK(!storage->hasArchive({Dsuid{controllerId1}, Category::consumption}));
    }
  }
  SECTION("OpenMeteringSystem") {
    auto openMeteringSystemId = OpenMeteringSystemId{"3153414733313030323938313039"s};

    storage->setMeteringStorageLocation("/storage");

    GIVEN("archive exists") {
      THEN("filename") {
        CHECK(storage->filepath(openMeteringSystemId) ==
              "/storage/open-metering-system-3153414733313030323938313039.rrd"s);
      }
      THEN("hasArchive succeeds") {
        REQUIRE_CALL(system, existsFile("/storage/open-metering-system-3153414733313030323938313039.rrd"s))
            .RETURN(true);
        CHECK(storage->hasArchive(openMeteringSystemId));
      }

      THEN("search archive succeeds") {
        REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s))
            .RETURN(std::vector{"/storage/open-metering-system-75fe16be7.rrd"s});
        CHECK(storage->searchOmsArchives() == std::vector{OpenMeteringSystemId{"75fe16be7"s}});
      }
    }
  }

  SECTION("searchArchiveFiles") {
    storage->setMeteringStorageLocation("/storage");
    using vector_t = decltype(std::declval<ArchivesStorage>().searchArchives());

    GIVEN("legacy dsid") {
      dsid_t id{.id = {1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12}};
      DS_NOTICE(id);
      REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s))
          .RETURN(std::vector{ds::str("/storage/01020304060708090a0b0c00.rrd"s)});
      CHECK(storage->searchArchives() ==
            std::vector{std::pair{CompoundId{id}, "/storage/01020304060708090a0b0c00.rrd"s}});
    }
    GIVEN("dsmeter consumption") {
      ArchiveId id{controllerId1, Category::consumption};
      REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s)).LR_RETURN(std::vector{storage->filepath(id)});
      CHECK(storage->searchArchives() ==
            std::vector{std::pair{CompoundId{id}, "/storage/1111111111111111111111111111111100.rrd"s}});
    }
    GIVEN("dsmeter production") {
      ArchiveId id{controllerId1, Category::production};
      REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s)).LR_RETURN(std::vector{storage->filepath(id)});
      CHECK(storage->searchArchives() ==
            std::vector{std::pair{CompoundId{id}, "/storage/1111111111111111111111111111111100-production.rrd"s}});

      GIVEN("incorrect suffix") {
        REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s))
            .LR_RETURN(std::vector{"/storage/1111111111111111111111111111111100-producxxxx.rrd"s});
        CHECK(storage->searchArchives() == vector_t{});
      }
    }
    GIVEN("open-metering-system") {
      OpenMeteringSystemId id{"3153414733313030323938313039"s};
      REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s)).LR_RETURN(std::vector{storage->filepath(id)});
      CHECK(storage->searchArchives() ==
            std::vector{std::pair{CompoundId{id}, "/storage/open-metering-system-3153414733313030323938313039.rrd"s}});
    }

    GIVEN("file with invalid dsuid") {
      REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s))
          .RETURN(std::vector{"/foo/xxxitex11zz11111111111111111111100.rrd"s});
      CHECK(storage->searchArchives() == vector_t{});
    }
    GIVEN("file with unknown name format") {
      REQUIRE_CALL(system, lsRegularFiles(_, ".rrd"s)).RETURN(std::vector{"/foo/futre-metering-system-3000.rrd"s});
      CHECK(storage->searchArchives() == vector_t{});
    }
  }
}

} // namespace
