/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "controller.h"

#include <ds/catch/catch.h>
#include <ds/system-mock.h>

#include <climits>
#include <trompeloeil.hpp>

#include "metering/archives-connection-mock.h"
#include "metering/archives-storage-mock.h"

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;
using namespace dss::metering::archives;

static ds::LogChannel dsLogChannel("dssMeteringControllerArchiveTest");

namespace {

static const char* TAGS = "[dSS][metering][controllerArchive]";

auto mockPath = [](const auto& archiveId) { return ds::str("/storage-location/", archiveId, ".rrd"s); };

SCENARIO("ControllerArchives", TAGS) {
  auto archivesConnection = ArchivesConnectionMock();
  auto archivesStorage = ArchivesStorageMock();
  auto system = ds::SystemMock();
  auto rrdInterface = RRDInterfaceMock();

  auto controllerArchives = makeControllerArchive(archivesConnection, archivesStorage, system);

  ALLOW_CALL(archivesConnection, rrdInterface()).LR_RETURN(rrdInterface);

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto toFilename = [controllerId1](Category category) { return mockPath(ArchiveId{controllerId1, category}); };

  WHEN("create consumption") {
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
        .RETURN(mockPath(_1));
    REQUIRE_CALL(rrdInterface, create(_, 1, _,
                                      std::vector{"DS:power:GAUGE:5:0:8000"s, "DS:energy:DERIVE:5:0:U"s,
                                                  "RRA:AVERAGE:0.5:1:600"s, "RRA:AVERAGE:0.5:60:720"s,
                                                  "RRA:AVERAGE:0.5:900:2976"s, "RRA:AVERAGE:0.5:86400:370"s,
                                                  "RRA:AVERAGE:0.5:604800:260"s, "RRA:AVERAGE:0.5:2592000:60"s}));
    controllerArchives->create({Dsuid{controllerId1}, Category::consumption});
  }

  WHEN("create production") {
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::production})).RETURN(mockPath(_1));
    REQUIRE_CALL(rrdInterface, create(_, 1, _,
                                      std::vector{"DS:power:GAUGE:5:0:50000"s, "DS:energy:DERIVE:5:0:U"s,
                                                  "RRA:AVERAGE:0.5:1:600"s, "RRA:AVERAGE:0.5:60:720"s,
                                                  "RRA:AVERAGE:0.5:900:2976"s, "RRA:AVERAGE:0.5:86400:370"s,
                                                  "RRA:AVERAGE:0.5:604800:260"s, "RRA:AVERAGE:0.5:2592000:60"s}));
    controllerArchives->create({Dsuid{controllerId1}, Category::production});
  }

  GIVEN("invalid rrd meta") {
    THEN("validateDBSeries fails") {
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
          .RETURN(mockPath(_1));
      REQUIRE_CALL(rrdInterface, info("/storage-location/(1111111111111111111111111111111100,consumption).rrd"s))
          .RETURN(std::vector<RRDInfo>{});
      bool updatePowerSetting = false;
      CHECK(!controllerArchives->validateDBSeries(ArchiveId{controllerId1, Category::consumption}, updatePowerSetting));
    }
  }

  GIVEN("rrd meta match spec") {
    auto createRRDInfo = [] {
      return std::vector<RRDInfo>{{.key = "ds[power].index"s, .type = RrdInfoType::counter, .value = counter_t{0}},
                                  {.key = "ds[energy].index"s, .type = RrdInfoType::counter, .value = counter_t{1}},
                                  {.key = "rra[0].rows"s, .type = RrdInfoType::counter, .value = counter_t{600}},
                                  {.key = "rra[1].rows"s, .type = RrdInfoType::counter, .value = counter_t{720}},
                                  {.key = "rra[2].rows"s, .type = RrdInfoType::counter, .value = counter_t{2976}},
                                  {.key = "rra[3].rows"s, .type = RrdInfoType::counter, .value = counter_t{370}},
                                  {.key = "rra[4].rows"s, .type = RrdInfoType::counter, .value = counter_t{260}},
                                  {.key = "rra[5].rows"s, .type = RrdInfoType::counter, .value = counter_t{60}}};
    };

    THEN("validateDBSeries succeeds") {
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
          .RETURN(mockPath(_1));
      REQUIRE_CALL(rrdInterface, info("/storage-location/(1111111111111111111111111111111100,consumption).rrd"s))
          .RETURN(createRRDInfo());
      bool updatePowerSetting = false;
      CHECK(controllerArchives->validateDBSeries(ArchiveId{controllerId1, Category::consumption}, updatePowerSetting));
      CHECK(!updatePowerSetting);
    }
    WHEN("power.max == legacy limit value") {
      auto y = GENERATE(Category::consumption, Category::production);
      auto x = GENERATE(4.5e3, // value used in first implementation, only for consumption
                        4e4    // 36978e9a08, no reason given
      );
      THEN("database power.max needs updating") {
        INFO(ds::str(x));
        REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, y})).RETURN(mockPath(_1));
        REQUIRE_CALL(rrdInterface, info(toFilename(y))).RETURN([x, createRRDInfo]() {
          auto info = createRRDInfo();
          info.emplace_back(
              RRDInfo{.key = "ds[power].max"s, .type = RrdInfoType::value, .value = gsl::narrow<rrd_value_t>(x)});
          return info;
        }());

        bool updatePowerSetting = false;
        CHECK(controllerArchives->validateDBSeries(ArchiveId{controllerId1, y}, updatePowerSetting));
        CHECK(updatePowerSetting);
      }
    }
    WHEN("power.max is correct") {
      struct X {
        Category category;
        int limit;
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{
          {Category::consumption, 8000},
          {Category::production, 50'000},
      }));

      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, x.category})).RETURN(mockPath(_1));
      REQUIRE_CALL(rrdInterface, info(toFilename(x.category))).RETURN([x, createRRDInfo]() {
        auto info = createRRDInfo();
        info.emplace_back(
            RRDInfo{.key = "ds[power].max"s, .type = RrdInfoType::value, .value = gsl::narrow<rrd_value_t>(x.limit)});
        return info;
      }());

      bool updatePowerSetting = false;
      CHECK(controllerArchives->validateDBSeries(ArchiveId{controllerId1, x.category}, updatePowerSetting));
      CHECK(!updatePowerSetting);
    }
  }

  SECTION("tune max power") {
    WHEN("consumption archive") {
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
          .RETURN(mockPath(_1));
      REQUIRE_CALL(rrdInterface,
                   tune(std::vector{"tune"s, "/storage-location/(1111111111111111111111111111111100,consumption).rrd"s,
                                    "--maximum"s, "power:8000"s}));
      controllerArchives->tuneDBPowerSettings(ArchiveId{controllerId1, Category::consumption});
    }
    WHEN("production archive") {
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::production}))
          .RETURN(mockPath(_1));
      REQUIRE_CALL(rrdInterface,
                   tune(std::vector{"tune"s, "/storage-location/(1111111111111111111111111111111100,production).rrd"s,
                                    "--maximum"s, "power:50000"s}));
      controllerArchives->tuneDBPowerSettings(ArchiveId{controllerId1, Category::production});
    }
  }

  WHEN("lastValue") {
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
        .RETURN(mockPath(_1));
    REQUIRE_CALL(rrdInterface, lastValue(_)).RETURN(std::make_pair(0, std::vector{std::make_pair("energy"s, 0.0)}));
    controllerArchives->lastValue({Dsuid{controllerId1}, Category::consumption});
  }

  SECTION("xport per category") {
    auto category = GENERATE(Category::consumption, Category::production);
    REQUIRE_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, category})).TIMES(1, 2).RETURN(mockPath(_1));
    WHEN("power") {
      WHEN("single dsm") {
        REQUIRE_CALL(rrdInterface,
                     xport(std::vector{"xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s,
                                       "3000"s, ds::str("DEF:raw0=", toFilename(category), ":power:AVERAGE"s),
                                       "CDEF:data0=raw0,UN,0,raw0,IF"s, "CDEF:sum=data0"s, "XPORT:sum"s}))
            .RETURN([]() -> ValueRange {
              return {1, 601, 1, std::vector<Value>{600, Value{1, ds::DateTime{1s}}}};
            }());
        auto range =
            controllerArchives->xport(1, 601, 1, {ArchiveId{controllerId1, category}}, SeriesTypes::power, false);
        CHECK(range.values.size() == 600);
      }
      WHEN("2 dsm queried together") {
        REQUIRE_CALL(rrdInterface,
                     xport(std::vector{"xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s,
                                       "3000"s, ds::str("DEF:raw0=", toFilename(category), ":power:AVERAGE"s),
                                       "CDEF:data0=raw0,UN,0,raw0,IF"s,
                                       ds::str("DEF:raw1=", toFilename(category), ":power:AVERAGE"s),
                                       "CDEF:data1=raw1,UN,0,raw1,IF"s, "CDEF:sum=data0,data1,+"s, "XPORT:sum"s}))
            .RETURN([]() -> ValueRange {
              return {1, 601, 1, std::vector<Value>{600, Value{}}};
            }());

        auto range = controllerArchives->xport(1, 601, 1,
                                               {ArchiveId{controllerId1, category}, ArchiveId{controllerId1, category}},
                                               SeriesTypes::power, false);
        CHECK(range.values.size() == 600); // results are not modified
      }
    }
    WHEN("energy") {
      WHEN("single dsm") {
        REQUIRE_CALL(rrdInterface,
                     xport(std::vector{"xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s,
                                       "3000"s, ds::str("DEF:raw0=", toFilename(category), ":energy:AVERAGE"s),
                                       "CDEF:data0=raw0,UN,0,raw0,IF"s, "CDEF:sum=data0"s,
                                       ds::str("CDEF:ratePerStep=sum,1,*"), "XPORT:ratePerStep"s}))
            .RETURN(ValueRange{});
        auto range =
            controllerArchives->xport(1, 601, 1, {ArchiveId{controllerId1, category}}, SeriesTypes::energy, false);
      }
      WHEN("2 dsm queried together") {
        REQUIRE_CALL(
            rrdInterface,
            xport(std::vector{
                "xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s, "3000"s,
                ds::str("DEF:raw0=", toFilename(category), ":energy:AVERAGE"s), "CDEF:data0=raw0,UN,0,raw0,IF"s,
                ds::str("DEF:raw1=", toFilename(category), ":energy:AVERAGE"s), "CDEF:data1=raw1,UN,0,raw1,IF"s,
                "CDEF:sum=data0,data1,+"s, ds::str("CDEF:ratePerStep=sum,1,*"), "XPORT:ratePerStep"s}))
            .RETURN([]() -> ValueRange {
              return {1, 601, 1, std::vector<Value>{600, Value{}}};
            }());

        auto range = controllerArchives->xport(1, 601, 1,
                                               {ArchiveId{controllerId1, category}, ArchiveId{controllerId1, category}},
                                               SeriesTypes::energy, false);
        CHECK(range.values.size() == 600); // results are not modified
      }
    }
  }

  SECTION("xport") {
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, Category::consumption})).RETURN(mockPath(_1));
    REQUIRE_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
    WHEN("non-fininte floating point returned") {
      const auto nonfinite = GENERATE(std::nan(""), std::numeric_limits<double>::infinity(), -INFINITY);
      REQUIRE_CALL(rrdInterface, xport(_))
          .RETURN(ValueRange{1, 61, 60, std::vector<Value>{{nonfinite, ds::DateTime{60s}}}});
      auto timeseries = controllerArchives->xport(1, 61, 60, {ArchiveId{controllerId1, Category::consumption}},
                                                  SeriesTypes::energy, false);
      THEN("values are mapped to 0") {
        CHECK(std::isfinite(timeseries.values[0].value));
        CHECK(timeseries.values[0].value == 0);
      }
    }
  }

  WHEN("update") {
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, Category::consumption}))
        .TIMES(1, 2)
        .RETURN(mockPath(_1));
    ALLOW_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
    REQUIRE_CALL(rrdInterface, update(std::vector{
                                   "update"s, "/storage-location/(1111111111111111111111111111111100,consumption).rrd"s,
                                   "100:75:300"s}));
    controllerArchives->addValue({Dsuid{controllerId1}, Category::consumption},
                                 {.power = 75.0, .energy = 300.0, .lastChanged = ds::DateTime(100s)});

    WHEN("using floating points") {
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, Category::consumption})).RETURN(mockPath(_1));
      THEN("energy values are rounded up (std::ceil)") {
        REQUIRE_CALL(
            rrdInterface,
            update(std::vector{"update"s, "/storage-location/(1111111111111111111111111111111100,consumption).rrd"s,
                               "100:75.1:301"s}));
        controllerArchives->addValue({Dsuid{controllerId1}, Category::consumption},
                                     {.power = 75.1, .energy = 300.3, .lastChanged = ds::DateTime(100s)});
      }
    }
    WHEN("when limitPower of the db is exceeded") {
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
          .RETURN(mockPath(_1));
      THEN("value is passed to the db") {
        REQUIRE_CALL(
            rrdInterface,
            update(std::vector{"update"s, "/storage-location/(1111111111111111111111111111111100,consumption).rrd"s,
                               "100:80000:301"s}));
        controllerArchives->addValue({Dsuid{controllerId1}, Category::consumption},
                                     {.power = 80000, .energy = 300.3, .lastChanged = ds::DateTime(100s)});
      }
    }
  }

  WHEN("energy is UINT_MAX") {
    REQUIRE_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
        .RETURN(mockPath(_1));
    THEN("energy is added to db") {
      REQUIRE_CALL(
          rrdInterface,
          update(std::vector{"update"s, "/storage-location/(1111111111111111111111111111111100,consumption).rrd"s,
                             "100:75.1:4294967295"s}));
      controllerArchives->addValue(
          {Dsuid{controllerId1}, Category::consumption},
          {.power = 75.1, .energy = static_cast<double>(UINT_MAX), .lastChanged = ds::DateTime(100s)});
    }
    WHEN("energy exceeds UINT_MAX + 1") {
      THEN("no limit is applied to power") {
        REQUIRE_CALL(
            rrdInterface,
            update(std::vector{"update"s, "/storage-location/(1111111111111111111111111111111100,consumption).rrd"s,
                               "100:75.1:4294967296"s}));
        controllerArchives->addValue(
            {Dsuid{controllerId1}, Category::consumption},
            {.power = 75.1, .energy = static_cast<double>(UINT_MAX) + 1, .lastChanged = ds::DateTime(100s)});
      }
    }
  }

  SECTION("lastTimestamp") {
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
        .RETURN(mockPath(_1));
    REQUIRE_CALL(archivesConnection,
                 lastTimestamp("/storage-location/(1111111111111111111111111111111100,consumption).rrd"s))
        .RETURN(ds::DateTime{10s});
    CHECK(controllerArchives->lastTimestamp({Dsuid{controllerId1}, Category::consumption}) == ds::DateTime{10s});
  }

  WHEN("flush") {
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, Category::consumption})).RETURN(mockPath(_1));
    REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, Category::production})).RETURN(mockPath(_1));
    REQUIRE_CALL(archivesConnection,
                 flush(std::vector{"/storage-location/(1111111111111111111111111111111100,consumption).rrd"s,
                                   "/storage-location/(1111111111111111111111111111111100,production).rrd"s}));
    controllerArchives->flush(
        std::vector{ArchiveId{controllerId1, Category::consumption}, ArchiveId{controllerId1, Category::production}});
  }
}

} // namespace
