/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "modbusmeteringsystem.h"

#include <ds/conf.h>
#include <ds/log.h>
#include <ds/system.h>

#include <regex>

#include "metering/archives-connection.h"
#include "metering/archives-storage.h"
#include "metering/config.h"
#include "metering/rrd-interface.h"

static ds::LogChannel dsLogChannel("dssMeteringModbusArchives");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

//resolution, max power
constexpr int maxPower = 13860; //63A, 220V

const std::unordered_map<int, double> maxEnergyLookupWs = {{60, maxPower * 60.0},
                                                        {900, maxPower * 900.0},
                                                        {86400, maxPower * 86400.0},
                                                        {604800, maxPower * 604800.0}};

namespace dss::metering::archives {

namespace {

class ModbusMeteringArchivesImpl final : public ModbusMeteringArchives {
  ds::System& m_system;
  ArchivesConnection& m_archivesConnection;
  ArchivesStorage& m_archivesStorage;
  const ElectricityMeteringConfig m_electricityMeteringConfig;
  mutable std::unordered_map<int, int> m_nonFiniteValueHistogram{}; // rate -> count

public:
  ModbusMeteringArchivesImpl(ArchivesConnection& archivesConnection, ArchivesStorage& archivesStorage, ds::System& system)
      : m_system(system),
        m_archivesConnection(archivesConnection),
        m_archivesStorage(archivesStorage),
        m_electricityMeteringConfig(m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {})) {}
  virtual ~ModbusMeteringArchivesImpl() override = default;

  void create(const ModbusMeteringArchiveId& modbusMeteringId) override;
  void addValue(const ModbusMeteringArchiveId& modbusMeteringId, ArchiveValue&& value) override;

  std::optional<ArchiveValue> lastValue(const ModbusMeteringArchiveId& modbusMeteringId) const override;

  ValueRange xport(time_t start, time_t end, int step, std::vector<ModbusMeteringArchiveId> modbusMeteringIds, SeriesTypes type,
                   bool energyInWh) const override;

  bool validateDBSeries(const ModbusMeteringArchiveId& modbusMeteringId, bool& _tunePowerMaxSetting) override;
  void tuneDBPowerSettings(const ModbusMeteringArchiveId& modbusMeteringId) override;

  std::optional<ds::DateTime> lastTimestamp(ModbusMeteringArchiveId modbusMeteringId) const override {
    return m_archivesConnection.lastTimestamp(m_archivesStorage.filepath(modbusMeteringId));
  }

  void flush(std::vector<ModbusMeteringArchiveId> modbusMeteringIds) const override {
    std::vector<std::string> files;
    std::transform(std::cbegin(modbusMeteringIds), std::cend(modbusMeteringIds), std::back_inserter(files),
                   [&](const auto elt) { return m_archivesStorage.filepath(elt); });
    m_archivesConnection.flush(files);
  }

private:
  auto limitPower(const ModbusMeteringArchiveId& modbusMeteringId) const {
    if (modbusMeteringId.second == Category::consumption) {
      return m_electricityMeteringConfig.discontinuityDetection.maxConsumption;
    } else {
      return m_electricityMeteringConfig.discontinuityDetection.maxProduction;
    }
  }
  auto limitEnergy(const ModbusMeteringArchiveId& modbusMeteringId) const {
    return limitPower(modbusMeteringId); // Ws value same as W
  }
};

void ModbusMeteringArchivesImpl::create(const ModbusMeteringArchiveId& modbusMeteringId) {
  DS_DEBUG("Creating new RRD database", modbusMeteringId);

  /* create new DB */
  std::vector<std::string> lines;
  lines.push_back(
      ds::str("DS:power:GAUGE:20:0:", limitPower(modbusMeteringId))); // ignored by unit tests on host, but enforced on platform
  lines.push_back("DS:energy:DERIVE:20:0:U");

  for (const auto& archiveConfig : m_electricityMeteringConfig.timeSeries) {
    std::stringstream sstream;
    sstream << "RRA:AVERAGE:0.5:" << std::chrono::seconds(archiveConfig.resolution).count() << ":"
            << archiveConfig.numberOfValues;
    lines.push_back(sstream.str());
  }

  auto iCurrentTimeStamp = std::chrono::system_clock::now();
  time_t currentInSecs = std::chrono::system_clock::to_time_t(iCurrentTimeStamp);
  if (currentInSecs > 10) { // huh??
    currentInSecs -= 10;
  }

  m_archivesConnection.rrdInterface().create(m_archivesStorage.filepath(modbusMeteringId), 1, currentInSecs, lines);
}

void ModbusMeteringArchivesImpl::addValue(const ModbusMeteringArchiveId& modbusMeteringId, ArchiveValue&& value) {
  DS_CONTEXT(ds::hideArg(modbusMeteringId));
  DS_REQUIRE(modbusMeteringId != ModbusMeteringArchiveId{}, "addValue invalid archive descriptor", modbusMeteringId);

  std::vector<std::string> lines;
  lines.push_back("update");
  if (auto rrdCachedPath = m_archivesConnection.rrdCachedPath(); !rrdCachedPath.empty()) {
    lines.push_back("--daemon");
    lines.push_back(rrdCachedPath);
  }
  lines.push_back(m_archivesStorage.filepath(modbusMeteringId));
  // ds::energy:DERIVE -- only accepts integer values
  lines.push_back(ds::str(std::chrono::system_clock::to_time_t(value.lastChanged), ":", value.power, ":",
                          gsl::narrow_cast<uint64_t>(std::ceil(value.energy))));

  DS_DEBUG(lines);
  m_archivesConnection.rrdInterface().update(lines);
}

std::optional<ArchiveValue> ModbusMeteringArchivesImpl::lastValue(const ModbusMeteringArchiveId& modbusMeteringId) const {
  try {
    auto [lastChanged, data] = m_archivesConnection.rrdInterface().lastValue(m_archivesStorage.filepath(modbusMeteringId));

    auto ret = ArchiveValue{.lastChanged = ds::DateTime{std::chrono::seconds{lastChanged}}};
    for (const auto& [name, value] : data) {
      if (name == "energy") {
        ret.energy = value;
      } else if (name == "power") {
        ret.power = value;
      }
    }
    // TODO rely on validateDBSeries to verify that energy and power data source is present
    return ret;

  } catch (const std::exception& e) {
    // probably unknown value in db, conversion to double failed
    DS_NOTICE("tryCreateCacheFromDB", e);
  }

  return std::nullopt;
}

ValueRange ModbusMeteringArchivesImpl::xport(time_t start, time_t end, int step, std::vector<ModbusMeteringArchiveId> modbusMeteringIds,
                                         SeriesTypes type, bool energyInWh) const {
  DS_DEBUG_ENTER();
  std::vector<std::string> lines;
  lines.push_back("xport");
  if (auto rrdCachedPath = m_archivesConnection.rrdCachedPath(); !rrdCachedPath.empty()) {
    lines.push_back("--daemon");
    lines.push_back(rrdCachedPath);
  }
  lines.push_back("--start");
  lines.push_back(ds::str(start));
  lines.push_back("--end");
  lines.push_back(ds::str(end));
  lines.push_back("--step");
  lines.push_back(ds::str(step));
  lines.push_back("--maxrows");
  lines.push_back("3000");
  {
    std::stringstream sstream;
    for (size_t i = 0; i < modbusMeteringIds.size(); ++i) {
      auto modbusMeteringId = modbusMeteringIds[i];
      sstream << "DEF:raw" << i << "=" << m_archivesStorage.filepath(modbusMeteringId);
      if ((step == 1) && ((type == SeriesTypes::power))) {
        sstream << ":power";
      } else {
        sstream << ":energy";
      }

      sstream << ":AVERAGE";
      lines.push_back(sstream.str());

      sstream.str(std::string());
      sstream << "CDEF:data" << i << "=raw" << i << ",UN,0,raw" << i << ",IF";
      lines.push_back(sstream.str());
      sstream.str(std::string());
    }
    sstream << "CDEF:sum=data0";
    for (size_t i = 1; i < modbusMeteringIds.size(); ++i) {
      sstream << ",data" << i << ",+";
    }
    lines.push_back(sstream.str());
  }
  switch (type) {
    case SeriesTypes::energy:
    case SeriesTypes::energyDelta: {
      lines.push_back(ds::str("CDEF:ratePerStep=sum,", step, ",*"));
      if (energyInWh) {
        lines.push_back("CDEF:adj=ratePerStep,3600,/");
        lines.push_back("XPORT:adj");
      } else {
        lines.push_back("XPORT:ratePerStep");
      }
      break;
    }
    case SeriesTypes::power:
    default: {
      lines.push_back("XPORT:sum");
      break;
    }
  }
  DS_DEBUG(lines);

  auto results = m_archivesConnection.rrdInterface().xport(lines);
  // DS_REQUIRE(start, end, step) -> move to rrd-interface

  for (auto& [value, timestamp] : results.values) {
    if (std::isfinite(value)) {
      continue;
    }
    auto total = m_nonFiniteValueHistogram[step]++;
    DS_NOTICE("NaN/Infinity value detected", timestamp, value, step, total);
    if (type == SeriesTypes::power) {
      if (value > maxPower) {
        DS_INFO("Trimming invalid value", timestamp, value, step, total);
        value = 0.0;
      }
    } else {
      double max = maxEnergyLookupWs.at(step);
      if (value > max) {
        DS_INFO("Trimming invalid value", timestamp, value, step, total);
        value = 0.0;
      }
    }
    value = 0.0;
  }

  DS_REQUIRE(std::none_of(std::cbegin(results.values), std::cend(results.values),
                          [](const auto& data) { return !std::isfinite(data.value); }),
             "values contain non-finite values");

  return results;
}

bool ModbusMeteringArchivesImpl::validateDBSeries(const ModbusMeteringArchiveId& modbusMeteringId, bool& _tunePowerMaxSetting) {
  static const auto rraIndexRegex = std::regex(R"(rra\[(\d+)\].rows)");
  unsigned long rrdMatchCount = 0;

  auto seriesConfig = m_electricityMeteringConfig.timeSeries;

  for (auto& rrdInfo : m_archivesConnection.rrdInterface().info(m_archivesStorage.filepath(modbusMeteringId))) {
    DS_DEBUG(rrdInfo);
    if ((rrdInfo.key == "ds[power].index" && (rrdInfo.type == RrdInfoType::counter) &&
         (std::get<counter_t>(rrdInfo.value) == 0)) ||
        (rrdInfo.key == "ds[energy].index" && (rrdInfo.type == RrdInfoType::counter) &&
         (std::get<counter_t>(rrdInfo.value) == 1))) {
      rrdMatchCount++;
    } else {
      std::smatch match;
      if (std::regex_search(rrdInfo.key, match, rraIndexRegex)) {
        auto index = std::stoul(match[1]);
        DS_REQUIRE(index < seriesConfig.size(), "RRA index exceeded");
        if (std::get<counter_t>(rrdInfo.value) == seriesConfig[index].numberOfValues) {
          rrdMatchCount++;
        }
      }
    }

    if (rrdInfo.key == "ds[power].max") {
      DS_DEBUG(rrdInfo);
      if ((rrdInfo.type == RrdInfoType::value) && (std::get<rrd_value_t>(rrdInfo.value) != limitPower(modbusMeteringId))) {
        DS_NOTICE("validateDBSeries: invalid upper power limit", rrdInfo, modbusMeteringId);
        _tunePowerMaxSetting = true;
      }
    }
  }

  DS_DEBUG("RRD MatchCount", rrdMatchCount, _tunePowerMaxSetting);
  return (rrdMatchCount == (2 + seriesConfig.size()));
}

void ModbusMeteringArchivesImpl::tuneDBPowerSettings(const ModbusMeteringArchiveId& modbusMeteringId) {
  DS_WARNING("tuning max acceptable power value in RRD", modbusMeteringId);
  std::vector<std::string> lines;
  lines.push_back("tune");
  lines.push_back(m_archivesStorage.filepath(modbusMeteringId));
  lines.push_back("--maximum");
  lines.push_back(ds::str("power:", limitPower(modbusMeteringId)));

  m_archivesConnection.rrdInterface().tune(lines);
}

} // namespace

std::unique_ptr<ModbusMeteringArchives> makeModbusArchive(ArchivesConnection& archivesConnection,
                                                          ArchivesStorage& archivesStorage, ds::System& system) {
  return std::make_unique<ModbusMeteringArchivesImpl>(archivesConnection, archivesStorage, system);
};

} // namespace dss::metering::archives
