/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "open-metering-system.h"

#include <ds/catch/catch.h>
#include <ds/date-time.h>
#include <ds/log.h>
#include <ds/system-mock.h>

#include <algorithm>
#include <filesystem>
#include <vector>

#include "metering/archives-connection-mock.h"
#include "metering/archives-storage-mock.h"
#include "metering/metering.h"
#include "metering/readings/open-metering-system.h"

static ds::LogChannel dsLogChannel("dssMeteringOpenMeteringSystemArchiveTest");

namespace {

using namespace trompeloeil;
using namespace std::literals;

using namespace dss::metering;
using namespace dss::metering::archives;

static const char* TAGS = "[dSS][metering][open-metering-system]";

using system_clock = std::chrono::system_clock;
using Reading = readings::OpenMeteringSystem;

auto mockPath = [](const auto& archiveId) { return ds::str("/storage-location/", archiveId, ".rrd"s); };

SCENARIO("openMeteringSystemArchive", TAGS) {
  auto system = ds::SystemMock();
  auto archivesConnection = ArchivesConnectionMock{};
  auto archivesStorage = ArchivesStorageMock{};

  auto omsArchive = makeOpenMeteringSystemArchive(archivesConnection, archivesStorage, system);

  auto rrdInterface = RRDInterfaceMock();
  ALLOW_CALL(archivesConnection, rrdInterface()).LR_RETURN(rrdInterface);

  auto openMeteringSystemId = OpenMeteringSystemId{"3153414733313030323938313039"s};

  SECTION("create") {
    REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(mockPath(_1));
    REQUIRE_CALL(
        rrdInterface,
        create(
            _, 1, _,
            std::vector{"DS:powerConsumed:GAUGE:20:0:U"s, "DS:powerProduced:GAUGE:20:0:U"s,
                        "DS:energyConsumedTari1:DCOUNTER:30:U:U"s, "DS:energyProducedTari1:DCOUNTER:30:U:U"s,
                        "DS:energyConsumedTari2:DCOUNTER:30:U:U"s, "DS:energyProducedTari2:DCOUNTER:30:U:U"s,
                        "RRA:AVERAGE:0.5:1:600"s, "RRA:AVERAGE:0.5:60:720"s, "RRA:AVERAGE:0.5:900:2976"s,
                        "RRA:AVERAGE:0.5:86400:370"s, "RRA:AVERAGE:0.5:604800:260"s, "RRA:AVERAGE:0.5:2592000:60"s}));

    omsArchive->create(openMeteringSystemId);
  }

  SECTION("add") {
    ALLOW_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
    REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(mockPath(_1));
    REQUIRE_CALL(rrdInterface, update(std::vector{"update"s, "/storage-location/3153414733313030323938313039.rrd"s,
                                                  "301:11:10:5:6:12:0"s}));
    omsArchive->addValue(openMeteringSystemId, {.lastChanged = ds::DateTime{301s},
                                                .powerConsumed = 11,
                                                .powerProduced = 10,
                                                .t1EnergyConsumed = 5,
                                                .t1EnergyProduced = 6,
                                                .t2EnergyConsumed = 12,
                                                .t2EnergyProduced = 0});
  }

  SECTION("lastValue/lastTimetstamp") {
    WHEN("archive exists") {
      ALLOW_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
      REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(mockPath(_1));

      THEN("lastValue returns valid data") {
        REQUIRE_CALL(rrdInterface, lastValue(mockPath(openMeteringSystemId)))
            .RETURN(std::make_pair(
                301, std::vector{
                         std::make_pair("powerConsumed"s, 11.0), std::make_pair("powerProduced"s, 10.0),
                         std::make_pair("energyConsumedTari1"s, 5.0), std::make_pair("energyProducedTari1"s, 6.0),
                         std::make_pair("energyConsumedTari2"s, 12.0), std::make_pair("energyProducedTari2"s, 1.0)}));
        CHECK(omsArchive->lastValue(openMeteringSystemId) == Reading{{}, ds::DateTime{301s}, 11, 10, 5, 6, 12, 1});
      }
      THEN("lastTimestamp is valid") {
        REQUIRE_CALL(archivesConnection, lastTimestamp(mockPath(openMeteringSystemId)))
            .RETURN(std::optional<ds::DateTime>(301s));
        CHECK(omsArchive->lastTimestamp(openMeteringSystemId) == ds::DateTime{301s});
      }
    }
    WHEN("no archive") {
      ALLOW_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
      REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(mockPath(_1));
      THEN("lastValue is nullopt") {
        // rrd returns UNKNOWNs, failure happens when converting them to double
        REQUIRE_CALL(rrdInterface, lastValue(mockPath(openMeteringSystemId)))
            .THROW(std::runtime_error("conversion error"));
        CHECK(omsArchive->lastValue(openMeteringSystemId) == std::nullopt);
      }
      THEN("lastTimestamp is nullopt") {
        REQUIRE_CALL(archivesConnection, lastTimestamp(mockPath(openMeteringSystemId))).RETURN(std::nullopt);
        CHECK(omsArchive->lastTimestamp(openMeteringSystemId) == std::nullopt);
      }
    }
  }

  SECTION("getSeries") {
    auto x = GENERATE(table<Category, SeriesTypes, std::vector<std::string>>(
        {{Category::consumption, SeriesTypes::energy,
          std::vector{"DEF:raw0=/storage-location/3153414733313030323938313039.rrd:energyConsumedTari1:AVERAGE"s,
                      "CDEF:data0=raw0,UN,0,raw0,IF"s,
                      "DEF:raw1=/storage-location/3153414733313030323938313039.rrd:energyConsumedTari2:AVERAGE"s,
                      "CDEF:data1=raw1,UN,0,raw1,IF"s, "CDEF:sum=data0,data1,+"s, "CDEF:ratePerStep=sum,1,*"s,
                      "CDEF:adj=ratePerStep,3600,/"s, "XPORT:adj"s}},
         {Category::production, SeriesTypes::power,
          std::vector{"DEF:raw0=/storage-location/3153414733313030323938313039.rrd:powerProduced:AVERAGE"s,
                      "CDEF:data0=raw0,UN,0,raw0,IF"s, "CDEF:sum=data0"s, "XPORT:sum"s}}}));

    INFO(x);
    ALLOW_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
    REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).TIMES(1, 2).RETURN(mockPath(_1));

    auto command =
        std::vector<std::string>{"xport"s, "--start"s, "0"s, "--end"s, "0"s, "--step"s, "1"s, "--maxrows"s, "3000"s};

    const auto& [category, seriesType, seriesCommand] = x;
    std::copy(std::cbegin(seriesCommand), std::cend(seriesCommand), std::back_inserter(command));

    REQUIRE_CALL(rrdInterface, xport(command)).RETURN(ValueRange{});
    omsArchive->getSeries(openMeteringSystemId, category, seriesType, 0, 0, 1, EnergyUnit::Wh);
  }

  SECTION("validation") { CHECK(omsArchive->validate(openMeteringSystemId) == ValidationResult::ok); }
  SECTION("migration") { CHECK_NOTHROW(omsArchive->migrate(openMeteringSystemId)); }
}

SCENARIO("openMeteringSystemArchive integration test", TAGS) {
  auto system = ds::SystemMock();
  auto archivesConnection = ArchivesConnectionMock{};
  auto archivesStorage = ArchivesStorageMock{};

  auto omsArchive = makeOpenMeteringSystemArchive(archivesConnection, archivesStorage, system);

  auto rrdInterface = makeRRDInterface();
  ALLOW_CALL(archivesConnection, rrdInterface()).LR_RETURN(*rrdInterface);

  auto filename = "oms-system-test.rrd";
  DS_DEFER(std::filesystem::remove(filename));

  auto openMeteringSystemId = OpenMeteringSystemId{"3153414733313030323938313039"s};

  GIVEN("rrd db") {
    ALLOW_CALL(archivesConnection, rrdCachedPath()).RETURN(""s);
    REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(filename);

    auto startTime = ds::DateTime{1658000000s};
    omsArchive->create(openMeteringSystemId, startTime - 10s);

    auto floorSeconds = [](auto timestamp) { return std::chrono::floor<std::chrono::seconds>(timestamp); };

    WHEN("no archive") {
      THEN("lastValue fails") {
        REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(filename);
        // integration test: file does not exist! -> warning in log message, about conversion failure
        CHECK(omsArchive->lastValue(openMeteringSystemId) == std::nullopt);
      }
    }

    WHEN("add value") {
      REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(filename);
      auto timestamp = startTime + 1s;
      omsArchive->addValue(openMeteringSystemId, {.lastChanged = timestamp,
                                                  .powerConsumed = 11,
                                                  .powerProduced = 10,
                                                  .t1EnergyConsumed = 5,
                                                  .t1EnergyProduced = 6,
                                                  .t2EnergyConsumed = 12,
                                                  .t2EnergyProduced = 0});
      THEN("value is read back") {
        REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(filename);
        CHECK(omsArchive->lastValue(openMeteringSystemId) == Reading{.lastChanged = floorSeconds(timestamp),
                                                                     .powerConsumed = 11,
                                                                     .powerProduced = 10,
                                                                     .t1EnergyConsumed = 5,
                                                                     .t1EnergyProduced = 6,
                                                                     .t2EnergyConsumed = 12,
                                                                     .t2EnergyProduced = 0});
      }
    }

    WHEN("adding multiple values") {
      ALLOW_CALL(archivesStorage, filepath(openMeteringSystemId)).RETURN(filename);

      auto endTime = startTime + 2min;
      for (auto cur = startTime; cur <= endTime; cur += 1s) {
        auto index = std::chrono::duration_cast<std::chrono::seconds>(cur - startTime).count();
        omsArchive->addValue(openMeteringSystemId, {.lastChanged = cur,
                                                    .powerConsumed = 100,
                                                    .powerProduced = 10,
                                                    .t1EnergyConsumed = 500.0 * index,
                                                    .t1EnergyProduced = 50.0 * index,
                                                    .t2EnergyConsumed = 200.0 * index,
                                                    .t2EnergyProduced = 20.0 * index});
      }

      THEN("series returns the same") {
        auto x = GENERATE(table<Category, SeriesTypes, double>({{Category::consumption, SeriesTypes::power, 100.0},
                                                                {Category::production, SeriesTypes::power, 10.0},
                                                                {Category::consumption, SeriesTypes::energy, 700.0},
                                                                {Category::production, SeriesTypes::energy, 70.0}}));

        INFO(x);
        REQUIRE_CALL(archivesStorage, filepath(openMeteringSystemId)).TIMES(1, 2).RETURN(filename);

        const auto& [category, seriesType, value] = x;
        auto series =
            omsArchive->getSeries(openMeteringSystemId, category, seriesType, system_clock::to_time_t(endTime - 2s),
                                  system_clock::to_time_t(endTime), 1, EnergyUnit::Ws);

        DS_INFO(series.values);
        CHECK(series.values == std::vector{Value{value, floorSeconds(endTime - 1s)}});
      }
    }
  }
}

} // namespace
