#include "rrd-interface.h"

#include <ds/enum.h>
#include <ds/log.h>

#include <mutex>

static ds::LogChannel dsLogChannel("dssMeteringRrdInterface");

using namespace std::literals;

namespace dss::metering {

namespace {

class RRDImplemenentation : public RRDInterface {
public:
  ValueRange xport(std::vector<std::string> args) override;
  void create(std::string filename, unsigned long pdpStep, time_t lastUpdate, std::vector<std::string> args) override;
  std::vector<RRDInfo> info(std::string filename) override;
  void tune(std::vector<std::string> arguments) override;
  void update(std::vector<std::string> arguments) override;
  void flush(std::vector<std::string> arguments) override;
  std::pair<time_t, std::vector<std::pair<std::string, double>>> lastValue(const std::string& filename) override;
  std::optional<time_t> lastTimestamp(std::string filename) override;
  mutable std::mutex m_mutex;
};

std::vector<const char*> toArgc(const std::vector<std::string>& arguments) {
  std::vector<const char*> ret;
  std::transform(arguments.begin(), arguments.end(), std::back_inserter(ret), [](const auto& x) { return x.c_str(); });
  return ret;
}

ValueRange RRDImplemenentation::xport(std::vector<std::string> args) {
  DS_DEBUG_ENTER(args);
  auto lock = std::scoped_lock(m_mutex);
  auto starts = toArgc(args);
  char** argString = (char**)&starts.front();

  int result = 0;
  int noOutput = 1;
  unsigned long dscount = 0;
  char** names = 0;
  rrd_value_t* data = 0;
  time_t start{}, end{};
  unsigned long step;

  rrd_clear_error();
  result = rrd_xport(starts.size(), argString, &noOutput, &start, &end, &step, &dscount, &names, &data);
  DS_REQUIRE(result == 0, "Error reading from rrd database", args, rrd_get_error());
  // rrd_lastupdate_r returns allocated memory only if result == 0
  DS_DEFER(for (unsigned int i = 0; i < dscount; ++i) { rrd_freemem(names[i]); } rrd_freemem(names);
           rrd_freemem(data););

  std::vector<Value> values;
  rrd_value_t* currentData = data;
  for (time_t timeStamp = start + step; timeStamp <= (time_t)(end - step); timeStamp += step) {
    values.push_back(Value{*currentData, std::chrono::system_clock::from_time_t(timeStamp)});
    currentData++;
  }

  DS_DEBUG_LEAVE();
  return {.start = start, .end = end, .step = step, .values = values};
}

void RRDImplemenentation::create(std::string filename, unsigned long pdpStep, time_t lastUpdate,
                                 std::vector<std::string> args) {
  DS_DEBUG_ENTER(args);
  auto lock = std::scoped_lock(m_mutex);
  auto starts = toArgc(args);
  const char** argString = &starts.front();

  rrd_clear_error();
  int result = rrd_create_r(filename.c_str(), pdpStep, lastUpdate, starts.size(), argString);
  DS_REQUIRE(result == 0, "Error create rrd database", filename, rrd_get_error());
  DS_DEBUG_LEAVE();
}

std::vector<RRDInfo> RRDImplemenentation::info(std::string filename) {
  DS_DEBUG_ENTER(filename);
  auto lock = std::scoped_lock(m_mutex);
  rrd_clear_error();
  rrd_info_t* rrdInfo = rrd_info_r(filename.c_str());
  DS_REQUIRE(rrdInfo != nullptr, "Error reading info rrd database", filename, rrd_get_error());

  std::vector<RRDInfo> metaInfos;
  rrd_info_t* rrdInfoOrig = rrdInfo;
  while (rrdInfo != NULL) {
    auto convertValue = [type = static_cast<RrdInfoType>(rrdInfo->type)](rrd_infoval_t value) -> RrdInfoValue {
      auto importBlob = [](const rrd_blob_t& blob) {
        return std::vector<unsigned char>{blob.ptr, blob.ptr + blob.size};
      };
      switch (type) {
        case RrdInfoType::value: return value.u_val;
        case RrdInfoType::counter: return value.u_cnt;
        case RrdInfoType::integer: return value.u_int;
        case RrdInfoType::string: return std::string(value.u_str);
        case RrdInfoType::blob: return importBlob(value.u_blo); // untested
        default: break;
      }
      DS_FAIL_REQUIRE("unknown type", type);
    };
    metaInfos.push_back(RRDInfo{.key = std::string(rrdInfo->key),
                                .type = static_cast<RrdInfoType>(rrdInfo->type),
                                .value = convertValue(rrdInfo->value)});
    rrdInfo = rrdInfo->next;
  }
  rrd_info_free(rrdInfoOrig);
  DS_DEBUG_LEAVE();
  return metaInfos;
}

void RRDImplemenentation::tune(std::vector<std::string> arguments) {
  DS_DEBUG_ENTER(arguments);
  auto lock = std::scoped_lock(m_mutex);
  auto starts = toArgc(arguments);
  const char** argString = &starts.front();

  rrd_clear_error();
  // cast-away the const, should be ok according to rrd_tune() sources
  int result = rrd_tune(starts.size(), (char**)argString);
  DS_REQUIRE(result == 0, "Error tune rrd database", arguments, rrd_get_error());
  DS_DEBUG_LEAVE();
}

void RRDImplemenentation::update(std::vector<std::string> arguments) {
  DS_DEBUG_ENTER(arguments);
  auto lock = std::scoped_lock(m_mutex);
  auto starts = toArgc(arguments);
  char** argString = (char**)&starts.front();

  rrd_clear_error();
  int result = rrd_update(starts.size(), argString);
  DS_REQUIRE(result == 0, "Error update rrd database", arguments, rrd_get_error());
  DS_DEBUG_LEAVE();
}

void RRDImplemenentation::flush(std::vector<std::string> arguments) {
  DS_DEBUG_ENTER(arguments);
  auto lock = std::scoped_lock(m_mutex);
  auto starts = toArgc(arguments);
  char** argString = (char**)&starts.front();
  int result = rrd_flushcached(starts.size(), argString);
  DS_REQUIRE(result == 0, "Error flushing cache to rrd database", arguments, rrd_get_error());
  DS_DEBUG_LEAVE();
}

std::pair<time_t, std::vector<std::pair<std::string, double>>> RRDImplemenentation::lastValue(
    const std::string& filename) {
  DS_DEBUG_ENTER(filename);
  auto lock = std::scoped_lock(m_mutex);
  char** names = 0;
  char** data = 0;
  time_t lastUpdate;
  unsigned long count = 0;

  // possible check for file existance or rely on rrd error handling?
  // rrd returns 'U' if file does not exist, resulting in ds::requireDouble to fail

  rrd_clear_error();
  int result = rrd_lastupdate_r(filename.c_str(), &lastUpdate, &count, &names, &data);
  DS_REQUIRE(result == 0, "Error reading last update from rrd database", filename, rrd_get_error());

  // rrd_lastupdate_r returns allocated memory only if result == 0
  DS_DEFER(for (unsigned int i = 0; i < count; ++i) {
    rrd_freemem(names[i]);
    rrd_freemem(data[i]);
  } rrd_freemem(names);
           rrd_freemem(data););

  std::vector<std::pair<std::string, double>> dataElement;
  for (unsigned int i = 0; i < count; ++i) {
    dataElement.push_back(std::make_pair(std::string(names[i]), ds::requireDouble(data[i])));
  }

  DS_DEBUG_LEAVE();
  return std::make_pair(lastUpdate, dataElement);
}

std::optional<time_t> RRDImplemenentation::lastTimestamp(std::string filename) {
  DS_DEBUG_ENTER(filename);
  auto lock = std::scoped_lock(m_mutex);
  rrd_clear_error();
  auto result = rrd_last_r(filename.c_str());

  if (result == static_cast<time_t>(-1)) {
    DS_NOTICE("Failed to reading last timestamp from rrd database", filename, rrd_get_error());
    return {};
  }
  DS_DEBUG_LEAVE();
  return result;
}

} // namespace

std::shared_ptr<RRDInterface> makeRRDInterface() {
  return std::make_shared<RRDImplemenentation>();
}

} // namespace dss::metering
