#include "socket-server.h"

#include <ds/log.h>

#include <boost/asio/local/stream_protocol.hpp>
#include <memory>

#include "metering.h"
#include "power-states.h"
#include "readings/open-metering-system.h"
#include "readings/modbusmeteringsystem.h"

using boost::asio::io_context;
using boost::asio::local::stream_protocol;

using dss::metering::Metering;
using dss::metering::MeteringSystem;

static ds::LogChannel dsLogChannel("dssMeteringSocketServer");

namespace {

using P1Reading = dss::metering::readings::OpenMeteringSystem;
using ModbusReading = dss::metering::readings::ModbusMeteringSystem;

class Session : public std::enable_shared_from_this<Session> {
public:
  Session(stream_protocol::socket&& socket, Metering& metering) : m_socket(std::move(socket)), m_metering(metering){};

  void start() { do_read(); }

  void doParse(int size) {
    DS_DEBUG("raw", std::string_view(m_buffer.data(), size), size);
    try {
      auto tmp = ds::fromJson<ModbusReading>(std::string(m_buffer.data()));
      DS_NOTICE("doParse", ds::toJson(tmp));
      if (tmp.meteringType == "modbus") {
        m_metering.triggerState("modbusMeteringSystem-" + std::string(tmp.id) + "-" + std::to_string(tmp.slaveAddress),
                              {.production = gsl::narrow_cast<unsigned int>(tmp.powerProduced),
                              .consumption = gsl::narrow_cast<unsigned int>(tmp.powerConsumed)});
        m_metering.addReading(std::move(tmp));
      } else {
        auto tmp = ds::fromJson<P1Reading>(std::string(m_buffer.data()));
        DS_NOTICE("doParse", ds::toJson(tmp));
        m_metering.triggerState("openMeteringSystem",
                                {.production = gsl::narrow_cast<unsigned int>(tmp.powerProduced),
                                .consumption = gsl::narrow_cast<unsigned int>(tmp.powerConsumed)});
        m_metering.addReading(std::move(tmp));
      }
      // self->do_read();
    } catch (std::exception& e) {
      DS_NOTICE("doParse", e.what());
    }
  }

  void do_read() {
    m_socket.async_read_some(boost::asio::buffer(m_buffer),
                             [self = shared_from_this()](boost::system::error_code ec, std::size_t size) {
                               if (ec) {
                                 DS_NOTICE("read: error:", ec);
                                 return;
                               }
                               self->doParse(size);
                             });
  }

private:
  stream_protocol::socket m_socket;
  std::array<char, 1024> m_buffer{};
  Metering& m_metering;
};

class SocketServerImpl : public dss::metering::SocketServer {
public:
  SocketServerImpl(io_context& io, const std::string& path, Metering& metering)
      : m_acceptor(io, stream_protocol::endpoint(path)), m_metering(metering), m_path(path) {}

  void start() override { do_accept(); }

private:
  void do_accept() {
    DS_NOTICE("wait for accept", ds::hideArg(m_path));
    m_acceptor.async_accept([this](boost::system::error_code ec, stream_protocol::socket socket) {
      if (!ec) {
        DS_NOTICE("accepted connection");
        std::make_shared<Session>(std::move(socket), m_metering)->start();
      }
      do_accept();
    });
  }

  stream_protocol::acceptor m_acceptor;
  Metering& m_metering;
  std::string m_path{};
};

} // namespace

namespace dss::metering {
std::unique_ptr<SocketServer> makeSocketServer(boost::asio::io_context& io, const std::string& path,
                                               Metering& metering) {
  DS_NOTICE("makeSocketServer", path);
  // TODO: m_system.remove(file);
  std::remove(path.c_str()); // remove dangling socket if process crashed
  return std::make_unique<SocketServerImpl>(io, path, metering);
}
} // namespace dss::metering
