/*
    Copyright (c) 2014 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey Jin Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "model-features.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#include <memory>
#include <stdexcept>

#include "base.h"
#include "model/device.h"
#include "model/modelconst.h"

#define ARRAY_END(mf) (mf + ds::arraySize(mf))

static ds::LogChannel dsLogChannel("modelFeatures");

namespace dss {

namespace {

// model strings
const char *KM220 = "KM:220";
const char *KM200 = "KM:200";
const char *KM2 = "KM:2";     // wildcard for all KM-2*
const char *KM300 = "KM:300"; // wildcard for all KM-3*
const char *KL200 = "KL:200";
const char *KL210 = "KL:210";
const char *KL220 = "KL:220";
const char *KL230 = "KL:230";
const char *KL300 = "KL:300";
const char *KL330 = "KL:330";
const char *KL2 = "KL:2"; // wildcard for all KL-2*
const char *TKM210 = "TKM:210";
const char *TKM220 = "TKM:220";
const char *TKM230 = "TKM:230";
const char *TKM300 = "TKM:300";
const char *TKM2 = "TKM:2";   // wildcard for all TKM-2*
const char *SDM20 = "SDM:20"; // wildcard for all SDM-20*
const char *SDM2 = "SDM:2";   // wildcard for all SDM-2*
const char *SDM300 = "SDM:300";
const char *SDM301 = "SDM:301";
const char *SDS210 = "SDS:210";
const char *SDS20 = "SDS:20"; // wildcard for SDS-20*
const char *SDS22 = "SDS:22"; // wildcard for SDS-22*
const char *SDS2 = "SDS:2";   // wildcard for all SDS-2*
const char *ZWS2 = "ZWS:2";   // wildcard for all ZWS-2*
const char *ZWS205 = "ZWS:205";
const char *ZWD300 = "ZWD:300";
const char *KL213 = "KL:213";
const char *KL214 = "KL:214";
const char *UMV204 = "UMV:204";
const char *UMV200 = "UMV:200";
const char *UMV210 = "UMV:210";
const char *UMR200 = "UMR:200";
const char *SK20 = "SK:20"; // wildcard for all SK-20*
const char *SK204 = "SK:204";
const char *AKM2 = "AKM:2"; // wildcard for all AKM-2*
const char *TNY2 = "TNY:2"; // wildcard for all TNY-*
const char *SKM300 = "SKM:300";


// model features
const ModelFeatureId MF_GE_KM220[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,        ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,     ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_KM2[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,        ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,     ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_KL200[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,        ModelFeatureId::outmodeswitch,
    ModelFeatureId::outvalue8,     ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_TKM210[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::transt,        ModelFeatureId::outmode,         ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,     ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_TKM220[] = {ModelFeatureId::blink,     ModelFeatureId::identification,
                                       ModelFeatureId::leddark,   ModelFeatureId::pushbutton,
                                       ModelFeatureId::pushbarea, ModelFeatureId::pushbadvanced};

const ModelFeatureId MF_GE_TKM230[] = {ModelFeatureId::blink,     ModelFeatureId::identification,
                                       ModelFeatureId::leddark,   ModelFeatureId::pushbutton,
                                       ModelFeatureId::pushbarea, ModelFeatureId::pushbadvanced};

const ModelFeatureId MF_GE_KM300[] = {ModelFeatureId::dontcare,        ModelFeatureId::blink,
                                      ModelFeatureId::identification,  ModelFeatureId::transt,
                                      ModelFeatureId::outmode,         ModelFeatureId::outvalue8,
                                      ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
                                      ModelFeatureId::blinkconfig,     ModelFeatureId::dimtimeconfig,
                                      ModelFeatureId::outmodeauto,     ModelFeatureId::dimmodeconfig,
                                      ModelFeatureId::ledauto,         ModelFeatureId::customtransitiontime};

const ModelFeatureId MF_GE_TKM300[] = {ModelFeatureId::dontcare,        ModelFeatureId::blink,
                                       ModelFeatureId::identification,  ModelFeatureId::transt,
                                       ModelFeatureId::outmode,         ModelFeatureId::outvalue8,
                                       ModelFeatureId::pushbutton,      ModelFeatureId::pushbdevice,
                                       ModelFeatureId::pushbarea,       ModelFeatureId::pushbadvanced,
                                       ModelFeatureId::pushbsensor,     ModelFeatureId::akminput,
                                       ModelFeatureId::akmsensor,       ModelFeatureId::akmdelay,
                                       ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
                                       ModelFeatureId::blinkconfig,     ModelFeatureId::dimtimeconfig,
                                       ModelFeatureId::outmodeauto,     ModelFeatureId::dimmodeconfig,
                                       ModelFeatureId::ledauto,         ModelFeatureId::customtransitiontime,
                                       ModelFeatureId::setumr200config};

const ModelFeatureId MF_GE_SDM300[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::transt,        ModelFeatureId::outmode,         ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,     ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig,   ModelFeatureId::dimtimeconfig,   ModelFeatureId::outmodeauto,
    ModelFeatureId::dimmodeconfig, ModelFeatureId::ledauto,         ModelFeatureId::customtransitiontime,
};

const ModelFeatureId MF_GE_SDM301[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::transt,        ModelFeatureId::outmode,         ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,     ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig,   ModelFeatureId::dimtimeconfig,   ModelFeatureId::outmodeauto,
    ModelFeatureId::dimmodeconfig, ModelFeatureId::ledauto,         ModelFeatureId::customtransitiontime,
};

const ModelFeatureId MF_GE_SDM20[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,        ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,     ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_SDS210[] = {
    ModelFeatureId::dontcare,        ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,         ModelFeatureId::transt,        ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,       ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,       ModelFeatureId::pushbadvanced, ModelFeatureId::twowayconfig,
    ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_SDS20[] = {
    ModelFeatureId::dontcare,     ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::ledauto,      ModelFeatureId::transt,          ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,    ModelFeatureId::pushbutton,      ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,    ModelFeatureId::pushbadvanced,   ModelFeatureId::pushbcombined,
    ModelFeatureId::twowayconfig, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_SDS22[] = {
    ModelFeatureId::dontcare,     ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::ledauto,      ModelFeatureId::transt,          ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,    ModelFeatureId::pushbutton,      ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,    ModelFeatureId::pushbadvanced,   ModelFeatureId::pushbcombined,
    ModelFeatureId::twowayconfig, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_SDS2[] = {
    ModelFeatureId::dontcare,        ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,         ModelFeatureId::transt,        ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,       ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,       ModelFeatureId::pushbadvanced, ModelFeatureId::twowayconfig,
    ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_ZWS2[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,          ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,     ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_UMV204[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,          ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,     ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_UMV200[] = {
    ModelFeatureId::dontcare,        ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,         ModelFeatureId::transt,        ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,      ModelFeatureId::pushbdevice,   ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced,   ModelFeatureId::extradimmer,   ModelFeatureId::umvrelay,
    ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_UMV210[] = {
    ModelFeatureId::dontcare,       ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::ledauto,        ModelFeatureId::transt,          ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,     ModelFeatureId::pushbdevice,     ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced,  ModelFeatureId::extradimmer,     ModelFeatureId::umvrelay,
    ModelFeatureId::outputchannels, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_GE_ZWD300[] = {ModelFeatureId::dontcare,        ModelFeatureId::blink,
                                       ModelFeatureId::identification,  ModelFeatureId::transt,
                                       ModelFeatureId::outmode,         ModelFeatureId::outvalue8,
                                       ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
                                       ModelFeatureId::blinkconfig,     ModelFeatureId::dimtimeconfig,
                                       ModelFeatureId::outmodeauto,     ModelFeatureId::dimmodeconfig,
                                       ModelFeatureId::ledauto,         ModelFeatureId::customtransitiontime};

const ModelFeatureId MF_GN_KM2[] = {ModelFeatureId::dontcare, ModelFeatureId::blink,  ModelFeatureId::identification,
                                    ModelFeatureId::ledauto,  ModelFeatureId::transt, ModelFeatureId::outmode,
                                    ModelFeatureId::outvalue8};

const ModelFeatureId MF_GN_TKM2[] = {ModelFeatureId::dontcare, ModelFeatureId::blink,   ModelFeatureId::identification,
                                     ModelFeatureId::transt,   ModelFeatureId::outmode, ModelFeatureId::outvalue8};

const ModelFeatureId MF_RT_KM2[] = {ModelFeatureId::dontcare, ModelFeatureId::blink,  ModelFeatureId::identification,
                                    ModelFeatureId::ledauto,  ModelFeatureId::transt, ModelFeatureId::outmode,
                                    ModelFeatureId::outvalue8};

const ModelFeatureId MF_RT_TKM2[] = {ModelFeatureId::dontcare, ModelFeatureId::blink,   ModelFeatureId::identification,
                                     ModelFeatureId::transt,   ModelFeatureId::outmode, ModelFeatureId::outvalue8};

const ModelFeatureId MF_RT_SDM2[] = {ModelFeatureId::blink, ModelFeatureId::identification, ModelFeatureId::leddark,
                                     ModelFeatureId::outmode, ModelFeatureId::outvalue8};

const ModelFeatureId MF_GR_KL210[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,          ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::pushbutton,     ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced,  ModelFeatureId::shadeprops,
    ModelFeatureId::shadeposition, ModelFeatureId::locationconfig, ModelFeatureId::windprotectionconfigawning};

const ModelFeatureId MF_GR_KL220[] = {ModelFeatureId::dontcare,       ModelFeatureId::blink,
                                      ModelFeatureId::identification, ModelFeatureId::ledauto,
                                      ModelFeatureId::pushbutton,     ModelFeatureId::pushbdevice,
                                      ModelFeatureId::pushbarea,      ModelFeatureId::pushbadvanced,
                                      ModelFeatureId::shadeprops,     ModelFeatureId::shadeposition,
                                      ModelFeatureId::motiontimefins, ModelFeatureId::shadebladeang,
                                      ModelFeatureId::locationconfig, ModelFeatureId::windprotectionconfigblind};

const ModelFeatureId MF_GR_KL230[] = {ModelFeatureId::dontcare,       ModelFeatureId::blink,
                                      ModelFeatureId::identification, ModelFeatureId::ledauto,
                                      ModelFeatureId::shadeprops,     ModelFeatureId::shadeposition,
                                      ModelFeatureId::motiontimefins, ModelFeatureId::shadebladeang,
                                      ModelFeatureId::locationconfig, ModelFeatureId::windprotectionconfigblind};

const ModelFeatureId MF_GR_KL300[] = {
    ModelFeatureId::dontcare,                 //
    ModelFeatureId::identification,           //
    ModelFeatureId::ledauto,                  //
    ModelFeatureId::shadeprops,               //
    ModelFeatureId::pushbutton,               //
    ModelFeatureId::pushbdevice,              //
    ModelFeatureId::pushbarea,                //
    ModelFeatureId::pushbadvanced,            //
    ModelFeatureId::shadeposition,            //
    ModelFeatureId::motiontimefins,           //
    ModelFeatureId::shadebladeang,            //
    ModelFeatureId::locationconfig,           //
    ModelFeatureId::windprotectionconfigblind //
};

const ModelFeatureId MF_GR_KL300_1[] = {
    ModelFeatureId::identification, //
    ModelFeatureId::pushbutton,     //
    ModelFeatureId::pushbdevice,    //
    ModelFeatureId::pushbarea,      //
    ModelFeatureId::pushbadvanced,  //
    ModelFeatureId::akmsensor,      //
    ModelFeatureId::akminput,       //
    ModelFeatureId::akmdelay,       //
};

const ModelFeatureId MF_GR_HKL330[] = {ModelFeatureId::dontcare,       ModelFeatureId::blink,
                                       ModelFeatureId::identification, ModelFeatureId::ledauto,
                                       ModelFeatureId::shadeprops,     ModelFeatureId::pushbutton,
                                       ModelFeatureId::pushbdevice,    ModelFeatureId::pushbarea,
                                       ModelFeatureId::pushbadvanced,  ModelFeatureId::shadeposition,
                                       ModelFeatureId::motiontimefins, ModelFeatureId::shadebladeang,
                                       ModelFeatureId::locationconfig, ModelFeatureId::windprotectionconfigblind};

const ModelFeatureId MF_GR_KL2[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,          ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::pushbutton,     ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced,  ModelFeatureId::shadeprops,
    ModelFeatureId::shadeposition, ModelFeatureId::locationconfig, ModelFeatureId::windprotectionconfigblind};

const ModelFeatureId MF_GR_TKM2[] = {ModelFeatureId::blink,     ModelFeatureId::identification,
                                     ModelFeatureId::leddark,   ModelFeatureId::pushbutton,
                                     ModelFeatureId::pushbarea, ModelFeatureId::pushbadvanced};

const ModelFeatureId MF_BL_KM200[] = {
    ModelFeatureId::dontcare,     ModelFeatureId::ledauto,        ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,   ModelFeatureId::pushbdevice,    ModelFeatureId::pushbadvanced,
    ModelFeatureId::heatinggroup, ModelFeatureId::heatingoutmode, ModelFeatureId::heatingprops,
    ModelFeatureId::pwmvalue,     ModelFeatureId::valvetype,      ModelFeatureId::identification};

const ModelFeatureId MF_BL_KM300[] = {
    ModelFeatureId::dontcare,     ModelFeatureId::ledauto,        ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,   ModelFeatureId::pushbdevice,    ModelFeatureId::pushbadvanced,
    ModelFeatureId::heatinggroup, ModelFeatureId::heatingoutmode, ModelFeatureId::heatingprops,
    ModelFeatureId::pwmvalue,     ModelFeatureId::valvetype,      ModelFeatureId::identification};

const ModelFeatureId MF_BL_SDS200[] = {
    ModelFeatureId::dontcare,     ModelFeatureId::ledauto,        ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,   ModelFeatureId::pushbdevice,    ModelFeatureId::pushbadvanced,
    ModelFeatureId::heatinggroup, ModelFeatureId::heatingoutmode, ModelFeatureId::heatingprops,
    ModelFeatureId::pwmvalue,     ModelFeatureId::valvetype,      ModelFeatureId::identification,
};

const ModelFeatureId MF_BL_UMV200[] = {
    ModelFeatureId::dontcare,        ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,         ModelFeatureId::transt,        ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,      ModelFeatureId::pushbdevice,   ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced,   ModelFeatureId::extradimmer,   ModelFeatureId::umvrelay,
    ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig,
    ModelFeatureId::heatinggroup,
};

const ModelFeatureId MF_BL_KL2[] = {
    ModelFeatureId::dontcare,
    ModelFeatureId::ledauto,
    ModelFeatureId::identification,
};

const ModelFeatureId MF_BL_SK204[] = {
    ModelFeatureId::temperatureoffset,  ModelFeatureId::ftwtempcontrolventilationselect,
    ModelFeatureId::ftwdisplaysettings, ModelFeatureId::ftwbacklighttimeout,
    ModelFeatureId::ventconfig,         ModelFeatureId::identification,
};

const ModelFeatureId MF_TK_TKM2[] = {ModelFeatureId::blink, ModelFeatureId::identification, ModelFeatureId::leddark,
                                     ModelFeatureId::pushbutton, ModelFeatureId::pushbarea};

const ModelFeatureId MF_GE_TKM2[] = {ModelFeatureId::blink,         ModelFeatureId::identification,
                                     ModelFeatureId::leddark,       ModelFeatureId::pushbutton,
                                     ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced,
                                     ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_SW_TKM2[] = {
    ModelFeatureId::blink,      ModelFeatureId::identification, ModelFeatureId::leddark,
    ModelFeatureId::pushbutton, ModelFeatureId::pushbarea,      ModelFeatureId::pushbadvanced,
    ModelFeatureId::highlevel,  ModelFeatureId::jokerconfig,    ModelFeatureId::twowayconfig};

const ModelFeatureId MF_SW_KL2[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,        ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,       ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,  ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::optypeconfig, ModelFeatureId::highlevel,
    ModelFeatureId::consumption,   ModelFeatureId::jokerconfig,  ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_SW_ZWS2[] = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,        ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,       ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,  ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::optypeconfig, ModelFeatureId::highlevel,
    ModelFeatureId::consumption,   ModelFeatureId::jokerconfig,  ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const ModelFeatureId MF_SW_ZWS205[] = {
    ModelFeatureId::dontcare,         ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,          ModelFeatureId::transt,        ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,       ModelFeatureId::pushbdevice,   ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced,    ModelFeatureId::pushbdisabled, ModelFeatureId::heatingprops,
    ModelFeatureId::highlevel,        ModelFeatureId::consumption,   ModelFeatureId::consumptioneventled,
    ModelFeatureId::consumptiontimer, ModelFeatureId::jokerconfig,   ModelFeatureId::jokertempcontrol,
    ModelFeatureId::outconfigswitch,  ModelFeatureId::impulseconfig, ModelFeatureId::valvetype,
    ModelFeatureId::blinkconfig,      ModelFeatureId::outmodeswitch, ModelFeatureId::outmodetempcontrol};

const ModelFeatureId MF_SW_KL213[] = {
    ModelFeatureId::dontcare,     ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::ledauto,      ModelFeatureId::transt,          ModelFeatureId::outvalue8,
    ModelFeatureId::optypeconfig, ModelFeatureId::highlevel,       ModelFeatureId::consumption,
    ModelFeatureId::jokerconfig,  ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_SW_KL214[] = {
    ModelFeatureId::dontcare,     ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::ledauto,      ModelFeatureId::transt,          ModelFeatureId::outvalue8,
    ModelFeatureId::optypeconfig, ModelFeatureId::highlevel,       ModelFeatureId::consumption,
    ModelFeatureId::jokerconfig,  ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const ModelFeatureId MF_SW_SDS20[] = {
    ModelFeatureId::blink,      ModelFeatureId::identification, ModelFeatureId::leddark,
    ModelFeatureId::pushbutton, ModelFeatureId::pushbarea,      ModelFeatureId::pushbadvanced,
    ModelFeatureId::highlevel,  ModelFeatureId::jokerconfig,    ModelFeatureId::twowayconfig};

const ModelFeatureId MF_SW_SDS22[] = {
    ModelFeatureId::blink,      ModelFeatureId::identification, ModelFeatureId::leddark,
    ModelFeatureId::pushbutton, ModelFeatureId::pushbarea,      ModelFeatureId::pushbadvanced,
    ModelFeatureId::highlevel,  ModelFeatureId::jokerconfig,    ModelFeatureId::twowayconfig};

const ModelFeatureId MF_SW_AKM2[] = {
    ModelFeatureId::akmsensor,
    ModelFeatureId::akminput,
    ModelFeatureId::akmdelay,
    ModelFeatureId::identification,
};

const ModelFeatureId MF_SW_UMR200[] = {ModelFeatureId::blink,
                                       ModelFeatureId::identification,
                                       ModelFeatureId::dontcare,
                                       ModelFeatureId::outvalue8,
                                       ModelFeatureId::pushbutton,
                                       ModelFeatureId::pushbsensor,
                                       ModelFeatureId::pushbarea,
                                       ModelFeatureId::pushbadvanced,
                                       ModelFeatureId::highlevel,
                                       ModelFeatureId::jokerconfig,
                                       ModelFeatureId::twowayconfig,
                                       ModelFeatureId::akminput,
                                       ModelFeatureId::akmsensor,
                                       ModelFeatureId::akmdelay,
                                       ModelFeatureId::blinkconfig,
                                       ModelFeatureId::impulseconfig,
                                       ModelFeatureId::umroutmode,
                                       ModelFeatureId::outconfigswitch,
                                       ModelFeatureId::outmodetempcontrol};

const ModelFeatureId MF_SW_SK20[] = {
    ModelFeatureId::temperatureoffset,
    ModelFeatureId::identification,
};

const ModelFeatureId MF_SW_TNY2[] = {
    ModelFeatureId::identification,
};

const ModelFeatureId MF_SW_SKM300[] = {
    ModelFeatureId::identification,  //
    ModelFeatureId::dontcare,        //
    ModelFeatureId::pushbutton,      //
    ModelFeatureId::pushbsensor,     //
    ModelFeatureId::pushbarea,       //
    ModelFeatureId::pushbadvanced,   //
    ModelFeatureId::highlevel,       //
    ModelFeatureId::jokerconfig,     //
    ModelFeatureId::twowayconfig,    //
    ModelFeatureId::akminput,        //
    ModelFeatureId::akmsensor,       //
    ModelFeatureId::akmdelay,        //
    ModelFeatureId::setumr200config, //
};

using matrix_t = std::unordered_map<std::string, std::vector<gsl::span<const ModelFeatureId>>>;

static const matrix_t yellowModelFeatures = {
    {KM220, {MF_GE_KM220}},   //
    {KM2, {MF_GE_KM2}},       //
    {KL200, {MF_GE_KL200}},   //
    {TKM210, {MF_GE_TKM210}}, //
    {TKM220, {MF_GE_TKM220}}, //
    {TKM230, {MF_GE_TKM230}}, //
    {KM300, {MF_GE_KM300}},   //
    {ZWD300, {MF_GE_ZWD300}}, //
    {TKM300, {MF_GE_TKM300}}, //
    {SDM20, {MF_GE_SDM20}},   //
    {SDM300, {MF_GE_SDM300}}, //
    {SDM301, {MF_GE_SDM301}}, //
    {SDS210, {MF_GE_SDS210}}, //
    {SDS20, {MF_GE_SDS20}},   //
    {SDS22, {MF_GE_SDS22}},   //
    {SDS2, {MF_GE_SDS2}},     //
    {ZWS2, {MF_GE_ZWS2}},     //
    {UMV204, {MF_GE_UMV204}}, //
    {UMV200, {MF_GE_UMV200}}, //
    {UMV210, {MF_GE_UMV210}}, //
    {TKM2, {MF_GE_TKM2}},     //
};

static const matrix_t greyModelFeatures = {
    {KL210, {MF_GR_KL210}},                //
    {KL220, {MF_GR_KL220}},                //
    {KL230, {MF_GR_KL230}},                //
    {KL300, {MF_GR_KL300, MF_GR_KL300_1}}, //
    {KL330, {MF_GR_HKL330}},               //
    {KL2, {MF_GR_KL2}},                    //
    {TKM2, {MF_GR_TKM2}},                  //
};

static const matrix_t blueModelFeatures = {
    {KM200, {MF_BL_KM200}},   //
    {KM300, {MF_BL_KM300}},   //
    {UMV200, {MF_BL_UMV200}}, //
    {SDS2, {MF_BL_SDS200}},   //
    {KL2, {MF_BL_KL2}},       //
    {SK204, {MF_BL_SK204}},   //
};

static matrix_t cyanModelFeatures = {
    {TKM2, {MF_TK_TKM2}}, //
};

static const matrix_t redModelFeatures = {
    {KM2, {MF_RT_KM2}},   //
    {TKM2, {MF_RT_TKM2}}, //
    {SDM2, {MF_RT_SDM2}}, //
};

static const matrix_t greenModelFeatures = {
    {KM2, {MF_GN_KM2}},   //
    {TKM2, {MF_GN_TKM2}}, //
};

static const matrix_t blackModelFeatures = {
    {TKM2, {MF_SW_TKM2}},     //
    {KL2, {MF_SW_KL2}},       //
    {ZWS2, {MF_SW_ZWS2}},     //
    {ZWS205, {MF_SW_ZWS205}}, //
    {KL213, {MF_SW_KL213}},   //
    {KL214, {MF_SW_KL214}},   //
    {SDS20, {MF_SW_SDS20}},   //
    {SDS22, {MF_SW_SDS22}},   //
    {AKM2, {MF_SW_AKM2}},     //
    {UMR200, {MF_SW_UMR200}}, //
    {SK20, {MF_SW_SK20}},     //
    {TNY2, {MF_SW_TNY2}},     //
    {SKM300, {MF_SW_SKM300}}, //
};

static auto staticFeatureMatrix = std::vector<const matrix_t *>{
    nullptr,              // off-by-one, zero-index unused
    &yellowModelFeatures, //
    &greyModelFeatures,   //
    &blueModelFeatures,   //
    &cyanModelFeatures,   //
    nullptr,              // ColorIDViolet
    &redModelFeatures,    //
    &greenModelFeatures,  //
    &blackModelFeatures,  //
    nullptr,              // ColorIDWhite
};

struct ModelFeaturesImpl final : public ModelFeatures {
  ModelFeaturesImpl() = default;
  void setFeatures(int _color, const std::string &model, std::vector<ModelFeatureId> &&features) override;
  std::vector<std::pair<std::string, features_t>> getFeatures(int color) override;
  std::vector<ModelFeatureId> getAvailableFeatures() const override;

private:
  using modelToFeatures_t = std::unordered_map<std::string, std::vector<std::vector<ModelFeatureId>>>;

  // color vector, each index represents one color according to definitions of modelconst.h
  std::unordered_map<int, modelToFeatures_t> m_dynamicFeatureMatrix;
  boost::mutex m_lock;
};

void ModelFeaturesImpl::setFeatures(int _color, const std::string &_model, std::vector<ModelFeatureId> &&features) {
  if (_model.empty()) {
    throw std::runtime_error("can't add model features: missing model key");
  }

  if (features.empty()) {
    throw std::runtime_error("invalid feature array\n");
  }

  if ((_color < ColorIDYellow) || (_color > ColorIDWhite)) {
    throw std::runtime_error("can not save feature: unsupported device color");
  }

  boost::mutex::scoped_lock lock(m_lock);

  auto contains = [&_model](const auto &map) { return (map.find(_model) != map.end()); };

  if ((staticFeatureMatrix[_color] != nullptr && contains(*staticFeatureMatrix[_color])) ||
      contains(m_dynamicFeatureMatrix[_color])) {
    DS_NOTICE("features for", _model, _color, "already present, ignoring");
    return;
  }

  // assume vdcs have only single virtual terminal block
  m_dynamicFeatureMatrix[_color][_model] = std::vector<std::vector<ModelFeatureId>>{std::move(features)};
}

std::vector<std::pair<std::string, ModelFeatures::features_t>> ModelFeaturesImpl::getFeatures(int _color) {
  if ((_color < ColorIDYellow) || (_color > ColorIDWhite)) {
    throw std::runtime_error("can not retrieve features: unsupported device color");
  }

  boost::mutex::scoped_lock lock(m_lock);

  std::vector<std::pair<std::string, features_t>> result;

  if (staticFeatureMatrix[_color] != nullptr) {
    features_t resultFeatures{};
    for (const auto &[model, featuresPerBlock] : *staticFeatureMatrix[_color]) {
      for (const auto &features : featuresPerBlock) {
        resultFeatures.push_back(std::vector(features.begin(), features.end()));
      }
      result.push_back(std::pair{model, std::move(resultFeatures)});
    }
  }

  for (auto const &[model, featuresPerBlock] : m_dynamicFeatureMatrix[_color]) {
    result.push_back(std::pair{model, featuresPerBlock});
  }

  return result;
}

std::vector<ModelFeatureId> ModelFeaturesImpl::getAvailableFeatures() const {
  std::vector<ModelFeatureId> available{};
  auto modelFeatureEnum = dsEnum(ModelFeatureId()).options;
  std::transform(std::cbegin(modelFeatureEnum), std::cend(modelFeatureEnum), std::back_inserter(available),
                 [](const auto &element) { return element.first; });
  return available;
}

} // namespace

std::string ModelFeatures::colorToString(int _color) {
  if (auto colorString = Device::getDeviceClassString(static_cast<DeviceClasses_t>(_color)); !colorString.empty()) {
    return colorString;
  }
  throw std::runtime_error("unknown color");
}

std::unique_ptr<ModelFeatures> makeModelFeatures() {
  return std::make_unique<ModelFeaturesImpl>();
}

}; // namespace dss
