/*
    Copyright (c) 2019 digitalSTROM.org," " Zurich," " Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/log.h>

#include "eventinterpreterplugins.h"
#include "businterface-mock.h"
#include "apartment.h"
#include "cluster.h"
#include "event.h"
#include "state.h"
#include "event/event_create.h"
#include "zone.h"

#include "tests/util/dss_instance_fixture.h"

using namespace std::literals;

static ds::LogChannel dsLogChannel("clusterTest");
static const char* TAGS = "[dss][cluster]"; // read more about this tags, maybe i should add weather tag? or weatherstation? what tags are?

using namespace dss;

namespace {

class ClusterMock : public Cluster {
public:
  using Cluster::Cluster;
  void callScene(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr, const std::string _token, const bool _force) override {
    setLastCalledScene(_sceneNr & 0x00ff);

    auto sharedPtrThis = DSS::getInstance()->getApartment().getCluster(this->getID());
    auto state =
        DSS::getInstance()->getApartment().allocateGroupState(sharedPtrThis, "wind.group" + ds::str(this->getID()));
    if (_sceneNr == SceneWindActive) {
      state->setState(CallOrigin::json, State_Active);
    } else if (_sceneNr == SceneWindInactive) {
      state->setState(CallOrigin::json, State_Inactive);
    }
  };
};

SCENARIO("cluster can be protected against weather", TAGS) {
  WindProtectionClass_t protectionClass = wpc_blind_class_1;
  double protectionClassValue = getClassValue(protectionClass);
  std::shared_ptr<Event> pEvent;
  BusInterfaceMock busInterface;
  DSSInstanceFixture fixture;
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;

  apt.setBusInterface(&busInterface);
  int clusterId = apt.getEmptyCluster()->getID();
  auto pCluster = std::make_shared<ClusterMock>(clusterId, apt);
  auto windState =
      DSS::getInstance()->getApartment().allocateGroupState(pCluster, "wind.group" + intToString(clusterId));

  ClusterWeatherProtectionPlugin interface(&DSS::getInstance()->getEventInterpreter());
  DSS::getInstance()->getEventInterpreter().initialize();  // needed to handle event

  pCluster->setProtectionClass(wpc_blind_class_1); // < 9.0 m/s

  WHEN("the gust value exceeded the wind protection class value") {
    pCluster->checkWindProtection(protectionClassValue * 1.1);
    pEvent = DSS::getInstance()->getEventRunner().getEvent(0).getEvent();

    THEN("the wind scene is called on the cluster group") {
      REQUIRE(pCluster->getLastCalledScene() == SceneWindActive);
    }

    THEN("event to check this cluster in 15 minutes is pushed") {
      REQUIRE(pEvent);
      REQUIRE(pEvent->getName() == EventName::ClusterCheckWind);
      REQUIRE(pEvent->getPropertyByName("clusterId") == intToString(clusterId));
      REQUIRE_FALSE(pEvent->getPropertyByName(EventProperty::Time).empty());
    }

    WHEN("the dss was restarted (loses internal event queue)") {
      while (DSS::getInstance()->getEventRunner().getSize() > 0) {
        auto event = DSS::getInstance()->getEventRunner().getEvent(0);
        DSS::getInstance()->getEventRunner().removeEvent(event.getID());
      }

      std::shared_ptr<Event> modelReadyEvent = std::make_shared<Event>(EventName::ModelReady);
      interface.handleEvent(*modelReadyEvent, EventSubscription(interface, EventName::ModelReady));
      THEN("check wind events are scheduled for clusters with active wind state") {
        auto clusterFromFixture = DSS::getInstance()->getApartment().getCluster(clusterId);
        REQUIRE(clusterFromFixture->getWindProtectionTimer().has_value());
        REQUIRE(DSS::getInstance()->getEventRunner().getSize() == 1);
        auto event = DSS::getInstance()->getEventRunner().getEvent(0).getEvent();
        REQUIRE(event->getName() == EventName::ClusterCheckWind);
        REQUIRE(event->getPropertyByName("clusterId") == intToString(clusterId));
        REQUIRE_FALSE(event->getPropertyByName(EventProperty::Time).empty());
      }
    }

    WHEN("gust value was 10% below the protection class for more than 15 minutes") {
      // checkWindProtection takes cluster from getInstance()->getApartment()
      auto clusterFromFixture = DSS::getInstance()->getApartment().getCluster(clusterId);
      clusterFromFixture->setWindProtectionTimer(DateTime().addMinute(-20));
      clusterFromFixture->setLastCalledScene(SceneWindActive);
      clusterFromFixture->setProtectionClass(wpc_blind_class_1);
      clusterFromFixture->checkWindProtection(protectionClassValue * 0.8);

      THEN("the wind inactive scene is called on the cluster group") {
        windState->setState(CallOrigin::json, State_Active); // due to simplified callScene we need to set it manually
        interface.handleEvent(*pEvent, EventSubscription(interface, EventName::ClusterCheckWind));
        REQUIRE(clusterFromFixture->getLastCalledScene() == SceneWindInactive);
      }
    }

    WHEN("gust value >= 10% below the protection class") {
      pCluster->checkWindProtection(protectionClassValue * 0.9);

      THEN("internal cluster timer is reset") {
        DateTime wpcTimer = pCluster->getWindProtectionTimer().value();
        REQUIRE(wpcTimer.difference(DateTime()) < 15); // timer has just been reset
      }
    }
  }
}

} // namespace
