/*
    Copyright (c) 2021 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "model/device-property-tree.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "model/apartment-mock.h"
#include "model/device-mock.h"
#include "model/shadeattributes.h"

static ds::LogChannel dsLogChannel("modelDevicePropertyTreeTest");
static const char* TAGS = "[model][propertytree]";

using namespace dss;
using namespace dss::model;
using namespace std::literals;
using namespace trompeloeil;
using trompeloeil::_;

SCENARIO("DevicePropertyTree", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto busInterface = BusInterfaceMock{};
  auto apartment = ApartmentMock{nullptr, &propertySystem, &busInterface};
  auto devicePropertyTree = makeDevicePropertyTree(apartment);

  SECTION("general properties") {
    auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL300});
    devicePropertyTree->publishToPropertyTree(device);

    THEN("root node is valid") {
      CHECK(device.getPropertyNode());
      auto& deviceRootNode = *device.getPropertyNode();
      CHECK(deviceRootNode.getProperty("DisplayID"));

      CHECK(deviceRootNode.getProperty("DeviceSpecification"));
      auto& deviceSpecifiactionNode = *deviceRootNode.getProperty("DeviceSpecification");
      CHECK(deviceSpecifiactionNode.getProperty("ActiveGroup"));
      CHECK(deviceSpecifiactionNode.getProperty("DefaultGroup"));
      CHECK(deviceSpecifiactionNode.getProperty("Groups")->getStringValue() == "[2]");
    }
  }

  SECTION("shades") {
    auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL300});
    auto& deviceRootNode = *device.getPropertyNode();

    WHEN("setting ShadeAttributes") {
      device.setCardinalDirection(cd_north);
      THEN("shade nodes exist") { CHECK(deviceRootNode.getProperty("CardinalDirection")->getStringValue() == "north"); }

      WHEN("update values") {
        device.setCardinalDirection(cd_south);
        THEN("values are updated") {
          CHECK(deviceRootNode.getProperty("CardinalDirection")->getStringValue() == "south");
        }
      }

      THEN("remove them") {
        REQUIRE(deviceRootNode.getProperty("CardinalDirection"));
        devicePropertyTree->removeShadeAttributes(*device.getPropertyNode());
        CHECK(!deviceRootNode.getProperty("CardinalDirection"));
      }
    }
    WHEN("deleting non-existant shade attributes") {
      REQUIRE(!device.getPropertyNode()->getProperty("shadeAttributes"s));
      devicePropertyTree->removeShadeAttributes(*device.getPropertyNode());
      CHECK(!device.getPropertyNode()->getProperty("shadeAttributes"s));
    }

    SECTION("AKM") {
      GIVEN("AKM device") {
        auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL300});
        auto& deviceRootNode = *device.getPropertyNode();
        devicePropertyTree->publishToPropertyTree(device);

        THEN("setting input mode to AKM type") {
          struct Y {
            std::string akmMode;
            ButtonInputMode buttonInputMode;
          };

          // strings are public API! careful!
          const auto& y =
              GENERATE(values(std::initializer_list<Y>{{"standard"s, ButtonInputMode::AKM_STANDARD},
                                                       {"inverted", ButtonInputMode::AKM_INVERTED},
                                                       {"on_rising_edge", ButtonInputMode::AKM_ON_RISING_EDGE},
                                                       {"on_falling_edge", ButtonInputMode::AKM_ON_FALLING_EDGE},
                                                       {"off_rising_edge", ButtonInputMode::AKM_OFF_RISING_EDGE},
                                                       {"off_falling_edge", ButtonInputMode::AKM_OFF_FALLING_EDGE},
                                                       {"rising_edge", ButtonInputMode::AKM_RISING_EDGE},
                                                       {"falling_edge", ButtonInputMode::AKM_FALLING_EDGE},
                                                       {""s, ButtonInputMode::STANDARD}}));

          REQUIRE_CALL(busInterface.device, setDeviceConfig(_, _, _, static_cast<uint8_t>(y.buttonInputMode)));
          device.setDeviceButtonInputMode(y.buttonInputMode);
          CHECK(deviceRootNode.getProperty("AKMInputProperty")->getStringValue() == y.akmMode);
        }
        GIVEN("akm input with valid mode set") {
          ALLOW_CALL(busInterface.device, setDeviceConfig(_, _, _, _));
          device.setDeviceButtonInputMode(ButtonInputMode::AKM_RISING_EDGE);
          WHEN("modifying AKM node directly") {
            deviceRootNode.getProperty("AKMInputProperty")->setStringValue("inverted");
            THEN("does not change the nodes value")
            CHECK(deviceRootNode.getProperty("AKMInputProperty")->getStringValue() == "rising_edge"s);
          }
        }
      }
    }
  }

  SECTION("groups") {
    GIVEN("shade device") {
      auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL300});
      auto& deviceRootNode = *device.getPropertyNode();
      devicePropertyTree->publishToPropertyTree(device);

      THEN("dfault groups are present") {
        CHECK(deviceRootNode.getProperty("groups/group2/id"));
        CHECK(deviceRootNode.getProperty("groups/group2/id")->getIntegerValue() == 2);
        CHECK(!deviceRootNode.getProperty("groups/group6"));
      }

      WHEN("adding to additional group") {
        device.addToGroup(GroupIDRed);
        THEN("group node is created") {
          CHECK(deviceRootNode.getProperty("groups/group6/id"));
          CHECK(deviceRootNode.getProperty("groups/group6/id")->getIntegerValue() == 6);
        }
        WHEN("removing additional group2") {
          device.removeFromGroup(GroupIDRed);
          THEN("node is removed again") {
            CHECK(!deviceRootNode.getProperty("groups/group6"));
            CHECK(deviceRootNode.getProperty("groups")); // but groups remain
          }
        }
      }
    }
  }
}
