/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MODELMAINTENANCE_H_
#define MODELMAINTENANCE_H_

#include <ds/asio/resource-fwd.h>

#include <condition_variable>
#include <memory>
#include <mutex>
#include <queue>

#include "device.h"
#include "dssfwd.h"
#include "metering/metering-fwd.h"
#include "model/modelevent.h"
#include "subsystem.h"
#include "taskprocessor.h"
#include "thread.h"

namespace ds {
  class Conf;
}

namespace dss {

struct ModelMaintenanceConfig {
  // minTransactionTime: transaction is held open to aggregate multiple actions into one transaction
  ds::Clock::duration minTransactionTime = std::chrono::seconds(30); // should cover ds485d crashes
  // maxAggregationWindow: if multiple changes arrive within, they are merged into the same transaction. This is not
  // maximum transaction time, you need to add duration of the last started action
  ds::Clock::duration maxAggregationWindow = std::chrono::minutes(3); // throw of a dice
};

  class Apartment;
  class Event;
  class MeterMaintenance;

  class ModelDeferredEvent {
  public:
    static const int kModelSceneTimeout = 2;
  private:
    dsuid_t m_Source;
    time_t m_Timestamp;
    bool m_IsCalled;
  public:
    /** Constructs a ModelDeferredEvent with timestamp */
    ModelDeferredEvent(dsuid_t _source) : m_Source(_source), m_IsCalled(false)
    {
      setTimestamp();
    }
    virtual ~ModelDeferredEvent() {}

    dsuid_t getSource() { return m_Source; }
    bool isOriginMyself() { return (m_Source == DSUID_NULL); }
    bool isDue() { time_t now = time(NULL); return (now - m_Timestamp >= kModelSceneTimeout); }
    bool isCalled() { return m_IsCalled; }

    void setTimestamp() { m_Timestamp = time(NULL); }
    void clearTimestamp() { m_Timestamp = 0; }
    void setCalled() { m_IsCalled = true; }
  };

  class ModelDeferredSceneEvent : public ModelDeferredEvent {
  private:
    int m_ZoneID;
    int m_GroupID;
    int m_OriginDeviceID;
    int m_SceneID;
    CallOrigin m_Origin;
    bool m_forcedFlag;
    std::string m_OriginToken;
  public:
    /** Constructs a ModelDeferredSceneEvent with timestamp */
    ModelDeferredSceneEvent(dsuid_t _source, int _zoneID, int _groupID, int _originDeviceID, int _sceneID, CallOrigin _origin, bool _forced, std::string _token) :
      ModelDeferredEvent(_source), m_ZoneID(_zoneID), m_GroupID(_groupID), m_OriginDeviceID(_originDeviceID),
      m_SceneID(_sceneID), m_Origin(_origin), m_forcedFlag(_forced), m_OriginToken(_token)
    {}
    virtual ~ModelDeferredSceneEvent() {}

    int getSceneID() { return m_SceneID; }
    int getGroupID() { return m_GroupID; }
    int getZoneID() { return m_ZoneID; }
    int getOriginDeviceID() { return m_OriginDeviceID; }
    bool getForcedFlag() { return m_forcedFlag; }
    CallOrigin getCallOrigin() { return m_Origin; }
    void setScene(int _sceneID) { m_SceneID = _sceneID; setTimestamp(); }
    std::string getOriginToken() { return m_OriginToken; }
  };

  class ModelDeferredButtonEvent : public ModelDeferredEvent {
  private:
    int m_DeviceID;
    int m_ButtonIndex;
    int m_ClickType;
    int m_HoldTime;
  public:
    /** Constructs a ModelDeferredButtonEvent with timestamp */
    ModelDeferredButtonEvent(dsuid_t _source, int _deviceID, int _buttonIndex, int _clickType) :
      ModelDeferredEvent(_source),
      m_DeviceID(_deviceID), m_ButtonIndex(_buttonIndex), m_ClickType(_clickType), m_HoldTime(0)
    {}
    virtual ~ModelDeferredButtonEvent() {}

    int getDeviceID() { return m_DeviceID; }
    int getButtonIndex() { return m_ButtonIndex; }
    int getClickType() { return m_ClickType; }
    int getRepeatCount() { return m_HoldTime; }
    void setClickType(int _clickType) { m_ClickType = _clickType; setTimestamp(); }
    void incRepeatCount() { m_HoldTime ++; }
  };

  //----------------------------------------------------------------------------

  struct ModelMaintenanceHooks {
    // to support unit tests
    std::function<void()> linkPresent{[] {}};
    std::function<void()> linkLost{[] {}};
    std::function<void(dsuid_t)> ds485MemberLost{[](auto arg) {}};
    std::function<void(dsuid_t)> ds485MemberJoin{[](auto arg) {}};

    std::function<void()> minimalTransactionTimeExpired{[] {}};
    std::function<void()> transactionDone{[] {}};
    std::function<void()> snapshotDone{[] {}};
  };

  class ModelMaintenance : public ThreadedSubsystem {
    struct Impl;
    /* const */ std::unique_ptr<Impl> m_impl; // PIMPL pattern
  public:
    ModelMaintenance(DSS* _pDSS, PropertySystem& propertySystem, EventSink&, const int _eventTimeoutMS = 1000);
    ~ModelMaintenance();

    int makeImpl(ds::asio::ResourceBase& parent, ModelMaintenanceConfig&& = {}, ModelMaintenanceHooks&& = {});

    virtual void shutdown();

    virtual void initialize();

    /** Adds a model event to the queue.
     * The ownership of the event will reside with the Apartment. ModelEvents arriving while initializing will be discarded.
     */
    virtual void addModelEvent(std::shared_ptr<const ModelEvent> event);
    virtual void addWork(std::function<void()>&& work);
    virtual void handleWork(void);
    void publishWebSocketEvent(std::shared_ptr<Event> _event);

    /** Starts the event-processing */
    virtual void execute();

    void onGroupCallScene(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const bool _forced, std::string _token, bool callSceneMin, ds::DateTime timeStamp);
    void onGroupUndoScene(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const std::string _token, ds::DateTime timeStamp);
    void onGroupSaveScene(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, std::string _token);

    void onDeviceNameChanged(dsuid_t _meterID, const devid_t _deviceID,
                             const std::string& _name);
    void onDsmNameChanged(dsuid_t _meterID, const std::string& _name);
    void onDsmFlagsChanged(dsuid_t _meterID, const std::bitset<8> _flags);
    virtual void muteFire(CallOrigin origin);
    bool isInitializing() const { return m_IsInitializing; }
    void setApartment(Apartment* _value);
    void setMetering(metering::Metering* _value);
    void setBusInterface(BusInterface& bus);

    virtual TaskProcessor& getTaskProcessor() { return m_taskProcessor; }
    TaskProcessor& getTaskProcessorMaySleep() { return m_taskProcessorMaySleep; }

    virtual void scheduleDeviceReadout(const dsuid_t& _dSMeterID, std::shared_ptr<Task> task);

    virtual void asyncDiscoverDS485Devices();

  protected:
    virtual void doStart();
    bool handleModelEvents(); //< access from unit test
  private:
    bool handleDeferredModelEvents();
    void handleDeferredModelStateChanges(CallOrigin _origin, int _zoneID, int _groupID, int _sceneID);
    void discoverDS485Devices();
    void readConfiguration();
    void writeConfiguration();
    void scheduleConfigWrite();
    void delayedConfigWrite();
    void migrateHeating();

    void raiseEvent(const std::shared_ptr<Event> &event);
    void notifyModelConsistent();

    void onDeviceCallScene(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const bool _forced, const std::string _token, ds::DateTime timeStamp);
    void onDeviceSaveScene(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const std::string _token);
    void onDeviceUndoScene(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const std::string _token, ds::DateTime timeStamp);
    void onDeviceBlink(const dsuid_t& _dsMeterID, const int _deviceID, const int _originDeviceID, const CallOrigin _origin, const std::string _token);
    void onDeviceActionEvent(const dsuid_t& _dsMeterID, const int _deviceID, const int _buttonNr, const int _clickType);
    void onDeviceDirectActionEvent(const dsuid_t& _dsMeterID, const int _deviceID, const int _actionID);

    void onGroupCallSceneFiltered(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, const CallOrigin _origin, const bool _forced, const std::string _token);
    void onDeviceActionFiltered(dsuid_t _source, const int _deviceID, const int _buttonNr, const int _clickType);
    void onGroupBlink(dsuid_t _source, const int _zoneID, const int _groupID, const int _originDeviceID, const CallOrigin _origin, const std::string _token);

    void onAddDevice(const dsuid_t& _dsMeterID, const int _zoneID, const int _devID);
    void onRemoveDevice(const dsuid_t& _dsMeterID, const int _zoneID, const int _devID);
    void onJoinedDSMeter(const dsuid_t& _dsMeterID);
    void onLinkLost();
    void onLostDSMeter(const dsuid_t& _dsMeterID);
    void onDeviceConfigChanged(const dsuid_t& _dsMeterID, int _deviceID,
                               int _configClass, int _configIndex, int _value);
    void onDeviceConfigRead(const dsuid_t& _dsMeterID, int _deviceID,
                               int _configClass, int _configIndex, int _value);

    void rescanDevice(const dsuid_t& _dsMeterID, const int _deviceID);
    void rescanDevice(const dsuid_t& dsuid, uint8_t value);
    void onDeviceSetOutputValue(const dsuid_t& dsuidMeter, int deviceID, uint8_t value, ds::DateTime timeStamp);
    void onZoneGroupSetOutputValue(int zoneID, int groupID, uint8_t value, ds::DateTime timeStamp);
    void onZoneGroupSetSceneLocalPrio(uint16_t zoneID, uint8_t groupID, uint8_t sceneID);
    void onZoneGroupCallSceneMin(uint16_t zoneID, uint8_t groupID, uint8_t sceneID, ds::DateTime timeStamp);
    void onGroupStopDimming(const int _zoneID, const int _groupID, const int _originDeviceID, const int _sceneID, ds::DateTime timeStamp);
    void onSensorEvent(dsuid_t _meterID, const devid_t _deviceID, const int& _eventIndex);
    void onBinaryInputEvent(dsuid_t _meterID, const devid_t _deviceID, const int& _eventIndex, BinaryInputType _eventType,
        BinaryInputStateValue _state);
    void onSensorValue(dsuid_t _meterID, const devid_t _deviceID, const int& _sensorIndex, const int& _sensorValue);
    void onSensorValueEx(dsuid_t _meterID, const devid_t _deviceID, const int& _sensorIndex, const double& _sensorValue, const uint32_t& _sensorAge, const uint32_t& _contextId, const std::string& _contextMsg);
    void onCircuitPowerStateChange(dsuid_t _meterID, const int& _baseIndex, const int& _stateMask, const int& _stateValue);
    void onZoneSensorValue(const dsuid_t &_meterID, const dsuid_t& _sourceDevice, const int& _zoneID, const int& _groupID, SensorType _sensorType, const int& _sensorValue, const int& _precision);
    void onEANReady(dsuid_t _dsMeterID, const devid_t _deviceID,
                      const DeviceOEMState_t _state, const DeviceOEMInetState_t _iNetState,
                      const unsigned long long _eanNumber,
                      const int _serialNumber, const int _partNumber,
                      const bool _isIndependent, const bool _isConfigLocked,
                      const int _pairedDevices, const bool _isVisible);
    void onOEMDataReady(dsuid_t _deviceID,
                           const DeviceOEMState_t _state, const std::string& _productName,
                           const std::string& _iconPath, const std::string& _productURL,
                           const std::string& _defaultName,
                           const std::string& _configLink);
    void onDeviceOEMDataUpdateProductInfoState(dsuid_t _deviceID,
                                               const DeviceOEMState_t _state);

    void onHeatingControllerConfig(dsuid_t _dsMeterID, const int _ZoneID, std::shared_ptr<void> _spec);
    void onHeatingControllerValues(dsuid_t _dsMeterID, const int _ZoneID, std::shared_ptr<void> _spec);
    void onHeatingControllerState(dsuid_t _dsMeterID, const int _ZoneID, const int _State);

    void onClusterConfigLock(const int _clusterID, const bool _configurationLock);
    void onClusterLockedScenes(const int _clusterID, const std::vector<int> &_lockedScenes);

    void onGenericEvent(const GenericEventType_t _eventType, const std::shared_ptr<GenericEventPayload_t> &_pPayload, const CallOrigin _origin);
    void onOperationLock(int _zoneID, int _groupID, bool _lock, CallOrigin _callOrigin);
    void onAutoClusterMaintenance(dsuid_t _deviceID);
    void onAutoClusterCleanup();
    void onMeterReady();
    void onDeviceDataReady(dsuid_t _meterID, const devid_t _deviceID,
                        const int& _pairedDevices, const bool& _visible);
    void onDsmStateChange(dsuid_t _meterID, const uint8_t& _state);
    void onVdceEvent(const VdceModelEvent& event);

    void onUpstreamShort(dsuid_t _meterID, uint16_t _deviceID, uint8_t _circuitID, uint8_t _isSensor, uint8_t _keyType,
        uint8_t _clickType);

    void setupWebUpdateEvent();
    void updateWebData(Event& _event, const EventSubscription& _subscription);
    void sendWebUpdateEvent(int _interval = 86400);
    void pollSensors(std::shared_ptr<DeviceReference> pDevRef);
    void synchronizeHeatingAssignment(const dsuid_t& _dSMeterID);

  protected:
    std::atomic<bool> m_IsInitializing; //< allow to clear from unit test
    unsigned m_processedEvents; //< actually dequeued, increased before processing

  private:
    PropertySystem& m_propertySystem;
    EventSink& m_eventSink;
    bool m_IsDirty;
    bool m_pendingSaveRequest;
    DateTime m_pendingSaveRequestTS;
    unsigned m_suppressSaveRequestNotify;
    std::mutex m_SaveRequestMutex;

    std::queue<std::shared_ptr<const ModelEvent>> m_ModelEvents;
    std::mutex m_ModelEventsMutex;
    std::condition_variable m_NewModelEvent;
    std::queue<std::function<void()>> m_workQueue;

    Apartment* m_pApartment;
    metering::Metering* m_pMetering;
    const std::chrono::milliseconds m_EventTimeoutMS;
    BusInterface* m_busInterface{nullptr};

    typedef std::vector<std::shared_ptr<ModelDeferredEvent> > m_DeferredEvents_t;
    std::vector<std::shared_ptr<ModelDeferredEvent> > m_DeferredEvents;

    TaskProcessor m_taskProcessor;
    TaskProcessor m_taskProcessorMaySleep;

    static const std::string kWebUpdateEventName;
    std::shared_ptr<InternalEventRelayTarget> m_pRelayTarget;

    std::mutex m_rvMut;
    std::condition_variable m_rvCond;

    std::shared_ptr<MeterMaintenance> m_pMeterMaintenance;

    std::mutex m_readoutTasksMutex;
    std::vector<std::pair<dsuid_t, std::shared_ptr<Task> > > m_deviceReadoutTasks;

  }; // ModelMaintenance

}

DS_REFLECT_STRUCT((dss), ModelMaintenanceConfig, (minTransactionTime)(maxAggregationWindow));

#endif /* MODELMAINTENANCE_H_ */
