/*
    Copyright (c) 2022 digitalSTROM.org,Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include "model/powerline.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "model/device-mock.h"

static ds::LogChannel dsLogChannel("modelPowerlineTest");
static const char* TAGS = "[model][powerline]";

using namespace dss;
using namespace dss::model;
using namespace trompeloeil;
using trompeloeil::_;

namespace {
TEST_CASE("powerlineTest", TAGS) {
  SECTION("decodeInputCount") {
    SECTION("v100") {
      SECTION("binary hardcoded values") {
        CHECK(decodeInputCount(0x1 << 12 | 0b0 << 6 | 0x7) == 0);
        CHECK(decodeInputCount(0x1000 | 0x0) == 1);
        CHECK(decodeInputCount(0x1000 | 0x1) == 2);
        CHECK(decodeInputCount(0x1000 | 0x2) == 4);
        CHECK(decodeInputCount(0x1000 | 0x7) == 0);
      }
      SECTION("decode-mock helpers") {
        struct X {
          uint16_t functionId;
          int inputCount;
        };
        // v100 only works for light devices, decodeInputCount also includes color in its filter
        const auto& x = GENERATE(
            values(std::initializer_list<X>{{functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100keys::_0), 0},
                                            {functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100keys::_1), 1},
                                            {functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100keys::_2), 2},
                                            {functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100keys::_4), 4}}));
        CHECK(decodeInputCount(x.functionId) == x.inputCount);
      }
    }
    SECTION("v105") {
      SECTION("binary hardcoded values") {
        CHECK(decodeInputCount(0b100 << 6 | 0x0) == 0);
        CHECK(decodeInputCount(0x0100 | 0x2) == 2);
      }
      SECTION("decode-mock helpers") {
        struct X {
          uint16_t functionId;
          int inputCount;
        };
        const auto& x = GENERATE(
            values(std::initializer_list<X>{{functionID105(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_0), 0},
                                            {functionID105(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_1), 1},
                                            {functionID105(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_2), 2},
                                            {functionID105(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_4), 4}}));

        CHECK(decodeInputCount(x.functionId) == x.inputCount);
      }
    }

    SECTION("v105v2") {
      SECTION("binary hardcoded values") {
        CHECK(decodeInputCount(0b111 << 6 | 0x3) == 4);
        CHECK(decodeInputCount(0x01c0 | 0x1) == 1);
      }
      SECTION("decode-mock helpers") {
        struct X {
          uint16_t functionId;
          int inputCount;
        };
        const auto& x = GENERATE(
            values(std::initializer_list<X>{{functionID105v2(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_0), 0},
                                            {functionID105v2(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_1), 1},
                                            {functionID105v2(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_2), 2},
                                            {functionID105v2(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_4), 4}}));

        CHECK(decodeInputCount(x.functionId) == x.inputCount);
      }
    }
    SECTION("weird clima subclasses") {
      // probably no device produced? busscanner was not dealing with it
      SECTION("binary hardcoded values") {
        // input deduction same as for 105
        CHECK(decodeInputCount(8 << 6 | 0x3) == 4);
        CHECK(decodeInputCount(9 << 6 | 0x1) == 1);
      }
      SECTION("decode-mock helpers") {
        CHECK(decodeInputCount(functionID105v2Heating(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_1)) == 1);
        CHECK(decodeInputCount(functionID105v2Climate(DeviceClasses_t::DEVICE_CLASS_BL, FID105keys::_2)) == 2);
      }
    }
  }
  SECTION("non-standard R100 devices (Roman)") { // probably pre-trial with bugs
    const auto& x = GENERATE(28688, // 303505d7f8000000000004c000095d7c00 -> GN-KM200, -> should have no input
                             32816  // 3504175fe0000000000000000000671f00 -> SW-TKM210, -> should have 2 inputs
    );
    INFO(x);
    CHECK(decodeInputCount(x) == 0);
  }
}

} // namespace
