/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <memory>

#include "device.h"
#include "nonaddressablemodelitem.h"

namespace dss {

class Group;
class DSMeter;

/** Abstract interface to select certain Devices from a set */
class IDeviceSelector {
public:
  /** The implementor should return true if the device should appear in the
   * resulting set. */
  virtual bool selectDevice(std::shared_ptr<const Device> _device) const = 0;
  virtual ~IDeviceSelector() {}
};

/** Abstract interface to perform an Action on each device of a set */
class IDeviceAction {
public:
  /** This action will be performed for every device contained in the set. */
  virtual bool perform(std::shared_ptr<Device> _device) = 0;
  virtual ~IDeviceAction() {}
};

/** A set holds an arbitrary list of devices.
 * A Command sent to an instance of this class will replicate the command to all
 * contained devices.
 * Only references to devices will be stored.
 */
class Set final : public NonAddressableModelItem {
private:
  std::vector<DeviceReference> m_ContainedDevices;

public:
  Set() = default;
  Set(const Set&) = default;
  Set(Set&&) = default;
  Set& operator=(const Set&) = default;
  Set& operator=(Set&&) = default;

  /** Constructor for a set containing only \a _device. */
  Set(std::shared_ptr<const Device> _device);
  /** Constructor for a set containing only \a _reference. */
  Set(DeviceReference& _reference);
  /** Constructor for a set containing \a _devices. */
  Set(std::vector<DeviceReference> _devices);

  /** Performs the given action on all contained devices */
  void perform(IDeviceAction& _deviceAction);

  /** Returns a subset selected by the given selector
   * A device will be included in the resulting set if _selector.selectDevice(...) return true.
   */
  Set getSubset(const IDeviceSelector& _selector) const;
  /** Returns a subset of the devices which are member of the given group
   * Note that these groups could be spanned over multiple dsMeters.
   */
  Set getByGroup(int _groupNr) const;
  /** Returns a subset of the devices which are member of the given group
   * Note that these groups could be spanned over multiple dsMeters.
   */
  Set getByGroup(std::shared_ptr<const Group> _group) const;
  /** Returns a subset of the devices which are member of the given group
   * Note that these groups could be spanned over multiple dsMeters.
   */
  Set getByGroup(const std::string& _name) const;

  Set getByColor(DeviceClasses_t _color) const;

  /** Returns a subset of the devices which have sensors of the given type */
  Set getBySensorType(SensorType _type) const;

  /** Returns a subset of devices with the given function-id. */
  Set getByFunctionID(const int _functionID) const;

  /** Returns a subset that contains all devices belonging to Zone \a _zoneID. */
  Set getByZone(int _zoneID) const;

  /** Returns a subset that contains all devices belonging to Zone \a _zoneName */
  Set getByZone(const std::string& _zoneName) const;

  /** Returns a subset that contains all devices belonging to DSMeter \a _dsMeter */
  Set getByDSMeter(std::shared_ptr<const DSMeter> _dsMeter) const;
  /** Returns a subset that contains all devices belonging to DSMeter identified by \a _dsMeterDSID */
  Set getByDSMeter(const dsuid_t& _dsMeterDSID) const;
  /** Returns a subset that contains all devices known to be registered last on DSMeter */
  Set getByLastKnownDSMeter(const dsuid_t& _dsMeterDSID) const;

  /** Returns a subset that contains all devices that have the presence state of \a _present */
  Set getByPresence(const bool _present) const;

  Set getByTag(const std::string& _tagName) const;

  /** Returns the device indicated by _name
   */
  DeviceReference getByName(const std::string& _name) const;
  /** Returns the device indicated by \a _busid */
  DeviceReference getByBusID(const devid_t _busid, const dsuid_t& _dsMeterID) const;
  DeviceReference getByBusID(const devid_t _busid, std::shared_ptr<const DSMeter> _meter) const;

  /** Returns the device indicated by \a _dsid */
  DeviceReference getByDSID(const dsuid_t _dsid) const;

  /** Returns the device indicated by \a _serial */
  DeviceReference getBySerial(const uint32_t _serial) const;

  /** Returns distance to the device in Set */
  int findByDSID(const dsuid_t& dsid) const;

  /** Returns the number of devices contained in this set */
  int length() const;
  /** Returns true if the set is empty */
  bool isEmpty() const;
  bool empty() const { return isEmpty(); } // for std compatibility
  size_t size() const { return length(); } // for std compatibility

  /** Returns a set that's combined with the set _other.
   * Duplicates get filtered out.
   */
  Set combine(Set& _other) const;
  /* Returns a set with all device in _other removed */
  Set remove(const Set& _other) const;

  /** Returns the \a _index'th device */
  const DeviceReference& get(int _index) const;
  /** @copydoc get */
  const DeviceReference& operator[](const int _index) const;

  /** @copydoc get */
  DeviceReference& get(int _index);
  /** @copydoc get */
  DeviceReference& operator[](const int _index);

  const std::vector<Dsuid> getDsuidList() const;

  /** Returns true if the set contains \a _device */
  bool contains(const DeviceReference& _device) const;
  /** Returns true if the set contains \a _device */
  bool contains(std::shared_ptr<const Device> _device) const;

  /** Adds the device \a _device to the set */
  void addDevice(const DeviceReference& _device);
  /** Adds the device \a _device to the set */
  void addDevice(std::shared_ptr<const Device> _device);

  /** Removes the device \a _device from the set */
  void removeDevice(const DeviceReference& _device);
  /** Removes the device \a _device from the set */
  void removeDevice(std::shared_ptr<const Device> _device);

  void nextScene(const CallOrigin _origin, const SceneAccessCategory _category) override;
  void previousScene(const CallOrigin _origin, const SceneAccessCategory _category) override;

  unsigned long getPowerConsumption() override;

protected:
  std::vector<std::shared_ptr<AddressableModelItem> > splitIntoAddressableItems() override;
}; // Set

class OutputPresentFilter : public IDeviceSelector {
public:
  virtual bool selectDevice(std::shared_ptr<const Device> _device) const { return _device->hasOutput(); }
};

class LegacyAwningDeviceFilter : public IDeviceSelector {
public:
  virtual bool selectDevice(std::shared_ptr<const Device> _device) const { return _device->isLegacyAwning(); }
};

class KLDeviceFilter : public IDeviceSelector {
public:
  virtual bool selectDevice(std::shared_ptr<const Device> _device) const {
    return _device->getDeviceType() == DEVICE_TYPE_KL;
  }
};

} // namespace dss
