/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland
    Copyright (c) 2022 Aizo AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "oemdatareader.h"

#include <ds/log.h>

#include "ds485/dsbusinterface.h"
#include "dss.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "model/set.h"

#define TASK_REQUE_ON_FAILURE_SLEEP_SECONDS 4

static ds::LogChannel dsLogChannel("OemDataReader");

namespace dss {
ConfigReaderHelper::ConfigReaderHelper(const std::string& busConnection)
    : m_busConnection(busConnection), m_dsmApiHandle(NULL), m_deviceAdress(0), m_dsmId(), m_revisionID(0) {}

ConfigReaderHelper::~ConfigReaderHelper() {
  if (m_dsmApiHandle != NULL) {
    DsmApiClose(m_dsmApiHandle);
    DsmApiCleanup(m_dsmApiHandle);
    m_dsmApiHandle = NULL;
  }
}

void ConfigReaderHelper::setup(std::shared_ptr<Device> device) {
  m_deviceAdress = device->getShortAddress();
  m_dsmId = device->getDSMeterDSID();
  m_revisionID = device->getRevisionID();
  if (DSS::hasInstance()) {
    m_dsm = DSS::getInstance()->getApartment().getDSMeterByDSID(m_dsmId);
  }
}

uint16_t ConfigReaderHelper::getDeviceConfigWord(uint8_t class_, uint8_t index) const {
  if (m_dsmApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  // wait until dsm has finished device registrations
  while (m_dsm && m_dsm->isPresent() && (m_dsm->getState() == DSM_STATE_REGISTRATION)) {
    sleep(5);
  }
  if ((m_dsm == NULL) || !m_dsm->isPresent()) {
    throw ItemNotFoundException("OEMDataReader: dsm is not present");
  }

  uint16_t retVal;
  int ret = DeviceConfig_get_sync_16(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  if (ret == ERROR_SYNC_RESPONSE_TIMEOUT || ret == ERROR_TIMEOUT) {
    ret = DeviceConfig_get_sync_16(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  }
  DSBusInterface::checkResultCode(ret);

  return retVal;
}

uint8_t ConfigReaderHelper::getDeviceConfig(uint8_t class_, uint8_t index) const {
  if (m_dsmApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  // wait until dsm has finished device registrations
  while (m_dsm && m_dsm->isPresent() && (m_dsm->getState() == DSM_STATE_REGISTRATION)) {
    sleep(5);
  }
  if ((m_dsm == NULL) || !m_dsm->isPresent()) {
    throw ItemNotFoundException("OEMDataReader: dsm is not present");
  }

  uint8_t retVal;
  int ret = DeviceConfig_get_sync_8(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  if (ret == ERROR_SYNC_RESPONSE_TIMEOUT || ret == ERROR_TIMEOUT) {
    ret = DeviceConfig_get_sync_8(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  }
  DSBusInterface::checkResultCode(ret);

  return retVal;
}

OEMDataReader::OEMDataReader(const std::string& busConnection) : ConfigReaderHelper(busConnection) {}

OEMDataReader::~OEMDataReader() {}

void OEMDataReader::run() {
  uint64_t ean = 0;
  uint16_t serialNumber = 0;
  uint8_t partNumber = 0;
  bool isVisible = false;
  bool isIndependent = false;
  bool isConfigLocked = false;
  uint8_t pairedDevices = 0;
  DeviceOEMState_t state = DEVICE_OEM_UNKNOWN;
  DeviceOEMInetState_t deviceInetState = DEVICE_OEM_EAN_NO_EAN_CONFIGURED;

  m_dsmApiHandle = DsmApiInitialize();
  if (!m_dsmApiHandle) {
    throw std::runtime_error("OEMDataReader: Unable to get dsmapi handle");
  }

  int result = DsmApiOpen(m_dsmApiHandle, m_busConnection.c_str(), 0);
  if (result < 0) {
    throw std::runtime_error(std::string("OEMDataReader: Unable to open connection to: ") + m_busConnection);
  }

  try {
    // read device-active data and EAN part number
    uint16_t result0;
    result0 = getDeviceConfigWord(1, 0x1e);
    partNumber = result0 & 0x7F;
    isIndependent = (result0 & 0x80);
    uint8_t upper = result0 >> 8;
    isVisible = !(upper & (1 << 4));
    pairedDevices = (upper & 0x0f);

    // optimize readout time: avoid switching banks
    serialNumber = getDeviceConfigWord(1, 0x1c);

    // check if EAN is programmed: Bank 3: 0x2e-0x2f 0x0000 < x < 0xffff
    uint16_t result;
    result = getDeviceConfigWord(3, 0x2e);
    deviceInetState = (DeviceOEMInetState_t)(result >> 12);

    if (deviceInetState == DEVICE_OEM_EAN_NO_EAN_CONFIGURED) {
      // no EAN programmed
      state = DEVICE_OEM_NONE;
    } else {
      // EAN programmed
      ean |= ((long long unsigned int)(result & 0xFFF) << 32);

      result = getDeviceConfigWord(3, 0x2a);
      ean |= result;

      result = getDeviceConfigWord(3, 0x2c);
      ean |= ((long long unsigned int)result << 16);

      state = DEVICE_OEM_VALID;
    }

    if (m_revisionID >= TBVersion_OemConfigLock) {
      uint8_t deviceActive;
      deviceActive = getDeviceConfig(3, 0x1f);
      if (std::bitset<8>(deviceActive).test(0)) {
        isConfigLocked = true;
      }
    }

    auto pEvent = std::make_shared<ModelEventWithDSID>(ModelEvent::etDeviceEANReady, m_dsmId);
    pEvent->addParameter(m_deviceAdress);
    pEvent->addParameter(state);
    pEvent->addParameter(deviceInetState);
    pEvent->addParameter(ean >> 32);
    pEvent->addParameter(ean & 0xFFFFFFFF);
    pEvent->addParameter(serialNumber);
    pEvent->addParameter(partNumber);
    pEvent->addParameter(isIndependent);
    pEvent->addParameter(isConfigLocked);
    pEvent->addParameter(pairedDevices);
    pEvent->addParameter(isVisible);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
    }
  } catch (ItemNotFoundException& infe) {
    Logger::getInstance()->log("OEMDataReader::run: dSM not found or left, aborting readout\n");
    if (DSS::hasInstance()) {
      if (m_dsm) {
        std::shared_ptr<Device> dev = m_dsm->getDevices().getByBusID(m_deviceAdress, m_dsm).getDevice();
        dev->setOemInfoState(DEVICE_OEM_UNKNOWN);
        DSS::getInstance()->getModelMaintenance().addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
      }
    }
  } catch (const std::exception& er) {
    Logger::getInstance()->log(std::string("OEMDataReader::run: bus error: ") + er.what() + " reading from dSM " +
                                   dsuid2str(m_dsmId) + " DeviceId " + intToString(m_deviceAdress) +
                                   ", requeueing task",
                               lsWarning);

    if (!DSS::hasInstance() || !m_dsm) {
      return;
    }

    DsmApiClose(m_dsmApiHandle);
    DsmApiCleanup(m_dsmApiHandle);
    m_dsmApiHandle = NULL;

    try {
      std::shared_ptr<Device> dev = m_dsm->getDevices().getByBusID(m_deviceAdress, m_dsm).getDevice();
      DS_REQUIRE(dev->isPresent(), "Device was physically removed");
      sleep(TASK_REQUE_ON_FAILURE_SLEEP_SECONDS);
      DSS::getInstance()->getModelMaintenance().scheduleDeviceReadout(
          m_dsmId, std::shared_ptr<OEMDataReader>(shared_from_this()));
    } catch (const std::exception& e) {
      DS_ERROR("OEMDataReader::run: bus error: ", e, " reading from dSM ", m_dsmId, " DeviceId ", m_deviceAdress);
    }
  }
}

TNYConfigReader::TNYConfigReader(const std::string& busConnection) : ConfigReaderHelper(busConnection) {}

TNYConfigReader::~TNYConfigReader() {}

void TNYConfigReader::run() {
  bool isVisible = false;
  uint8_t pairedDevices = 0;

  m_dsmApiHandle = DsmApiInitialize();
  if (!m_dsmApiHandle) {
    throw std::runtime_error("TNYConfigReader: Unable to get dsmapi handle");
  }

  int result = DsmApiOpen(m_dsmApiHandle, m_busConnection.c_str(), 0);
  if (result < 0) {
    throw std::runtime_error(std::string("TNYConfigReader: Unable to open connection to: ") + m_busConnection);
  }

  try {
    // read device-active data and pairing data
    uint8_t cfg;
    cfg = getDeviceConfig(1, CfgFunction_DeviceActive);
    isVisible = !(cfg & (1 << 4));
    pairedDevices = (cfg & 0x0f);

    auto pEvent = std::make_shared<ModelEventWithDSID>(ModelEvent::etDeviceDataReady, m_dsmId);
    pEvent->addParameter(m_deviceAdress);
    pEvent->addParameter(pairedDevices);
    pEvent->addParameter(isVisible);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
    }
  } catch (ItemNotFoundException& infe) {
    Logger::getInstance()->log("TNYConfigReader::run: dSM not found or left, aborting readout\n");
  } catch (std::runtime_error& er) {
    // Bus error
    Logger::getInstance()->log(std::string("TNYConfigReader::run: bus error: ") + er.what() + " reading from dSM " +
                                   dsuid2str(m_dsmId) + " DeviceId " + intToString(m_deviceAdress) +
                                   ", requeueing task",
                               lsWarning);

    if (!DSS::hasInstance() || !m_dsm) {
      return;
    }

    DsmApiClose(m_dsmApiHandle);
    DsmApiCleanup(m_dsmApiHandle);
    m_dsmApiHandle = NULL;

    try {
      std::shared_ptr<Device> dev = m_dsm->getDevices().getByBusID(999, m_dsm).getDevice();
      DS_REQUIRE(dev->isPresent(), "Device was physically removed");
      sleep(TASK_REQUE_ON_FAILURE_SLEEP_SECONDS);
      DSS::getInstance()->getModelMaintenance().scheduleDeviceReadout(
          m_dsmId, std::shared_ptr<TNYConfigReader>(shared_from_this()));
    } catch (const std::exception& e) {
      DS_ERROR("OEMDataReader::run: bus error: ", e, " reading from dSM ", m_dsmId, " DeviceId ", m_deviceAdress,
               ", requeueing task");
    }
  }
}

} // namespace dss
