/*
    Copyright (c) 2010,2011,2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "vdcdataquery.h"

#include <ds/log.h>

#include <boost/filesystem.hpp>

#include "dss.h"
#include "model/device.h"
#include "model/modelmaintenance.h"
#include "oemwebquery.h"
#include "propertysystem.h"
#include "vdc-connection.h"

static ds::LogChannel dsLogChannel("scannerVdcDataQuery");

namespace dss {
VdcDataQuery::VdcDataQuery(std::shared_ptr<Device> _device, PropertySystem& propertySystem)
    : Task(), m_propertySystem(propertySystem), m_Device(_device) {}

void VdcDataQuery::run() {
  try {
    const std::string& oemModelGuid = m_Device->getVdsdSpec().oemModelGuid;
    if (beginsWith(oemModelGuid, "gs1:")) {
      // For the time being assume that all IP Devices with an oemModelGuid/GTIN
      // are registered at the dS-Article Db. This might be a configurable property in the future.
      const DeviceOEMInetState_t iNetState = DEVICE_OEM_EAN_INTERNET_ACCESS_MANDATORY;
      try {
        // format of oemModelGuid: gs1:(01)123456789
        std::string eanString = oemModelGuid.substr(4);
        if (beginsWith(eanString, "(01)")) {
          eanString.erase(0, 4);
        }
        unsigned long long eanNumber = strToULongLong(eanString);
        m_Device->setOemInfo(eanNumber, 0, 1, iNetState, true);
        if ((iNetState == DEVICE_OEM_EAN_INTERNET_ACCESS_OPTIONAL) ||
            (iNetState == DEVICE_OEM_EAN_INTERNET_ACCESS_MANDATORY)) {
          TaskProcessor& tp(DSS::getInstance()->getModelMaintenance().getTaskProcessor());
          tp.addEvent(std::make_shared<OEMWebQuery>(WebserviceConnection::getInstanceMsHub(), m_Device,
                                                    m_Device->getOemProductInfoState(), m_propertySystem));
          m_Device->setOemProductInfoState(DEVICE_OEM_LOADING);
        } else {
          m_Device->setOemProductInfoState(DEVICE_OEM_NONE);
        }
        m_Device->setOemInfoState(DEVICE_OEM_VALID);
      } catch (std::invalid_argument& e) {
        DS_WARNING("VdcDataQuery: could not convert gtin for device", m_Device->getDSID(), oemModelGuid);
      }
    }

  } catch (std::runtime_error& e) {
    DS_WARNING("VdcDataQuery: could not query properties from", m_Device->getDSID(), "Message:", e.what());
  }

  uint8_t* data;
  size_t dataSize = 0;

  try {
    VdcHelper::getIcon(m_Device->getDSMeterDSID(), m_Device->getDSID(), &dataSize, &data);
  } catch (std::runtime_error& e) {
    DS_WARNING("VdcDataQuery: could not query icon data from", m_Device->getDSID(), "Message:", e.what());
  }
  if (dataSize > 0) {
    boost::filesystem::path iconBasePath;
    boost::filesystem::path iconFile;
    iconBasePath = m_propertySystem.getStringValue("/config/subsystems/Apartment/iconBasePath");
    iconFile = dsuid2str(m_Device->getDSID()) + ".png";
    boost::filesystem::path iconPath = iconBasePath / iconFile;

    /* write binary file from memory */
    FILE* binFile = fopen(iconPath.c_str(), "wb");
    if (binFile) {
      fwrite(data, sizeof(uint8_t), dataSize, binFile);
      fclose(binFile);
      m_Device->setVdcIconPath(iconFile.string());
    }
  }
  free(data);
}
} // namespace dss
