/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "apartmentrequesthandler.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

#include "event-mock.h"
#include "metering/metering-mock.h"
#include "model/apartment-mock.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance-mock.h"
#include "propertysystem-mock.h"
#include "web/restful.h"

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;
using namespace std::literals;

namespace {

static ds::LogChannel dsLogChannel("webHandlerApartmentRequestHandlerTest");
static const char* TAGS = "[web][handler][apartmentrequesthandler]";


TEST_CASE("dssWebHandlerApartmentRequestHandler", TAGS) {
  PropertySystemMock propertySystem{};
  EventSinkMock eventSink{};
  auto apartment = ApartmentMock{};

  auto modelMaintenance = ModelMaintenanceMock{propertySystem, eventSink};
  auto apartmentRequestHandler = ApartmentRequestHandler(apartment, modelMaintenance);

  dss::metering::MeteringMock meteringMock{};
  ALLOW_CALL(apartment, getMetering()).LR_RETURN(meteringMock);

  SECTION("Requests") {
    std::stringstream ss;

    SECTION("getConsumption") {
      REQUIRE_CALL(meteringMock, getConsumedPowerApartment()).RETURN(77);
      ss << apartmentRequestHandler.jsonHandleRequest(RestfulRequest{"foo/getConsumption", ""s}, nullptr, nullptr);
      CHECK(ss.str() == R"({"result":{"consumption":77},"ok":true})");
    }
    SECTION("getModelFeatures") {
      REQUIRE_CALL(apartment.modelFeatures, getFeatures(_))
          .TIMES(9)
          .RETURN(std::vector{std::pair{
              "foo"s,
              ModelFeatures::features_t{{ModelFeatureId::dontcare, ModelFeatureId::ledauto, ModelFeatureId{1000}}}}});
      REQUIRE_CALL(apartment.modelFeatures, getAvailableFeatures())
          .RETURN(std::vector{ModelFeatureId::dontcare, ModelFeatureId::pushbdisabled});
      ss << apartmentRequestHandler.jsonHandleRequest(RestfulRequest{"foo/getModelFeatures", ""s}, nullptr, nullptr);
      auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
      REQUIRE(reply.ok);

      auto modelFeaturesReply = ds::fromJson<std::map<std::string, ds::JsonFragment>>(reply.result);
      auto modelFeatures = ds::fromJson<std::map<std::string, ds::JsonFragment>>(modelFeaturesReply["modelFeatures"]);
      CHECK(ds::str(modelFeatures["GE"]) == R"({"foo":{"dontcare":true,"ledauto":true}})");

      auto reference = ds::fromJson<std::vector<std::pair<ModelFeatureId, bool>>>(modelFeaturesReply["reference"]);
      CHECK(reference ==
            std::vector{std::pair{ModelFeatureId::dontcare, false}, std::pair{ModelFeatureId::pushbdisabled, false}});
    }
    SECTION("getModelFeatures2") {
      REQUIRE_CALL(apartment.modelFeatures, getFeatures(_))
          .TIMES(9)
          .RETURN(std::vector{
              std::pair{"foo"s, ModelFeatures::features_t{{ModelFeatureId::dontcare, ModelFeatureId::highlevel},
                                                          {ModelFeatureId::akmdelay, ModelFeatureId{1000}}}}});
      REQUIRE_CALL(apartment.modelFeatures, getAvailableFeatures())
          .RETURN(std::vector{ModelFeatureId::dontcare, ModelFeatureId::pushbdisabled});
      ss << apartmentRequestHandler.jsonHandleRequest(RestfulRequest{"foo/getModelFeatures2", ""s}, nullptr, nullptr);
      auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
      REQUIRE(reply.ok);

      auto modelFeaturesReply = ds::fromJson<std::map<std::string, ds::JsonFragment>>(reply.result);
      auto modelFeatures = ds::fromJson<std::map<std::string, ds::JsonFragment>>(modelFeaturesReply["modelFeatures"]);
      CHECK(ds::str(modelFeatures["GE"]) == R"({"foo":[{"dontcare":true,"highlevel":true},{"akmdelay":true}]})");

      auto reference = ds::fromJson<std::vector<ModelFeatureId>>(modelFeaturesReply["reference"]);
      CHECK(reference == std::vector{ModelFeatureId::dontcare, ModelFeatureId::pushbdisabled});
    }
  }
}

} // namespace
