/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "meteringrequesthandler.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "metering/metering-mock.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "web/restful.h"

static ds::LogChannel dsLogChannel("webHandlerMeteringRequestHandlerTest");

using namespace dss;
using namespace dss::metering;
using namespace trompeloeil;
using trompeloeil::_;
using namespace std::literals;

namespace {

static const char* TAGS = "[web][handler][metering]";

TEST_CASE("dssWebHandlerMeteringRequest", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apartment = *pApartment;
  auto metering = MeteringMock{};

  auto meteringRequestHandler = MeteringRequestHandler(apartment, metering);

  auto controllerId1 = dsModel::ControllerId{"3504175fe000000000000010000003dd00"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  controller1->setCapability_HasMetering(true);

  std::stringstream ss; // helper to

  /*SECTION("getResolutions") {
    REQUIRE_CALL(metering, getConfig()).RETURN(std::vector<TimeSeriesConfig>{{1s, 600}, {1min, 720}, {15min, 2976}});

    ss << meteringRequestHandler.jsonHandleRequest(RestfulRequest{"foo/getResolutions", {}}, nullptr, nullptr);
    CHECK(ss.str() ==
          R"({"result":{"resolutions":[{"resolution":"1"},{"resolution":"60"},{"resolution":"900"}]},"ok":true})");
  }

  SECTION("getSeries") {
    ss << meteringRequestHandler.jsonHandleRequest(RestfulRequest{"foo/getSeries", {}}, nullptr, nullptr);
    CHECK(
        ss.str() ==
        R"({"result":{"series":[{"dSUID":"3504175fe000000000000010000003dd00","type":"energy"},{"dSUID":"3504175fe000000000000010000003dd00","type":"energyDelta"},{"dSUID":"3504175fe000000000000010000003dd00","type":"power"}]},"ok":true})");
  }

  SECTION("getValues") {
    SECTION("power") {
      SECTION("single meterid") {
        REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=power",
                                                    "&startTime=300", "&endTime=360")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"power","unit":"W","resolution":"1","values":[[301,13.5]]},"ok":true})");
      }
      SECTION(".meters(all)") {
        WHEN("meter present") {
          REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
              .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
              .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

          ss << meteringRequestHandler.jsonHandleRequest(
              RestfulRequest{"foo/getValues", ds::str("dsuid=.meters(all)", "&resolution=1", "&type=power",
                                                      "&startTime=300", "&endTime=360")},
              nullptr, nullptr);
          CHECK(
              ss.str() ==
              R"({"result":{"meterID":["3504175fe000000000000010000003dd00"],"type":"power","unit":"W","resolution":"1","values":[[301,13.5]]},"ok":true})");
        }
        WHEN("no meters present") {
          apartment.removeDSMeter(Dsuid{controllerId1});
          REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _)).RETURN(ValueRange{});
          ss << meteringRequestHandler.jsonHandleRequest(
              RestfulRequest{"foo/getValues", ds::str("dsuid=.meters(all)", "&resolution=1", "&type=power",
                                                      "&startTime=300", "&endTime=360")},
              nullptr, nullptr);
          CHECK(ss.str() == R"({"ok":false,"message":"Could not find data for 'power' and resolution '1'"})");
        }
      }
    }
    SECTION("energyDelta") {
      REQUIRE_CALL(metering, getSeries(_, _, _, true, _, _))
          .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
          .LR_SIDE_EFFECT(CHECK(_2 == 1s));

      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=energyDelta",
                                                  "&startTime=300", "&endTime=360")},
          nullptr, nullptr);
      CHECK(
          ss.str() ==
          R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"energyDelta","unit":"Wh","resolution":"1","values":[[301,13.5]]},"ok":true})");
    }
    SECTION("unit=Ws") {
      REQUIRE_CALL(metering, getSeries(_, _, _, false, _, _))
          .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
          .LR_SIDE_EFFECT(CHECK(_2 == 1s));

      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=energyDelta",
                                                  "&startTime=300", "&endTime=360", "&unit=Ws")},
          nullptr, nullptr);
      CHECK(
          ss.str() ==
          R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"energyDelta","unit":"Ws","resolution":"1","values":[[301,13.5]]},"ok":true})");
    }
    SECTION("unit=W for energy") {
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=energyDelta",
                                                  "&startTime=300", "&endTime=360", "&unit=W")},
          nullptr, nullptr);
      CHECK(ss.str() == R"({"ok":false,"message":"Invalid unit 'W'"})");
    }
    SECTION("unit=Wh for power") {
      REQUIRE_CALL(metering, getSeries(_, _, SeriesTypes::power, false, _, _))
          .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}});
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=power", "&unit=Wh")},
          nullptr, nullptr);

      constexpr auto expect =
          R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"power","unit":"W","resolution":"1","values":[[301,13.5]]},"ok":true})"sv;
      CHECK(ss.str() == expect);
    }
    SECTION("type=schmack") {
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=schmack",
                                                  "&startTime=300", "&endTime=360")},
          nullptr, nullptr);
      constexpr auto expect =
          R"({"ok":false,"message":"Unknown ::dss::metering::SeriesTypes enum option. value:schmack)"sv;
      CHECK(ss.str().substr(0, expect.size()) == expect);
    }
    SECTION("type=consumption (legacy power)") {
      REQUIRE_CALL(metering, getSeries(_, _, SeriesTypes::power, false, _, _))
          .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}});
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=consumption",
                                                  "&startTime=300", "&endTime=360")},
          nullptr, nullptr);
      constexpr auto expect =
          R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"power","unit":"W","resolution":"1","values":[[301,13.5]]},"ok":true})"sv;
      CHECK(ss.str().substr(0, expect.size()) == expect);
    }
    SECTION("type missing") {
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1")}, nullptr, nullptr);
      constexpr auto expect = R"({"ok":false,"message":"Missing parameter name:type)"sv;
      CHECK(ss.str().substr(0, expect.length()) == expect);
    }
    SECTION("dsuid missing") {
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("&resolution=1", "&type=power", "&startTime=300", "&endTime=360")},
          nullptr, nullptr);
      constexpr auto expect = R"({"ok":false,"message":"Missing parameter name:dsuid)"sv;
      CHECK(ss.str().substr(0, expect.length()) == expect); // strip file/line number
    }
    SECTION("dsuid invalid") {
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues",
                         ds::str("dsuid=.meters(aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa00)", "&resolution=1", "&type=power")},
          nullptr, nullptr);
      constexpr auto expect = R"({"ok":false,"message":"Could not find dsMeter with given dsuid."})"sv;
      CHECK(ss.str().substr(0) == expect); // strip file/line number
    }
    SECTION("valueCount=10") {
      REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
          .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
          .LR_SIDE_EFFECT(CHECK(_6 == 10));
      meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues",
                         ds::str("dsuid=", controllerId1, "&resolution=1", "&type=energy", "&valueCount=10")},
          nullptr, nullptr);
    }
    SECTION("resolution=60") {
      REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
          .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
          .LR_SIDE_EFFECT(CHECK(_2 == 60s));
      meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=60", "&type=energy")}, nullptr,
          nullptr);
    }
    SECTION("resolution missing") {
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&type=energy")}, nullptr, nullptr);
      constexpr auto expect = R"({"ok":false,"message":"Missing parameter name:resolution)"sv;
      CHECK(ss.str().substr(0, expect.size()) == expect); // strip file/line number
    }
    SECTION("resolution invalid") {
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&type=energy", "&resolution=NaN")}, nullptr,
          nullptr);
      constexpr auto expect =
          R"({"ok":false,"message":"bad lexical cast: source type value could not be interpreted as target"})"sv;
      CHECK(ss.str().substr(0) == expect); // strip file/line number
    }
    SECTION("with production") {
      SECTION("normal") {
        REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        REQUIRE_CALL(metering, getProducedSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 73.5, .timeStamp = ds::DateTime{301s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=power",
                                                    "&startTime=300", "&endTime=360", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"power","unit":"W","resolution":"1","values":[[301,13.5,73.5]]},"ok":true})");
      }
      SECTION("no production values") {
        REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        REQUIRE_CALL(metering, getProducedSeries(_, _, _, _, _, _))
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}))
            .RETURN(ValueRange{});

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=power",
                                                    "&startTime=300", "&endTime=360", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"power","unit":"W","resolution":"1","values":[[301,13.5,0.0]]},"ok":true})");
      }
      SECTION("production last value missing") {
        REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}},
                                                     Value{.value = 14.5, .timeStamp = ds::DateTime{302s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        REQUIRE_CALL(metering, getProducedSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 0.0, .timeStamp = ds::DateTime{301s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=power",
                                                    "&startTime=300", "&endTime=360", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"power","unit":"W","resolution":"1","values":[[301,13.5,0.0],[302,14.5,0.0]]},"ok":true})");
      }
      SECTION("consumption last value missing") {
        REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 0.0, .timeStamp = ds::DateTime{301s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        REQUIRE_CALL(metering, getProducedSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}},
                                                     Value{.value = 14.5, .timeStamp = ds::DateTime{302s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=power",
                                                    "&startTime=300", "&endTime=360", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"meterID":"3504175fe000000000000010000003dd00","type":"power","unit":"W","resolution":"1","values":[[301,0.0,13.5],[302,0.0,14.5]]},"ok":true})");
      }
      SECTION("production/consumption values timestamps shifted by 1s") {
        REQUIRE_CALL(metering, getSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 0.0, .timeStamp = ds::DateTime{300s}},
                                                     Value{.value = 14.5, .timeStamp = ds::DateTime{301s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        REQUIRE_CALL(metering, getProducedSeries(_, _, _, _, _, _))
            .RETURN(ValueRange{.values = std::vector{Value{.value = 13.5, .timeStamp = ds::DateTime{301s}},
                                                     Value{.value = 14.5, .timeStamp = ds::DateTime{302s}}}})
            .LR_SIDE_EFFECT(CHECK(_5 == TimeDuration{.start = ds::DateTime{300s}, .end = ds::DateTime(360s)}));

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getValues", ds::str("dsuid=", controllerId1, "&resolution=1", "&type=power",
                                                    "&startTime=300", "&endTime=360", "&withProduction=true")},
            nullptr, nullptr);
      }
    }
  }

  SECTION("getLatest") {
    SECTION("type=power") {
      SECTION("single meterid") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=", controllerId1, "&type=power")}, nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":10.0,"date":"1970-01-01 00:05:01"}]},"ok":true})");
      }
      SECTION(".meters(all)") {
        WHEN("meter present") {
          REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
          ss << meteringRequestHandler.jsonHandleRequest(
              RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=power", "&unit=W")}, nullptr,
              nullptr);
          CHECK(
              ss.str() ==
              R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":10.0,"date":"1970-01-01 00:05:01"}]},"ok":true})");
        }
        WHEN("no meters present") {
          apartment.removeDSMeter(Dsuid{controllerId1});
          ss << meteringRequestHandler.jsonHandleRequest(
              RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=power", "&unit=W")}, nullptr,
              nullptr);
          CHECK(ss.str() == R"({"result":{"values":[]},"ok":true})");
        }
      }
      SECTION(".meters(dsid1,dsid2)") {
        auto controllerId2 = dsModel::ControllerId{"3504175fe0000000000000000000d4da00"};
        auto controller2 = apartment.allocateDSMeter(dss::Dsuid{controllerId2});
        controller2->setCapability_HasMetering(true);

        REQUIRE_CALL(metering, getEnergy(Dsuid{controllerId1})).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        REQUIRE_CALL(metering, getEnergy(Dsuid{controllerId2})).RETURN(ArchiveValue{20, 200, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest",
                           ds::str("from=.meters(", controllerId1, ",", controllerId2, ")", "&type=power")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":10.0,"date":"1970-01-01 00:05:01"},{"dSUID":"3504175fe0000000000000000000d4da00","dsid":"3504175fe00000000000d4da","value":20.0,"date":"1970-01-01 00:05:01"}]},"ok":true})");
      }
    }
    SECTION("type=energy") {
      REQUIRE_CALL(metering, getEnergy(Dsuid{controllerId1})).RETURN(ArchiveValue{10, 3600, ds::DateTime{301s}});
      ss << meteringRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energy", "&unit=Wh")}, nullptr, nullptr);
      CHECK(
          ss.str() ==
          R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":1.0,"date":"1970-01-01 00:05:01"}]},"ok":true})");
    }
    SECTION("check valid type") {
      SECTION("type=power") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=power")}, nullptr, nullptr);
        constexpr auto expect =
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":10.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv;
        CHECK(ss.str() == expect);
      }
      SECTION("type=consumption (legacy power)") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=consumption")}, nullptr, nullptr);
        constexpr auto expect =
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":10.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv;
        CHECK(ss.str() == expect);
      }
      SECTION("type=energy") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energy", "&unit=Ws")}, nullptr,
            nullptr);
        constexpr auto expect =
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":100.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv;
        CHECK(ss.str() == expect);
      }
      SECTION("type=energyDelta") {
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energyDelta")}, nullptr, nullptr);
        CHECK(ss.str() == R"({"ok":false,"message":"Invalid or missing type parameter"})");
      }
      SECTION("type=schmack") {
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=schmack", "&unit=W")}, nullptr,
            nullptr);
        constexpr auto expect =
            R"({"ok":false,"message":"Unknown ::dss::metering::SeriesTypes enum option. value:schmack)"sv;
        CHECK(ss.str().substr(0, expect.size()) == expect);
      }
    }
    SECTION("check valid unit for energy queries") {
      SECTION("unit=Ws") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energy", "&unit=Ws")}, nullptr,
            nullptr);
        constexpr auto expect =
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":100.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv;
        CHECK(ss.str() == expect);
      }
      SECTION("dsd-9878 / sven queries energy with type=consumption and unit=Ws") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=consumption", "&unit=Ws")}, nullptr,
            nullptr);
        constexpr auto expect =
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":10.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv;
        CHECK(ss.str() == expect);
      }
      SECTION("unit=Wh") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 3600, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energy", "&unit=Wh")}, nullptr,
            nullptr);
        constexpr auto expect =
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":1.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv;
        CHECK(ss.str() == expect);
      }
      SECTION("unit=W") {
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energy", "&unit=W")}, nullptr, nullptr);
        CHECK(ss.str() == R"({"ok":false,"message":"Invalid unit 'W'"})");
      }
      SECTION("unit=default, argument is optional, check default") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 3600, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energy")}, nullptr, nullptr);
        constexpr auto expect =
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","dsid":"3504175fe0000010000003dd","value":1.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv;
        CHECK(ss.str() == expect);
      }
      SECTION("unit=xx") {
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&type=energy", "&unit=xx")}, nullptr,
            nullptr);
        constexpr auto expect = R"({"ok":false,"message":"Unknown ::dss::metering::Unit enum option. value:xx)"sv;
        CHECK(ss.str().substr(0, expect.size()) == expect);
      }
    }
    SECTION("missing arguments") {
      SECTION("missing 'from' argument") {
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("type=energy", "&unit=W")}, nullptr, nullptr);
        constexpr auto expect = R"({"ok":false,"message":"Missing parameter name:from)"sv;
        CHECK(ss.str().substr(0, expect.size()) == expect);
      }
      SECTION("missing 'type' argument") {
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=.meters(all)", "&unit=W")}, nullptr, nullptr);
        constexpr auto expect = R"({"ok":false,"message":"Missing parameter name:type)"sv;
        CHECK(ss.str().substr(0, expect.size()) == expect);
      }
    }

    SECTION("with production") {
      SECTION("normal") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 3600, ds::DateTime{301s}});
        REQUIRE_CALL(metering, getProducedEnergy(_)).RETURN(ArchiveValue{20, 7200, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=", controllerId1, "&type=power", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","consumption":10.0,"production":20.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv);
      }
      SECTION("no production values") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 3600, ds::DateTime{301s}});
        REQUIRE_CALL(metering, getProducedEnergy(_)).RETURN(std::nullopt);

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getLatest", ds::str("from=", controllerId1, "&type=power", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":"3504175fe000000000000010000003dd00","consumption":10.0,"production":0.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv);
      }
    }
  }

  SECTION("getAggregatedLatest") {
    SECTION("type=power") {
      SECTION(".meters(all)") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getAggregatedLatest", ds::str("from=.meters(all)", "&type=power", "&unit=W")}, nullptr,
            nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":["3504175fe000000000000010000003dd00"],"dsid":["3504175fe0000010000003dd"],"value":10,"date":"1970-01-01 00:05:01"}]},"ok":true})");
      }

      SECTION(".meters(dsid1,dsid2)") {
        auto controllerId2 = dsModel::ControllerId{"3504175fe0000000000000000000d4da00"};
        auto controller2 = apartment.allocateDSMeter(dss::Dsuid{controllerId2});
        controller2->setCapability_HasMetering(true);

        REQUIRE_CALL(metering, getEnergy(Dsuid{controllerId1})).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        REQUIRE_CALL(metering, getEnergy(Dsuid{controllerId2})).RETURN(ArchiveValue{20, 200, ds::DateTime{302s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getAggregatedLatest",
                           ds::str("from=.meters(", controllerId1, ",", controllerId2, ")", "&type=power")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":["3504175fe000000000000010000003dd00","3504175fe0000000000000000000d4da00"],"dsid":["3504175fe0000010000003dd","3504175fe00000000000d4da"],"value":30,"date":"1970-01-01 00:05:02"}]},"ok":true})");
      }

      SECTION(".meters(dsid1,dsid2) but 2nd meter returns nullopt") {
        auto controllerId2 = dsModel::ControllerId{"3504175fe0000000000000000000d4da00"};
        auto controller2 = apartment.allocateDSMeter(dss::Dsuid{controllerId2});
        controller2->setCapability_HasMetering(true);

        REQUIRE_CALL(metering, getEnergy(Dsuid{controllerId1})).RETURN(ArchiveValue{10, 100, ds::DateTime{301s}});
        REQUIRE_CALL(metering, getEnergy(Dsuid{controllerId2})).RETURN(std::nullopt);

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getAggregatedLatest",
                           ds::str("from=.meters(", controllerId1, ",", controllerId2, ")", "&type=power")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":["3504175fe000000000000010000003dd00"],"dsid":["3504175fe0000010000003dd"],"value":10,"date":"1970-01-01 00:05:01"}]},"ok":true})");
      }
    }

    SECTION("with production") {
      SECTION("normal") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 3600, ds::DateTime{301s}});
        REQUIRE_CALL(metering, getProducedEnergy(_)).RETURN(ArchiveValue{20, 7200, ds::DateTime{301s}});
        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getAggregatedLatest",
                           ds::str("from=", controllerId1, "&type=power", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":["3504175fe000000000000010000003dd00"],"consumption":10.0,"production":20.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv);
      }
      SECTION("no production values") {
        REQUIRE_CALL(metering, getEnergy(_)).RETURN(ArchiveValue{10, 3600, ds::DateTime{301s}});
        REQUIRE_CALL(metering, getProducedEnergy(_)).RETURN(std::nullopt);

        ss << meteringRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getAggregatedLatest",
                           ds::str("from=", controllerId1, "&type=power", "&withProduction=true")},
            nullptr, nullptr);
        CHECK(
            ss.str() ==
            R"({"result":{"values":[{"dSUID":["3504175fe000000000000010000003dd00"],"consumption":10.0,"production":0.0,"date":"1970-01-01 00:05:01"}]},"ok":true})"sv);
      }
    }
  }*/
}

} // namespace
