#include "base.h"
#include "datetools.h"
#include "ds-backend/connector.h"
#include "ds/date-time.h"
#include "ds/str.h"
#include "http_client.h"
#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <uuid.h>

#include <json.h>
#include <stdio.h>
#include <iostream>
#include <cstring>
#include <ds/log.h>
#include <chrono>

#include "dss.h"
#include "event.h"
#include "exception.h"
#include "event/event_fields.h"
#include "propertysystem.h"
#include "model/devicereference.h"
#include "model/group.h"
#include "model/state.h"
#include "webservice_api.h"
#include "web/webrequests.h"
#include "propertysystem_common_paths.h"
#include "model/apartment.h"
#include "model/set.h"
#include "model/zone.h"

using namespace std::literals;

static ds::LogChannel dsLogChannel("dssWebserviceMsHub");

namespace dss {

typedef std::vector<std::shared_ptr<Event> >::iterator ItEvent;

ParseError::ParseError(const std::string& _message) : runtime_error( _message )
{
}

/****************************************************************************/
/* Helpers of Ms Hub                                                        */
/****************************************************************************/
namespace MsHub {

const static std::string evtGroup_Metering = "Metering";
const static std::string evtGroup_Activity = "Activity";
const static std::string evtGroup_ApartmentAndDevice = "ApartmentAndDevice";

/*
 * #22209 Remove device sensor value events from MSHub uploads
const static std::string evtCategory_DeviceSensorValue = "DeviceSensorValue";
const static std::string evtCategory_DeviceStatusReport = "DeviceStatusReport";
const static std::string evtCategory_DeviceSensorError = "DeviceSensorError";
 */
const static std::string evtCategory_ZoneSensorValue = "ZoneSensorValue";
const static std::string evtCategory_ZoneSensorError = "ZoneSensorError";
const static std::string evtCategory_ZoneGroupCallScene = "ZoneSceneCall";
const static std::string evtCategory_ZoneGroupUndoScene = "ZoneUndoScene";
const static std::string evtCategory_ApartmentState = "ApartmentStateChange";
const static std::string evtCategory_ZoneGroupState = "ZoneStateChange";
const static std::string evtCategory_DeviceInputState = "DeviceInputState";
const static std::string evtCategory_HeatingControllerSetup = "HeatingControllerSetup";
const static std::string evtCategory_HeatingControllerValue = "HeatingControllerValue";
const static std::string evtCategory_HeatingControllerState = "HeatingControllerState";
const static std::string evtCategory_HeatingEnabled = "HeatingEnabled";
const static std::string evtCategory_HeatingSystemCapability = "HeatingSystemCapability";
const static std::string evtCategory_HeatingZoneCapability = "HeatingZoneCapability";
const static std::string evtCategory_AddonToCloud = "AddOnToCloud";

//const static int dsEnum_SensorError_invalidValue = 1;
const static int dsEnum_SensorError_noValue = 2;

void appendCommon(JSONWriter& json, const std::string& group,
                  const std::string& category, const Event *event)
{
  json.add("EventGroup", group);
  json.add("EventCategory", category);
  json.add("Timestamp", event->getTimestamp().toISO8601_ms());
}

/* mshub */
std::vector<std::string> uploadEvents()
{
  std::vector<std::string> events;
  events.push_back(EventName::ZoneSensorValue);
  events.push_back(EventName::ZoneSensorError);
  events.push_back(EventName::CallScene);
  events.push_back(EventName::UndoScene);
  events.push_back(EventName::StateChange);
  events.push_back(EventName::AddonStateChange);
  events.push_back(EventName::HeatingEnabled);
  events.push_back(EventName::HeatingSystemCapability);
  events.push_back(EventName::HeatingZoneCapability);
  events.push_back(EventName::HeatingControllerSetup);
  events.push_back(EventName::HeatingControllerValue);
  events.push_back(EventName::HeatingControllerState);
  events.push_back(EventName::OldStateChange);
  events.push_back(EventName::AddonToCloud);
  return events;
}

void toJson(const std::shared_ptr<Event> &event, JSONWriter& json) {
  std::shared_ptr<const DeviceReference> pDeviceRef;
  std::string propValue;

  try {
    if ((event->getName() == EventName::ZoneSensorValue) && (event->getRaiseLocation() == erlGroup)) {
      std::shared_ptr<const Group> pGroup = event->getRaisedAtGroup();
      appendCommon(json, evtGroup_Metering, evtCategory_ZoneSensorValue,
                   event.get());
      json.add("SensorType", event->getPropertyByName<SensorType>("sensorType"));
      json.add("SensorValue", event->getPropertyByName("sensorValueFloat"));
      json.add("ZoneID",  pGroup->getZoneID());
      json.add("GroupID", pGroup->getID());
    } else if ((event->getName() == EventName::ZoneSensorError) && (event->getRaiseLocation() == erlGroup)) {
      std::shared_ptr<const Group> pGroup = event->getRaisedAtGroup();
      appendCommon(json, evtGroup_ApartmentAndDevice,
                   evtCategory_ZoneSensorError, event.get());
      json.add("SensorType", event->getPropertyByName<SensorType>("sensorType"));
      json.add("ZoneID",  pGroup->getZoneID());
      json.add("GroupID", pGroup->getID());
      json.add("StatusCode", dsEnum_SensorError_noValue);
    } else if ((event->getName() == EventName::CallScene) && (event->getRaiseLocation() == erlGroup)) {
      std::shared_ptr<const Group> pGroup = event->getRaisedAtGroup();
      appendCommon(json, evtGroup_Activity, evtCategory_ZoneGroupCallScene,
                   event.get());
      json.add("ZoneID",  pGroup->getZoneID());
      json.add("GroupID", pGroup->getID());
      json.add("SceneID", strToInt(event->getPropertyByName(ef_sceneID)));
      json.add("Force", event->hasPropertySet(ef_forced));
      json.add("Origin", event->getPropertyByName(ef_callOrigin));
    } else if ((event->getName() == EventName::UndoScene) && (event->getRaiseLocation() == erlGroup)) {
      std::shared_ptr<const Group> pGroup = event->getRaisedAtGroup();
      appendCommon(json, evtGroup_Activity, evtCategory_ZoneGroupUndoScene,
                   event.get());
      json.add("ZoneID",  pGroup->getZoneID());
      json.add("GroupID", pGroup->getID());
      json.add("SceneID", strToInt(event->getPropertyByName(ef_sceneID)));
      json.add("Origin", event->getPropertyByName(ef_callOrigin));
    } else if (event->getName() == EventName::StateChange) {
      std::shared_ptr<const State> pState = event->getRaisedAtState();
      switch (pState->getType()) {
      case StateType_Apartment:
      case StateType_Service:
      case StateType_Script:
        appendCommon(json, evtGroup_Activity, evtCategory_ApartmentState,
                     event.get());
        break;
      case StateType_Group:
        appendCommon(json, evtGroup_Activity, evtCategory_ZoneGroupState,
                     event.get());
        break;
      case StateType_Device:
        appendCommon(json, evtGroup_Activity, evtCategory_DeviceInputState,
                     event.get());
        break;
      case StateType_Circuit:
        appendCommon(json, evtGroup_Activity, evtCategory_DeviceInputState,
                     event.get());
        break;
      case StateType_SensorZone:
      case StateType_SensorDevice:
      default:
        break;
      }
      json.add("StateName", pState->getName());
      json.add("StateValue", pState->toString());
      json.add("Origin", event->getPropertyByName(ef_callOrigin));
    } else if (event->getName() == EventName::OldStateChange) {
      appendCommon(json, evtGroup_Activity, evtCategory_ApartmentState, event.get());
      json.add("StateName", event->getPropertyByName("statename"));
      json.add("StateValue", event->getPropertyByName("state"));
      json.add("Origin", event->getPropertyByName(ef_callOrigin));
    } else if (event->getName() == EventName::AddonStateChange) {
      std::shared_ptr<const State> pState = event->getRaisedAtState();
      std::string addonName = event->getPropertyByName("scriptID");
      appendCommon(json, evtGroup_Activity, evtCategory_ApartmentState, event.get());
      json.add("StateName", addonName + "." + pState->getName());
      json.add("StateValue", pState->toString());
      json.add("Origin", event->getPropertyByName("scriptID"));
    } else if (event->getName() == EventName::HeatingControllerSetup) {
      appendCommon(json, evtGroup_ApartmentAndDevice,
                   evtCategory_HeatingControllerSetup, event.get());
      for (auto&& param : event->getProperties().getContainer()) {
        json.add(param.first, param.second);
      }
    } else if (event->getName() == EventName::HeatingControllerValue) {
      appendCommon(json, evtGroup_ApartmentAndDevice,
                   evtCategory_HeatingControllerValue, event.get());
      for (auto&& param : event->getProperties().getContainer()) {
        json.add(param.first, param.second);
      }
    } else if (event->getName() == EventName::HeatingControllerState) {
      appendCommon(json, evtGroup_ApartmentAndDevice,
                   evtCategory_HeatingControllerState, event.get());
      for (auto&& param : event->getProperties().getContainer()) {
        json.add(param.first, param.second);
      }
    } else if (event->getName() == EventName::HeatingEnabled) {
      appendCommon(json, evtGroup_ApartmentAndDevice,
                   evtCategory_HeatingEnabled, event.get());
      for (auto&& param : event->getProperties().getContainer()) {
        json.add(param.first, param.second);
      }
    } else if (event->getName() == EventName::HeatingSystemCapability) {
      appendCommon(json, evtGroup_ApartmentAndDevice,
                   evtCategory_HeatingSystemCapability, event.get());
      for (auto&& param : event->getProperties().getContainer()) {
        json.add(param.first, param.second);
      }
    } else if (event->getName() == EventName::HeatingZoneCapability) {
      appendCommon(json, evtGroup_ApartmentAndDevice,
                   evtCategory_HeatingZoneCapability, event.get());
      for (auto&& param : event->getProperties().getContainer()) {
        json.add(param.first, param.second);
      }
    } else if (event->getName() == EventName::AddonToCloud) {
      appendCommon(json, evtGroup_Activity, evtCategory_AddonToCloud, event.get());
      json.add("EventName", event->getPropertyByName("EventName"));
      json.startObject("parameter");
      for (auto&& param : event->getProperties().getContainer()) {
        json.add(param.first, param.second);
      }
      json.endObject();

    } else {
      DS_FAIL_REQUIRE("Unhandled event.", event);
    }
  } catch (std::invalid_argument& e) {
    DS_FAIL_REQUIRE("Error converting event.", event);
  }
  return;
}

} /* namespace MsHub */


/****************************************************************************/
/* Ms Hub Reply Checker                                                     */
/****************************************************************************/

class MsHubReplyChecker : public URLRequestCallback {
  __DECL_LOG_CHANNEL__
public:
  /**
   * MsHubReplyChecker - triggered by http client, will parse the returned
   * JSON as WebserviceReply structure, and trigger provided callback
   * @callback: trigger
   */
  MsHubReplyChecker(WebserviceCallDone_t callback) : m_callback(callback) {};
  virtual ~MsHubReplyChecker() {};
  virtual void result(long code, const std::string &result);
private:
  WebserviceCallDone_t m_callback;
};

__DEFINE_LOG_CHANNEL__(MsHubReplyChecker, lsInfo);

void MsHubReplyChecker::result(long code, const std::string &result) {

  if (code != 200) {
    log("HTTP POST failed " + intToString(code), lsError);
    if (m_callback) {
      m_callback->done(NETWORK_ERROR, WebserviceReply());
    }
    return;
  }

  try {
    WebserviceReply resp = WebserviceMsHub::parseReply(result.c_str());
    if (resp.code != 0) {
      log("Webservice complained: <" + intToString(resp.code) + "> " + resp.desc,
          lsWarning);
    }

    if (m_callback) {
      m_callback->done(REST_OK, resp);
    }
  } catch (ParseError &ex) {
    log(std::string("ParseError: <") + ex.what() + "> " + result, lsError);
    if (m_callback) {
      m_callback->done(JSON_ERROR, WebserviceReply());
    }
    return;
  }
}


/****************************************************************************/
/* Webservice MS Hub                                                        */
/****************************************************************************/

__DEFINE_LOG_CHANNEL__(WebserviceMsHub, lsInfo)

bool WebserviceMsHub::isAuthorized() {
  PropertySystem &propSystem = DSS::getInstance()->getPropertySystem();
  return propSystem.getBoolValue(pp_websvc_mshub_active);
}

template <class Iterator>
bool WebserviceMsHub::doUploadSensorData(Iterator begin, Iterator end,
                                         WebserviceCallDone_t callback) {

  JSONWriter json(JSONWriter::jsonNoneResult);
  int ct = 0;

  std::string parameters;
  // AppToken is piggy backed with websvc_connection::request(.., authenticated=true)
  parameters += "dssid=" + DSS::getInstance()->getPropertySystem().getProperty(pp_sysinfo_dsid)->getStringValue();
  parameters += "&source=dss"; /* EventSource_dSS */

  json.startArray("eventsList");
  for (; begin != end; begin++) {
    try {
      json.startObject();
      MsHub::toJson(*begin, json);
      json.endObject();
      ct++;
    } catch (DSSException& e) {
      json.endObject();
      log(e.what(), lsWarning);
    }
  }
  json.endArray();

  if (ct == 0) {
    return false;
  }

  std::string postdata = json.successJSON();
  // TODO eventually emit summary: 10 callsceene, 27 metering, ...
  // dumping whole postdata leads to log rotation and is seldomly needed
  // unless for solving a blame war with cloud team
  DS_INFO("upload events:", ct, "bytes:", postdata.length());
  DS_DEBUG("event data:", postdata);

  // https://devdsservices.aizo.com/Help/Api/POST-public-dss-v1_0-DSSEventData-SaveEvent_token_apartmentId_dssid_source
  std::shared_ptr<MsHubReplyChecker> mcb = std::make_shared<MsHubReplyChecker>(callback);
  std::unordered_map<std::string, std::string> sensorUploadHeaders;
  sensorUploadHeaders["Content-Type"] = "application/json;charset=UTF-8";

  /*WebserviceConnection::getInstanceMsHub()->request("public/dss/v1_0/DSSEventData/SaveEvent",
                                                    parameters,
                                                    sensorUploadHeaders,
                                                    postdata,
                                                    mcb,
                                                    true);*/
  return true;
}

template bool WebserviceMsHub::doUploadSensorData<ItEvent>(
      ItEvent begin, ItEvent end, WebserviceCallDone_t callback);

void WebserviceMsHub::doDssBackAgain(WebserviceCallDone_t callback)
{
  std::string parameters;

  if (!WebserviceMsHub::isAuthorized()) {
    return;
  }

  // AppToken is piggy backed with websvc_connection::request(.., authenticated=true)
  parameters += "&dssid=" + DSS::getInstance()->getPropertySystem().getProperty(pp_sysinfo_dsid)->getStringValue();
  std::shared_ptr<MsHubReplyChecker> mcb = std::make_shared<MsHubReplyChecker>(callback);
  log("sending DSSBackAgain", lsInfo);
  //WebserviceConnection::getInstanceMsHub()->request(
  //    "internal/dss/v1_0/DSSApartment/DSSBackAgain", parameters, HttpMethod::POST, mcb, true);
}

void WebserviceMsHub::doModelChanged(ChangeType type,
                                     WebserviceCallDone_t callback) {
  PropertySystem &propSystem = DSS::getInstance()->getPropertySystem();
  std::string url;
  std::string params;

  if (!WebserviceMsHub::isAuthorized()) {
    return;
  }

  url = propSystem.getStringValue(pp_websvc_apartment_changed_url_path);
  params = "dssid=" + propSystem.getStringValue(pp_sysinfo_dsid);
  params += "&apartmentChangeType=";
  switch (type) {
  case ApartmentChange:
      params += "Apartment";
      break;
  case TimedEventChange:
      params += "TimedEvent";
      break;
  case UDAChange:
      params += "UserDefinedAction";
      break;
  }

  log("execute: " + url + "?" + params, lsDebug);
  std::shared_ptr<MsHubReplyChecker> mcb = std::make_shared<MsHubReplyChecker>(callback);
  //WebserviceConnection::getInstanceMsHub()->request(url, params, HttpMethod::POST, mcb, true);
}

void WebserviceMsHub::doNotifyTokenDeleted(const std::string &token,
                                           WebserviceCallDone_t callback) {
  PropertySystem &propSystem = DSS::getInstance()->getPropertySystem();
  std::string url;
  std::string params;

  if (!WebserviceMsHub::isAuthorized()) {
    return;
  }

  url += propSystem.getStringValue(pp_websvc_access_mgmt_delete_token_url_path);
  params = "dsid=" + propSystem.getStringValue(pp_sysinfo_dsid);
  params += "&token=" + token;

  // webservice is fire and forget, so use shared ptr for life cycle mgmt
  std::shared_ptr<MsHubReplyChecker> cont = std::make_shared<MsHubReplyChecker>(callback);
  //WebserviceConnection::getInstanceMsHub()->request(url, params, HttpMethod::POST, cont, false);
}

//TODO - Refactor: Remove from MsHub
void WebserviceMsHub::doGetWeatherInformation() {
  PropertySystem &propSystem = DSS::getInstance()->getPropertySystem();
  
  HttpClient client;
  
  std::string params;
  std::string cityName = propSystem.getStringValue(pp_geolocation_city);
  int posParenthesis = cityName.find ("(");
  int posSlash = cityName.find ("/");
  int posEnd = cityName.size();
  if (posSlash != std::string::npos) {
      posEnd = posSlash;
  }
  if (posParenthesis != std::string::npos) {
      if (posParenthesis < posEnd) {
        posEnd = posParenthesis;
      }
  }
  if (cityName.substr(0,posEnd).back()  == ' ') {
      posEnd--;
  }

  params = "city=" + urlEncode(cityName.substr(0,posEnd));
  params += "&country=" + propSystem.getStringValue(pp_geolocation_country);

  dss::HttpRequest request{.url = "http://localhost:8089/backend/v1/apartment/weather?" + params,
                            .type = dss::HttpMethod::GET};

  std::string result;
  long resultCode = client.get(request.url, &result);


  if (resultCode == 200) {
    auto weatherInfo = ds::fromJson<WeatherInfo>(ds::parseJsonFragment(result.c_str()));

    ds::DateTime now;
    if (now - weatherInfo.data.lastChanged > 24h) {
      return;
    }

    std::string weatherIconID = ds::str(weatherInfo.data.includes.weather.icon);
    std::string weatherConditionID = ds::str(weatherInfo.data.includes.weather.id);
    std::string weatherService = "7"; // Is this really needed??

    Apartment& apt = DSS::getInstance()->getApartment();
    Set outdoorTemperatureSensors = apt.getDevices().getBySensorType(SensorType::TemperatureOutdoors);
    apt.setWeatherInformation(std::move(weatherIconID), std::move(weatherConditionID),
                    std::move(weatherService),
                    dss::DateTime());

    if (outdoorTemperatureSensors.empty() && weatherInfo.data.includes.main.temperature != 0.0) {
      float tempInDegreesCelsius = weatherInfo.data.includes.main.temperature - 273.15f;
      DS_INFO("GetWeatherInformation: Outdoor Temperature: " + doubleToString(tempInDegreesCelsius));
      apt.getZone(0)->pushSensor(CallOrigin::system, SceneAccessCategory::manual, DSUID_NULL,
          SensorType::TemperatureOutdoors, tempInDegreesCelsius, "");
    }
  } else {
    DS_WARNING("Falied to get weather data: Code " + ds::str(resultCode) + " : " + result);
  }
}

void WebserviceMsHub::doVdcStoreVdcToken(const std::string &vdcToken, WebserviceCallDone_t callback) {
  PropertySystem &propSystem = DSS::getInstance()->getPropertySystem();
  std::string url;
  std::string params;

  if (!WebserviceMsHub::isAuthorized()) {
    callback->done(NETWORK_ERROR, WebserviceReply({0, ""}));
    return;
  }

  url += "public/vDC/v1_0/Device/StoreVDCToken";
  params += "dssId=" + propSystem.getStringValue(pp_sysinfo_dsid);
  params += "&ospToken=" + propSystem.getStringValue(pp_websvc_rc_osptoken);
  params += "&vdcToken=" + vdcToken;

  // webservice is fire and forget, so use shared ptr for life cycle mgmt
  std::shared_ptr<MsHubReplyChecker> cont = std::make_shared<MsHubReplyChecker>(callback);
  //WebserviceConnection::getInstanceMsHub()->request(url, params, HttpMethod::POST, cont, false);
}

WebserviceReply WebserviceMsHub::parseReply(const char* buf) {
  WebserviceReply resp;
  bool return_code_seen = false;

  if (!buf) {
    throw ParseError("buffer is NULL");
  }

  json_object * jobj = json_tokener_parse(buf);
  if (!jobj) {
    throw ParseError("invalid JSON");
  }
  json_object_object_foreach(jobj, key, val) { /*Passing through every array element*/
    enum json_type type = json_object_get_type(val);
    if (!strcmp(key, "ReturnCode")) {
      if (type != json_type_int) {
        throw ParseError("invalid type for ReturnCode");
      }
      resp.code = json_object_get_int(val);
      return_code_seen = true;
    } else if (!strcmp(key, "ReturnMessage")) {
      if (type == json_type_null) {
        // empty string
        continue;
      }
      if (type != json_type_string) {
        throw ParseError("invalid type for ReturnMessage");
      }
      resp.desc = json_object_get_string(val);
    } else {
      // ignore, unkown keys
    }
  }
  json_object_put(jobj);

  if (!return_code_seen) {
    throw ParseError("missing mandatory element: ReturnCode");
  }

  return resp;
}
} /* namespace dss */
