/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN

#include <digitalSTROM/dsuid.h>
#include <ds-model/model.h>

#include <boost/test/unit_test.hpp>

#include "businterface-mock.h"
#include "cache/apartment-proxy-mock.h"
#include "cache/apartment-proxy.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/zone.h"
#include "propertysystem.h"
#include "vdc-messages.pb.h"
#include "web/handler/devicerequesthandler.h"
#include "web/restful.h"
#include "webfixture.h"

using namespace dss;
using namespace trompeloeil;

static ds::LogChannel dsLogChannel("testsDeviceRequestHandlerTest");

BOOST_AUTO_TEST_SUITE(WebDevice)

class Fixture : public WebFixture {
public:
  Fixture() {
    m_pPropertySystem.reset(new PropertySystem());
    m_pApartment = makeApartment(nullptr, m_pPropertySystem.get(), &m_busInterface);
    m_pHandler.reset(new DeviceRequestHandler(*m_pApartment));

    memset(&m_ValidDSID, 0, 12);
    m_ValidDSID.id[12 - 1] =  1;
    m_ValidDSUID = DSUID_NULL;
    m_ValidDSUID.id[DSUID_SIZE - 1] =  1;
    m_ValidName = "device1";
    m_InvalidName = "nonexisting";
    m_ValidZoneID = 2;
    m_ValidGroupID = 1;
    m_FunctionID = 6;
    m_ProductID = 9;
    m_RevisionID = 3;
    DeviceSpec spec = {};
    spec.FunctionID = m_FunctionID;
    spec.ProductID = m_ProductID;
    spec.revisionId = m_RevisionID;

    std::shared_ptr<Zone> firstZone = m_pApartment->allocateZone(m_ValidZoneID);
    std::shared_ptr<Device> dev = m_pApartment->allocateDevice(m_ValidDSUID);
    DeviceReference devRef(dev, m_pApartment.get());
    firstZone->addDevice(devRef);
    dev->setName(m_ValidName);
    dev->addToGroup(m_ValidGroupID);
    dev->setPartiallyFromSpec(spec);
    std::shared_ptr<Device> dev2 = m_pApartment->allocateDevice(dsuid_from_dsid(m_ValidDSID));
    DeviceReference devRef2(dev2, m_pApartment.get());
    firstZone->addDevice(devRef2);
    dev2->setName(m_ValidName);
    dev2->addToGroup(m_ValidGroupID);
    dev2->setPartiallyFromSpec(spec);
  }
protected:
  BusInterfaceMock m_busInterface{};
  std::shared_ptr<Apartment> m_pApartment;
  std::shared_ptr<DeviceRequestHandler> m_pHandler;
  std::shared_ptr<PropertySystem> m_pPropertySystem;
  dsid_t m_ValidDSID;
  dsuid_t m_ValidDSUID;
  std::string m_ValidName;
  std::string m_InvalidName;
  int m_ValidZoneID;
  int m_ValidGroupID;
  int m_FunctionID;
  int m_ProductID;
  int m_RevisionID;
};

BOOST_FIXTURE_TEST_CASE(testNoParam, Fixture) {
  RestfulRequest req("device/bla", "");
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testEmptyNameParam, Fixture) {
  RestfulRequest req("device/bla", "name=");
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testInvalidNameParam, Fixture) {
  RestfulRequest req("device/bla", "name=" + urlEncode(m_InvalidName));
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testEmptyDSIDParam, Fixture) {
  RestfulRequest req("device/bla", "dsid=");
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testInvalidDSIDParam, Fixture) {
  RestfulRequest req("device/bla", "dsid=asdfasdf");
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testInvalidDSUIDParam, Fixture) {
  RestfulRequest req("device/bla", "dsuid=asdfasdf");
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testTooLongDSIDParam, Fixture) {
  RestfulRequest req("device/bla",
                     "dsid=3504175fe0000000ffc000113504175fe0000000ffc00011");
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testTooLongDSUIDParam, Fixture) {
  RestfulRequest req("device/bla",
                     "dsuid=3504175fe0000000ffc000113504175fe0000000ffc00011");
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testRightDSID, Fixture) {
  std::string params = "dsid=" + urlEncode(dsid2str(m_ValidDSID));
  RestfulRequest req("device/bla", params);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), std::runtime_error);
}

BOOST_FIXTURE_TEST_CASE(testRightDSUID, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/bla", params);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), std::runtime_error);
}

BOOST_FIXTURE_TEST_CASE(testDeviceGetSpec, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getSpec", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  checkPropertyEquals("functionID", m_FunctionID, response);
  checkPropertyEquals("revisionID", m_RevisionID, response);
  checkPropertyEquals("productID", m_ProductID, response);
}

BOOST_FIXTURE_TEST_CASE(testDeviceGetGroups, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getGroups", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  BOOST_CHECK_EQUAL(response.getResponse(), "{\"result\":{\"groups\":[{\"id\":"+intToString(m_ValidGroupID)+",\"name\":\"yellow\"}]},\"ok\":true}");
}

BOOST_FIXTURE_TEST_CASE(testDeviceGetState, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getState", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  checkPropertyEquals("isOn", false, response);
}

BOOST_FIXTURE_TEST_CASE(testDeviceGetName, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getName", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  checkPropertyEquals("name", m_ValidName, response);
}

BOOST_FIXTURE_TEST_CASE(testDeviceSetNameMissingNewName, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/setName", params);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), std::exception);
}

BOOST_FIXTURE_TEST_CASE(testDeviceSetName, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  std::string kNewName = "renamed";
  params += "&newName=" + urlEncode(kNewName);

  REQUIRE_CALL(m_busInterface.structureModify, deviceSetName(_, _, _));
  RestfulRequest req("device/setName", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, true);

  RestfulRequest req2("device/getName", params);
  WebServerResponse response2 = m_pHandler->jsonHandleRequest(req2, std::shared_ptr<Session>(), NULL);

  checkPropertyEquals("name", kNewName, response2);
}

BOOST_FIXTURE_TEST_CASE(testDeviceAddTagMissingTag, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/addTag", params);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), std::exception);
}

BOOST_FIXTURE_TEST_CASE(testDeviceHasTagMissingTag, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/hasTag", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testDeviceRemoveTagMissingTag, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/removeTag", params);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), std::exception);
}

BOOST_FIXTURE_TEST_CASE(testSetGetFirstSeenRightDate, Fixture) {
  std::shared_ptr<Device> device = m_pApartment->getDeviceByDSID(m_ValidDSUID);
  static const DateTime date = DateTime::parseISO8601("2000-01-01T01:00:01Z");
  device->setFirstSeen(date);
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  {
    RestfulRequest req("device/getFirstSeen", params);
    WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
    checkPropertyEquals("time", date.toISO8601(), response);
  }
}

BOOST_FIXTURE_TEST_CASE(testDeviceTestTags, Fixture) {
  const std::string& kTagName = "justATag";
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += "&tag=" + urlEncode(kTagName);
  RestfulRequest reqHasTag("device/hasTag", params);
  RestfulRequest reqAddTag("device/addTag", params);
  RestfulRequest reqRemoveTag("device/removeTag", params);

  // check that tag doesn't exist
  WebServerResponse response = m_pHandler->jsonHandleRequest(reqHasTag, std::shared_ptr<Session>(), NULL);
  checkPropertyEquals("hasTag", false, response);

  // add tag
  WebServerResponse response2 = m_pHandler->jsonHandleRequest(reqAddTag, std::shared_ptr<Session>(), NULL);
  testOkIs(response2, true);

  // check that tag does exist
  WebServerResponse response3 = m_pHandler->jsonHandleRequest(reqHasTag, std::shared_ptr<Session>(), NULL);
  checkPropertyEquals("hasTag", true, response3);

  // remove tag
  WebServerResponse response4 = m_pHandler->jsonHandleRequest(reqRemoveTag, std::shared_ptr<Session>(), NULL);
  testOkIs(response4, true);

  // check that tag doesn't exist
  WebServerResponse response5 = m_pHandler->jsonHandleRequest(reqHasTag, std::shared_ptr<Session>(), NULL);
  checkPropertyEquals("hasTag", false, response5);
}

BOOST_FIXTURE_TEST_CASE(testDeviceGetTagsNoTags, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getTags", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);

  checkPropertyEquals("tags", std::string("[ ]"), response);
}

BOOST_FIXTURE_TEST_CASE(testDeviceGetTagsOneTag, Fixture) {
  const std::string& kTagName = "testing";
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getTags", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->addTag(kTagName);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);

  checkPropertyEquals("tags", "[ \""+kTagName+"\" ]", response);
}

BOOST_FIXTURE_TEST_CASE(testEnable, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/enable", params);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), std::runtime_error);
}

BOOST_FIXTURE_TEST_CASE(testDisable, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/disable", params);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), std::runtime_error);
}

BOOST_FIXTURE_TEST_CASE(testLockNotPresent, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/lock", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testLockPresent, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/lock", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsConnected(true);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);

  REQUIRE_CALL(m_busInterface.device, lockOrUnlockDevice(_, true));
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, true);
}

BOOST_FIXTURE_TEST_CASE(testUnlockNotPresent, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/unlock", params);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testUnlockPresent, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/unlock", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsConnected(true);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  REQUIRE_CALL(m_busInterface.device, lockOrUnlockDevice(_, false));
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, true);
}

BOOST_FIXTURE_TEST_CASE(testSetConfigNoParameters, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/setConfig", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testSetConfigInvalidParameters, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += "&class=469";
  params += "&index=452";
  RestfulRequest req("device/setConfig", params);

  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testSetConfig, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += "&class=3";
  params += "&index=0";
  params += "&value=21";
  RestfulRequest req("device/setConfig", params);

  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  REQUIRE_CALL(m_busInterface.device, setDeviceConfig(_, 3, 0, 21));
  auto result = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  BOOST_CHECK_EQUAL(ds::str(result), R"({"ok":true})");
}

BOOST_FIXTURE_TEST_CASE(testGetConfigInvalidParameters, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += "&class=469";
  params += "&index=452";
  RestfulRequest req("device/getConfig", params);

  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testGetConfigNoParameters, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getConfig", params);

  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testGetConfig, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += "&class=3";
  params += "&index=0";
  RestfulRequest req("device/getConfig", params);

  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  REQUIRE_CALL(m_busInterface.device, getDeviceConfig(_, 3, 0)).RETURN(0x7);
  auto result = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  BOOST_CHECK_EQUAL(ds::str(result), R"({"result":{"class":3,"index":0,"value":7},"ok":true})");
}

BOOST_FIXTURE_TEST_CASE(testGetConfigWordInvalidParameters, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += "&class=469";
  params += "&index=452";

  RestfulRequest req("device/getConfigWord", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testGetConfigWordNoParameters, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getConfigWord", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testGetConfigWord, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += "&class=3";
  params += "&index=0";
  RestfulRequest req("device/getConfigWord", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  REQUIRE_CALL(m_busInterface.device, getDeviceConfigWord(_, 3, 0)).RETURN(0x7f7f);
  auto result = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  BOOST_CHECK_EQUAL(ds::str(result), R"({"result":{"class":3,"index":0,"value":32639},"ok":true})");
}

BOOST_FIXTURE_TEST_CASE(testGetTransmissionQuality, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  RestfulRequest req("device/getTransmissionQuality", params);
  m_pApartment->getDeviceByDSID(m_ValidDSUID)->setIsPresent(true);
  REQUIRE_CALL(m_busInterface.device, getTransmissionQuality(_)).RETURN(std::make_pair(0xf7, 0x7f));
  auto result = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  BOOST_CHECK_EQUAL(ds::str(result), R"({"result":{"upstream":127,"downstream":247},"ok":true})");
}

BOOST_FIXTURE_TEST_CASE(testSetOutputChannel2, Fixture) {
  BusInterfaceMock busInterface;
  m_pApartment->setBusInterface(&busInterface);
  auto proxyPtr = std::make_unique<ApartmentProxyMock>();
  auto& proxy = *proxyPtr;
  m_pApartment->setProxy(std::move(proxyPtr));

  sequence seq;
  REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_,0,_,10, false))
  .IN_SEQUENCE(seq);
  REQUIRE_CALL(proxy, setOutputValue(_, ChannelType::brightness, _, _))
  .IN_SEQUENCE(seq);
  REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_,1,_,15, false))
  .IN_SEQUENCE(seq);
  REQUIRE_CALL(proxy, setOutputValue(_, ChannelType::hue, _, _))
  .IN_SEQUENCE(seq);
  REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_,2,_,5,true))
  .IN_SEQUENCE(seq);
  REQUIRE_CALL(proxy, setOutputValue(_, ChannelType::saturation, _, _))
  .IN_SEQUENCE(seq);

  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += R"(&channels={ "brightness": { "value": 10 }, "hue": { "value": 15 }, "saturation": { "value": 5 }})";
  RestfulRequest req("device/setOutputChannelValue2", params);
  auto device = m_pApartment->getDeviceByDSID(m_ValidDSUID);
  device->setOutputChannels({ChannelType::brightness, ChannelType::hue, ChannelType::saturation});
  device->setIsPresent(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, true);
}

BOOST_FIXTURE_TEST_CASE(testNonVdcSetOutputChannelStringValue, Fixture) {
  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += R"(&channels={ "testId": { "value": "testValue" }})";
  RestfulRequest req("device/setOutputChannelValue2", params);
  auto device = m_pApartment->getDeviceByDSID(m_ValidDSUID);
  device->setOutputChannels({ChannelType::none});
  device->setOutputChannelInfo(0, Channel{0, "testId", "testTechnicalName", static_cast<ChannelType>(200)});
  device->setIsPresent(true);
  BOOST_CHECK_THROW(m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL), ds::Exception);
}

BOOST_FIXTURE_TEST_CASE(testVdcSetOutputChannel, Fixture) {
  BusInterfaceMock busInterface;
  m_pApartment->setBusInterface(&busInterface);
  auto proxyPtr = std::make_unique<ApartmentProxyMock>();
  auto& proxy = *proxyPtr;
  m_pApartment->setProxy(std::move(proxyPtr));

  ::google::protobuf::RepeatedPtrField<::vdcapi::PropertyElement> protoPayload;
  sequence seq;
  REQUIRE_CALL(proxy, setOutputValue(_, ChannelType::brightness, _, _))
  .IN_SEQUENCE(seq);
  REQUIRE_CALL(busInterface.device, setProperty(_,_))
    .LR_SIDE_EFFECT(protoPayload = _2)
    .IN_SEQUENCE(seq);

  std::string params = "dsuid=" + urlEncode(dsuid2str(m_ValidDSUID));
  params += R"(&channels={)"
            R"("testId0": {"value": "testValue"}, )"
            R"("testId1": {"value": 1}, )"
            R"("testId2": {"value": true}, )"
            R"("brightness": {"value": 12.5}})";
  RestfulRequest req("device/setOutputChannelValue2", params);
  auto device = m_pApartment->getDeviceByDSID(m_ValidDSUID);
  device->setOutputChannels({ChannelType::none, ChannelType::none, ChannelType::none, ChannelType::brightness});
  device->setOutputChannelInfo(0, Channel{0, "testId0", "testTechnicalName0", static_cast<ChannelType>(200)});
  device->setOutputChannelInfo(1, Channel{1, "testId1", "testTechnicalName1", static_cast<ChannelType>(201)});
  device->setOutputChannelInfo(2, Channel{2, "testId2", "testTechnicalName2", static_cast<ChannelType>(202)});
  device->setIsPresent(true);
  device->setVdcDevice(true);
  WebServerResponse response = m_pHandler->jsonHandleRequest(req, std::shared_ptr<Session>(), NULL);
  testOkIs(response, true);
  BOOST_CHECK_EQUAL(
    protoPayload[0].ShortDebugString(),
    R"(name: "channelStates" )"
    R"(elements { name: "brightness" elements { name: "value" value { v_double: 12.5 } } } )"
    R"(elements { name: "testId0" elements { name: "value" value { v_string: "testValue" } } } )"
    R"(elements { name: "testId1" elements { name: "value" value { v_double: 1 } } } )"
    R"(elements { name: "testId2" elements { name: "value" value { v_bool: true } } })"
  );
}

BOOST_AUTO_TEST_SUITE_END()
