/*
    Copyright (c) 2009,2011 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>


#include "model/apartment.h"
#include "model/zone.h"
#include "model/floor.h"

using namespace dss;

BOOST_AUTO_TEST_SUITE(zone)

DSUID_DEFINE(dsuid1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1);
DSUID_DEFINE(dsuid2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2);
DSUID_DEFINE(dsuid3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3);

BOOST_AUTO_TEST_CASE(testSensorSetReset) {
  Zone zone(2, NULL);
  MainZoneSensor_t ms1 = { dsuid1, SensorType::CO2Concentration, 1 };
  MainZoneSensor_t ms3 = { dsuid3, SensorType::HumidityIndoors, 3 };

  zone.setSensor(ms1);
  zone.setSensor(ms3);

  BOOST_CHECK(zone.isSensorAssigned(ms1.m_sensorType));
  BOOST_CHECK(zone.isSensorAssigned(ms3.m_sensorType));
  zone.resetSensor(ms1.m_sensorType);
  BOOST_CHECK(!zone.isSensorAssigned(ms1.m_sensorType));
  BOOST_CHECK(zone.isSensorAssigned(ms3.m_sensorType));
}

BOOST_AUTO_TEST_CASE(testRemovInvalidSensors) {
  // test: remove sensors that don't have the hosting device in the zone

  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;

  Zone zone(7, &apt);
  auto dev1 = DeviceReference(apt.allocateDevice(dsuid1), &apt);
  zone.addDevice(dev1);
  auto dev3 = DeviceReference(apt.allocateDevice(dsuid3), &apt);
  zone.addDevice(dev3);

  MainZoneSensor_t ms1 = { dsuid1, SensorType::CO2Concentration, 1 };
  MainZoneSensor_t ms2 = { dsuid2, SensorType::HumidityIndoors, 2 };
  MainZoneSensor_t ms3 = { dsuid3, SensorType::TemperatureIndoors, 3 };

  zone.setSensor(ms1);
  zone.setSensor(ms2);
  zone.setSensor(ms3);

  // device with dsuid2 is not in zone
  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 3);
  BOOST_CHECK(zone.isSensorAssigned(SensorType::HumidityIndoors));
  zone.removeInvalidZoneSensors();
  BOOST_CHECK(zone.isSensorAssigned(SensorType::CO2Concentration));
  BOOST_CHECK(zone.isSensorAssigned(SensorType::TemperatureIndoors));
  BOOST_CHECK(!zone.isSensorAssigned(SensorType::HumidityIndoors));
}

BOOST_AUTO_TEST_CASE(testAssignedUnassignedSensorTypes) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;

  Zone zone(7, &apt);
  BOOST_CHECK_GE(zone.getUnassignedSensorTypes().size(), 4);
  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 0);
  auto maxZoneSensors = zone.getUnassignedSensorTypes().size();

  auto dev1 = DeviceReference(apt.allocateDevice(dsuid1), &apt);
  zone.addDevice(dev1);
  auto dev3 = DeviceReference(apt.allocateDevice(dsuid3), &apt);
  zone.addDevice(dev3);

  MainZoneSensor_t ms1_1 = { dsuid1, SensorType::CO2Concentration, 1 };
  MainZoneSensor_t ms1_2 = { dsuid1, SensorType::HumidityIndoors, 2 };
  MainZoneSensor_t ms3 = { dsuid3, SensorType::TemperatureIndoors, 3 };
  zone.setSensor(ms1_1);
  zone.setSensor(ms1_2);
  zone.setSensor(ms3);

  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 3);
  BOOST_CHECK_EQUAL(zone.getAssignedSensorTypes(*dev1.getDevice()).size(), 2);
  BOOST_CHECK_EQUAL(zone.getUnassignedSensorTypes().size(), maxZoneSensors - 3);

  zone.removeDevice(dev1);
  zone.removeInvalidZoneSensors();
  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 1);
  BOOST_CHECK_EQUAL(zone.getAssignedSensorTypes(*dev1.getDevice()).size(), 0);
  BOOST_CHECK_EQUAL(zone.getUnassignedSensorTypes().size(), maxZoneSensors - 1);
}

BOOST_AUTO_TEST_CASE(testParseControlMode) {
  TemperatureConfig x;
  BOOST_CHECK_NO_THROW(
      x.parseControlMode("{\
        \"emergencyValue\": 50,\
        \"ctrlKp\": 5.2,\
        \"ctrlTs\": 240.1,\
        \"ctrlTi\": 1.2,\
        \"ctrlKd\": 1.3,\
        \"ctrlImin\": 600.0,\
        \"ctrlImax\": 2400.0,\
        \"ctrlYmin\": 0.5,\
        \"ctrlYmax\": 100.4,\
        \"ctrlAntiWindUp\": true\
    }"));

  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"emergencyValue\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlTs\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlTi\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlKd\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlYmin\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlYmax\": 50.0}"));
  BOOST_CHECK_THROW(x.parseControlMode("{\"ctrlAntiWindUp\": 1}"), std::runtime_error);
}

BOOST_AUTO_TEST_CASE(testParseFollowerMode) {
  TemperatureConfig x;
  BOOST_CHECK_NO_THROW(
      x.parseFollowerMode("{\
        \"referenceZone\": 38523,\
        \"ctrlOffset\": 10\
    }"));

  BOOST_CHECK_NO_THROW(x.parseFollowerMode("{\"referenceZone\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseFollowerMode("{\"ctrlOffset\": 50.0}"));
}

BOOST_AUTO_TEST_CASE(testParseManualMode) {
  TemperatureConfig x;
  BOOST_CHECK_NO_THROW(
      x.parseManualMode("{\
        \"controlValue\": 10\
    }"));

  BOOST_CHECK_NO_THROW(x.parseManualMode("{\"controlValue\": 50.0}"));
}

BOOST_AUTO_TEST_CASE(testFloors) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;

  std::shared_ptr<Floor> f1 = apt.allocateFloor();
  std::shared_ptr<Floor> f2 = apt.allocateFloor();

  std::shared_ptr<Zone> z1 = apt.allocateZone(7);
  std::shared_ptr<Zone> z2 = apt.allocateZone(8);
  std::shared_ptr<Zone> z3 = apt.allocateZone(9);

  f1->addZone(z1->getID());
  f1->addZone(z2->getID());
  f2->addZone(z3->getID());

  std::vector<int> f1Zones = f1->getZones();
  std::vector<int> f2Zones = f2->getZones();

  BOOST_CHECK(f1Zones.size() == 2);
  BOOST_CHECK(f2Zones.size() == 1);

  // z1 should be on the first floor
  BOOST_CHECK(std::find(f1Zones.begin(), f1Zones.end(), z1->getID()) != f1Zones.end());
  // z3 should be on the second floor
  BOOST_CHECK(std::find(f2Zones.begin(), f2Zones.end(), z3->getID()) != f2Zones.end());
  f2->addZone(z1->getID());
  f1Zones = f1->getZones();
  f2Zones = f2->getZones();
  // z1 moved away from first floor
  BOOST_CHECK(std::find(f1Zones.begin(), f1Zones.end(), z1->getID()) == f1Zones.end());
  // z1 was moved to the second floor
  BOOST_CHECK(std::find(f2Zones.begin(), f2Zones.end(), z1->getID()) != f2Zones.end());
  // z2 should still be on the first floor
  BOOST_CHECK(std::find(f1Zones.begin(), f1Zones.end(), z2->getID()) != f1Zones.end());
}

BOOST_AUTO_TEST_CASE(testFloorOrderWrap) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;

  std::shared_ptr<Floor> f1 = apt.allocateFloor();
  std::shared_ptr<Floor> f2 = apt.allocateFloor();
  std::shared_ptr<Floor> f3 = apt.allocateFloor();

  // remove a floor to create an "hole" in the order ID sequence
  apt.removeFloor(f2->getID());

  // pretend order ID's reached the limit where they would wrap to force
  // the allocate function to sort and reset them
  f3->setOrder(INT_MAX);

  // this allocation will force the recalculation of order IDs
  std::shared_ptr<Floor> f4 = apt.allocateFloor();

  // order IDs were recalculated, eliminating sequence holes,
  // we removed floor 2, so we have the default apartment floor (order 0)
  // and floors 1, 3 and 4 left, this should give floor 4 the order id of 3
  BOOST_CHECK(f4->getOrder() == 3);
}

BOOST_AUTO_TEST_SUITE_END()
